package prototype;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.List;
import java.util.Properties;
import java.util.Scanner;
import java.util.Vector;

import prototype.database.postgresql.PostgreSqlConfiguration;
import prototype.objects.ITrigger;
import prototype.parser.StatementParser;
import prototype.pdp.ConfidentialityAndIntegrityPdp;
import prototype.pdp.IAccessControlMechanism;
import prototype.results.IResult;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.implementation.mysql.MySqlStatementFactory;
import prototype.statements.implementation.postgresql.PostgreSqlStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.system.implementation.ExecutionException;
import prototype.system.implementation.System;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.postgresql.PostgreSqlDatabaseEngine;
import prototype.system.postgresql.PostgreSqlStateFactory;
import prototype.users.UserFactory;
import prototype.util.Logger;
import prototype.util.Pair;

/**
 * Class containing the main method.
 * 
 * 
 *
 */
public class Simulator {

	public static void automatic() {
		System lts = null;
		Scanner in = null;
		boolean verbose = true;
		try {
			StatementParser parser = StatementParser.getInstance();

			// MySqlStateFactory stateFactory = new MySqlStateFactory();
			PostgreSqlStateFactory stateFactory = new PostgreSqlStateFactory();
			// MySqlConfiguration configuration = new MySqlConfiguration("localhost:3306/prototype", "root", "password");
			// PostgreSqlConfiguration configuration = new PostgreSqlConfiguration("localhost:5432/prototype", "postgres", "password");
			// //MySqlStatementFactory statementFactory = MySqlStatementFactory.getInstance();

			Properties prop = readProperties();

			PostgreSqlConfiguration configuration = new PostgreSqlConfiguration(prop.getProperty("database"), prop.getProperty("dbUser"),
					prop.getProperty("dbPassword"));

			PostgreSqlStatementFactory statementFactory = PostgreSqlStatementFactory.getInstance();

			UserFactory userFactory = UserFactory.getInstance();

			// IDatabaseEngine engine = new MySqlDatabaseEngine(configuration, stateFactory);
			IDatabaseEngine engine = new PostgreSqlDatabaseEngine(configuration, stateFactory);

			Logger logger = new Logger(new File("logs.txt"), true);

			IAccessControlMechanism mechanism = new ConfidentialityAndIntegrityPdp(stateFactory, engine, userFactory, statementFactory, logger);

			lts = new System(mechanism, engine, statementFactory, logger);
			in = new Scanner(java.lang.System.in);

			List<String> test = attack(in); // getTestVector();

			for (String input : test) {
				if (verbose)
					java.lang.System.out.println("########## " + input + " ##########");
				execute(lts, parser, statementFactory, userFactory, input, verbose);
			}

			logger.close();

			// startInteraction(lts, in, parser, statementFactory, userFactory);

		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			try {
				lts.close();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
			in.close();
		}
	}
	
	public static void manual() {
		System lts = null;
		Scanner in = null;
		boolean verbose = true;
		try {
			StatementParser parser = StatementParser.getInstance();

			PostgreSqlStateFactory stateFactory = new PostgreSqlStateFactory();


			Properties prop = readProperties();

			PostgreSqlConfiguration configuration = new PostgreSqlConfiguration(prop.getProperty("database"), prop.getProperty("dbUser"),
					prop.getProperty("dbPassword"));

			PostgreSqlStatementFactory statementFactory = PostgreSqlStatementFactory.getInstance();

			UserFactory userFactory = UserFactory.getInstance();

			IDatabaseEngine engine = new PostgreSqlDatabaseEngine(configuration, stateFactory);

			Logger logger = new Logger(new File("logs.txt"), true);

			IAccessControlMechanism mechanism = new ConfidentialityAndIntegrityPdp(stateFactory, engine, userFactory, statementFactory, logger);

			lts = new System(mechanism, engine, statementFactory, logger);
			in = new Scanner(java.lang.System.in);

			startInteraction(lts, in, parser, statementFactory, userFactory);

			logger.close();


		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			try {
				lts.close();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
			in.close();
		}
	}
	
	private static Properties readProperties() throws ExecutionException {
		Properties prop = new Properties();
		InputStream input = null;

		try {

			input = new FileInputStream("config.properties");

			// load a properties file
			prop.load(input);

			return prop;

		} catch (IOException ex) {
			throw new ExecutionException(
					"Property file config.properties not found.");
		} finally {
			if (input != null) {
				try {
					input.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	private static void startInteraction(System lts, Scanner in, StatementParser parser, PostgreSqlStatementFactory statementFactory,
			UserFactory userFactory) throws MalformedStatementException, ExecutionException {
		String input = null;
		java.lang.System.out.println("Write a command. Write 'Stop' to conclude.");
		input = in.nextLine();

		while (!input.equalsIgnoreCase("Stop")) {

			execute(lts, parser, statementFactory, userFactory, input, true);

			java.lang.System.out.println("Write a command. Write 'Stop' to conclude.");
			input = in.nextLine();
		}
	}

	private static List<String> getTestVector() {
		Vector<String> test = new Vector<String>();

		test.add("AS admin : CREATE TABLE S (x,y) ");
		test.add("AS admin : CREATE TABLE R (x) ");
		test.add("AS admin: CREATE VIEW V (x,y) AS OWNER OF {x,y | (S(x,y) AND (x = 'a' OR y = 'b'))}");
		// test.add("AS admin:  INSERT IN S ['c','a']");
		test.add("AS admin:  INSERT IN S ['a','c']");
		test.add("AS admin:  INSERT IN S ['c','b']");
		test.add("AS admin:  INSERT IN S ['d','b']");
		// test.add("AS admin:  INSERT IN R ['b']");
		// test.add("AS admin: EXISTS x1. (  ( EXISTS x2. ( (V(x2) AND x1 = x2) )   AND EXISTS x3. (R(x1,x3))   )  )");

		test.add("AS admin: ADD USER user1");
		test.add("AS admin: GRANT SELECT ON V TO user1");
		// test.add("AS admin: GRANT SELECT ON R TO user1");
		// test.add("AS admin:   EXISTS x,y. ( (S(x,y) AND x = 'a' AND y = 'c' ) )  ");
		// test.add("AS admin:   EXISTS x,y. ( (S(x,y) AND x = 'c' AND y = 'b' ) )  ");
		test.add("AS user1:   EXISTS x,y. ( (S(x,y) AND x = 'd' ) )  ");
		// test.add("AS user1:  ( EXISTS x,y. ( (S(x,y) AND x = 'c' ) )  AND ( EXISTS w. ( (R(w) AND w='b') ) OR EXISTS q,z. ( (S(z,q) AND z='c') ) ) )");

		// test.add("AS admin : CREATE TABLE S (x, y, z) WITH fk : FOREIGN KEY (y,z) REFERENCES R (x,y)");

		// test.add("AS admin: CREATE VIEW v2 (x,y,z) AS OWNER OF {x,y,z | (R(x,y,z) OR S(z,y,x)) }");

		// test.add("AS admin : CREATE TABLE t2 (id, name) WITH fk : FOREIGN KEY (name) REFERENCES t1 (id)");
		// test.add("AS admin : INSERT IN t1 ['aaa','ab']");
		// test.add("AS admin : ('a'='a' OR t1('a','b')");

		// test.add("AS admin : ( EXISTS x1,x2. (  t1(x1, x2) ) OR EXISTS  v1,v2. (  t2(v1, v2) ) )");
		// test.add("AS admin : CREATE TABLE t2 (id, name) WITH fk : FOREIGN KEY (name) REFERENCES t1 (id)");
		// test.add("AS admin : CREATE VIEW v1 (id, name) AS OWNER OF { v1, v2 | t2(v1,v2) }");
		// test.add("AS admin : CREATE VIEW v2 (id, name) AS INVOKER OF { v1, v2 | t1(v1,v2) }");
		// test.add("AS admin : ADD USER user1");
		// test.add("AS admin : ADD USER user2");
		// //test.add("AS user1 : INSERT IN t1 ['aaa','ab']");
		// test.add("AS admin : DELETE FROM t1 ['aaa','acd']");
		// test.add("AS admin : GRANT SELECT ON t1 TO user1");
		// test.add("AS admin : INSERT IN t1 ['aaa','venticinque']");
		// test.add("AS user1 : INSERT IN t1 ['aaa','ab']");
		// test.add("AS admin : GRANT* CREATE TRIGGER ON t1 TO user1");
		// test.add("AS admin : GRANT* CREATE VIEW TO user1");
		// test.add("AS user1 : GRANT* CREATE VIEW TO user2");
		// test.add("AS user2 : GRANT* CREATE VIEW TO user1");
		//
		// test.add("AS admin : REVOKE CREATE VIEW FROM user1");

		//
		// test.add("AS admin : INSERT IN t1 ['aaa','ab']");
		// test.add("AS admin : CREATE TRIGGER tri1 AFTER INSERT ON t1 AS OWNER WHEN (EXISTS x1,x2. (t1(x1,x2)) OR 'a' = 'a')  DO INSERT IN t2 ['a',id]");
		// test.add("AS admin : INSERT IN t1 ['aaa','ab']");
		// test.add("AS admin : INSERT IN t1 ['aaa','c']");
		//
		// test.add("AS admin : CREATE TRIGGER tri2 AFTER DELETE ON t1 AS INVOKER WHEN ( t2(id, name) AND id = 'a' ) DO INSERT IN t2 [id,name]");
		// test.add("AS admin : ( EXISTS x1,x2. (  t1(x1, x2) ) OR EXISTS  v1,v2. (  t2(v1, v2) ) )");
		// test.add("AS admin : { x1 | ( EXISTS x2. (  t1(x1, x2) ) OR EXISTS  v2. (  t2(x1, v2) ) ) }");
		// test.add("AS admin :  EXISTS x1. ( (EXISTS x2. ( t1(x1,x2) ) OR EXISTS x3.( t2(x1, x3))) )   ");

		return test;
	}

	private static List<String> attack(Scanner in) throws ExecutionException {

		java.lang.System.out.println("ATTACK NUMBER");
		int attackNumber = in.nextInt();

		Vector<String> attack = new Vector<String>();

		switch (attackNumber) {
		case 0:
			return getTestVector();
		case 1:
			attack.add("AS admin : CREATE TABLE secret (id)");
			attack.add("AS admin : INSERT IN secret ['a']");
			attack.add("AS admin : INSERT IN secret ['b']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT SELECT ON secret TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			attack.add("AS u1: CREATE VIEW v (id) AS OWNER OF {x | (secret(x) AND x = 'a')}");
			attack.add("AS u1: GRANT SELECT ON v TO u2");
			attack.add("AS admin : GRANT* SELECT ON secret TO u1");
			attack.add("AS u1: GRANT SELECT ON v TO u2");
			attack.add("AS u2: {x | v(x)}");
			attack.add("AS u2: {x | secret(x)}");
			attack.add("AS u2: EXISTS x. ((secret(x) AND x = 'a'))");
			attack.add("AS u2: EXISTS x. ((secret(x) AND x = 'b'))");
			break;

		case 2:
			attack.add("AS admin : CREATE TABLE secret (id)");
			attack.add("AS admin : INSERT IN secret ['a']");
			attack.add("AS admin : INSERT IN secret ['b']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT* SELECT ON secret TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			attack.add("AS u1: CREATE VIEW v (id) AS OWNER OF {x | (secret(x) AND x = 'a')}");
			attack.add("AS u1: GRANT SELECT ON v TO u2");
			attack.add("AS u2: {x | v(x)}");
			attack.add("AS admin : REVOKE SELECT ON secret FROM u1");
			attack.add("AS u2: {x | v(x)}");
			break;

		case 3:
			attack.add("AS admin : CREATE TABLE salaries (id)");
			attack.add("AS admin : CREATE TABLE projects (id)");
			attack.add("AS admin : INSERT IN salaries ['a']");
			attack.add("AS admin : INSERT IN projects ['b']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT SELECT ON projects TO u1");
			attack.add("AS admin : GRANT INSERT ON projects TO u1");
			attack.add("AS admin : GRANT CREATE TRIGGER ON projects TO u1");
			attack.add("AS admin : GRANT SELECT ON salaries TO u2");
			attack.add("AS admin : GRANT DELETE ON salaries TO u2");
			attack.add("AS admin : GRANT SELECT ON projects TO u2");
			attack.add("AS admin : GRANT INSERT ON projects TO u2");
			attack.add("AS u1 : CREATE TRIGGER tri1 AFTER INSERT ON projects AS INVOKER WHEN TRUE DO DELETE FROM salaries ['a']");
			// attack.add("AS u2: INSERT IN projects ['b']");
			attack.add("AS u2: INSERT IN projects ['c']");
			// attack.add("AS u1: {x | salaries(x)}");
			break;
		case 4:
			attack.add("AS admin : CREATE TABLE patient (id) WITH pk : PRIMARY KEY (id)");
			attack.add("AS admin : CREATE TABLE sick (id, sickness) WITH pk : PRIMARY KEY (id), fk : FOREIGN KEY (id) REFERENCES patient (id)");
			attack.add("AS admin : INSERT IN patient ['bob']");
			attack.add("AS admin : INSERT IN sick ['bob', 'aids']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON patient TO u1");
			attack.add("AS admin : GRANT INSERT ON sick TO u1");
			attack.add("AS u1: INSERT IN sick ['bob', 'malaria']");

			break;
		case 5:
			attack.add("AS admin : CREATE TABLE projectManager (id)");
			attack.add("AS admin : CREATE TABLE transferRequest (id)");
			attack.add("AS admin : CREATE TABLE notification (id)");
			attack.add("AS admin : INSERT IN transferRequest ['a']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON projectManager TO u1");
			attack.add("AS admin : GRANT INSERT ON projectManager TO u1");
			attack.add("AS admin : GRANT DELETE ON projectManager TO u1");
			attack.add("AS admin : GRANT SELECT ON notification TO u1");
			attack.add("AS admin : GRANT INSERT ON notification TO u1");
			attack.add("AS admin : GRANT DELETE ON notification TO u1");
			attack.add("AS admin : CREATE TRIGGER tri1 AFTER INSERT ON projectManager AS OWNER WHEN transferRequest(id) DO INSERT IN notification [id]");
			attack.add("AS u1: DELETE FROM notification ['a']");
			attack.add("AS u1: INSERT IN projectManager ['a']");
			attack.add("AS u1: EXISTS x. ((notification(x) AND x = 'a'))");
			break;

		case 6:
			attack.add("AS admin : CREATE TABLE t1 (id)");
			attack.add("AS admin : CREATE TABLE t2 (id)");
			attack.add("AS admin: CREATE VIEW v (id) AS OWNER OF {x | (t1(x) AND t2(x))}");
			attack.add("AS admin: INSERT IN t1 ['a']");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON v TO u1");
			attack.add("AS admin : GRANT SELECT ON t1 TO u1");
			attack.add("AS admin : GRANT INSERT ON t1 TO u1");
			attack.add("AS u1: INSERT IN t1 ['a']");
			attack.add("AS u1: INSERT IN t1 ['b']");
			break;

		case 7:
			attack.add("AS admin : CREATE TABLE R (id)");
			attack.add("AS admin : CREATE TABLE T (id)");
			attack.add("AS admin : CREATE TABLE Z (id)");
			attack.add("AS admin: CREATE VIEW V (id) AS OWNER OF {x | (T(x) AND Z(x))}");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : ADD USER u3");
			attack.add("AS admin : GRANT SELECT ON R TO u1");
			attack.add("AS admin : GRANT* SELECT ON T TO u1");
			attack.add("AS admin : GRANT* SELECT ON Z TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			attack.add("AS admin : GRANT* SELECT ON R TO u2");
			attack.add("AS admin : GRANT* SELECT ON T TO u2");
			attack.add("AS admin : GRANT* SELECT ON Z TO u2");
			attack.add("AS admin : GRANT CREATE VIEW TO u2");
			attack.add("AS u1 : CREATE VIEW W1 (id) AS OWNER OF {x | (R(x) OR V(x))}");
			attack.add("AS u1 : GRANT SELECT ON W1 TO u3");
			attack.add("AS u2 : CREATE VIEW W2 (id) AS OWNER OF {x | (R(x) OR V(x))}");
			attack.add("AS u2 : GRANT SELECT ON W2 TO u3");
			break;

		case 8:
			attack.add("AS admin : CREATE TABLE S (x,y) ");
			attack.add("AS admin : CREATE TABLE R (x) ");
			attack.add("AS admin : CREATE TABLE Q (x) ");
			attack.add("AS admin: CREATE VIEW V (x,y) AS OWNER OF {x,y | (S(x,y) AND (x = 'a' OR y = 'c'))}");
			attack.add("AS admin: CREATE VIEW W (x) AS OWNER OF {x | (R(x) OR Q(x))}");

			attack.add("AS admin:  INSERT IN S ['a','a']");
			attack.add("AS admin:  INSERT IN S ['b','c']");
			attack.add("AS admin:  INSERT IN S ['d','b']");
			attack.add("AS admin:  INSERT IN R ['d']");
			attack.add("AS admin:  INSERT IN Q ['f']");

			attack.add("AS admin: ADD USER user1");
			attack.add("AS admin: GRANT SELECT ON V TO user1");
			attack.add("AS admin: GRANT SELECT ON W TO user1");

			attack.add("AS user1:  ( EXISTS x,y. ( (S(x,y) AND x = 'b') ) AND ( NOT(EXISTS z. ( (R(z) AND z = 'c') )) OR EXISTS q,w. ( (S(q,w) AND q = 'd') )) )");
			break;
		default:
			throw new ExecutionException("There is no Attack " + in + ".");
		}

		return attack;
	}

	private static void execute(System lts, StatementParser parser, IStatementFactory statementFactory, UserFactory userFactory, String input,
			boolean verbose) throws MalformedStatementException, ExecutionException {
		// parse the statement
		IStatement statement = parser.create(input, statementFactory, userFactory, lts.getSystemState());

		// execute the statement
		IResult output = lts.execute(statement);
		if (verbose)
			java.lang.System.out.println(output.toString());

		// execute the scheduled triggers
		while (!lts.noScheduledTriggers()) {
			Pair<ITrigger, IResult> p = lts.executeScheduledTrigger();
			if (verbose) {
				java.lang.System.out.println("EXECUTE " + p.getFirst());
				java.lang.System.out.println(p.getSecond().toString());
			}
		}
	}

	public static void main(String[] args) {
		if (args[0].equals("manual"))
			manual();
		else if (args[0].equals("automatic"))
			automatic();
		else if (args[0].equals("experiments"))
			experiments(Integer.parseInt(args[1]), Integer.parseInt(args[2]), Integer.parseInt(args[3]));
	}

	public static void experiments(int size, int times, int attackNr) {
		for (int i = 0; i < times; i++) {
			java.lang.System.out.println("Attack " + attackNr + " size " + size + " " + (i + 1) + "/" + times);
			System lts = null;
			Scanner in = null;
			try {
				StatementParser parser = StatementParser.getInstance();

				// MySqlStateFactory stateFactory = new MySqlStateFactory();
				PostgreSqlStateFactory stateFactory = new PostgreSqlStateFactory();
				// MySqlConfiguration configuration = new MySqlConfiguration("localhost:3306/prototype", "root", "password");
				// PostgreSqlConfiguration configuration = new PostgreSqlConfiguration("localhost:5432/prototype", "postgres", "password");
				// MySqlStatementFactory statementFactory = MySqlStatementFactory.getInstance();

				Properties prop = readProperties();

				PostgreSqlConfiguration configuration = new PostgreSqlConfiguration(prop.getProperty("database"), prop.getProperty("dbUser"),
						prop.getProperty("dbPassword"));
				PostgreSqlStatementFactory statementFactory = PostgreSqlStatementFactory.getInstance();

				UserFactory userFactory = UserFactory.getInstance();

				// IDatabaseEngine engine = new MySqlDatabaseEngine(configuration, stateFactory);
				PostgreSqlDatabaseEngine engine = new PostgreSqlDatabaseEngine(configuration, stateFactory);

				engine.disableRefreshOfMateriliazedViews();
				Logger logger = new Logger(new File("attack_" + attackNr + "_setup_size_" + size + "_" + i + ".txt"), true);

				IAccessControlMechanism mechanism = new ConfidentialityAndIntegrityPdp(stateFactory, engine, userFactory, statementFactory, logger);

				lts = new System(mechanism, engine, statementFactory, logger);

				java.lang.System.out.println("start create");
				List<String> createTables = setup(attackNr);
				for (String input : createTables) {
					execute(lts, parser, statementFactory, userFactory, input, false);
				}
				java.lang.System.out.println("end create");

				java.lang.System.out.println("start fill table");
				List<String> fillTables = fillTables(attackNr, size);
				for (String input : fillTables) {
					execute(lts, parser, statementFactory, userFactory, input, false);
				}
				java.lang.System.out.println("end fill table");

				java.lang.System.out.println("start materialization");
				engine.enableRefreshOfMaterializedViews();
				engine.refreshMaterializedViews(lts.getSystemState().getViews());
				java.lang.System.out.println("end materialization");

				java.lang.System.out.println("start attack");
				logger.changeFile(new File("attack_" + attackNr + "_attack_size_" + size + "_" + i + ".txt"), true);

				List<String> attack = attack(attackNr);
				for (String input : attack) {
					execute(lts, parser, statementFactory, userFactory, input, false);
				}

				java.lang.System.out.println("end attack");

				logger.close();

			} catch (Exception e) {
				StringWriter writer = new StringWriter();
				PrintWriter printWriter = new PrintWriter(writer);
				e.printStackTrace(printWriter);
				java.lang.System.out.println(writer.toString());
			} finally {
				try {
					lts.close();
				} catch (ExecutionException e) {
					e.printStackTrace();
				}
			}
		}
	}

	private static List<String> fillTables(int attackNumber, int size) throws ExecutionException {

		Vector<String> attack = new Vector<String>();

		switch (attackNumber) {
		case 1:
			attack.add("AS admin : INSERT IN secret ['a']");
			attack.add("AS admin : INSERT IN secret ['b']");

			for (int i = 0; i < size - 2; i++)
				attack.add("AS admin : INSERT IN secret ['a_" + i + "']");
			break;

		case 2:
			attack.add("AS admin : INSERT IN secret ['a']");
			attack.add("AS admin : INSERT IN secret ['b']");

			for (int i = 0; i < size - 2; i++)
				attack.add("AS admin : INSERT IN secret ['a_" + i + "']");
			break;

		case 3:

			attack.add("AS admin : INSERT IN salaries ['a']");
			attack.add("AS admin : INSERT IN projects ['b']");

			for (int i = 0; i < (size / 2) - 1; i++)
				attack.add("AS admin : INSERT IN salaries ['a_" + i + "']");

			for (int i = 0; i < (size / 2) - 1; i++)
				attack.add("AS admin : INSERT IN projects ['b_" + i + "']");
			break;
		case 4:
			attack.add("AS admin : INSERT IN patient ['bob']");
			attack.add("AS admin : INSERT IN sick ['bob', 'asshole']");

			for (int i = 0; i < (size / 2) - 1; i++)
				attack.add("AS admin : INSERT IN patient ['pat_" + i + "']");

			for (int i = 0; i < (size / 2) - 1; i++) {
				attack.add("AS admin : INSERT IN sick ['pat_" + i + "', 'dis_" + ((size / 2) + i) + "']");
			}

			break;
		case 5:

			attack.add("AS admin : INSERT IN transferRequest ['a']");

			for (int i = 0; i < size - 1; i++)
				attack.add("AS admin : INSERT IN transferRequest ['a_" + i + "']");
			break;

		case 6:
			attack.add("AS admin: INSERT IN t1 ['a']");

			for (int i = 0; i < size - 1; i++)
				attack.add("AS admin : INSERT IN t1 ['a_" + i + "']");
			break;

		case 7:
			for (int i = 0; i < (size / 2); i++)
				attack.add("AS admin : INSERT IN T ['a_" + i + "']");

			for (int i = 0; i < (size / 4); i++) {
				attack.add("AS admin : INSERT IN Z ['a_" + i + "']");
				attack.add("AS admin : INSERT IN R ['b_" + i + "']");
			}
			break;
		case 8:
			attack.add("AS admin:  INSERT IN S ['a','a']");
			attack.add("AS admin:  INSERT IN S ['b','c']");
			attack.add("AS admin:  INSERT IN S ['d','b']");
			attack.add("AS admin:  INSERT IN R ['d']");

			for (int i = 0; i < (size / 2) - 2; i++)
				attack.add("AS admin : INSERT IN S ['a_" + i + "','b_" + i + "']");

			for (int i = 0; i < (size / 4) - 1; i++)
				attack.add("AS admin : INSERT IN R ['c_" + i + "']");

			for (int i = 0; i < (size / 4) - 1; i++)
				attack.add("AS admin : INSERT IN Q ['d_" + i + "']");
			break;
		default:
			throw new ExecutionException("There is no Attack " + attackNumber + ".");
		}

		return attack;
	}

	private static List<String> setup(int attackNumber) throws ExecutionException {

		Vector<String> attack = new Vector<String>();

		switch (attackNumber) {
		case 1:
			attack.add("AS admin : CREATE TABLE secret (id)");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT SELECT ON secret TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			break;

		case 2:
			attack.add("AS admin : CREATE TABLE secret (id)");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT* SELECT ON secret TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			break;

		case 3:
			attack.add("AS admin : CREATE TABLE salaries (id)");
			attack.add("AS admin : CREATE TABLE projects (id)");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : GRANT SELECT ON projects TO u1");
			attack.add("AS admin : GRANT INSERT ON projects TO u1");
			attack.add("AS admin : GRANT CREATE TRIGGER ON projects TO u1");
			attack.add("AS admin : GRANT* SELECT ON salaries TO u2");
			attack.add("AS admin : GRANT SELECT ON projects TO u2");
			attack.add("AS admin : GRANT INSERT ON projects TO u2");
			break;
		case 4:
			attack.add("AS admin : CREATE TABLE patient (id) WITH pk : PRIMARY KEY (id)");
			attack.add("AS admin : CREATE TABLE sick (id, sickness) WITH pk : PRIMARY KEY (id), fk : FOREIGN KEY (id) REFERENCES patient (id)");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON patient TO u1");
			attack.add("AS admin : GRANT INSERT ON sick TO u1");
			break;
		case 5:
			attack.add("AS admin : CREATE TABLE projectManager (id)");
			attack.add("AS admin : CREATE TABLE transferRequest (id)");
			attack.add("AS admin : CREATE TABLE notification (id)");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON projectManager TO u1");
			attack.add("AS admin : GRANT INSERT ON projectManager TO u1");
			attack.add("AS admin : GRANT DELETE ON projectManager TO u1");
			attack.add("AS admin : GRANT SELECT ON notification TO u1");
			attack.add("AS admin : GRANT INSERT ON notification TO u1");
			attack.add("AS admin : GRANT DELETE ON notification TO u1");
			attack.add("AS admin : CREATE TRIGGER tri1 AFTER INSERT ON projectManager AS OWNER WHEN transferRequest(id) DO INSERT IN notification [id]");
			break;

		case 6:
			attack.add("AS admin : CREATE TABLE t1 (id)");
			attack.add("AS admin : CREATE TABLE t2 (id)");
			attack.add("AS admin: CREATE VIEW v (id) AS OWNER OF {x | (t1(x) AND t2(x))}");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : GRANT SELECT ON v TO u1");
			attack.add("AS admin : GRANT SELECT ON t1 TO u1");
			attack.add("AS admin : GRANT INSERT ON t1 TO u1");
			break;

		case 7:
			attack.add("AS admin : CREATE TABLE R (id)");
			attack.add("AS admin : CREATE TABLE T (id)");
			attack.add("AS admin : CREATE TABLE Z (id)");
			attack.add("AS admin: CREATE VIEW V (id) AS OWNER OF {x | (T(x) AND Z(x))}");
			attack.add("AS admin : ADD USER u1");
			attack.add("AS admin : ADD USER u2");
			attack.add("AS admin : ADD USER u3");
			attack.add("AS admin : GRANT SELECT ON R TO u1");
			attack.add("AS admin : GRANT* SELECT ON T TO u1");
			attack.add("AS admin : GRANT* SELECT ON Z TO u1");
			attack.add("AS admin : GRANT CREATE VIEW TO u1");
			attack.add("AS admin : GRANT* SELECT ON R TO u2");
			attack.add("AS admin : GRANT* SELECT ON T TO u2");
			attack.add("AS admin : GRANT* SELECT ON Z TO u2");
			attack.add("AS admin : GRANT CREATE VIEW TO u2");
			break;
		case 8:
			attack.add("AS admin : CREATE TABLE S (x,y) ");
			attack.add("AS admin : CREATE TABLE R (x) ");
			attack.add("AS admin : CREATE TABLE Q (x) ");
			attack.add("AS admin: CREATE VIEW V (x,y) AS OWNER OF {x,y | (S(x,y) AND (x = 'a' OR y = 'c'))}");
			attack.add("AS admin: CREATE VIEW W (x) AS OWNER OF {x | (R(x) OR Q(x))}");
			attack.add("AS admin: ADD USER u1");
			attack.add("AS admin: GRANT SELECT ON V TO u1");
			attack.add("AS admin: GRANT SELECT ON W TO u1");
			break;
		default:
			throw new ExecutionException("There is no Attack " + attackNumber + ".");
		}

		return attack;
	}

	private static List<String> attack(int attackNumber) throws ExecutionException {
		Vector<String> attack = new Vector<String>();

		switch (attackNumber) {
		case 1:
			attack.add("AS u1: CREATE VIEW v (id) AS OWNER OF {x | (secret(x) AND x = 'a')}");
			attack.add("AS u1: GRANT SELECT ON v TO u2");
			attack.add("AS u2: {x | v(x)}");
			attack.add("AS u2: {x | secret(x)}");
			attack.add("AS u2: EXISTS x. ((secret(x) AND x = 'a'))");
			attack.add("AS u2: EXISTS x. ((secret(x) AND x = 'b'))");
			break;

		case 2:
			attack.add("AS u1: CREATE VIEW v (id) AS OWNER OF {x | (secret(x) AND x = 'a')}");
			attack.add("AS u1: GRANT SELECT ON v TO u2");
			attack.add("AS u2: {x | v(x)}");
			attack.add("AS admin : REVOKE SELECT ON secret FROM u1");
			attack.add("AS u2: {x | v(x)}");
			break;

		case 3:
			attack.add("AS u1 : CREATE TRIGGER tri1 AFTER INSERT ON projects AS INVOKER WHEN TRUE DO DELETE FROM salaries ['a']");
			attack.add("AS u2: INSERT IN projects ['c']");

			// attack.add("AS u1 : CREATE TRIGGER tri1 AFTER INSERT ON projects AS INVOKER WHEN TRUE DO GRANT SELECT ON salaries TO u1");
			// //attack.add("AS u2: INSERT IN projects ['b']");
			// attack.add("AS u2: INSERT IN projects ['c']");
			// //attack.add("AS u1: {x | salaries(x)}");
			break;
		case 4:
			attack.add("AS u1: INSERT IN sick ['bob', 'malaria']");
			break;
		case 5:
			attack.add("AS u1: DELETE FROM notification ['a']");
			attack.add("AS u1: INSERT IN projectManager ['a']");
			attack.add("AS u1: EXISTS x. ((notification(x) AND x = 'a'))");
			break;

		case 6:
			attack.add("AS u1: INSERT IN t1 ['a']");
			attack.add("AS u1: INSERT IN t1 ['b']");
			break;

		case 7:
			attack.add("AS u1 : CREATE VIEW W1 (id) AS OWNER OF {x | (R(x) OR V(x))}");
			attack.add("AS u1 : GRANT SELECT ON W1 TO u3");
			attack.add("AS u2 : CREATE VIEW W2 (id) AS OWNER OF {x | (R(x) OR V(x))}");
			attack.add("AS u2 : GRANT SELECT ON W2 TO u3");
			break;
		case 8:
			attack.add("AS u1:  ( EXISTS x,y. ( (S(x,y) AND x = 'b') ) AND ( NOT(EXISTS z. ( (R(z) AND z = 'c') )) OR EXISTS q,w. ( (S(q,w) AND q = 'd') )) )");
			break;
		default:
			throw new ExecutionException("There is no Attack " + attackNumber + ".");
		}

		return attack;
	}

}
