package prototype.database.postgresql;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Savepoint;
import java.sql.Statement;

import prototype.system.implementation.ExecutionException;

/**
 * Class managing the connection with the PostgreSql database.
 * 
 * 
 *
 */
public class PostgreSqlConnectionManager {
	/**
	 * Database connection.
	 */
	private Connection connection;

	/**
	 * The code associated to the integrity exceptions in getSqlState().
	 */
	private static String INTEGRITY_EXCEPTION_CODE = "23000";

	/**
	 * The id of the next savepoint.
	 */
	private long savepointId = initialId;

	/**
	 * Initial savepoint id.
	 */
	public static long initialId = 0;

	/**
	 * Constructor.
	 * 
	 * @param configuration
	 *            - database configuration.
	 * @throws ExecutionException
	 */
	public PostgreSqlConnectionManager(PostgreSqlConfiguration configuration) throws ExecutionException {
		if (configuration != null) {
			try {
				// Load the MySql driver
				Class.forName("org.postgresql.Driver");
				// Setup the connection with the DB

				connection = DriverManager.getConnection(configuration.getConnectionString());
			} catch (SQLException | ClassNotFoundException e) {
				throw new ExecutionException(e);
			}
		} else
			throw new ExecutionException("Database configuration cannot be null.");
	}

	/**
	 * Close the connection.
	 * 
	 * @throws ExecutionException
	 */
	public void close() throws ExecutionException {
		if (connection != null) {
			try {
				connection.close();

			} catch (SQLException e) {
				throw new ExecutionException(e);
			}

			this.savepointId = 0;
			connection = null;
		}
	}

	/**
	 * Create a new SQL statement.
	 * 
	 * @return The new statement.
	 */
	public Statement createStatement() throws ExecutionException {
		try {
			return connection.createStatement();
		} catch (SQLException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Get the code associated with Integrity exceptions.
	 * 
	 * @return The code returned by the method getSqlCode() of the SqlException in case there has been an integrity exception.
	 */
	public static String getIntegrityExceptionCode() {
		return INTEGRITY_EXCEPTION_CODE;
	}

	/**
	 * Set the auto-commit flag.
	 * 
	 * @param b
	 *            - The flag.
	 * @throws ExecutionException
	 */
	public void setAutoCommit(boolean b) throws ExecutionException {
		try {
			connection.setAutoCommit(b);
		} catch (SQLException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Get the id for the next savepoint.
	 * 
	 * @return The id.
	 */
	public String getNewSavepointId() {
		String id = "" + savepointId;
		savepointId++;
		return id;
	}

	/**
	 * Create a new savepoint with a given Id.
	 * 
	 * @param id
	 *            - The id.
	 * @return The savepoint.
	 * @throws ExecutionException
	 */
	public Savepoint createNewSavepoint(String id) throws ExecutionException {
		try {
			Savepoint save = connection.setSavepoint(id);
			
			return save;
		} catch (SQLException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Commit the transaction.
	 * 
	 * @throws ExecutionException
	 */
	public void commit() throws ExecutionException {
		try {
			connection.commit();
			this.savepointId = 0;
		} catch (SQLException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Rollback to the given savepoint. If the savepoint is null, it just rollbacks the whole transaction.
	 * 
	 * @param savepoint
	 *            - The savepoint.
	 * @throws ExecutionException
	 */
	public void rollback(Savepoint savepoint) throws ExecutionException {
		if (savepoint == null) {
			try {
				connection.rollback();

				this.savepointId = 0;
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		} else {
			try {
				connection.rollback(savepoint);
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		}
	}

}