package prototype.objects.constraints;

import java.util.List;
import java.util.Vector;

import prototype.interfaces.CopyException;
import prototype.objects.ITable;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.AbstractObject;
import prototype.users.IUser;
import prototype.util.Util;

/**
 * Class representing foreign key constraints.
 * 
 * 
 *
 */
public class ForeignKeyConstraint extends AbstractObject implements IConstraint {

	/**
	 * Destination table for the foreing key constraints.
	 */
	private ITable table;

	/**
	 * Fields in table.
	 */
	private List<String> destinationFields;

	/**
	 * Fields from which the foreing key starts.
	 */
	private List<String> fields;

	private String tablePrefix;

	public ForeignKeyConstraint(String id, IUser owner, PRIVILEGE_MODE mode, List<String> fields, List<String> destinationFields,
			ITable destinationTable) throws MalformedDbObjectException {
		super(id, owner, mode);

		if (fields == null || fields.isEmpty())
			throw new MalformedDbObjectException("Fields cannot be null or empty.");
		else if (mode != PRIVILEGE_MODE.OWNER)
			throw new MalformedDbObjectException("Mode can only be OWNER.");
		else if (destinationFields == null || destinationFields.isEmpty())
			throw new MalformedDbObjectException("DestinationField cannot be null or empty.");
		else if (destinationTable == null)
			throw new MalformedDbObjectException("destinationTable cannot be null.");
		else if (fields.size() != destinationFields.size())
			throw new MalformedDbObjectException("The size of fields and destinationfields is different.");
		else if (destinationTable.getFields().containsAll(destinationFields)) {
			this.destinationFields = destinationFields;
			this.fields = fields;
			this.table = destinationTable;
		} else
			throw new MalformedDbObjectException("destinationFields are not in destinationTable.getFields().");

	}

	/**
	 * Get the destination fields.
	 * 
	 * @return The destination fields.
	 */
	public List<String> getDestinationFields() {
		return destinationFields;
	}

	/**
	 * Get the source fields.
	 * 
	 * @return The fields.
	 */
	public List<String> getFields() {
		return fields;
	}

	/**
	 * Get the table.
	 * 
	 * @return The table.
	 */
	public ITable getDestinationTable() {
		return table;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.IDbObject#copy()
	 */
	@Override
	public ForeignKeyConstraint copy() throws CopyException {
		try {
			ForeignKeyConstraint fk = new ForeignKeyConstraint(getId(), getOwner(), getMode(), new Vector<String>(this.fields), new Vector<String>(
					this.destinationFields), table.copy());
			fk.tablePrefix = this.tablePrefix;
			return fk;
		} catch (MalformedDbObjectException e) {
			throw new CopyException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.constraints.IConstraint#toSql()
	 */
	@Override
	public String toSql() {
		String def = "CONSTRAINT " + (tablePrefix != null? tablePrefix+"_":"")+ this.getId() + " FOREIGN KEY (";

		for (String f : fields)
			def += f + ", ";

		def = Util.replaceLast(def, ", ", "");

		def += ") REFERENCES " + this.table.getId() + "(";
		for (String f : destinationFields)
			def += f + ", ";

		def = Util.replaceLast(def, ", ", ")");

		return def;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.implementation.AbstractObject#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String def = (tablePrefix != null ? tablePrefix : "") + this.getId() + ": FK (";
		for (String f : fields)
			def += f + ", ";
		def = Util.replaceLast(def, ", ", ")");

		def += " TO TABLE " + getDestinationTable().getId() + " (";
		for (String f : destinationFields)
			def += f + ", ";
		def = Util.replaceLast(def, ", ", ")");

		return def;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof ForeignKeyConstraint) {
			if (obj == this)
				return true;
			else
				return ((ForeignKeyConstraint) obj).getId().equals(this.getId())
						&& ((ForeignKeyConstraint) obj).getFields().equals(this.getFields())
						&& ((ForeignKeyConstraint) obj).getOwner().equals(this.getOwner())
						&& ((ForeignKeyConstraint) obj).getMode().equals(this.getMode())
						&& ((ForeignKeyConstraint) obj).getDestinationTable().equals(this.getDestinationTable())
						&& ((ForeignKeyConstraint) obj).getDestinationFields().equals(this.getDestinationFields())
						&& ((tablePrefix != null && ((ForeignKeyConstraint) obj).tablePrefix != null && ((ForeignKeyConstraint) obj).tablePrefix
								.equals(tablePrefix)) || (tablePrefix == null && ((ForeignKeyConstraint) obj).tablePrefix == null));
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getId().hashCode();
		result = 31 * result + getFields().hashCode();
		result = 31 * result + getOwner().hashCode();
		result = 31 * result + getMode().hashCode();
		result = 31 * result + getDestinationTable().hashCode();
		result = 31 * result + getDestinationFields().hashCode();
		result = 31 * result + (tablePrefix != null ? tablePrefix.hashCode() : 0);
		return result;
	}

	@Override
	public void setTablePrefix(String table) {
		this.tablePrefix = table;
	}

}
