package prototype.objects.constraints;

import java.util.HashSet;
import java.util.Set;

import prototype.interfaces.CopyException;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.AbstractObject;
import prototype.users.IUser;
import prototype.util.Util;

/**
 * Class representing primary key constraints.
 * 
 * 
 *
 */
public class UniqueConstraint extends AbstractObject implements IConstraint {

	/**
	 * Set of fields on which the constraint is defined.
	 */
	private Set<String> fields;
	private String tablePrefix;

	/**
	 * Constructor.
	 * 
	 * @param id
	 *            - Constraint's id.
	 * @param owner
	 *            - Constraint's owner.
	 * @param mode
	 *            - Constraint's mode (can be only OWNER).
	 * @param fields
	 *            - Set of fields on which the constraint is defined.
	 * @throws MalformedDbObjectException
	 */
	public UniqueConstraint(String id, IUser owner, PRIVILEGE_MODE mode, Set<String> fields) throws MalformedDbObjectException {
		super(id, owner, mode);
		if (fields != null && !fields.isEmpty()) {
			if (mode == PRIVILEGE_MODE.OWNER)
				this.fields = fields;
			else
				throw new MalformedDbObjectException("Mode must be OWNER.");
		} else
			throw new MalformedDbObjectException("Fields cannot be null or empty.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.IDbObject#copy()
	 */
	@Override
	public UniqueConstraint copy() throws CopyException {
		try {
			UniqueConstraint uc = new UniqueConstraint(getId(), getOwner(), getMode(), new HashSet<String>(fields));
			uc.tablePrefix = this.tablePrefix;
			return uc;
		} catch (MalformedDbObjectException e) {
			throw new CopyException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.constraints.IConstraint#toSql()
	 */
	@Override
	public String toSql() {
		String def = "CONSTRAINT " + (tablePrefix != null? tablePrefix+"_":"")+ this.getId() + " UNIQUE (";

		for (String f : fields)
			def += f + ", ";

		def = Util.replaceLast(def, ", ", ")");


		return def;
	}

	/* (non-Javadoc)
	 * @see guarnieri.marco.ethz.objects.implementation.AbstractObject#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String def = (tablePrefix != null? tablePrefix : "") + this.getId()+": UNIQUE(";
		for (String f : fields)
			def += f + ", ";
		def = Util.replaceLast(def, ", ", ")");
		return def;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof UniqueConstraint) {
			if (obj == this)
				return true;
			else
				return ((UniqueConstraint) obj).getId().equals(this.getId()) && ((UniqueConstraint) obj).getFields().equals(this.getFields())
						&& ((UniqueConstraint) obj).getOwner().equals(this.getOwner())
						&& ((UniqueConstraint) obj).getMode().equals(this.getMode())
						&& ((tablePrefix != null && ((UniqueConstraint) obj).tablePrefix != null && ((UniqueConstraint) obj).tablePrefix
						.equals(tablePrefix)) || (tablePrefix == null && ((UniqueConstraint) obj).tablePrefix == null));
		} else
			return false;
	}
	
	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getId().hashCode();
		result = 31 * result + getFields().hashCode();
		result = 31 * result + getOwner().hashCode();
		result = 31 * result + getMode().hashCode();
		result = 31 * result + (tablePrefix != null? tablePrefix.hashCode() : 0);
		return result;
	}

	/**
	 * Get the fields on which the constraint has been defined.
	 * 
	 * @return The set of fields.
	 */
	public Set<String> getFields() {
		return this.fields;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.getDefinition();
	}


	@Override
	public void setTablePrefix(String table) {
		this.tablePrefix = table;
	}
}
