package prototype.objects.implementation;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import prototype.interfaces.CopyException;
import prototype.objects.ITable;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.constraints.IConstraint;
import prototype.objects.constraints.PrimaryKeyConstraint;
import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.users.IUser;

/**
 * Class representing tables (relation schemas) in the database.
 * 
 * 
 *
 */
public class Table extends AbstractTypedObject implements ITable {

	/**
	 * The set of integrity constraints.
	 */
	private Set<IConstraint> constraints;

	/**
	 * Constructor.
	 * 
	 * @param id
	 *            - table's id.
	 * @param fields
	 *            - table's fields.
	 * @param owner
	 *            - table's owner.
	 * @param constraints
	 *            - table's integrity constraints.
	 * @throws MalformedDbObjectException
	 */
	public Table(String id, List<String> fields, IUser owner, Set<IConstraint> constraints) throws MalformedDbObjectException {
		super(id, owner, PRIVILEGE_MODE.OWNER, fields); // the privilege mode of a table is always OWNER
		if (constraints != null) {
			int i = 0;
			for (IConstraint c : constraints)
				if (c instanceof PrimaryKeyConstraint)
					i++;

			if (i > 1)
				throw new MalformedDbObjectException("The table can have at most one primary key constraint");

			for (IConstraint c : constraints)
				for (IConstraint c1 : constraints)
					if (c != c1 && c.getId().equalsIgnoreCase(c1.getId()))
						throw new MalformedDbObjectException("The constraints do not have unique identifiers.");

			this.constraints = new HashSet<IConstraint>();
			for (IConstraint c : constraints) {

				try {
					IConstraint copy = c.copy();
					copy.setTablePrefix(this.getId());
					this.constraints.add(copy);
				} catch (CopyException e) {
					throw new MalformedDbObjectException(e);
				}

			}
		} else
			throw new MalformedDbObjectException("Constraints cannot be null.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.ITable#getConstraints()
	 */
	@Override
	public Set<IConstraint> getConstraints() {
		return constraints;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof ITable) {
			if (obj == this)
				return true;
			else
				return ((ITable) obj).getId().equals(this.getId()) && ((ITable) obj).getFields().equals(this.getFields())
						&& ((ITable) obj).getConstraints().equals(this.getConstraints()) && ((ITable) obj).getOwner().equals(this.getOwner())
						&& ((ITable) obj).getMode().equals(this.getMode());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getId().hashCode();
		result = 31 * result + getFields().hashCode();
		result = 31 * result + getOwner().hashCode();
		result = 31 * result + getMode().hashCode();
		result = 31 * result + getFields().hashCode();
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.implementation.AbstractTypedObject#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String def = "TABLE " + this.getId() + " (";

		for (int i = 0; i < this.getCardinality(); i++) {
			if (i != 0)
				def += ", ";
			def += this.getFields().get(i);
		}

		def += ") OWNED BY " + getOwner().getUsername();
		for (IConstraint cstr : this.getConstraints())
			def += cstr.getDefinition();

		return def;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.getDefinition();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.IDbObject#copy()
	 */
	@Override
	public Table copy() throws CopyException {
		Set<IConstraint> newCstrs = new HashSet<IConstraint>();
		for (IConstraint c : getConstraints())
			newCstrs.add((IConstraint) c.copy());
		try {
			return new Table(getId(), new Vector<String>(getFields()), getOwner().copy(), constraints);
		} catch (MalformedDbObjectException e) {
			throw new CopyException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.ITable#getFreeVariables()
	 */
	@Override
	public List<Variable> getFreeVariables() {

		List<Variable> res = new Vector<Variable>();
		for (String f : this.getFields())
			try {
				res.add(new Variable(f));
			} catch (MalformedExpressionException e) {
				assert (false); // unreachable code.
			}

		return res;
	}

}
