package prototype.objects.implementation;

import java.util.List;

import prototype.interfaces.CopyException;
import prototype.objects.ITrigger;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.triggers.actions.ITriggerAction;
import prototype.objects.implementation.triggers.actions.UpdateTriggerAction;
import prototype.objects.implementation.triggers.events.ITriggerEvent;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.BooleanQuery;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.FreeVariableVisitor;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Class representing triggers.
 * 
 * 
 *
 */
public class Trigger extends AbstractObject implements ITrigger {

	/**
	 * Trigger's action.
	 */
	private ITriggerAction action;

	/**
	 * Trigger's event.
	 */
	private ITriggerEvent event;

	/**
	 * Trigger's condition.
	 */
	private IExpression condition;

	/**
	 * Free variables used in condition and action.
	 */
	private List<Variable> freeVariables;

	public Trigger(String id, IUser owner, PRIVILEGE_MODE mode, ITriggerAction action, ITriggerEvent event, IExpression condition)
			throws MalformedDbObjectException {
		super(id, owner, mode);

		if (action != null) {
			if (event != null) {
				if (condition != null) {

					// check the free variables in the condition
					FreeVariableVisitor v = new FreeVariableVisitor();
					condition.accept(v);

					if (event.getRelationSchema().getFreeVariables().containsAll(v.getFreeVariables())) {

						// check the free variables in the update action
						if (action instanceof UpdateTriggerAction) {
							for (IOperand op : ((UpdateTriggerAction) action).getTuple().getContent())
								if (op instanceof Variable)
									if (!event.getRelationSchema().getFreeVariables().contains(op))
										throw new MalformedDbObjectException("Action can contains just free variables derived from the fields of "
												+ "the relation schema where the trigger is defined");
						}

						// here everything should be ok
						this.action = action;
						this.event = event;
						this.condition = condition;
						this.freeVariables = event.getRelationSchema().getFreeVariables();

					} else
						throw new MalformedDbObjectException("Condition can contains just free variables derived from the fields of "
								+ "the relation schema where the trigger is defined");

				} else
					throw new MalformedDbObjectException("Condition cannot be null");
			} else
				throw new MalformedDbObjectException("Event cannot be null");
		} else
			throw new MalformedDbObjectException("Action cannot be null");

	}

	@Override
	public ITriggerEvent getEvent() {
		return this.event;
	}

	@Override
	public IExpression getCondition() {
		return this.condition;
	}

	@Override
	public ITriggerAction getAction() {
		return this.action;
	}

	@Override
	public ISelectStatement getConcreteCondition(IUser triggerInvoker, ITuple concreteTuple, IStatementFactory factory)
			throws MalformedStatementException {
		if (triggerInvoker != null)
			if (concreteTuple != null)
				if (concreteTuple.getSize() == this.getEvent().getRelationSchema().getCardinality()) {

					// replace free variables x_{1}, ..., x_{|table|} with v_{1}, ..., v_{|table|} in concrete tuple
					ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(freeVariables, concreteTuple.getContent());
					condition.accept(v);

					IExpression replaced = v.getResult();

					IQuery query;
					try {
						query = new BooleanQuery(replaced);
					} catch (MalformedQueryException e) {
						throw new MalformedStatementException(e);
					}
					IUser user = null;
					if (getMode() == PRIVILEGE_MODE.OWNER)
						user = getOwner();
					else {
						assert (getMode() == PRIVILEGE_MODE.INVOKER);
						user = triggerInvoker;
					}

					return factory.createSelectStatement(user, query);

				} else
					throw new MalformedStatementException(
							"ConcreteTuple's size does not match the cardinality of the relation schema on which the trigger is defined.");
			else
				throw new MalformedStatementException("ConcreteTuple cannot be null.");
		else
			throw new MalformedStatementException("TriggerInvoker cannot be null.");
	}

	@Override
	public IStatement getConcreteAction(IUser triggerInvoker, ITuple concreteTuple, IStatementFactory factory) throws MalformedStatementException {
		if (triggerInvoker != null)
			if (concreteTuple != null)
				if (concreteTuple.getSize() == this.getEvent().getRelationSchema().getCardinality()) {
					IUser user = null;
					if (getMode() == PRIVILEGE_MODE.OWNER)
						user = getOwner();
					else {
						assert (getMode() == PRIVILEGE_MODE.INVOKER);
						user = triggerInvoker;
					}

					return this.action.concretize(user, concreteTuple, factory);
				} else
					throw new MalformedStatementException(
							"ConcreteTuple's size does not match the cardinality of the relation schema on which the trigger is defined.");
			else
				throw new MalformedStatementException("ConcreteTuple cannot be null.");
		else
			throw new MalformedStatementException("TriggerInvoker cannot be null.");
	}

	@Override
	public String getDefinition() {
		String def = "TRIGGER " + getId() + " " + getEvent().getDefinition()  + " WHEN "
				+ condition.toString() + " DO " + getAction().getDefinition();

		return def;
	}

	@Override
	public ITrigger copy() throws CopyException {
		try {
			return new Trigger(getId(), getOwner().copy(), getMode(), getAction().copy(), getEvent().copy(), getCondition().copy());
		} catch (MalformedDbObjectException | MalformedExpressionException e) {
			throw new CopyException(e);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof Trigger) {
			if (obj == this)
				return true;
			else
				return ((Trigger) obj).getId().equals(this.getId()) && ((Trigger) obj).getAction().equals(this.getAction())
						&& ((Trigger) obj).getOwner().equals(this.getOwner()) && ((Trigger) obj).getEvent().equals(this.getEvent())
						&& ((Trigger) obj).getCondition().equals(this.getCondition()) && ((Trigger) obj).getMode().equals(this.getMode());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getId().hashCode();
		result = 31 * result + getAction().hashCode();
		result = 31 * result + getOwner().hashCode();
		result = 31 * result + getEvent().hashCode();
		result = 31 * result + getCondition().hashCode();
		result = 31 * result + getMode().hashCode();
		return result;
	}

}
