package prototype.objects.implementation;

import java.util.List;
import java.util.Vector;

import prototype.interfaces.CopyException;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.queries.domain.relational.calculus.INonBooleanQuery;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.users.IUser;

/**
 * Class representing database views'.
 * 
 * 
 *
 */
public class View extends AbstractTypedObject implements IView {

	/**
	 * Query defining the view.
	 */
	private ISelectStatement definition;

	/**
	 * Constructor.
	 * 
	 * @param id
	 *            - View's identifier.
	 * @param owner
	 *            - View's owner.
	 * @param mode
	 *            - View's privilege mode.
	 * @param fields
	 *            - View's fields.
	 * @param definition
	 *            - View's definition.
	 * @throws MalformedDbObjectException
	 */
	public View(String id, IUser owner, PRIVILEGE_MODE mode, List<String> fields, ISelectStatement definition) throws MalformedDbObjectException {
		super(id, owner, mode, fields);
		if (definition != null) {
			if (definition.getQuery() instanceof INonBooleanQuery) {
				if (((INonBooleanQuery) definition.getQuery()).getOutput().size() == fields.size())
					if (owner.equals(definition.getInvoker())) {
						try {
							this.definition = definition.copy();
						} catch (CopyException e) {
							assert (false); // unreachable code
						}
					} else
						throw new MalformedDbObjectException("The invoker of the definition must be the owner of the view.");
				else
					throw new MalformedDbObjectException("The fields in definition and in fields do not match (different size).");
			} else
				throw new MalformedDbObjectException("Definition must be a non-boolean query.");

		} else
			throw new MalformedDbObjectException("Definition cannot be null.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.IView#getQuery()
	 */
	@Override
	public ISelectStatement getSelectQuery() {
		return this.definition;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.implementation.AbstractTypedObject#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String def = "VIEW " + this.getId() + " (";

		for (int i = 0; i < this.getCardinality(); i++) {
			if (i != 0)
				def += ", ";
			def += this.getFields().get(i);
		}
		String privMode = "";
		switch (this.getMode()) {
		case INVOKER:
			privMode = "INVOKER PRIVILEGES";
			break;
		case OWNER:
			privMode = "OWNER PRIVILEGES";
			break;
		}
		def += ") OWNED BY " + getOwner().getUsername() + " " + privMode + definition.getDefinition();

		return def;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.getDefinition();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.interfaces.ICopiable#copy()
	 */
	@Override
	public View copy() throws CopyException {
		try {
			return new View(getId(), getOwner(), getMode(), new Vector<String>(getFields()), getSelectQuery().copy());
		} catch (MalformedDbObjectException e) {
			throw new CopyException(e);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof View) {
			if (obj == this)
				return true;
			else
				return ((View) obj).getId().equals(this.getId()) && ((View) obj).getFields().equals(this.getFields())
						&& ((View) obj).getOwner().equals(this.getOwner()) && ((View) obj).getSelectQuery().equals(this.getSelectQuery())
						&& ((View) obj).getMode().equals(this.getMode());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getId().hashCode();
		result = 31 * result + getFields().hashCode();
		result = 31 * result + getOwner().hashCode();
		result = 31 * result + getSelectQuery().hashCode();
		result = 31 * result + getMode().hashCode();
		return result;
	}

}
