package prototype.objects.implementation.triggers.actions;

import prototype.interfaces.CopyException;
import prototype.privileges.IPrivilege;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Trigger action representing GRANT, GRANT WITH OPTION, and REVOKE statements.
 * 
 * 
 *
 */
public class ChangePolicyTriggerAction implements ITriggerAction {

	/**
	 * The statement's target.
	 */
	private IUser target;
	/**
	 * The statement's privilege.
	 */
	private IPrivilege privilege;
	/**
	 * The statement's mode.
	 */
	private CHANGE_POLICY_MODE mode;

	/**
	 * Constructor.
	 * 
	 * @param mode
	 *            - The mode.
	 * @param privilege
	 *            - The privilege.
	 * @param target
	 *            - The target of the statement.
	 * @throws MalformedStatementException
	 */
	public ChangePolicyTriggerAction(CHANGE_POLICY_MODE mode, IPrivilege privilege, IUser target) throws MalformedStatementException {
		if (mode == null) {
			throw new MalformedStatementException("Mode cannot be null.");
		} else if (privilege == null) {
			throw new MalformedStatementException("Privilege cannot be null.");
		} else if (target == null) {
			throw new MalformedStatementException("Target cannot be null.");
		} else {
			this.mode = mode;
			try {
				this.privilege = privilege.copy();
				this.target = target.copy();
			} catch (CopyException e) {
				throw new MalformedStatementException(e);
			}
		}
	}

	/**
	 * Get the privilege used in the statement.
	 * 
	 * @return the privilege used in the statement.
	 */
	public IPrivilege getPrivilege() {
		return this.privilege;
	}

	/**
	 * Get the user that is the target of the statement.
	 * 
	 * @return the user that is the target of the statement.
	 */
	public IUser getTarget() {
		return this.target;
	}

	/**
	 * Get the mode associated with the statement.
	 * 
	 * @return GRANT if the stmt is a GRANT statement, GRANT_WITH_OPTION if the stmt is a GRANT WITH GRANT OPTION statement, and REVOKE if the stmt is
	 *         a REVOKE statement.
	 */
	public CHANGE_POLICY_MODE getMode() {
		return this.mode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof ChangePolicyTriggerAction) {
			if (obj == this)
				return true;
			else
				return ((ChangePolicyTriggerAction) obj).getTarget().equals(this.getTarget())
						&& ((ChangePolicyTriggerAction) obj).getMode().equals(this.getMode())
						&& ((ChangePolicyTriggerAction) obj).getPrivilege().equals(this.getPrivilege());
		} else
			return false;
	}
	
	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getTarget().hashCode();
		result = 31 * result + getPrivilege().hashCode();
		result = 31 * result + getMode().hashCode();
		return result;
	}

	public String getDefinition() {
		return this.getMode().toString() + ", " + this.getTarget() + ", " + this.getPrivilege();
	}

	@Override
	public IStatement concretize(IUser user, ITuple concreteTuple, IStatementFactory factory) throws MalformedStatementException {
		return factory.createChangePolicyStatement(user, mode, privilege, target);
	}

	@Override
	public ITriggerAction copy() throws CopyException {
		try {
			return new ChangePolicyTriggerAction(getMode(), getPrivilege().copy(), getTarget().copy());
		} catch (MalformedStatementException e) {
			throw new CopyException(e);
		}
	}
}
