package prototype.objects.implementation.triggers.actions;

import java.util.List;
import java.util.Vector;

import prototype.interfaces.CopyException;
import prototype.objects.ITable;
import prototype.objects.MalformedDbObjectException;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.modes.UPDATE_MODE;
import prototype.tuple.ITuple;
import prototype.tuple.Tuple;
import prototype.users.IUser;

/**
 * Trigger action representing update statements. The tuple may contain variables referring to the schema fields.
 * 
 * 
 *
 */
public class UpdateTriggerAction implements ITriggerAction {

	/**
	 * The update mode (either INSERT or DELETE)
	 */
	private UPDATE_MODE mode;

	/**
	 * The relation schema the statement is going to modify.
	 */
	private ITable schema;

	/**
	 * The tuple the statement is going to add/remove.
	 */
	private ITuple tuple;

	/**
	 * Constructor.
	 * 
	 * @param mode
	 *            - The update mode (INSERT or DELETE)
	 * @param schema
	 *            - The relation schema the statement is going to modify.
	 * @param tuple
	 *            - The tuple the statement is going to add or remove.
	 * @throws MalformedStatementException
	 */
	public UpdateTriggerAction(UPDATE_MODE mode, ITable schema, ITuple tuple) throws MalformedDbObjectException {
		if (mode == null) {
			throw new MalformedDbObjectException("Mode cannot be null.");
		} else if (schema == null) {
			throw new MalformedDbObjectException("Schema cannot be null.");
		} else if (tuple == null) {
			throw new MalformedDbObjectException("Tuple cannot be null.");
		} else if (schema.getFields().size() != tuple.getSize()) {
			throw new MalformedDbObjectException("Schema size and tuple size are different.");
		} else {

			for (IOperand o : tuple.getContent()) {
				if (o instanceof Variable)
					if (!schema.getFields().contains(((Variable) o).getIdentifier()))
						throw new MalformedDbObjectException("The tuple contain a variable that is not bound by the schema's fields.");
			}

			this.mode = mode;
			try {
				this.schema = schema.copy();
				this.tuple = tuple.copy();
			} catch (CopyException e) {
				throw new MalformedDbObjectException(e);
			}
		}
	}

	/**
	 * Get the relation schema.
	 * 
	 * @return The relation schema.
	 */
	public ITable getRelationSchema() {
		return this.schema;
	}

	/**
	 * Get the tuple.
	 * 
	 * @return The tuple.
	 */
	public ITuple getTuple() {
		return this.tuple;
	}

	/**
	 * Get the update mode.
	 * 
	 * @return The update mode.
	 */
	public UPDATE_MODE getMode() {
		return this.mode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof UpdateTriggerAction) {
			if (obj == this)
				return true;
			else
				return ((UpdateTriggerAction) obj).getRelationSchema().equals(this.getRelationSchema())
						&& ((UpdateTriggerAction) obj).getTuple().equals(this.getTuple())
						&& ((UpdateTriggerAction) obj).getMode().equals(this.getMode());
		} else
			return false;
	}
	
	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getRelationSchema().hashCode();
		result = 31 * result + getTuple().hashCode();
		result = 31 * result + getMode().hashCode();
		return result;
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.implementation.triggers.actions.ITriggerAction#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String columns = "[ ";
		for (int i = 0; i < this.getRelationSchema().getCardinality(); i++) {
			if (i != 0)
				columns += ", ";
			columns += this.getRelationSchema().getFields().get(i);
		}
		columns += "]";

		String values = "( ";
		for (int i = 0; i < this.getTuple().getSize(); i++) {
			if (i != 0)
				values += ", ";
			values += this.getTuple().getContent().get(i);
		}
		values += ")";

		return this.getMode().toString() + ", " + this.getRelationSchema().getId() + columns + ", " + values;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.objects.implementation.triggers.actions.ITriggerAction#concretize(guarnieri.marco.ethz.users.IUser,
	 * guarnieri.marco.ethz.tuple.Tuple)
	 */
	@Override
	public IStatement concretize(IUser user, ITuple concreteTuple, IStatementFactory factory) throws MalformedStatementException {
		Tuple newTuple = replaceVariables(this.getTuple(), this.getRelationSchema().getFields(), concreteTuple);
		return factory.createUpdateStatement(user, mode, schema, newTuple);
	}

	private Tuple replaceVariables(ITuple tuple, List<String> variables, ITuple concreteTuple) {
		List<IOperand> newValues = new Vector<IOperand>();

		for (int i = 0; i < tuple.getSize(); i++) {
			if (tuple.getContent().get(i) instanceof Constant)
				newValues.add(tuple.getContent().get(i));
			else if (tuple.getContent().get(i) instanceof Variable) {
				Variable v = (Variable) tuple.getContent().get(i);
		
				int j = variables.indexOf(v.getIdentifier()); // Assumption: variables does not contain duplicates.
				newValues.add(concreteTuple.getContent().get(j));
			}
		}

		return new Tuple(newValues);
	}

	@Override
	public ITriggerAction copy() throws CopyException {
		try {
			return new UpdateTriggerAction(getMode(), getRelationSchema().copy(), getTuple().copy());
		} catch (MalformedDbObjectException | CopyException e) {
			throw new CopyException(e);
		}
	}
}
