package prototype.parser;

import prototype.grammar.DrcqlBaseVisitor;
import prototype.grammar.DrcqlParser.Add_userContext;
import prototype.grammar.DrcqlParser.BooleanSelectContext;
import prototype.grammar.DrcqlParser.CommandContext;
import prototype.grammar.DrcqlParser.ConstraintContext;
import prototype.grammar.DrcqlParser.ConstraintListContext;
import prototype.grammar.DrcqlParser.Create_tableContext;
import prototype.grammar.DrcqlParser.Create_triggerContext;
import prototype.grammar.DrcqlParser.Create_viewContext;
import prototype.grammar.DrcqlParser.DeleteContext;
import prototype.grammar.DrcqlParser.EqualityContext;
import prototype.grammar.DrcqlParser.EventContext;
import prototype.grammar.DrcqlParser.ExpressionContext;
import prototype.grammar.DrcqlParser.GrantContext;
import prototype.grammar.DrcqlParser.Grant_with_optionContext;
import prototype.grammar.DrcqlParser.InsertContext;
import prototype.grammar.DrcqlParser.MixListContext;
import prototype.grammar.DrcqlParser.NonBooleanSelectContext;
import prototype.grammar.DrcqlParser.PolyadicAndExpressionContext;
import prototype.grammar.DrcqlParser.PolyadicOrExpressionContext;
import prototype.grammar.DrcqlParser.PredicateContext;
import prototype.grammar.DrcqlParser.PrivilegeContext;
import prototype.grammar.DrcqlParser.PrivilegeModeContext;
import prototype.grammar.DrcqlParser.QuantifiedExpressionContext;
import prototype.grammar.DrcqlParser.RevokeContext;
import prototype.grammar.DrcqlParser.StatementContext;
import prototype.grammar.DrcqlParser.TupleContext;
import prototype.grammar.DrcqlParser.UnaryOpExpressionContext;
import prototype.grammar.DrcqlParser.VarListContext;

import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.antlr.v4.runtime.misc.ParseCancellationException;
import org.antlr.v4.runtime.tree.ParseTree;

import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.constraints.ForeignKeyConstraint;
import prototype.objects.constraints.IConstraint;
import prototype.objects.constraints.PrimaryKeyConstraint;
import prototype.objects.constraints.UniqueConstraint;
import prototype.objects.implementation.Table;
import prototype.objects.implementation.Trigger;
import prototype.objects.implementation.View;
import prototype.objects.implementation.triggers.actions.ChangePolicyTriggerAction;
import prototype.objects.implementation.triggers.actions.ITriggerAction;
import prototype.objects.implementation.triggers.actions.UpdateTriggerAction;
import prototype.objects.implementation.triggers.events.EVENT;
import prototype.objects.implementation.triggers.events.TIME;
import prototype.objects.implementation.triggers.events.TriggerEvent;
import prototype.privileges.CreateTriggerPrivilege;
import prototype.privileges.CreateViewPrivilege;
import prototype.privileges.DeletePrivilege;
import prototype.privileges.IPrivilege;
import prototype.privileges.InsertPrivilege;
import prototype.privileges.MalformedPrivilege;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.BooleanQuery;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.Operator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IGrantStatement;
import prototype.statements.interfaces.changePolicy.IGrantWithOptionStatement;
import prototype.statements.interfaces.changePolicy.IRevokeStatement;
import prototype.statements.interfaces.create.ICreateTableStatement;
import prototype.statements.interfaces.create.ICreateTriggerStatement;
import prototype.statements.interfaces.create.ICreateViewStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.statements.modes.UPDATE_MODE;
import prototype.system.interfaces.IState;
import prototype.tuple.ITuple;
import prototype.tuple.Tuple;
import prototype.users.IUser;
import prototype.users.IUserFactory;
import prototype.users.MalformedUserException;

/**
 * Visitor for the parse tree.
 * 
 * 
 *
 */
public class AstVisitor extends DrcqlBaseVisitor<Object> {

	/**
	 * The statement factory.
	 */
	private IStatementFactory stmtFactory;
	/**
	 * The user factory
	 */
	private IUserFactory userFactory;

	/**
	 * The invoker.
	 */
	private IUser invoker;
	/**
	 * The system's state.
	 */
	private IState state;

	/**
	 * Constructor.
	 * 
	 * @param factory
	 */
	public AstVisitor(IStatementFactory stmtFactory, IUserFactory userFactory, IState state) {
		this.stmtFactory = stmtFactory;
		this.userFactory = userFactory;
		this.state = state;
	}

	@Override
	public IStatement visitCommand(CommandContext ctx) {

		IUser invoker;
		try {
			invoker = userFactory.createUser(ctx.u.getText());
		} catch (MalformedUserException e) {
			throw new ParseCancellationException(e);
		}

		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		return (IStatement) ctx.s.accept(v);

	}

	@Override
	public IStatement visitStatement(StatementContext ctx) {

		if (ctx.children.size() == 1) {
			for (ParseTree ch : ctx.children) {
				// visit the command
				AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
				v.setInvoker(this.getInvoker());
				return (IStatement) ch.accept(v);
			}
		} else
			assert (false); // it should be unreachable.
		return null;

	}

	@Override
	public IUpdateStatement visitInsert(InsertContext ctx) {

		// get the table object.
		ITable table = null;
		for (ITable t : state.getTables())
			if (t.getId().equals(ctx.table.getText()))
				table = t;
		if (table == null)
			throw new ParseCancellationException("The table " + ctx.table.getText() + " does not exist in the current state.");

		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(this.getInvoker());
		ITuple tuple = (ITuple) ctx.tpl.accept(v);

		try {
			return this.stmtFactory.createUpdateStatement(getInvoker(), UPDATE_MODE.INSERT, table, tuple);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public IUpdateStatement visitDelete(DeleteContext ctx) {
		// get the table object.
		ITable table = null;
		for (ITable t : state.getTables())
			if (t.getId().equals(ctx.table.getText()))
				table = t;
		if (table == null)
			throw new ParseCancellationException("The table " + ctx.table.getText() + " does not exist in the current state.");

		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(this.getInvoker());
		ITuple tuple = (ITuple) ctx.tpl.accept(v);

		try {
			return this.stmtFactory.createUpdateStatement(getInvoker(), UPDATE_MODE.DELETE, table, tuple);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public ITuple visitTuple(TupleContext ctx) {
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		@SuppressWarnings("unchecked")
		List<IOperand> values = (List<IOperand>) ctx.v.accept(v);

		return new Tuple(values);
	}

	@Override
	public List<String> visitVarList(VarListContext ctx) {
		if (ctx.list != null) {
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			@SuppressWarnings("unchecked")
			List<String> list = (List<String>) ctx.list.accept(v);
			list.add(ctx.id.getText());
			return list;
		} else {
			List<String> result = new Vector<String>();
			result.add(ctx.id.getText());
			return result;
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public List<IOperand> visitMixList(MixListContext ctx) {

		List<IOperand> list;
		if (ctx.list != null) {
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			list = (List<IOperand>) ctx.list.accept(v);
		} else
			list = new Vector<IOperand>();

		if (ctx.id != null) {
			try {
				list.add(new Variable(ctx.id.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.c != null) {
			try {
				list.add(new Constant(ctx.c.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}
		}

		return list;

	}

	@Override
	public IAddUser visitAdd_user(Add_userContext ctx) {
		// get the user.
		IUser user;
		try {
			user = userFactory.createUser(ctx.id.getText());
		} catch (MalformedUserException e1) {
			throw new ParseCancellationException(e1);
		}

		try {
			return this.stmtFactory.createAddUserStatement(getInvoker(), user);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public IGrantStatement visitGrant(GrantContext ctx) {
		// get the target.
		IUser target;
		try {
			target = userFactory.createUser(ctx.target.getText());
		} catch (MalformedUserException e1) {
			throw new ParseCancellationException(e1);
		}

		// get the privilege
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IPrivilege privilege = (IPrivilege) ctx.p.accept(v);

		try {
			return (IGrantStatement) this.stmtFactory.createChangePolicyStatement(getInvoker(), CHANGE_POLICY_MODE.GRANT, privilege, target);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public IGrantWithOptionStatement visitGrant_with_option(Grant_with_optionContext ctx) {
		// get the target.
		IUser target;
		try {
			target = userFactory.createUser(ctx.target.getText());
		} catch (MalformedUserException e1) {
			throw new ParseCancellationException(e1);
		}

		// get the privilege
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IPrivilege privilege = (IPrivilege) ctx.p.accept(v);

		try {
			return (IGrantWithOptionStatement) this.stmtFactory.createChangePolicyStatement(getInvoker(), CHANGE_POLICY_MODE.GRANT_WITH_OPTION,
					privilege, target);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public IRevokeStatement visitRevoke(RevokeContext ctx) {
		// get the target.
		IUser target;
		try {
			target = userFactory.createUser(ctx.target.getText());
		} catch (MalformedUserException e1) {
			throw new ParseCancellationException(e1);
		}

		// get the privilege
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IPrivilege privilege = (IPrivilege) ctx.p.accept(v);

		try {
			return (IRevokeStatement) this.stmtFactory.createChangePolicyStatement(getInvoker(), CHANGE_POLICY_MODE.REVOKE, privilege, target);
		} catch (MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public IPrivilege visitPrivilege(PrivilegeContext ctx) {
		if (ctx.cv != null) {
			return CreateViewPrivilege.getInstance();
		} else if (ctx.obj != null) {
			// get the table object.
			IType obj = null;
			for (ITable t : state.getTables())
				if (t.getId().equals(ctx.obj.getText()))
					obj = t;
			for (IView v : state.getViews())
				if (v.getId().equals(ctx.obj.getText()))
					obj = v;

			if (obj == null)
				throw new ParseCancellationException("The object " + ctx.obj.getText() + " does not exist in the current state.");

			if (ctx.s != null) {
				try {
					return new SelectPrivilege(obj);
				} catch (MalformedPrivilege e) {
					throw new ParseCancellationException(e);
				}
			} else if (ctx.i != null) {
				try {
					return new InsertPrivilege(obj);
				} catch (MalformedPrivilege e) {
					throw new ParseCancellationException(e);
				}
			} else if (ctx.d != null) {
				try {
					return new DeletePrivilege(obj);
				} catch (MalformedPrivilege e) {
					throw new ParseCancellationException(e);
				}
			} else if (ctx.ct != null) {
				try {
					return new CreateTriggerPrivilege(obj);
				} catch (MalformedPrivilege e) {
					throw new ParseCancellationException(e);
				}
			} else {
				assert (false);// unreachable
				return null;
			}

		}

		else {
			assert (false);// unreachable
			return null;
		}
	}

	/**
	 * Set the invoker.
	 * 
	 * @param invoker
	 */
	private void setInvoker(IUser invoker) {
		this.invoker = invoker;
	}

	/**
	 * Get the invoker.
	 * 
	 * @return The invoker.
	 */
	private IUser getInvoker() {
		return this.invoker;
	}

	@Override
	public ICreateViewStatement visitCreate_view(Create_viewContext ctx) {
		String id = ctx.id.getText();

		// get the mode.
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		PRIVILEGE_MODE mode = (PRIVILEGE_MODE) ctx.pmode.accept(v);

		// get the fields
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		@SuppressWarnings("unchecked")
		List<String> fields = (List<String>) ctx.fields.accept(v);

		// get the definition
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		ISelectStatement query = (ISelectStatement) ctx.query.accept(v);

		try {
			ISelectStatement definition = stmtFactory.createSelectStatement(getInvoker(), query.getQuery());
			IView view = new View(id, getInvoker(), mode, fields, definition);
			return (ICreateViewStatement) this.stmtFactory.createCreateStatement(getInvoker(), view);
		} catch (MalformedStatementException | MalformedDbObjectException e) {
			throw new ParseCancellationException(e);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public ICreateTableStatement visitCreate_table(Create_tableContext ctx) {
		String id = ctx.id.getText();

		// get the fields
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		List<String> fields = (List<String>) ctx.fields.accept(v);

		// get the constraints
		Set<IConstraint> constraints;
		if (ctx.cstrs != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			constraints = (Set<IConstraint>) ctx.cstrs.accept(v);
		} else
			constraints = new HashSet<IConstraint>();

		try {

			ITable table = new Table(id, fields, getInvoker(), constraints);

			return (ICreateTableStatement) this.stmtFactory.createCreateStatement(getInvoker(), table);
		} catch (MalformedStatementException | MalformedDbObjectException e) {
			throw new ParseCancellationException(e);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public Set<IConstraint> visitConstraintList(ConstraintListContext ctx) {

		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);

		Set<IConstraint> previous = null;
		if (ctx.list != null) {

			previous = (Set<IConstraint>) ctx.list.accept(v);

		} else {
			previous = new HashSet<IConstraint>();

		}

		// get the next one
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IConstraint cstr = (IConstraint) ctx.cstr.accept(v);

		previous.add(cstr);
		return previous;
	}

	@Override
	public IConstraint visitConstraint(ConstraintContext ctx) {
		if (ctx.pk != null) {
			// primary key

			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			@SuppressWarnings("unchecked")
			List<String> fields = (List<String>) ctx.source.accept(v);

			try {
				return new PrimaryKeyConstraint(ctx.id.getText(), getInvoker(), PRIVILEGE_MODE.OWNER, new HashSet<String>(fields));
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}

		} else if (ctx.uq != null) {
			// unique
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			@SuppressWarnings("unchecked")
			List<String> fields = (List<String>) ctx.source.accept(v);

			try {
				return new UniqueConstraint(ctx.id.getText(), getInvoker(), PRIVILEGE_MODE.OWNER, new HashSet<String>(fields));
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.fk != null) {
			// foreing key

			// get source fields
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			@SuppressWarnings("unchecked")
			List<String> sourceFields = (List<String>) ctx.source.accept(v);

			// get destination fields
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			@SuppressWarnings("unchecked")
			List<String> destinationFields = (List<String>) ctx.target.accept(v);

			// get the table object.
			ITable table = null;
			for (ITable t : state.getTables())
				if (t.getId().equals(ctx.table.getText()))
					table = t;
			if (table == null)
				throw new ParseCancellationException("The table " + ctx.table.getText() + " does not exist in the current state.");

			try {
				return new ForeignKeyConstraint(ctx.id.getText(), getInvoker(), PRIVILEGE_MODE.OWNER, sourceFields, destinationFields, table);
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}

		} else {
			assert (false); // unreachable
			return null;
		}
	}

	@Override
	public PRIVILEGE_MODE visitPrivilegeMode(PrivilegeModeContext ctx) {
		if (ctx.ow != null)
			return PRIVILEGE_MODE.OWNER;
		else if (ctx.in != null)
			return PRIVILEGE_MODE.INVOKER;
		else {
			assert (false);// unreachable code
			return null;
		}
	}

	@Override
	public ICreateTriggerStatement visitCreate_trigger(Create_triggerContext ctx) {

		String id = ctx.id.getText();

		// get the mode.
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		PRIVILEGE_MODE mode = (PRIVILEGE_MODE) ctx.pmode.accept(v);

		// get the event
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		TriggerEvent event = (TriggerEvent) ctx.e.accept(v);

		// get the action
		ITriggerAction action;
		if (ctx.actionI != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IInsertStatement stmt = (IInsertStatement) ctx.actionI.accept(v);
			try {
				action = new UpdateTriggerAction(UPDATE_MODE.INSERT, stmt.getRelationSchema(), stmt.getTuple());
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.actionD != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IDeleteStatement stmt = (IDeleteStatement) ctx.actionD.accept(v);
			try {
				action = new UpdateTriggerAction(UPDATE_MODE.DELETE, stmt.getRelationSchema(), stmt.getTuple());
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.actionG != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IGrantStatement stmt = (IGrantStatement) ctx.actionG.accept(v);
			try {
				action = new ChangePolicyTriggerAction(CHANGE_POLICY_MODE.GRANT, stmt.getPrivilege(), stmt.getTarget());
			} catch (MalformedStatementException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.actionGO != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IGrantWithOptionStatement stmt = (IGrantWithOptionStatement) ctx.actionGO.accept(v);
			try {
				action = new ChangePolicyTriggerAction(CHANGE_POLICY_MODE.GRANT_WITH_OPTION, stmt.getPrivilege(), stmt.getTarget());
			} catch (MalformedStatementException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.actionR != null) {
			v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IGrantWithOptionStatement stmt = (IGrantWithOptionStatement) ctx.actionR.accept(v);
			try {
				action = new ChangePolicyTriggerAction(CHANGE_POLICY_MODE.REVOKE, stmt.getPrivilege(), stmt.getTarget());
			} catch (MalformedStatementException e) {
				throw new ParseCancellationException(e);
			}
		} else {
			assert (false); // unreachable
			action = null;
		}

		// get condition
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression condition = (IExpression) ctx.w.accept(v);

		try {
			ITrigger trigger = new Trigger(id, getInvoker(), mode, action, event, condition);

			return (ICreateTriggerStatement) stmtFactory.createCreateStatement(getInvoker(), trigger);
		} catch (MalformedDbObjectException | MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public TriggerEvent visitEvent(EventContext ctx) {
		// get the table object.
		ITable table = null;
		for (ITable t : state.getTables())
			if (t.getId().equals(ctx.t.getText()))
				table = t;
		if (table == null)
			throw new ParseCancellationException("The table " + ctx.t.getText() + " does not exist in the current state.");
		if (ctx.i != null) {
			try {
				return new TriggerEvent(TIME.AFTER, EVENT.INSERT, table);
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.d != null) {
			try {
				return new TriggerEvent(TIME.AFTER, EVENT.DELETE, table);
			} catch (MalformedDbObjectException e) {
				throw new ParseCancellationException(e);
			}
		} else {
			assert (false); // unreachable code
			return null;
		}
	}

	@Override
	public Object visitBooleanSelect(BooleanSelectContext ctx) {
		// get the variables
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression = (IExpression) ctx.exp.accept(v);

		try {
			IQuery query = new BooleanQuery(expression);
			return stmtFactory.createSelectStatement(getInvoker(), query);
		} catch (MalformedQueryException | MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}

	}

	@Override
	public ISelectStatement visitNonBooleanSelect(NonBooleanSelectContext ctx) {

		// get the variables
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		@SuppressWarnings("unchecked")
		List<String> fields = (List<String>) ctx.var.accept(v);
		List<Variable> output = new Vector<Variable>();
		for (String s : fields)
			try {
				output.add(new Variable(s));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression = (IExpression) ctx.exp.accept(v);

		try {
			IQuery query = new NonBooleanQuery(output, expression);
			return stmtFactory.createSelectStatement(getInvoker(), query);
		} catch (MalformedQueryException | MalformedStatementException e) {
			throw new ParseCancellationException(e);
		}

	}

	@Override
	public IExpression visitExpression(ExpressionContext ctx) {

		if (ctx.e != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.e.accept(v);
		} else if (ctx.p != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.p.accept(v);
		} else if (ctx.p1 != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.p1.accept(v);
		} else if (ctx.p2 != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.p2.accept(v);
		} else if (ctx.q != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.q.accept(v);
		} else if (ctx.u != null) {
			// visit the command
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(this.getInvoker());
			return (IExpression) ctx.u.accept(v);
		} else if (ctx.f != null) {
			return FalseExpression.getInstance();
		} else if (ctx.t != null) {
			return TrueExpression.getInstance();
		} else
			assert (false); // it should be unreachable.
		return null;
	}

	@Override
	public QuantifiedExpression visitQuantifiedExpression(QuantifiedExpressionContext ctx) {

		// get the quantifier
		Operator quantifier;
		if (ctx.ex != null) {
			// existential quantifier
			try {
				quantifier = ExistentialQuantifier.getInstance();
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}
		} else if (ctx.fa != null) {
			// universal quantifier
			try {
				quantifier = UniversalQuantifier.getInstance();
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}
		} else {
			assert (false); // unreachable
			quantifier = null;
		}

		// get the variables
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		@SuppressWarnings("unchecked")
		List<String> fields = (List<String>) ctx.var.accept(v);
		LinkedHashSet<Variable> variables = new LinkedHashSet<Variable>();
		for (String s : fields)
			try {
				variables.add(new Variable(s));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		// get expression
		v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression = (IExpression) ctx.expr.accept(v);

		try {
			return new QuantifiedExpression(quantifier, variables, expression);
		} catch (MalformedExpressionException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public UnaryExpression visitUnaryOpExpression(UnaryOpExpressionContext ctx) {
		// get expression
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression = (IExpression) ctx.expr.accept(v);

		try {
			return new UnaryExpression(NotOperator.getInstance(), expression);
		} catch (MalformedExpressionException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public PolyadicExpression visitPolyadicAndExpression(PolyadicAndExpressionContext ctx) {
		List<IExpression> expressions = null;
		if (ctx.other != null) {
			// get other
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			PolyadicExpression other = (PolyadicExpression) ctx.other.accept(v);
			expressions = other.getExpressions();

		} else if (ctx.exp1 != null) {
			// get expression1
			expressions = new Vector<IExpression>();

			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IExpression expression1 = (IExpression) ctx.exp1.accept(v);
			expressions.add(expression1);
		} else {
			assert (false);// unrachable.
		}

		// get expression2
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression2 = (IExpression) ctx.exp2.accept(v);
		expressions.add(expression2);

		try {
			return new PolyadicExpression(AndOperator.getInstance(), expressions);
		} catch (MalformedExpressionException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public PolyadicExpression visitPolyadicOrExpression(PolyadicOrExpressionContext ctx) {
		List<IExpression> expressions = null;
		if (ctx.other != null) {
			// get other
			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			PolyadicExpression other = (PolyadicExpression) ctx.other.accept(v);
			expressions = other.getExpressions();

		} else if (ctx.exp1 != null) {
			// get expression1
			expressions = new Vector<IExpression>();

			AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
			v.setInvoker(invoker);
			IExpression expression1 = (IExpression) ctx.exp1.accept(v);
			expressions.add(expression1);
		} else {
			assert (false);// unrachable.
		}

		// get expression2
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		IExpression expression2 = (IExpression) ctx.exp2.accept(v);
		expressions.add(expression2);

		try {
			return new PolyadicExpression(OrOperator.getInstance(), expressions);
		} catch (MalformedExpressionException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public PredicateExpression visitPredicate(PredicateContext ctx) {

		// get the table object.
		IType type = null;
		for (ITable t : state.getTables())
			if (t.getId().equals(ctx.obj.getText()))
				type = t;
		for (IView v : state.getViews())
			if (v.getId().equals(ctx.obj.getText()))
				type = v;

		if (type == null)
			throw new ParseCancellationException("The type " + ctx.obj.getText() + " does not exist in the current state.");

		// get the variables
		AstVisitor v = new AstVisitor(stmtFactory, userFactory, state);
		v.setInvoker(invoker);
		@SuppressWarnings("unchecked")
		List<String> fields = (List<String>) ctx.var.accept(v);
		List<IOperand> variables = new Vector<IOperand>();
		for (String s : fields)
			try {
				variables.add(new Variable(s));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		try {
			return new PredicateExpression(type, variables);
		} catch (MalformedExpressionException e) {
			throw new ParseCancellationException(e);
		}
	}

	@Override
	public EqualityExpression visitEquality(EqualityContext ctx) {

		if (ctx.var1 != null && ctx.var2 != null) {

			try {
				return new EqualityExpression(new Variable(ctx.var1.getText()), new Variable(ctx.var2.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		} else if (ctx.var1 != null && ctx.const2 != null) {
			try {
				return new EqualityExpression(new Variable(ctx.var1.getText()), new Constant(ctx.const2.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		} else if (ctx.const1 != null && ctx.const2 != null) {

			try {
				return new EqualityExpression(new Constant(ctx.const1.getText()), new Constant(ctx.const2.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		} else if (ctx.const1 != null && ctx.var2 != null) {

			try {
				return new EqualityExpression(new Constant(ctx.const1.getText()), new Variable(ctx.var2.getText()));
			} catch (MalformedExpressionException e) {
				throw new ParseCancellationException(e);
			}

		} else {
			assert (false);
			return null;
		}

	}

}
