package prototype.parser;

import prototype.grammar.DrcqlLexer;
import prototype.grammar.DrcqlParser;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.TokenStream;
import org.antlr.v4.runtime.misc.ParseCancellationException;
import org.antlr.v4.runtime.tree.ParseTree;

import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUserFactory;

/**
 * Singleton that creates the statements from the commands in string form.
 * 
 * 
 *
 */
public class StatementParser {

	/**
	 * The singleton.
	 */
	private static StatementParser singleton_;

	/**
	 * Private constructor.
	 */
	private StatementParser() {
	}

	/**
	 * Get the singleton instance.
	 * 
	 * @return The instance.
	 */
	public static StatementParser getInstance() {
		if (singleton_ == null)
			singleton_ = new StatementParser();
		return singleton_;
	}

	/**
	 * Convert the command as string to the corresponding IStatement object.
	 * 
	 * @param command
	 *            - The command as string.
	 * @param statementFactory
	 *            - The statement factory.
	 * @param userFactory
	 *            - The user factory.
	 * @param state
	 *            - The state.
	 * @return The IStatement object.
	 * @throws MalformedStatementException
	 */
	public IStatement create(String command, IStatementFactory statementFactory, IUserFactory userFactory, IState state)
			throws MalformedStatementException {
		CharStream input = new ANTLRInputStream(command);
		DrcqlLexer lexer = new DrcqlLexer(input);
		TokenStream tokens = new CommonTokenStream(lexer);
		DrcqlParser parser = new DrcqlParser(tokens);

		ParseTree tree = parser.command();
		AstVisitor v = new AstVisitor(statementFactory, userFactory, state);
		try {
			return (IStatement) tree.accept(v);
		} catch (ParseCancellationException e) {
			throw new MalformedStatementException(e);
		}
	}

}
