package prototype.pdp;

import prototype.pdp.database.integrity.IntegrityAccessControlFunction;
import prototype.pdp.secure.newPdp.ConfidentialityAccessControlFunction;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.system.implementation.ExecutionException;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.interfaces.IState;
import prototype.system.interfaces.IStateFactory;
import prototype.users.IUserFactory;
import prototype.util.Logger;
import prototype.util.Logger.BooleanProperty;
import prototype.util.Logger.TimeProperty;

/**
 * A PDP offering both confidentiality and integrity.
 * 
 * 
 *
 */
public class ConfidentialityAndIntegrityPdp implements IAccessControlMechanism {

	/**
	 * The access control function for confidentiality.
	 */
	private ConfidentialityAccessControlFunction confidentiality;
	/**
	 * The access control function for integrity.
	 */
	private IntegrityAccessControlFunction integrity;

	/**
	 * The logger.
	 */
	private Logger logger;
	/**
	 * Logging Flag.
	 */
	private boolean logging;
	
	/**
	 * Constructor.
	 * 
	 * @param stmtFactory
	 *            - The statement's factory.
	 * @param userFactory
	 *            - The user factory.
	 * @param engine
	 *            - The database engine.
	 * @param stateFactory
	 *            - The state factory.
	 */
	public ConfidentialityAndIntegrityPdp(IStateFactory stateFactory, IDatabaseEngine engine, IUserFactory userFactory,
			IStatementFactory stmtFactory, Logger logger) {
		this.integrity = new IntegrityAccessControlFunction(stmtFactory, userFactory, engine, stateFactory);
		this.confidentiality = new ConfidentialityAccessControlFunction(stmtFactory, userFactory, engine, stateFactory, logger);

		if (logger != null)
			this.logging = true;
		else
			this.logging = false;
		this.logger = logger;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.pdp.IAccessControlMechanism#authorizes(guarnieri.marco.ethz.statements.interfaces.IStatement,
	 * guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public boolean authorizes(IStatement statement, IState systemState) throws ExecutionException {
		if (logging)
			logger.startMeasure(TimeProperty.INTEGRITY_TOTAL_TIME);
		boolean integrityRes = integrity.authorizes(statement, systemState);
		if (logging)
			logger.endMeasure(TimeProperty.INTEGRITY_TOTAL_TIME);
		
		if (logging)
			logger.startMeasure(TimeProperty.CONFIDENTIALITY_TOTAL_TIME);
		boolean confidentialityRes = confidentiality.authorizes(statement, systemState);
		if (logging)
			logger.endMeasure(TimeProperty.CONFIDENTIALITY_TOTAL_TIME);
		
		if(logging)
		{
			logger.setBooleanValue(BooleanProperty.INTEGRITY_RESULT, integrityRes);
			logger.setBooleanValue(BooleanProperty.CONFIDENTIALITY_RESULT, confidentialityRes);
		}
		
		return integrityRes && confidentialityRes;
	}

}
