package prototype.pdp.database.integrity;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.View;
import prototype.pdp.secure.NormalizeVisitor;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.select.ISelectStatement;

/**
 * Visitor implementing the apprDet algorithm (see paper section 6.1).
 * 
 * 
 *
 */
public class DeterminesVisitor implements IVisitor {

	/**
	 * Algorithm's result.
	 */
	private boolean result = false;
	/**
	 * The set of types.
	 */
	private Set<IType> types;

	/**
	 * The set of extended types.
	 */
	private Set<IExpression> extendedTypes;
	/**
	 * The statement factory.
	 */
	private IStatementFactory factory;

	public DeterminesVisitor(Set<? extends IType> types, IStatementFactory factory) {
		this.types = new HashSet<IType>();
		this.extendedTypes = new HashSet<IExpression>();
		this.factory = factory;
		// we first normalize all the views in types
		for (IType t : types) {
			if (t instanceof ITable)
				this.types.add(t);
			else if (t instanceof IView) {
				IExpression normalized = ((IView) t).getSelectQuery().getQuery().getExpression();
				IExpression old = null;

				IncrementalIdentifiers vInc = new IncrementalIdentifiers();
				normalized.accept(vInc);
				normalized = vInc.getRenamedExpression();
				int min = vInc.getMaximumValue();

				List<Variable> vars = new Vector<Variable>();
				for (Variable v : ((NonBooleanQuery) ((IView) t).getSelectQuery().getQuery()).getOutput())
					try {
						vars.add(new Variable(IncrementalIdentifiers.getPrefix() + vInc.getAssignment().get(v)));
					} catch (MalformedExpressionException e1) {
						assert (false);
					}

				do {
					NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
					normalized.accept(vNorm);

					old = normalized;
					normalized = vNorm.getNormalizedExpression();
					min = vNorm.getMaximumValue();
				} while (!old.equals(normalized));

				try {
					ISelectStatement stmt = factory.createSelectStatement(((IView) t).getSelectQuery().getInvoker(), new NonBooleanQuery(vars, normalized));

					IView normalizedView = new View(t.getId(), ((IView) t).getOwner(), ((IView) t).getMode(), t.getFields(), stmt);
					this.types.add(normalizedView);
				} catch (MalformedStatementException | MalformedQueryException | MalformedDbObjectException e) {
					assert (false);
				}
			} else
				assert (false); // unreachable
		}

		// compute the extend set
		for (IType t : this.types) {
			if (t instanceof IView) {
				IExpression normalized = ((IView) t).getSelectQuery().getQuery().getExpression();
				IExpression old = null;

				IncrementalIdentifiers vInc = new IncrementalIdentifiers();
				normalized.accept(vInc);
				normalized = vInc.getRenamedExpression();
				extendedTypes.add(normalized);
				int min = vInc.getMaximumValue();

				do {
					NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.OWNER);
					normalized.accept(vNorm);

					old = normalized;
					normalized = vNorm.getNormalizedExpression();

					extendedTypes.add(normalized);
					min = vNorm.getMaximumValue();
				} while (!old.equals(normalized));

			}
		}

	}

	/**
	 * Private constructor. Does not normalize the set of types nor computes the extended types.
	 * 
	 * @param types
	 * @param extendedTypes
	 * @param factory
	 */
	private DeterminesVisitor(Set<IType> types, Set<IExpression> extendedTypes, IStatementFactory factory) {
		this.types = types;
		this.extendedTypes = extendedTypes;
		this.factory = factory;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (types.contains(expr.getType()) || containedInExtend(expr))
			result = true;
	}

	@Override
	public void visit(EqualityExpression expr) {
		result = true;
	}
	
	@Override
	public void visit(FalseExpression expr) {
		result = true;
	}
	
	@Override
	public void visit(TrueExpression expr) {
		result = true;
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		DeterminesVisitor v = new DeterminesVisitor(types, extendedTypes, this.factory);
		expr.getExpression().accept(v);
		if (containedInExtend(expr) || v.getResult())
			result = true;
	}

	@Override
	public void visit(UnaryExpression expr) {
		DeterminesVisitor v = new DeterminesVisitor(types, extendedTypes, this.factory);
		expr.getExpression().accept(v);
		if (containedInExtend(expr) || v.getResult())
			result = true;
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {
		boolean allAuthorized = true;
		for (IExpression e : expr.getExpressions()) {
			DeterminesVisitor v = new DeterminesVisitor(types, extendedTypes, this.factory);
			e.accept(v);
			if (!v.getResult())
				allAuthorized = false;
		}

		if (allAuthorized || containedInExtend(expr))
			result = true;
	}

	/**
	 * Get the algorithm's result.
	 * 
	 * @return the result.
	 */
	public boolean getResult() {
		return result;
	}

	/**
	 * Check whether in the result of extend(types) there is a view whose definition is of the form { x | expression }.
	 * 
	 * @param expression
	 *            - The expression to analyze.
	 * @return True if there a view in extend(types) whose definition is "equivalent" to the expression.
	 */
	private boolean containedInExtend(IExpression expression) {
		for (IExpression eView : extendedTypes) {

			IncrementalIdentifiers vInc = new IncrementalIdentifiers();
			eView.accept(vInc);

			eView = vInc.getRenamedExpression();

			vInc = new IncrementalIdentifiers(); 
			expression.accept(vInc);

			if (eView.equals(vInc.getRenamedExpression()))
				return true;

		}
		return false;
	}
}
