package prototype.pdp.database.integrity;

import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;

import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.PRIVILEGE_MODE;
import prototype.pdp.IAccessControlMechanism;
import prototype.pdp.secure.NormalizeVisitor;
import prototype.privileges.CreateTriggerPrivilege;
import prototype.privileges.CreateViewPrivilege;
import prototype.privileges.DeletePrivilege;
import prototype.privileges.InsertPrivilege;
import prototype.privileges.MalformedPrivilege;
import prototype.privileges.SelectPrivilege;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IGrantWithOptionStatement;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.changePolicy.IRevokeStatement;
import prototype.statements.interfaces.create.ICreateTableStatement;
import prototype.statements.interfaces.create.ICreateTriggerStatement;
import prototype.statements.interfaces.create.ICreateViewStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.system.MalformedStateException;
import prototype.system.implementation.ExecutionException;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.interfaces.IState;
import prototype.system.interfaces.IStateFactory;
import prototype.tuple.ITuple;
import prototype.users.IUser;
import prototype.users.IUserFactory;
import prototype.users.MalformedUserException;

/**
 * The integrity access control function. (see paper).
 * 
 * 
 *
 */
public class IntegrityAccessControlFunction implements IAccessControlMechanism {

	/**
	 * The statement's factory.
	 */
	private IStatementFactory stmtFactory;

	/**
	 * The user's factory.
	 */
	private IUserFactory userFactory;

	/**
	 * The database engine.
	 */
	private IDatabaseEngine engine;

	/**
	 * The state factory
	 */
	private IStateFactory stateFactory;
	


	/**
	 * Constructor.
	 * 
	 * @param stmtFactory
	 *            - The statement's factory.
	 * @param userFactory
	 *            - The user factory.
	 * @param engine
	 *            - The database engine.
	 */
	public IntegrityAccessControlFunction(IStatementFactory stmtFactory, IUserFactory userFactory, IDatabaseEngine engine, IStateFactory stateFactory) {
		this.stmtFactory = stmtFactory;
		this.userFactory = userFactory;
		this.engine = engine;
		this.stateFactory = stateFactory;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.pdp.IAccessControlMechanism#authorizes(guarnieri.marco.ethz.statements.interfaces.IStatement,
	 * guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public boolean authorizes(IStatement statement, IState systemState) throws ExecutionException {

		try {
			if (!systemState.isTransaction())
				return apprxAuthorized(systemState, statement);
			else if (systemState.getCurrentTrigger().getConcreteCondition(systemState.getInvoker(), systemState.getTuple(), stmtFactory)
					.equals(statement))
				// we are executing a trigger (if systemState.transition == true AND systemState.getCurrentTrigger() == null, then we are executing
				// the action).
				return true; // the trigger's condition is always satisfied for integrity

			else if (systemState.getCurrentTrigger().getMode() == PRIVILEGE_MODE.OWNER
					&& systemState.getCurrentTrigger()
							.getConcreteAction(systemState.getCurrentTrigger().getOwner(), systemState.getTuple(), stmtFactory).equals(statement))
				// rule EXECUTE TRIGGER - 1
				return apprxAuthorized(
						systemState,
						systemState.getCurrentTrigger().getConcreteAction(systemState.getCurrentTrigger().getOwner(), systemState.getTuple(),
								stmtFactory));
			else if (systemState.getCurrentTrigger().getMode() == PRIVILEGE_MODE.INVOKER
					&& systemState.getCurrentTrigger().getConcreteAction(systemState.getInvoker(), systemState.getTuple(), stmtFactory)
							.equals(statement))
				// rule EXECUTE TRIGGER - 2
				return apprxAuthorized(
						systemState,
						systemState.getCurrentTrigger().getConcreteAction(systemState.getCurrentTrigger().getOwner(), systemState.getTuple(),
								stmtFactory))
						&& apprxAuthorized(systemState,
								systemState.getCurrentTrigger().getConcreteAction(systemState.getInvoker(), systemState.getTuple(), stmtFactory));
			else
				return false;
		} catch (MalformedStatementException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Takes a systemState, applies the statement, and returns the new systemState. It implements the function apply defined in Sect. 6 of the paper.
	 * Note that we ignore changes due to INSERT and DELETE statements, since they do not affect the computation of authorized.
	 * 
	 * @param systemState
	 *            - The initial system state.
	 * @param statement
	 *            - The statement that must be executed.
	 * @return The new state obtained after the execution of the statement.
	 * @throws ExecutionException
	 */
	private IState apply(IState systemState, IStatement statement) throws ExecutionException {

		Set<IUser> users = systemState.getUsers();
		Set<ITable> tables = systemState.getTables();
		Set<IPositiveGrantStatement> policy = systemState.getPolicy();
		Set<IView> views = systemState.getViews();
		Set<ITrigger> triggers = systemState.getTriggers();
		IUser invoker = systemState.getInvoker();
		IState rollbackState = systemState.getRollbackState();
		SortedSet<ITrigger> scheduledTriggers = systemState.getScheduledTriggers();
		ITuple tuple = systemState.getTuple();

		if (statement instanceof ISelectStatement || statement instanceof IUpdateStatement)
			return systemState;
		else if (statement instanceof IPositiveGrantStatement) {
			policy.add((IPositiveGrantStatement) statement);
		} else if (statement instanceof IRevokeStatement) {
			policy = engine.revoke(policy, (IRevokeStatement) statement);
		} else if (statement instanceof ICreateTriggerStatement) {
			triggers.add(((ICreateTriggerStatement) statement).getObject());
		} else if (statement instanceof ICreateTableStatement) {
			tables.add(((ICreateTableStatement) statement).getObject());
		} else if (statement instanceof ICreateViewStatement) {
			views.add(((ICreateViewStatement) statement).getObject());
		} else if (statement instanceof IAddUser) {
			users.add(((IAddUser) statement).getUser());
		} else
			throw new ExecutionException("This statement is not supported.");

		try {
			return stateFactory.createNewState(users, tables, policy, views, triggers, invoker, rollbackState, scheduledTriggers, tuple, null);
		} catch (MalformedStateException e) {
			throw new ExecutionException(e);
		}

	}

	/**
	 * Checks whether the statement is authorized in the state according to the $\auth^{\mathit{approx}}$ relation (see paper sect. 7).
	 * 
	 * @param systemState
	 *            - The state.
	 * @param statement
	 *            - The statement.
	 * @return True if statement is authorized according to state, false otherwise.
	 * @throws ExecutionException
	 */
	private boolean apprxAuthorized(IState systemState, IStatement statement) throws ExecutionException {
		try {
			Set<IPositiveGrantStatement> policy = systemState.getPolicy();
			if (statement instanceof IUpdateStatement)
				// rules INSERT-DELETE and INSERT-DELETE admin
				if (statement.getInvoker().equals(userFactory.getAdmin()))
					// rule INSERT-DELETE admin
					return true;
				else {
					// rule INSERT-DELETE

					for (IPositiveGrantStatement p : policy)
						if (statement instanceof IInsertStatement && p.getTarget().equals(statement.getInvoker())
								&& p.getPrivilege() instanceof InsertPrivilege
								&& ((InsertPrivilege) p.getPrivilege()).getObj().equals(((IInsertStatement) statement).getRelationSchema()))
							if (apprxAuthorized(systemState, p))
								// the statement is authorized and the chain of "Grants" is well-formed
								return true;
							else if (statement instanceof IDeleteStatement && p.getTarget().equals(statement.getInvoker())
									&& p.getPrivilege() instanceof DeletePrivilege
									&& ((DeletePrivilege) p.getPrivilege()).getObj().equals(((IInsertStatement) statement).getRelationSchema()))
								if (apprxAuthorized(systemState, p))
									// the statement is authorized and the chain of "Grants" is well-formed
									return true;

					// there is no well-formed chain of grants authorizing the statement
					return false;
				}
			else if (statement instanceof ICreateTriggerStatement)
				// rules CREATE TRIGGER and CREATE TRIGGER admin
				if (statement.getInvoker().equals(userFactory.getAdmin()))
					// rule CREATE TRIGGER admin
					return true;
				else {
					// rule CREATE TRIGGER
					for (IPositiveGrantStatement p : policy)
						if (p.getTarget().equals(statement.getInvoker())
								&& p.getPrivilege() instanceof CreateTriggerPrivilege
								&& ((CreateTriggerPrivilege) p.getPrivilege()).getObj().equals(
										((ICreateTriggerStatement) statement).getObject().getEvent().getRelationSchema()))
							if (apprxAuthorized(systemState, p))
								// the statement is authorized and the chain of "Grants" is well-formed
								return true;

					// there is no well-formed chain of grants authorizing the statement
					return false;
				}
			else if (statement instanceof ICreateTableStatement)
				// this is not in the paper :-) only for administrator
				if (statement.getInvoker().equals(userFactory.getAdmin()))
					return true;
				else
					return false;
			else if (statement instanceof ICreateViewStatement)
				// rules CREATE VIEW and CREATE VIEW admin
				if (statement.getInvoker().equals(userFactory.getAdmin()))
					// rule CREATE VIEW admin
					return true;
				else {
					// rule CREATE VIEW

					for (IPositiveGrantStatement p : policy)
						if (p.getTarget().equals(statement.getInvoker()) && p.getPrivilege() instanceof CreateViewPrivilege)
							if (apprxAuthorized(systemState, p))
								// the statement is authorized and the chain of "Grants" is well-formed
								return true;

					// there is no well-formed chain of grants authorizing the statement
					return false;
				}
			else if (statement instanceof IAddUser)
				// rule ADD USER
				return statement.getInvoker().equals(userFactory.getAdmin());
			else if (statement instanceof ISelectStatement)
				// rule SELECT
				return true;
			else if (statement instanceof IRevokeStatement) {
				// rule REVOKE
				IState apply = apply(systemState, statement);
				for (IPositiveGrantStatement p : apply.getPolicy())
					if (!apprxAuthorized(apply, p))
						return false;

			} else if (statement instanceof IPositiveGrantStatement) {
				
				// we first check whether we can apply the rule GRANT-1
				for (IPositiveGrantStatement g : systemState.getPolicy())
					if (g instanceof IGrantWithOptionStatement && g.getPrivilege().equals(((IPositiveGrantStatement) statement).getPrivilege())
							&& g.getTarget().equals(statement.getInvoker()) && apprxAuthorized(systemState, g))
						return true;

				// rule GRANT - 5
				if (((IPositiveGrantStatement) statement).getPrivilege() instanceof SelectPrivilege
						&& ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj() instanceof IView
						&& ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getMode() == PRIVILEGE_MODE.INVOKER
						&& statement.getInvoker().equals(
								((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner()))
					return true;

				// rule GRANT - 2
				if (statement.getInvoker().equals(userFactory.getAdmin())
						&& (((IPositiveGrantStatement) statement).getPrivilege() instanceof InsertPrivilege
								|| ((IPositiveGrantStatement) statement).getPrivilege() instanceof DeletePrivilege
								|| ((IPositiveGrantStatement) statement).getPrivilege() instanceof CreateTriggerPrivilege
								|| ((IPositiveGrantStatement) statement).getPrivilege() instanceof CreateViewPrivilege || (((IPositiveGrantStatement) statement)
								.getPrivilege() instanceof SelectPrivilege && ((((SelectPrivilege) ((IPositiveGrantStatement) statement)
								.getPrivilege()).getObj() instanceof ITable || (((SelectPrivilege) ((IPositiveGrantStatement) statement)
								.getPrivilege()).getObj() instanceof IView && ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement)
								.getPrivilege()).getObj()).getMode() == PRIVILEGE_MODE.OWNER))))))
					return true;

				// rule GRANT - 3
				if (((IPositiveGrantStatement) statement).getPrivilege() instanceof SelectPrivilege
						&& ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj() instanceof IView
						&& ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getMode() == PRIVILEGE_MODE.OWNER
						&& statement.getInvoker().equals(
								(((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner()))) {

					Set<IType> types = authorized(systemState.getTables(), CHANGE_POLICY_MODE.GRANT_WITH_OPTION,
							((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner(), systemState);
					types.addAll(authorized(systemState.getViews(), CHANGE_POLICY_MODE.GRANT_WITH_OPTION,
							((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner(), systemState));
					return determines(types, ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getSelectQuery());
				}

				// rule GRANT - 4
				if (((IPositiveGrantStatement) statement).getPrivilege() instanceof SelectPrivilege
						&& ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj() instanceof IView
						&& ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getMode() == PRIVILEGE_MODE.OWNER
						&& statement.getInvoker().equals(userFactory.getAdmin())
						&& !((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner().equals(
								userFactory.getAdmin())) {
					Set<IType> types = authorized(systemState.getTables(), CHANGE_POLICY_MODE.GRANT,
							((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner(), systemState);
					types.addAll(authorized(systemState.getViews(), CHANGE_POLICY_MODE.GRANT,
							((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getOwner(), systemState));
					return determines(types, ((IView) ((SelectPrivilege) ((IPositiveGrantStatement) statement).getPrivilege()).getObj()).getSelectQuery());
				}

			}

			return false; // no rule can be applied
		} catch (MalformedUserException | MalformedPrivilege | MalformedStatementException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Execute the determines procedure (see paper).
	 * 
	 * @param types
	 *            - The set of authorized types.
	 * @param query
	 *            - The query to check.
	 * @return True if the query is fully determined by the authorized types, false otherwise.
	 */
	private boolean determines(Set<IType> types, ISelectStatement query) {
		IExpression expression = query.getQuery().getExpression();

		// unique ids
		IncrementalIdentifiers vInc = new IncrementalIdentifiers();
		expression.accept(vInc);

		expression = vInc.getRenamedExpression();

		// normalize invoker views
		int min = vInc.getMaximumValue();

		IExpression old = null;

		IExpression normalized = vInc.getRenamedExpression();
		do {
			NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
			normalized.accept(vNorm);
			IExpression ex = vNorm.getNormalizedExpression();
			vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), vNorm.getMaximumValue(), PRIVILEGE_MODE.OWNER);
			ex.accept(vNorm);

			old = normalized;
			normalized = vNorm.getNormalizedExpression();
			min = vNorm.getMaximumValue();
		} while (!old.equals(normalized));

		// check determines
		DeterminesVisitor vDet = new DeterminesVisitor(types, this.stmtFactory);
		normalized.accept(vDet);

		return vDet.getResult();
	}

	/**
	 * Extract the set of types that the user is authorized to read with the proper GRANT mode (see aV and aT functions in the paper).
	 * 
	 * @param types
	 *            - The set of types.
	 * @param grantMode
	 *            - The grant mode.
	 * @param user
	 *            - The user.
	 * @return The set of authorized types.
	 * @throws MalformedUserException
	 * @throws MalformedPrivilege
	 * @throws ExecutionException
	 * @throws MalformedStatementException
	 */
	private Set<IType> authorized(Set<? extends IType> types, CHANGE_POLICY_MODE grantMode, IUser user, IState state) throws MalformedUserException,
			MalformedPrivilege, ExecutionException, MalformedStatementException {
		Set<IType> authorized = new HashSet<IType>();
		if (user.equals(userFactory.getAdmin())) {

			for (IType t : types) {
				if (t instanceof ITable || (t instanceof IView && ((IView) t).getMode() == PRIVILEGE_MODE.OWNER)) {

					SelectPrivilege priv = new SelectPrivilege(t);
					Set<CHANGE_POLICY_MODE> modalities = new HashSet<CHANGE_POLICY_MODE>();
					modalities.add(grantMode);
					modalities.add(CHANGE_POLICY_MODE.GRANT_WITH_OPTION);

					for (CHANGE_POLICY_MODE mode : modalities)
						for (IUser u : state.getUsers()) {
							IPositiveGrantStatement g = (IPositiveGrantStatement) stmtFactory.createChangePolicyStatement(u, mode, priv, user);
							if (apprxAuthorized(state, g))
								authorized.add(t);
						}

				}
			}

		} else {
			for (IType t : types) {
				if (t instanceof ITable || (t instanceof IView && ((IView) t).getMode() == PRIVILEGE_MODE.OWNER)) {

					SelectPrivilege priv = new SelectPrivilege(t);
					INNER: for (IPositiveGrantStatement g : state.getPolicy())
						if (g.getPrivilege().equals(priv) && g.getTarget().equals(user)
								&& (g.getMode() == CHANGE_POLICY_MODE.GRANT_WITH_OPTION || g.getMode() == grantMode) && apprxAuthorized(state, g)) {
							authorized.add(t);
							break INNER;
						}

				}
			}
		}

		return authorized;

	}
}
