package prototype.pdp.secure;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Vector;

import prototype.objects.MalformedDbObjectException;
import prototype.objects.implementation.AnonymousView;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.FreeVariableVisitor;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that replaces SRNF-expressions of the form EXISTS xj1,...,jm. R(x1, ..., xN), where xQ1,...xQl are the variables in x1, ..., xN and not in
 * xj1,...,jm, with the corresponding anonymous views R_{xQ1,...xQl}(xQ1,...xQl). It also pushes existential quantifiers as much as possible.
 * 
 * The visitor is part of the pre-processing steps for the "secure" procedure.
 * 
 * 
 *
 */
public class AddAnonymousViewsVisitor implements IVisitor {

	/**
	 * The transformed expression.
	 */
	private IExpression transformed;

	@Override
	public void visit(PredicateExpression expr) {
		this.transformed = expr;
	}

	@Override
	public void visit(EqualityExpression expr) {
		this.transformed = expr;
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (expr.getQuantifier() instanceof UniversalQuantifier) {
			// visit the sub-expression
			AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
			expr.getExpression().accept(v);

			try {
				this.transformed = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.transformed);
			} catch (MalformedExpressionException e) {
				assert (false);
			}

		} else if (expr.getQuantifier() instanceof ExistentialQuantifier) {
			if (expr.getExpression() instanceof PredicateExpression) {
				// replace the quantified expression with the corresponding anonymous view

				// compute the set of projected fields
				List<String> fields = new Vector<String>();
				List<Variable> freeVars = new Vector<Variable>();
				List<IOperand> operands = new Vector<IOperand>();
				for (int i = 0; i < ((PredicateExpression) expr.getExpression()).getOperands().size(); i++)
					if (!expr.getVariables().contains(((PredicateExpression) expr.getExpression()).getOperands().get(i))) {

						fields.add(((PredicateExpression) expr.getExpression()).getType().getFields().get(i));
						freeVars.add((Variable) ((PredicateExpression) expr.getExpression()).getOperands().get(i));
						operands.add((Variable) ((PredicateExpression) expr.getExpression()).getOperands().get(i));
					} 
				try {
					AnonymousView v = new AnonymousView(((PredicateExpression) expr.getExpression()).getType().getId() + "_"
							+ (fields.toString().replace("[", "(").replace("]", ")")), fields, new NonBooleanQuery(freeVars, expr));

					this.transformed = new PredicateExpression(v, operands);
				} catch (MalformedDbObjectException | MalformedQueryException | MalformedExpressionException e) {
					assert (false);
				}
			} else if (expr.getExpression() instanceof PolyadicExpression) {
				// check whether we can push the quantifier

				List<IExpression> exprs = new Vector<IExpression>();
				for (IExpression e : ((PolyadicExpression) expr.getExpression()).getExpressions()) {
					AddAnonymousViewsVisitor vAdd = new AddAnonymousViewsVisitor();
					e.accept(vAdd);
					exprs.add(vAdd.getTransformed());
				}

				IExpression result = null;
				for (Variable var : expr.getVariables()) {
					List<IExpression> affectedBy = new Vector<IExpression>();
					for (IExpression e : exprs) {
						FreeVariableVisitor vFree = new FreeVariableVisitor();
						e.accept(vFree);

						if (vFree.getFreeVariables().contains(var))
							affectedBy.add(e);
					}

					if (!affectedBy.isEmpty() && !affectedBy.equals(exprs)) {
						// create the new expression
						LinkedHashSet<Variable> vars = new LinkedHashSet<Variable>();
						vars.add(var);

						IExpression subExpr;
						if (affectedBy.size() > 1)
							try {
								subExpr = new PolyadicExpression(((PolyadicExpression) expr.getExpression()).getOperator(), affectedBy);
							} catch (MalformedExpressionException e4) {
								assert (false);
								subExpr = null;
							}
						else
							subExpr = affectedBy.get(0);
						QuantifiedExpression q;
						try {
							q = new QuantifiedExpression(ExistentialQuantifier.getInstance(), vars, subExpr);
						} catch (MalformedExpressionException e3) {
							assert (false);
							q = null;
						}

						List<IExpression> exs = new Vector<IExpression>();
						exs.add(q);
						for (IExpression e : exprs)
							if (!affectedBy.contains(e))
								exs.add(e);

						IExpression subExpr1;
						try {
							subExpr1 = new PolyadicExpression(((PolyadicExpression) expr.getExpression()).getOperator(), exs);
						} catch (MalformedExpressionException e2) {
							assert (false);
							subExpr1 = null;
						}

						LinkedHashSet<Variable> vars1 = new LinkedHashSet<Variable>(expr.getVariables());
						vars1.remove(var);

						try {
							result = new QuantifiedExpression(ExistentialQuantifier.getInstance(), vars1, subExpr1);
						} catch (MalformedExpressionException e1) {
							assert (false);
						}
						break;
					}
				}

				if (result == null) {
					// visit the sub-expression
					AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
					expr.getExpression().accept(v);

					try {
						this.transformed = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.transformed);
					} catch (MalformedExpressionException e) {
						assert (false);
					}
				} else
					this.transformed = result;

			} else if (expr.getExpression() instanceof QuantifiedExpression
					&& ((QuantifiedExpression) expr.getExpression()).getQuantifier() instanceof ExistentialQuantifier) {
				// flatten existential quantifiers

				try {
					QuantifiedExpression exprCopy = expr.copy();
					QuantifiedExpression nestedCopy = (QuantifiedExpression) exprCopy.getExpression();
					LinkedHashSet<Variable> newVars = expr.copy().getVariables();

					// merge the variables
					newVars.addAll(nestedCopy.getVariables());

					// modify the sub-sub expression
					AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
					nestedCopy.getExpression().accept(v);

					this.transformed = new QuantifiedExpression(ExistentialQuantifier.getInstance(), newVars, v.getTransformed());
				} catch (MalformedExpressionException e) {
					assert (false); // UNREACHABLE
				}
			} else {
				// visit the sub-expression
				AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
				expr.getExpression().accept(v);

				try {
					this.transformed = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.transformed);
				} catch (MalformedExpressionException e) {
					assert (false);
				}
			}
		}

	}

	@Override
	public void visit(UnaryExpression expr) {

		AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
		expr.getExpression().accept(v);

		try {
			this.transformed = new UnaryExpression(expr.getOperator(), v.transformed);
		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {
		List<IExpression> exprs = new Vector<IExpression>();

		for (IExpression ex : expr.getExpressions()) {
			AddAnonymousViewsVisitor v = new AddAnonymousViewsVisitor();
			ex.accept(v);

			exprs.add(v.transformed);
		}

		try {
			this.transformed = new PolyadicExpression(expr.getOperator(), exprs);
		} catch (MalformedExpressionException e) {
			assert (false);
		}
	}

	@Override
	public void visit(FalseExpression expr) {
		this.transformed = expr;
	}

	@Override
	public void visit(TrueExpression expr) {
		this.transformed = expr;
	}

	/**
	 * Get the transformed expression.
	 * 
	 * @return The transformed expression.
	 */
	public IExpression getTransformed() {
		return transformed;
	}

}
