package prototype.pdp.secure;

import prototype.objects.IType;
import prototype.privileges.SelectPrivilege;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;

public class AllAuthorizedVisitor implements IVisitor {

	/**
	 * True if everything is ok.
	 */
	private boolean ok;
	/**
	 * The result.
	 */
	private boolean result;
	/**
	 * The state.
	 */
	private IState state;
	/**
	 * The user.
	 */
	private IUser user;

	public AllAuthorizedVisitor(IUser user, IState state) {
		if (user != null && state.getUsers().contains(user)) {
			this.ok = true;
			this.user = user;
			this.state = state;
			this.result = false;
		} else
			this.ok = false;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok)
			this.result = authorized(state, user, expr.getType());
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok)
			this.result = true;
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			AllAuthorizedVisitor v = new AllAuthorizedVisitor(user, state);
			expr.getExpression().accept(v);

			this.result = v.allAuthorized();
		}
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			AllAuthorizedVisitor v = new AllAuthorizedVisitor(user, state);
			expr.getExpression().accept(v);

			this.result = v.allAuthorized();
		}
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			boolean allAuthorized = true;
			for (IExpression expression : expr.getExpressions()) {
				AllAuthorizedVisitor v = new AllAuthorizedVisitor(user, state);
				expression.accept(v);

				allAuthorized = allAuthorized && v.allAuthorized();
			}
			this.result = allAuthorized;
		}
	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			this.result = true;
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			this.result = true;
	}

	/**
	 * Check whether the give user is authorized to read the content of the given type in the given state.
	 * 
	 * @param state
	 *            - The state.
	 * @param user
	 *            - The user.
	 * @param type
	 *            - The type.
	 * @return True if user is authorized to read the content of type in state. False otherwise.
	 */
	private boolean authorized(IState state, IUser user, IType type) {
		for (IPositiveGrantStatement g : state.getPolicy())
			if (g.getTarget().equals(user) && (g.getPrivilege() instanceof SelectPrivilege)
					&& ((SelectPrivilege) g.getPrivilege()).getObj().equals(type))
				return true;
		return false;
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	/**
	 * True if all predicates used in the visited are authorized.
	 * 
	 * @return
	 */
	public boolean allAuthorized() {
		return result;
	}

}
