package prototype.pdp.secure;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.View;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.OrOperator;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;
import prototype.util.Pair;

/**
 * The class representing the authorization graph (See appendix C of our paper).
 * 
 * 
 *
 */
public class AuthorizationGraph {

	/**
	 * The graph's nodes representing ITypes.
	 */
	private Set<IType> typeNodes;

	/**
	 * The directed edges (from/to) between ITypes.
	 */
	private Set<Pair<IType, IType>> typeEdges;

	/**
	 * The directed edges (from/to) from users to types.
	 */
	private Set<Pair<IUser, IType>> userEdges;

	/**
	 * Labels for the edges between types.
	 */
	private Map<Pair<IType, IType>, IExpression> typeEdgeLabels;

	/**
	 * Labels for the edges between users and types.
	 */
	private Map<Pair<IUser, IType>, IExpression> userEdgeLabels;

	/**
	 * Constructor.
	 * 
	 * @param state
	 *            - The state.
	 */
	public AuthorizationGraph(IState state, IStatementFactory factory) {
		if (state != null) {
			int min = 0;
			this.typeNodes = new HashSet<IType>(state.getTables());
			this.userEdges = new HashSet<Pair<IUser, IType>>();
			this.typeEdges = new HashSet<Pair<IType, IType>>();
			this.typeEdgeLabels = new HashMap<Pair<IType, IType>, IExpression>();
			this.userEdgeLabels = new HashMap<Pair<IUser, IType>, IExpression>();
			// compute the set of normalized views
			for (IView view : state.getViews()) {
				if (view.getMode() == PRIVILEGE_MODE.OWNER) {

					IExpression old = null;
					IExpression normalized = view.getSelectQuery().getQuery().getExpression();
					// translate the expression to one with incremental identifiers
					IncrementalIdentifiers vInc = new IncrementalIdentifiers(min);
					normalized.accept(vInc);

					min = vInc.getMaximumValue();

					normalized = vInc.getRenamedExpression();

					List<Variable> vars = new Vector<Variable>();
					for (Variable v : ((NonBooleanQuery) view.getSelectQuery().getQuery()).getOutput())
						try {
							vars.add(new Variable(IncrementalIdentifiers.getPrefix() + vInc.getAssignment().get(v)));
						} catch (MalformedExpressionException e1) {
							assert (false);
						}

					do {
						NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
						normalized.accept(vNorm);

						old = normalized;
						normalized = vNorm.getNormalizedExpression();
						min = vNorm.getMaximumValue();
					} while (!old.equals(normalized));

					ISelectStatement stmt;
					try {
						stmt = factory.createSelectStatement(view.getSelectQuery().getInvoker(), new NonBooleanQuery(vars, normalized));
						IView normalizedView = new View(view.getId(), view.getOwner(), view.getMode(), view.getFields(), stmt);
						typeNodes.add(normalizedView);
					} catch (MalformedStatementException | MalformedQueryException | MalformedDbObjectException e) {
						assert (false);
					}

				}

			}

			// create edges and add labels
			for (IType n : typeNodes) {
				if (n instanceof IView) {
					if (((IView) n).getSelectQuery().getQuery().getExpression() instanceof PredicateExpression) {

						if (typeNodes.contains(((PredicateExpression) ((IView) n).getSelectQuery().getQuery().getExpression()).getType())) {
							Pair<IType, IType> edge = new Pair<IType, IType>(n, ((PredicateExpression) ((IView) n).getSelectQuery().getQuery()
									.getExpression()).getType());
							typeEdges.add(edge);

							IExpression label = TrueExpression.getInstance();
							typeEdgeLabels.put(edge, label);

						} else
							assert (false); // it should be unreachable

					} else if (((IView) n).getSelectQuery().getQuery().getExpression() instanceof PolyadicExpression) {
						PolyadicExpression expr = (PolyadicExpression) ((IView) n).getSelectQuery().getQuery().getExpression();
						if (expr.getOperator() instanceof AndOperator) {
							for (IExpression e : expr.getExpressions()) {
								if (e instanceof PredicateExpression) {

									List<Variable> output = ((NonBooleanQuery) ((IView) n).getSelectQuery().getQuery()).getOutput();

									// INNER1: for (IOperand o : ((PredicateExpression) e).getOperands())
									// if (o instanceof Constant) {
									// onlyOutput = false;
									// break INNER1;
									// } else if (o instanceof Variable) {
									// if (!output.contains(o)) {
									// onlyOutput = false;
									// break INNER1;
									// }
									// }

									// check that the predicate uses the variables in the view's definition.
									if (output.containsAll(((PredicateExpression) e).getOperands())
											&& ((PredicateExpression) e).getOperands().containsAll(output)) {

										// List<Variable> toReplace = new Vector<Variable>();
										// for (IOperand o : ((PredicateExpression) e).getOperands())
										// toReplace.add((Variable) o);
										//
										// List<IOperand> replacements = new Vector<IOperand>();
										// for (String f : ((PredicateExpression) e).getType().getFields())
										// try {
										// replacements.add(new Variable(((PredicateExpression) e).getType().getId() + "_" + f));
										// } catch (MalformedExpressionException e2) {
										// assert (false); // unreachable
										// }

										List<Variable> toReplace = output;

										List<IOperand> replacements = new Vector<IOperand>();

										if (((PredicateExpression) e).getType() instanceof IView)
											for (String f : ((PredicateExpression) e).getType().getFields())
												try {
													replacements.add(new Variable(f));
												} catch (MalformedExpressionException e2) {
													assert (false); // unreachable
												}
										else if (((PredicateExpression) e).getType() instanceof ITable)
											for (Variable v : ((ITable) ((PredicateExpression) e).getType()).getFreeVariables())
												replacements.add(v);
										else
											assert (false);

										Pair<IType, IType> edge = new Pair<IType, IType>(n, ((PredicateExpression) e).getType());
										typeEdges.add(edge);
										IExpression label = null;

										if (expr.getExpressions().size() == 2) {
											Vector<IExpression> vect = new Vector<IExpression>(expr.getExpressions());
											vect.remove(e);
											label = vect.get(0);
										} else {
											Vector<IExpression> vect = new Vector<IExpression>(expr.getExpressions());
											vect.remove(e);
											try {
												label = new PolyadicExpression(AndOperator.getInstance(), vect);
											} catch (MalformedExpressionException e1) {
												assert (false); // unreachable
											}
										}

										// translate the label
										ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(toReplace, replacements, true,
												IncrementalIdentifiers.getPrefix(), min);
										label.accept(v);
										typeEdgeLabels.put(edge, v.getResult());
										min = v.getMinValue();
									}

								}
							}

						}
					}
				}
			}

			// create edges and add labels
			for (IPositiveGrantStatement g : state.getPolicy()) {
				if (g.getPrivilege() instanceof SelectPrivilege) {
					Pair<IUser, IType> edge = new Pair<IUser, IType>(g.getTarget(), ((SelectPrivilege) g.getPrivilege()).getObj());
					userEdges.add(edge);

					userEdgeLabels.put(edge, TrueExpression.getInstance());

				}
			}

		}
		
		
	}

	/**
	 * Get the constraint. (It terminates iff the graph does not contain cycles!!!)
	 * 
	 * @param type
	 *            - The type associated with the constraint.
	 * @param user
	 *            - The user associated with the constraint.
	 */
	public IExpression getConstraint(IType type, IUser user) {

		Vector<IExpression> disjuncts = new Vector<IExpression>();

		for (Pair<IUser, IType> p : userEdges) {
			if (p.getFirst().equals(user)) {

				disjuncts.addAll(getConstraint(type, p.getSecond(), userEdgeLabels.get(p)));

			}
		}

		try {
			if (disjuncts.size() > 1)
				return new PolyadicExpression(OrOperator.getInstance(), disjuncts);
			else if (disjuncts.size() == 1)
				return disjuncts.get(0);
			else
				return FalseExpression.getInstance();
		} catch (MalformedExpressionException e) {
			assert (false);
			return null;
		}
	}

	/**
	 * Recursive function that compute the constraint.
	 * 
	 * @param type
	 *            - the type
	 * @param iExpression
	 *            - the expression so far.
	 * @return The set of expressions.
	 */
	private List<IExpression> getConstraint(IType toReach, IType type, IExpression iExpression) {
		Vector<IExpression> v = new Vector<IExpression>();

		if (type.equals(toReach))
			v.add(iExpression);
		else {

			IncrementalIdentifiers vInc = new IncrementalIdentifiers();
			iExpression.accept(vInc);

			int min = vInc.getMaximumValue();
			for (Pair<IType, IType> p : typeEdges) {
				if (p.getFirst().equals(type)) {
					Vector<IExpression> v1 = new Vector<IExpression>();

					IExpression expr = iExpression;
					// replace the variables
					if (type instanceof IView) {
						List<Variable> toReplace = ((NonBooleanQuery) ((IView) type).getSelectQuery().getQuery()).getOutput();

						List<IOperand> replacements = new Vector<IOperand>();
						if (toReach instanceof ITable)
							for (Variable var : ((ITable) toReach).getFreeVariables())
								replacements.add(var);
						else if (toReach instanceof IView)
							for (Variable var : ((NonBooleanQuery) ((IView) toReach).getSelectQuery().getQuery()).getOutput())
								replacements.add(var);

						ReplaceFreeVariablesVisitor vRepl = new ReplaceFreeVariablesVisitor(toReplace, replacements, true,
								IncrementalIdentifiers.getPrefix(), min);
						iExpression.accept(vRepl);
						expr = vRepl.getResult();
						min = vRepl.getMinValue();
					}

					v1.add(expr);
					v1.add(typeEdgeLabels.get(p));
					PolyadicExpression newExpr = null;
					try {
						newExpr = new PolyadicExpression(AndOperator.getInstance(), v1);
					} catch (MalformedExpressionException e) {
						assert (false);
					}

					v.addAll(getConstraint(toReach, p.getSecond(), newExpr));
				}
			}
			// if (v.isEmpty())
			// v.add(iExpression);
		}
		return v;
	}
}
