package prototype.pdp.secure;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.constraints.ForeignKeyConstraint;
import prototype.objects.constraints.IConstraint;
import prototype.objects.implementation.AnonymousView;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.FreeVariableVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.OrOperator;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;
import prototype.util.Pair;
import prototype.util.Util;

/**
 * The class representing the authorization graph (See appendix C of our paper).
 * 
 * 
 *
 */
public class AuthorizationGraphV3 {

	/**
	 * Abstract class representing all the nodes.
	 * 
	 * 
	 *
	 */
	abstract class Node {

		/**
		 * The object associated with the node.
		 */
		private Object obj;

		private Node() {
		}

		public Node(Object obj) {
			this.obj = obj;
		}

		public Object obj() {
			return obj;
		}

		@Override
		public String toString() {
			return obj.toString();
		}

		@Override
		public int hashCode() {
			return obj.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof Node)
				return this.obj.equals(((Node) obj).obj);
			else
				return false;
		}

	}

	/**
	 * Nodes containing users.
	 * 
	 * 
	 */
	class UserNode extends Node {

		public UserNode(IUser user) {
			super(user);
		}

		@Override
		public IUser obj() {
			return (IUser) super.obj();
		}

	}

	/**
	 * Nodes containing expressions.
	 * 
	 * 
	 */
	class ExpressionNode extends Node {

		public ExpressionNode(AnonymousView expr) throws MalformedExpressionException, MalformedDbObjectException, MalformedQueryException {
			super.obj = AuthorizationGraphV3.inlineDefs(expr);
		}

		public ExpressionNode(LinkedHashSet<AnonymousView> exprs) throws MalformedExpressionException, MalformedDbObjectException,
				MalformedQueryException {
			super.obj = inlineDefs(exprs);
		}

		private LinkedHashSet<AnonymousView> inlineDefs(Set<AnonymousView> exprs) throws MalformedExpressionException, MalformedDbObjectException,
				MalformedQueryException {
			LinkedHashSet<AnonymousView> expanded = new LinkedHashSet<AnonymousView>();
			for (AnonymousView expr : exprs) {
				expanded.addAll(AuthorizationGraphV3.inlineDefs(expr));
			}
			return expanded;
		}

		@SuppressWarnings("unchecked")
		@Override
		public LinkedHashSet<AnonymousView> obj() {
			return (LinkedHashSet<AnonymousView>) super.obj();
		}

		@SuppressWarnings("unchecked")
		@Override
		public String toString() {
			return ((LinkedHashSet<AnonymousView>) super.obj()).iterator().next().toString();
		}
	}

	class Edge {
		private Node from;

		private Node to;

		private Map<Variable, Variable> variables;

		public Edge(Node from, Node to, Map<Variable, Variable> variables) {
			this.from = from;
			this.to = to;
			this.variables = variables;
		}

		public Node from() {
			return from;
		}

		public Node to() {
			return to;
		}

		public Map<Variable, Variable> variables() {
			return variables;
		}

		@Override
		public String toString() {
			return from + " ---" + variables + "---> " + to;
		}

		@Override
		public int hashCode() {
			int result = 17;
			result = 31 * result + from.hashCode();
			result = 31 * result + to.hashCode();
			result = 31 * result + variables.hashCode();
			return result;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			else if (obj instanceof Edge) {
				if (obj == this)
					return true;
				else
					return ((Edge) obj).from.equals(this.from) && ((Edge) obj).to.equals(this.to) && ((Edge) obj).variables.equals(this.variables);
			} else
				return false;
		}
	}

	/**
	 * The graph's nodes.
	 */
	private Set<Node> nodes;

	/**
	 * The directed edges (from/to) between Nodes.
	 */
	private Set<Edge> edges;

	/**
	 * Constructor.
	 * 
	 * @param state
	 *            - The state.
	 * @throws MalformedExpressionException
	 * @throws MalformedQueryException
	 * @throws MalformedDbObjectException
	 */
	public AuthorizationGraphV3(IState state, IStatementFactory factory) throws MalformedExpressionException, MalformedQueryException,
			MalformedDbObjectException {
		if (state != null) {
			this.nodes = new LinkedHashSet<Node>();
			this.edges = new LinkedHashSet<Edge>();

			// add the users
			for (IUser user : state.getUsers())
				nodes.add(new UserNode(user));

			// add the tables
			for (ITable table : state.getTables()) {
				List<IOperand> operands = new Vector<IOperand>(table.getFreeVariables());
				PredicateExpression expr = new PredicateExpression(table, operands);

				// create the nodes
				for (Set<Variable> quantVars : Util.powerset(new LinkedHashSet<Variable>(table.getFreeVariables()))) {
					Set<Variable> vars = new TreeSet<Variable>(table.getFreeVariables());

					// remove the quantified variables
					vars.removeAll(quantVars);

					// if (!quantVars.containsAll(table.getFreeVariables())) SUPPORT ALSO BOOLEAN ANONYMOUS VIEWS
					NonBooleanQuery query;
					if (!quantVars.isEmpty())
						query = new NonBooleanQuery(new Vector<Variable>(vars), new QuantifiedExpression(ExistentialQuantifier.getInstance(),
								new LinkedHashSet<Variable>(quantVars), expr));
					else
						query = new NonBooleanQuery(new Vector<Variable>(vars), expr);

					AnonymousView v = createAnonymousView(table.getId(), new Vector<Variable>(vars), query);

					nodes.add(new ExpressionNode(v));
				}
			}

			// add the views
			for (IView table : state.getViews()) {
				List<IOperand> operands = new Vector<IOperand>(((NonBooleanQuery) table.getSelectQuery().getQuery()).getOutput());
				PredicateExpression expr = new PredicateExpression(table, operands);

				// create the nodes
				for (Set<Variable> quantVars : Util.powerset(new LinkedHashSet<Variable>(((NonBooleanQuery) table.getSelectQuery().getQuery())
						.getOutput()))) {
					Set<Variable> vars = new TreeSet<Variable>(((NonBooleanQuery) table.getSelectQuery().getQuery()).getOutput());

					// remove the quantified variables
					vars.removeAll(quantVars);

					// if (!quantVars.containsAll(((NonBooleanQuery) table.getSelectQuery().getQuery()).getOutput())) SUPPORT ALSO BOOLEAN ANONYMOUS
					// VIEWS
					NonBooleanQuery query;
					if (!quantVars.isEmpty())
						query = new NonBooleanQuery(new Vector<Variable>(vars), new QuantifiedExpression(ExistentialQuantifier.getInstance(),
								new LinkedHashSet<Variable>(quantVars), expr));
					else
						query = new NonBooleanQuery(new Vector<Variable>(vars), expr);

					AnonymousView v = createAnonymousView(table.getId(), new Vector<Variable>(vars), query);

					nodes.add(new ExpressionNode(v));
				}
			}

			// merge the "equivalent" nodes
			while (!merge())
				;

			// create edges

			// RULE - 1
			for (Node n : nodes)
				if (n instanceof ExpressionNode) {
					// get free variables
					Set<AnonymousView> exprs = ((ExpressionNode) n).obj();
					AnonymousView candidate = exprs.iterator().next();

					// we extract the free variables of the first expression in exprs
					// NOTE: all the expressions in exprs have the same free variables
					LinkedHashSet<Variable> freeVars = new LinkedHashSet<Variable>(candidate.getSelectQuery().getOutput());

					if (freeVars.size() > 0)// 1) SUPPORT ALSO BOOLEAN ANONYMOUS VIEWS
						for (Variable var : freeVars) {
							LinkedHashSet<Variable> quantVar = new LinkedHashSet<Variable>();
							quantVar.add(var);
							QuantifiedExpression quantExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantVar, candidate
									.getSelectQuery().getExpression());

							Set<Variable> actualFreeVars = new TreeSet<Variable>(freeVars);
							actualFreeVars.remove(var);

							NonBooleanQuery q = new NonBooleanQuery(new Vector<Variable>(actualFreeVars), quantExpr);

							Pair<ExpressionNode, Map<Variable, Variable>> res = getNodeFromExpression(q);

							for (Variable variable : freeVars) {
								if (!variable.equals(var))
									res.getSecond().put(var, var);
							}

							// create the edge
							createEdge(n, res.getFirst(),  res.getSecond());
						}

				}

			// RULE - 2
			for (Node n : nodes)

				if (n instanceof ExpressionNode) {
					// get free variables
					Set<AnonymousView> exprs = ((ExpressionNode) n).obj();
					AnonymousView candidate = exprs.iterator().next();

					// we extract the free variables of the first expression in exprs
					// NOTE: all the expressions in exprs have the same free variables
					Set<Variable> freeVars = new HashSet<Variable>(candidate.getSelectQuery().getOutput());

					for (AnonymousView view : exprs)
						// we apply the rule for each expression in the node.. This might create new edges
						if (view.getSelectQuery().getExpression() instanceof PolyadicExpression
								&& ((PolyadicExpression) view.getSelectQuery().getExpression()).getOperator() instanceof AndOperator) {

							for (IExpression expr : ((PolyadicExpression) view.getSelectQuery().getExpression()).getExpressions()) {
								if (expr instanceof PredicateExpression) {

									// check that the predicate uses the variables in the view's definition.
									if (freeVars.containsAll(((PredicateExpression) expr).getOperands())
											&& ((PredicateExpression) expr).getOperands().containsAll(freeVars)) {

										Pair<ExpressionNode, Map<Variable, Variable>> res = getNodeFromType(((PredicateExpression) expr).getType());

										AnonymousView c1 = res.getFirst().obj().iterator().next();

										Map<Variable, Variable> globalMap = new HashMap<Variable, Variable>();
										// set the map
										for (int i = 0; i < c1.getCardinality(); i++) {
											// ASSUMPTION : there are no constants in predicates
											Variable from = (Variable) ((PredicateExpression) expr).getOperands().get(i);// (Variable)
																															// ((PredicateExpression)
											// view.getSelectQuery().getExpression()).getOperands().get(i);

											Variable to = c1.getSelectQuery().getOutput().get(i);

											globalMap.put(from, to);
										}

										// create the label

										IExpression label = null;

										if (((PolyadicExpression) view.getSelectQuery().getExpression()).getExpressions().size() == 2) {
											Vector<IExpression> vect = new Vector<IExpression>(((PolyadicExpression) view.getSelectQuery()
													.getExpression()).getExpressions());
											vect.remove(expr);
											label = vect.get(0);
										} else {
											Vector<IExpression> vect = new Vector<IExpression>(((PolyadicExpression) view.getSelectQuery()
													.getExpression()).getExpressions());
											vect.remove(expr);
											try {
												label = new PolyadicExpression(AndOperator.getInstance(), vect);
											} catch (MalformedExpressionException e1) {
												assert (false); // unreachable
											}
										}

										// createEdge(n, res.getFirst(), label, globalMap);

										FreeVariableVisitor vFree = new FreeVariableVisitor();
										label.accept(vFree);

										Set<Variable> labelFreeVars = vFree.getFreeVariables();

										// create the edges for all the projections
										for (Set<Variable> vars : Util.powerset(new HashSet<Variable>(freeVars)))
											// if (!vars.isEmpty() && vars.containsAll(labelFreeVars)) {
											if (vars.containsAll(labelFreeVars)) { // SUPPORT ALSO BOOLEAN ANONYMOUS VIEWS

												List<Variable> quantVarsFrom = new Vector<Variable>(candidate.getSelectQuery().getOutput());
												quantVarsFrom.removeAll(vars);
												IExpression fromExpr;
												if (!quantVarsFrom.isEmpty())
													fromExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(),
															new LinkedHashSet<Variable>(quantVarsFrom), candidate.getSelectQuery().getExpression());
												else
													fromExpr = candidate.getSelectQuery().getExpression();

												List<Variable> correspondingVars = new Vector<Variable>();
												for (Variable var : vars)
													correspondingVars.add(globalMap.get(var));

												List<Variable> quantVarsDest = new Vector<Variable>(c1.getSelectQuery().getOutput());
												quantVarsDest.removeAll(correspondingVars);
												IExpression destExpr;
												if (!quantVarsDest.isEmpty())
													destExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(),
															new LinkedHashSet<Variable>(quantVarsDest), c1.getSelectQuery().getExpression());
												else
													destExpr = c1.getSelectQuery().getExpression();

												NonBooleanQuery destQuery = new NonBooleanQuery(new Vector<Variable>(new TreeSet<Variable>(vars)),
														destExpr);
												NonBooleanQuery fromQuery = new NonBooleanQuery(new Vector<Variable>(new TreeSet<Variable>(
														correspondingVars)), fromExpr);

												Pair<ExpressionNode, Map<Variable, Variable>> fromRes = getNodeFromExpression(fromQuery);
												Pair<ExpressionNode, Map<Variable, Variable>> destRes = getNodeFromExpression(destQuery);

												Map<Variable, Variable> map = new HashMap<Variable, Variable>();

												for (int i = 0; i < fromQuery.getOutput().size(); i++)
													map.put(fromRes.getSecond().get(fromQuery.getOutput().get(i)),
															globalMap.get(destRes.getSecond().get(destQuery.getOutput().get(i))));

												// create the edge
												createEdge(fromRes.getFirst(), destRes.getFirst(),  globalMap);

											}

									}
								}
							}

						}

				}

			// RULE - 3

			for (Node n : nodes)
				if (n instanceof UserNode)
					for (IPositiveGrantStatement g : state.getPolicy())
						if (g.getPrivilege() instanceof SelectPrivilege && g.getTarget().equals(n.obj())) {
							Pair<ExpressionNode, Map<Variable, Variable>> res = getNodeFromType(((SelectPrivilege) g.getPrivilege()).getObj());

							// create the edge
							createEdge(n, res.getFirst(),  res.getSecond());

						}

			// RULE - 4
			for (Node n : nodes)
				if (n instanceof ExpressionNode)
					for (AnonymousView candidate : ((ExpressionNode) n).obj())
						if (candidate.getSelectQuery().getExpression() instanceof PredicateExpression
								&& ((PredicateExpression) candidate.getSelectQuery().getExpression()).getType() instanceof ITable) {
							ITable dest = (ITable) ((PredicateExpression) candidate.getSelectQuery().getExpression()).getType();

							// check the foreign key constraints
							for (ITable from : state.getTables())
								for (IConstraint cstr : from.getConstraints())
									if (cstr instanceof ForeignKeyConstraint && ((ForeignKeyConstraint) cstr).getDestinationTable().equals(dest)) {

										Pair<ExpressionNode, Map<Variable, Variable>> res = getNodeFromType(from);
										AnonymousView destCandidate = res.getFirst().obj().iterator().next();

										// Create the set of variables composing the FK
										List<Variable> fkFromFields = new Vector<Variable>();
										List<Variable> fkDestFields = new Vector<Variable>();
										for (int i = 0; i < from.getFields().size(); i++)
											if (((ForeignKeyConstraint) cstr).getDestinationFields().contains(from.getFields().get(i))) {
												fkDestFields.add(destCandidate.getSelectQuery().getOutput().get(i));

												fkFromFields.add(destCandidate.getSelectQuery().getOutput()
														.get(dest.getFields().indexOf(((ForeignKeyConstraint) cstr).getFields().get(i))));
											}

										// create the edges for all the projections
										for (Set<Variable> vars : Util.powerset(new HashSet<Variable>(fkDestFields)))
										// if (!vars.isEmpty()) { SUPPORT ALSO BOOLEAN ANONYMOUS VIEWS
										{
											List<Variable> quantVarsDest = new Vector<Variable>(destCandidate.getSelectQuery().getOutput());
											quantVarsDest.removeAll(vars);
											IExpression destExpr;
											if (!quantVarsDest.isEmpty())
												destExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(), new LinkedHashSet<Variable>(
														quantVarsDest), destCandidate.getSelectQuery().getExpression());
											else
												destExpr = destCandidate.getSelectQuery().getExpression();

											List<Variable> varsAsList = new Vector<Variable>(destCandidate.getSelectQuery().getOutput());
											varsAsList.retainAll(vars);

											List<Variable> varsFrom = new Vector<Variable>();
											for (Variable v : vars)
												varsFrom.add(fkFromFields.get(fkDestFields.indexOf(v)));

											LinkedHashSet<Variable> quantVarsFrom = new LinkedHashSet<Variable>(candidate.getSelectQuery()
													.getOutput());
											quantVarsFrom.removeAll(varsFrom);

											IExpression fromExpr;
											if (!quantVarsFrom.isEmpty())
												fromExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantVarsFrom, candidate
														.getSelectQuery().getExpression());
											else
												fromExpr = candidate.getSelectQuery().getExpression();

											NonBooleanQuery destQuery = new NonBooleanQuery(new Vector<Variable>(new TreeSet<Variable>(varsAsList)),
													destExpr);
											NonBooleanQuery fromQuery = new NonBooleanQuery(new Vector<Variable>(new TreeSet<Variable>(varsFrom)),
													fromExpr);

											Pair<ExpressionNode, Map<Variable, Variable>> fromRes = getNodeFromExpression(fromQuery);
											Pair<ExpressionNode, Map<Variable, Variable>> destRes = getNodeFromExpression(destQuery);

											Map<Variable, Variable> map = new HashMap<Variable, Variable>();

											for (int i = 0; i < fromQuery.getOutput().size(); i++)
												map.put(fromRes.getSecond().get(fromQuery.getOutput().get(i)),
														destRes.getSecond().get(destQuery.getOutput().get(i)));

											// create the edge
											createEdge(fromRes.getFirst(), destRes.getFirst(),  map);

										}

									}
						}

			// RULE - 5
			for (Node n : nodes)
				if (n instanceof ExpressionNode) {
					// get free variables
					Set<AnonymousView> exprs = ((ExpressionNode) n).obj();
					AnonymousView candidate = exprs.iterator().next();

					// we extract the free variables of the first expression in exprs
					// NOTE: all the expressions in exprs have the same free variables
					Set<Variable> freeVars = new HashSet<Variable>(candidate.getSelectQuery().getOutput());

					for (AnonymousView view : exprs)
						// we apply the rule for each expression in the node.. This might create new edges
						if (view.getSelectQuery().getExpression() instanceof PolyadicExpression
								&& ((PolyadicExpression) view.getSelectQuery().getExpression()).getOperator() instanceof OrOperator) {

							for (IExpression expr : ((PolyadicExpression) view.getSelectQuery().getExpression()).getExpressions()) {
								if (expr instanceof PredicateExpression) {

									// check that the predicate uses the variables in the view's definition.
									if (freeVars.containsAll(((PredicateExpression) expr).getOperands())
											&& ((PredicateExpression) expr).getOperands().containsAll(freeVars)) {

										Pair<ExpressionNode, Map<Variable, Variable>> res = getNodeFromType(((PredicateExpression) expr).getType());

										AnonymousView c1 = res.getFirst().obj().iterator().next();

										Map<Variable, Variable> globalMap = new HashMap<Variable, Variable>();
										// set the map
										for (int i = 0; i < c1.getCardinality(); i++) {
											// ASSUMPTION : there are no constants in predicates
											Variable from = (Variable) ((PredicateExpression) expr).getOperands().get(i);

											Variable to = c1.getSelectQuery().getOutput().get(i);

											globalMap.put(to, from);
										}

										// create the edges for all the projections
										for (Set<Variable> vars : Util.powerset(new HashSet<Variable>(freeVars)))
											if (!vars.isEmpty()) {

												List<Variable> quantVarsDest = new Vector<Variable>(candidate.getSelectQuery().getOutput());
												quantVarsDest.removeAll(vars);
												IExpression destExpr;
												if (!quantVarsDest.isEmpty())
													destExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(),
															new LinkedHashSet<Variable>(quantVarsDest), candidate.getSelectQuery().getExpression());
												else
													destExpr = candidate.getSelectQuery().getExpression();

												List<Variable> varsAsList = new Vector<Variable>(candidate.getSelectQuery().getOutput());
												varsAsList.retainAll(vars);

												LinkedHashSet<Variable> correspondingVars = new LinkedHashSet<Variable>();
												for (Variable var : vars)
													correspondingVars.add(res.getFirst().obj().iterator().next().getSelectQuery().getOutput()
															.get(candidate.getSelectQuery().getOutput().indexOf(var)));

												LinkedHashSet<Variable> quantVarsFrom = new LinkedHashSet<Variable>(c1.getSelectQuery().getOutput());
												quantVarsFrom.removeAll(correspondingVars);

												IExpression fromExpr;
												if (!quantVarsFrom.isEmpty())
													fromExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantVarsFrom, c1
															.getSelectQuery().getExpression());
												else
													fromExpr = c1.getSelectQuery().getExpression();

												NonBooleanQuery destQuery = new NonBooleanQuery(new Vector<Variable>(
														new TreeSet<Variable>(varsAsList)), destExpr);
												NonBooleanQuery fromQuery = new NonBooleanQuery(new Vector<Variable>(new TreeSet<Variable>(
														correspondingVars)), fromExpr);

												Pair<ExpressionNode, Map<Variable, Variable>> fromRes = getNodeFromExpression(fromQuery);
												Pair<ExpressionNode, Map<Variable, Variable>> destRes = getNodeFromExpression(destQuery);

												Map<Variable, Variable> map = new HashMap<Variable, Variable>();

												for (int i = 0; i < fromQuery.getOutput().size(); i++)
													map.put(fromRes.getSecond().get(fromQuery.getOutput().get(i)),
															globalMap.get(destRes.getSecond().get(destQuery.getOutput().get(i))));

												// create the edge
												createEdge(fromRes.getFirst(), destRes.getFirst(), map);

											}

									}
								}
							}

						}
				}

		}
	}

	/**
	 * Merge all the nodes that are syntactically equivalent modulo variable identifiers. Returns true if we modified nodes. False otherwise.
	 * 
	 * @throws MalformedQueryException
	 * @throws MalformedDbObjectException
	 * @throws MalformedExpressionException
	 */
	private boolean merge() throws MalformedExpressionException, MalformedDbObjectException, MalformedQueryException {
		boolean modified = false;
		for (Node n1 : new LinkedHashSet<Node>(nodes))
			for (Node n2 : new LinkedHashSet<Node>(nodes))
				if (n1 instanceof ExpressionNode && n2 instanceof ExpressionNode)
					for (AnonymousView v1 : ((ExpressionNode) n1).obj())
						for (AnonymousView v2 : ((ExpressionNode) n2).obj()) {
							// check whether v1 and v2 are equivalent modulo variable identifiers

							boolean equal = false;

							UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor();
							v1.getSelectQuery().getExpression().accept(v);

							IExpression v1Expr = v.getRenamedExpression();
							Map<Variable, Integer> v1Map = v.getAssignment();

							v = new UniformVariableAssignmentVisitor();
							v2.getSelectQuery().getExpression().accept(v);

							IExpression v2Expr = v.getRenamedExpression();
							Map<Variable, Integer> v2Map = v.getAssignment();
							if (v1Expr.equals(v2Expr)) {
								Map<Variable, Variable> map = new HashMap<Variable, Variable>();

								for (Variable from : v1Map.keySet())
									INNER: for (Variable to : v2Map.keySet())
										if (v2Map.get(to) == v1Map.get(from)) {
											map.put(from, to);
											break INNER;
										}

								// check whether the variables used in the two queries also agree with the mapping

								List<Variable> v1Vars = new Vector<Variable>();
								for (Variable var : v1.getSelectQuery().getOutput())
									v1Vars.add(map.get(var));

								List<Variable> v2Vars = v2.getSelectQuery().getOutput();

								if (v1Vars.equals(v2Vars)) // the two queries are identical modulo variable identifiers :-)
									equal = true;
							}

							if (equal) {
								modified = true;

								// remove the existing nodes
								nodes.remove(n1);
								nodes.remove(n2);

								// merge the nodes
								LinkedHashSet<AnonymousView> views = new LinkedHashSet<AnonymousView>();
								views.addAll(((ExpressionNode) n1).obj());
								views.addAll(((ExpressionNode) n2).obj());
								Node merged = new ExpressionNode(views);
								nodes.add(merged);
							}
						}
		return modified;
	}

	private void createEdge(Node from, Node to, Map<Variable, Variable> map) {
		if (!from.equals(to)) {
			// create the edge
			Edge edge = new Edge(from, to,  map);
			edges.add(edge);
		}
	}

	/**
	 * Create an anonymous view associated with the query.
	 * 
	 * @param id
	 *            - The original type's id.
	 * @param query
	 *            - The query associated with the view.
	 * @return The anonymous view.
	 */
	private AnonymousView createAnonymousView(String id, List<Variable> vars, NonBooleanQuery query) {

		List<String> fields = new Vector<String>();
		String viewId = id;
		for (int i = 0; i < query.getOutput().size(); i++) {
			if (vars != null) {
				if (i == 0)
					viewId = viewId + "_(";
				else
					viewId = viewId + ", ";

				viewId += vars.get(i).toString();
			}

			fields.add(query.getOutput().get(i).toString());

			if (vars != null && i == query.getOutput().size() - 1)
				viewId = viewId + ")";
		}

		try {
			return new AnonymousView(viewId, fields, query);
		} catch (MalformedDbObjectException e) {
			assert (false);
		}
		return null;
	}

	// /**
	// * Extract the node associated with the given type.
	// *
	// * @param type
	// * - The type.
	// * @return The node containing the type and the mapping between variables.
	// */
	private Pair<ExpressionNode, Map<Variable, Variable>> getNodeFromType(IType type) throws MalformedExpressionException, MalformedQueryException {

		// create the non boolean query
		Vector<IOperand> operands = new Vector<IOperand>();
		Vector<Variable> variables = new Vector<Variable>();
		for (String op : type.getFields()) {
			operands.add(new Variable(op));
			variables.add(new Variable(op));
		}

		PredicateExpression pred = new PredicateExpression(type, operands);
		NonBooleanQuery q = new NonBooleanQuery(variables, pred);

		return getNodeFromExpression(q);

		// for (Node n : nodes) {
		// if (n instanceof ExpressionNode) {
		// ExpressionNode node = (ExpressionNode) n;
		//
		// for (AnonymousView view : node.obj()) {
		// if (view.equals(type) || view.getId().equals(type.getId())) {
		// Map<Variable, Variable> map = new HashMap<Variable, Variable>();
		//
		// return new Pair<ExpressionNode, Map<Variable, Variable>>((ExpressionNode) n, map); // this must be set outside :-)
		// }
		// }
		// }
		// }
		// return null;
	}

	/**
	 * Extract the node associated with the given expression (modulo variable identifiers).
	 * 
	 * @param expr
	 *            - The expression.
	 * @return The node containing the expression and the mapping between variables.
	 */
	private Pair<ExpressionNode, Map<Variable, Variable>> getNodeFromExpression(NonBooleanQuery q) {
		UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor();
		q.getExpression().accept(v);

		IExpression key = v.getRenamedExpression();

		// we remove anonymous views
		IExpression old = null;
		int min = v.getMaximumValue();
		IExpression normalized = key;
		do {
			RemoveAnonymousViewsVisitor vAnonymViews = new RemoveAnonymousViewsVisitor(min);
			normalized.accept(vAnonymViews);

			old = normalized;
			normalized = vAnonymViews.getTransformed();
			min = vAnonymViews.getMinValue();
		} while (!old.equals(normalized));

		// rename again
		v = new UniformVariableAssignmentVisitor();
		normalized.accept(v);

		key = v.getRenamedExpression();

		Map<Variable, Integer> keyMap = v.getAssignment();
		for (Node n : nodes) {
			if (n instanceof ExpressionNode) {
				ExpressionNode node = (ExpressionNode) n;
				for (AnonymousView view : node.obj()) {
					NonBooleanQuery q1 = view.getSelectQuery();
					UniformVariableAssignmentVisitor v1 = new UniformVariableAssignmentVisitor();
					q1.getExpression().accept(v1);

					IExpression candidate = v1.getRenamedExpression();
					Map<Variable, Integer> candidateMap = v1.getAssignment();
					if (key.equals(candidate)) {
						Map<Variable, Variable> map = new HashMap<Variable, Variable>();

						for (Variable from : keyMap.keySet())
							INNER: for (Variable to : candidateMap.keySet())
								if (candidateMap.get(to) == keyMap.get(from)) {
									map.put(from, to);
									break INNER;
								}

						// check whether the variables used in the two queries also agree with the mapping

						List<Variable> keyVars = new Vector<Variable>();
						for (Variable var : q.getOutput())
							keyVars.add(map.get(var));

						List<Variable> candidateVars = q1.getOutput();

						if (candidateVars.equals(keyVars)) // the two queries are identical modulo variable identifiers :-)
							return new Pair<ExpressionNode, Map<Variable, Variable>>((ExpressionNode) n, map);
					}
				}
			}
		}
		return null;
	}

	

	@Override
	public String toString() {
		String print = "NODES:" + System.getProperty("line.separator");
		for (Node n : nodes)
			print += n + System.getProperty("line.separator");
		print += "EDGES: " + System.getProperty("line.separator");
		for (Edge e : edges)
			print += e + System.getProperty("line.separator");
		return print;
	}

	private static LinkedHashSet<AnonymousView> inlineDefs(AnonymousView expr) throws MalformedExpressionException, MalformedDbObjectException,
			MalformedQueryException {
		LinkedHashSet<AnonymousView> expanded = new LinkedHashSet<AnonymousView>();
		String id = expr.getId();
		List<String> fields = expr.getFields();
		AnonymousView normalized = expr;
		AnonymousView old = null;

		IncrementalIdentifiers vInc = new IncrementalIdentifiers();
		normalized.getSelectQuery().getExpression().accept(vInc);
		IExpression normalizedExpr = vInc.getRenamedExpression();

		List<Variable> normalizedVars = new Vector<Variable>();
		for (Variable var : normalized.getSelectQuery().getOutput())
			normalizedVars.add(new Variable(IncrementalIdentifiers.getPrefix() + vInc.getAssignment().get(var)));
		normalized = new AnonymousView(id, fields, new NonBooleanQuery(normalizedVars, normalizedExpr));

		int min = vInc.getMaximumValue();

		do { // remove the views with INVOKER privileges
			NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
			normalized.getSelectQuery().getExpression().accept(vNorm);

			old = normalized;
			normalizedExpr = vNorm.getNormalizedExpression();
			normalized = new AnonymousView(id, fields, new NonBooleanQuery(normalizedVars, normalizedExpr));

			expanded.add(normalized);
			min = vNorm.getMaximumValue();
		} while (!old.equals(normalized));

		expanded.add(normalized);

		do {
			NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.OWNER);
			normalized.getSelectQuery().getExpression().accept(vNorm);

			old = normalized;
			normalizedExpr = vNorm.getNormalizedExpression();
			normalized = new AnonymousView(id, fields, new NonBooleanQuery(normalizedVars, normalizedExpr));

			do { // remove the views with INVOKER privileges
				vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
				normalized.getSelectQuery().getExpression().accept(vNorm);

				old = normalized;
				normalizedExpr = vNorm.getNormalizedExpression();
				normalized = new AnonymousView(id, fields, new NonBooleanQuery(normalizedVars, normalizedExpr));

				expanded.add(normalized);
				min = vNorm.getMaximumValue();
			} while (!old.equals(normalized));

			expanded.add(normalized);
			min = vNorm.getMaximumValue();
		} while (!old.equals(normalized));

		return expanded;
	}
}
