package prototype.pdp.secure;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.constraints.ForeignKeyConstraint;
import prototype.objects.constraints.IConstraint;
import prototype.objects.constraints.PrimaryKeyConstraint;
import prototype.objects.constraints.UniqueConstraint;
import prototype.pdp.IAccessControlMechanism;
import prototype.pdp.database.integrity.DeterminesVisitor;
import prototype.privileges.MalformedPrivilege;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.BooleanQuery;
import prototype.queries.domain.relational.calculus.INonBooleanQuery;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.interfaces.changePolicy.IChangePolicyStatement;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.system.MalformedStateException;
import prototype.system.implementation.ExecutionException;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.interfaces.IState;
import prototype.system.interfaces.IStateFactory;
import prototype.users.IUser;
import prototype.users.IUserFactory;
import prototype.users.MalformedUserException;
import prototype.util.Logger;
import prototype.util.Util;
import prototype.util.Logger.TimeProperty;

/**
 * The confidentiality access control function. (see paper).
 * 
 * 
 *
 */
public class ConfidentialityAccessControlFunction implements IAccessControlMechanism {

	/**
	 * The statement's factory.
	 */
	private IStatementFactory stmtFactory;

	/**
	 * The user's factory.
	 */
	private IUserFactory userFactory;

	/**
	 * The database engine.
	 */
	private IDatabaseEngine engine;

	/**
	 * The state factory
	 */
	private IStateFactory stateFactory;

	private static final boolean anonymousViews = true;

	/**
	 * The logger.
	 */
	private Logger logger;
	/**
	 * Logging Flag.
	 */
	private boolean logging;

	/**
	 * Constructor.
	 * 
	 * @param stmtFactory
	 *            - The statement's factory.
	 * @param userFactory
	 *            - The user factory.
	 * @param engine
	 *            - The database engine.
	 * @param stateFactory
	 *            - The state factory.
	 */
	public ConfidentialityAccessControlFunction(IStatementFactory stmtFactory, IUserFactory userFactory, IDatabaseEngine engine,
			IStateFactory stateFactory, Logger logger) {
		this.stmtFactory = stmtFactory;
		this.userFactory = userFactory;
		this.engine = engine;
		this.stateFactory = stateFactory;
		if (logger != null)
			this.logging = true;
		else
			this.logging = false;
		this.logger = logger;
	}

	@Override
	public boolean authorizes(IStatement statement, IState systemState) throws ExecutionException {
		try {

			IUser user = null;
			if (systemState.isTransaction())
				user = systemState.getInvoker();
			else
				user = statement.getInvoker();

			if (user.equals(userFactory.getAdmin()))
				return true;

			if (statement instanceof ISelectStatement) {

				// for boolean queries we just need to check secure.
				return secure(user, ((ISelectStatement) statement).getQuery(), systemState);

			} else if (statement instanceof IInsertStatement || statement instanceof IDeleteStatement) {

				if (pec((IUpdateStatement) statement, systemState, user)) {

					boolean secure = secure(user, fromAction((IUpdateStatement) statement), systemState);
					for (IConstraint c : getConstraints((IUpdateStatement) statement, systemState)) {
						boolean secureFs = secure(user, fromSuccess((IUpdateStatement) statement, c, systemState), systemState);
						boolean secureFd = secure(user, fromDeny((IUpdateStatement) statement, c, systemState), systemState);
						secure = secure && secureFs && secureFd;
					}
					return secure;
				} else {
					return false;
				}

			} else if (statement instanceof IChangePolicyStatement) {

				if (statement instanceof IPositiveGrantStatement && ((IChangePolicyStatement) statement).getPrivilege() instanceof SelectPrivilege) {

					if (((IChangePolicyStatement) statement).getTarget().equals(user))
						return authorizedTypes(systemState, user).contains(
								((SelectPrivilege) ((IChangePolicyStatement) statement).getPrivilege()).getObj());

					return true;

				} else
					return true;

			} else
				return true;

		} catch (MalformedUserException | MalformedQueryException | MalformedExpressionException | MalformedStatementException | MalformedPrivilege
				| MalformedStateException e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * Extract the formula representing what the adversary can learn in case a deny happens. (see paper)
	 * 
	 * @param statement
	 * @param c
	 * @return The formula.
	 * @throws MalformedExpressionException
	 * @throws MalformedQueryException
	 */
	private IQuery fromDeny(IUpdateStatement statement, IConstraint c, IState systemState) throws MalformedExpressionException,
			MalformedQueryException {

		if (statement instanceof IInsertStatement) {

			if (c instanceof PrimaryKeyConstraint || c instanceof UniqueConstraint) {

				List<IOperand> operands = new Vector<IOperand>();
				List<Variable> vars = statement.getRelationSchema().getFreeVariables();
				LinkedHashSet<Variable> quantifiedVars = new LinkedHashSet<Variable>();
				List<IExpression> disjuncts = new Vector<IExpression>();
				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++) {
					if (c instanceof PrimaryKeyConstraint
							&& ((PrimaryKeyConstraint) c).getFields().contains(statement.getRelationSchema().getFields().get(i)))
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We directly add it to the tuple.
						operands.add(statement.getTuple().getContent().get(i));
					else if (c instanceof UniqueConstraint
							&& ((UniqueConstraint) c).getFields().contains(statement.getRelationSchema().getFields().get(i)))
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We directly add it to the tuple.
						operands.add(statement.getTuple().getContent().get(i));
					else {
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We add the corresponding variable.
						operands.add(vars.get(i));
						quantifiedVars.add(vars.get(i));
						// we also create the corresponding inequality
						disjuncts.add(new UnaryExpression(NotOperator.getInstance(), new EqualityExpression(vars.get(i), statement.getTuple()
								.getContent().get(i))));

					}
				}

				PredicateExpression pred = new PredicateExpression(statement.getRelationSchema(), operands);

				IExpression expr1 = null;
				if (disjuncts.size() > 1)
					expr1 = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
				else if (disjuncts.size() == 1)
					expr1 = disjuncts.get(0);

				List<IExpression> exprs = new Vector<IExpression>();
				exprs.add(pred);

				if (expr1 != null)
					exprs.add(expr1);

				IExpression expr;
				if (exprs.size() > 1)
					expr = new PolyadicExpression(AndOperator.getInstance(), exprs);
				else
					expr = pred;

				IExpression result;
				if (!quantifiedVars.isEmpty())
					result = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantifiedVars, expr);
				else
					result = expr;

				return new BooleanQuery(result);

			} else if (c instanceof ForeignKeyConstraint) {
				ForeignKeyConstraint fk = (ForeignKeyConstraint) c;
				List<Variable> vars = fk.getDestinationTable().getFreeVariables();
				List<IExpression> disjuncts = new Vector<IExpression>();

				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
					if (fk.getFields().contains(statement.getRelationSchema().getFields().get(i))) {
						// the i-th value in the tuple is part of the foreign key
						// get the corresponding field in destinationTable

						String correspondingField = fk.getDestinationFields().get(
								fk.getFields().indexOf(statement.getRelationSchema().getFields().get(i)));

						// get the corresponding variable in destinationTable
						Variable var = null;
						for (Variable v : vars)
							if (v.getIdentifier().equals(correspondingField))
								var = v;

						// extract the corresponding value in the tuple
						IOperand value = statement.getTuple().getContent().get(i);

						// create the corresponding inequality
						disjuncts.add(new UnaryExpression(NotOperator.getInstance(), new EqualityExpression(var, value)));
					} else {
						// the i-th value is not part of the foreign key
						// do nothing
					}

				PredicateExpression pred = new PredicateExpression(fk.getDestinationTable(), new Vector<IOperand>(vars));
				disjuncts.add(new UnaryExpression(NotOperator.getInstance(), pred));

				IExpression expr;
				if (disjuncts.size() > 1)
					expr = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
				else
					expr = pred;

				QuantifiedExpression result = new QuantifiedExpression(UniversalQuantifier.getInstance(), new LinkedHashSet<Variable>(vars), expr);

				return new BooleanQuery(result);
			}

		} else if (statement instanceof IDeleteStatement) {
			if (c instanceof ForeignKeyConstraint) {
				ForeignKeyConstraint fk = (ForeignKeyConstraint) c;
				ITable sourceTable = null;
				for (ITable t : systemState.getTables()) {
					if (t.getConstraints().contains(fk)) {
						sourceTable = t;
						break;
					}
				}

				List<Variable> vars = sourceTable.getFreeVariables();
				LinkedHashSet<Variable> quantifiedVars = new LinkedHashSet<Variable>();
				Map<String, IOperand> map = new HashMap<String, IOperand>();
				List<IOperand> operands = new Vector<IOperand>();

				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
					if (fk.getDestinationFields().contains(statement.getRelationSchema().getFields().get(i))) {
						// the i-th value in the tuple is part of the foreign key
						// get the corresponding field in destinationTable

						String correspondingField = fk.getFields().get(
								fk.getDestinationFields().indexOf(statement.getRelationSchema().getFields().get(i)));

						// extract the corresponding value in the tuple
						IOperand value = statement.getTuple().getContent().get(i);

						map.put(correspondingField, value);

					} else {
						// the i-th value is not part of the foreign key
						// do nothing
					}

				for (Variable var : vars)
					if (map.containsKey(var.getIdentifier()))
						operands.add(map.get(var.getIdentifier()));
					else {
						quantifiedVars.add(var);
						operands.add(var);
					}

				PredicateExpression pred = new PredicateExpression(sourceTable, operands);

				IExpression exist;
				if (!quantifiedVars.isEmpty())
					exist = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantifiedVars, pred);
				else
					exist = pred;

				List<IOperand> operands2 = new Vector<IOperand>();
				List<IExpression> disjuncts = new Vector<IExpression>();
				LinkedHashSet<Variable> variables = new LinkedHashSet<Variable>();
				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
					if (fk.getDestinationFields().contains(statement.getRelationSchema().getFields().get(i))) {
						// the i-th value is part of the foreign key
						operands2.add(statement.getTuple().getContent().get(i));

					} else {
						// the i-th value is not part of the key

						IOperand value = statement.getTuple().getContent().get(i);
						IOperand var = vars.get(i);

						operands2.add(var);

						variables.add((Variable) var);

						EqualityExpression eq = new EqualityExpression(var, value);
						UnaryExpression not = new UnaryExpression(NotOperator.getInstance(), eq);

						disjuncts.add(not);
					}

				PredicateExpression pred1 = new PredicateExpression(fk.getDestinationTable(), operands2);
				UnaryExpression not = new UnaryExpression(NotOperator.getInstance(), pred1);
				disjuncts.add(not);

				IExpression or;
				if (disjuncts.size() > 1)
					or = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
				else
					or = disjuncts.get(0);

				IExpression forall;
				if (!variables.isEmpty())
					forall = new QuantifiedExpression(UniversalQuantifier.getInstance(), variables, or);
				else
					forall = or;

				List<IExpression> conjuncts = new Vector<IExpression>();
				conjuncts.add(forall);
				conjuncts.add(exist);
				PolyadicExpression result = new PolyadicExpression(AndOperator.getInstance(), conjuncts);
				return new BooleanQuery(result);
			} else
				assert (false);

		} else
			assert (false);

		return null;
	}

	/**
	 * Extract the formula representing what the adversary can learn in case a success happens. (see paper)
	 * 
	 * @param statement
	 * @param c
	 * @param systemState
	 * @return The formula.
	 * @throws MalformedExpressionException
	 * @throws MalformedQueryException
	 */
	private IQuery fromSuccess(IUpdateStatement statement, IConstraint c, IState systemState) throws MalformedExpressionException,
			MalformedQueryException {
		if (statement instanceof IInsertStatement) {

			if (c instanceof PrimaryKeyConstraint || c instanceof UniqueConstraint) {

				List<IOperand> operands = new Vector<IOperand>();
				List<Variable> vars = statement.getRelationSchema().getFreeVariables();
				LinkedHashSet<Variable> quantifiedVars = new LinkedHashSet<Variable>();
				List<IExpression> disjuncts = new Vector<IExpression>();
				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++) {
					if (c instanceof PrimaryKeyConstraint
							&& ((PrimaryKeyConstraint) c).getFields().contains(statement.getRelationSchema().getFields().get(i)))
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We directly add it to the tuple.
						operands.add(statement.getTuple().getContent().get(i));
					else if (c instanceof UniqueConstraint
							&& ((UniqueConstraint) c).getFields().contains(statement.getRelationSchema().getFields().get(i)))
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We directly add it to the tuple.
						operands.add(statement.getTuple().getContent().get(i));
					else {
						// statement.getTuple().getContent().get(i) corresponds to one of the key values.
						// We add the corresponding variable.
						operands.add(vars.get(i));
						quantifiedVars.add(vars.get(i));
						// we also create the corresponding inequality
						disjuncts.add(new UnaryExpression(NotOperator.getInstance(), new EqualityExpression(vars.get(i), statement.getTuple()
								.getContent().get(i))));

					}
				}

				PredicateExpression pred = new PredicateExpression(statement.getRelationSchema(), operands);
				IExpression expr1 = null;
				if (disjuncts.size() > 1)
					expr1 = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
				else if (disjuncts.size() == 1)
					expr1 = disjuncts.get(0);

				List<IExpression> exprs = new Vector<IExpression>();
				exprs.add(pred);
				if (expr1 != null)
					exprs.add(expr1);

				IExpression expr;
				if (exprs.size() > 1)
					expr = new PolyadicExpression(AndOperator.getInstance(), exprs);
				else
					expr = pred;

				IExpression quantExpr;
				if (!quantifiedVars.isEmpty())
					quantExpr = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantifiedVars, expr);
				else
					quantExpr = expr;
				UnaryExpression result = new UnaryExpression(NotOperator.getInstance(), quantExpr);
				return new BooleanQuery(result);

			} else if (c instanceof ForeignKeyConstraint) {

				ForeignKeyConstraint fk = (ForeignKeyConstraint) c;
				List<Variable> vars = fk.getDestinationTable().getFreeVariables();
				Map<String, IOperand> map = new HashMap<String, IOperand>();

				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
					if (fk.getFields().contains(statement.getRelationSchema().getFields().get(i))) {
						// the i-th value in the tuple is part of the foreign key
						// get the corresponding field in destinationTable

						String correspondingField = fk.getDestinationFields().get(
								fk.getFields().indexOf(statement.getRelationSchema().getFields().get(i)));

						map.put(correspondingField, statement.getTuple().getContent().get(i));
					} else {
						// the i-th value is not part of the foreign key
						// do nothing
					}

				List<IOperand> operands = new Vector<IOperand>();
				LinkedHashSet<Variable> quantifiedVars = new LinkedHashSet<Variable>();
				for (Variable var : vars)
					if (map.containsKey(var.getIdentifier()))
						operands.add(map.get(var.getIdentifier()));

					else {
						quantifiedVars.add(var);
						operands.add(var);
					}

				PredicateExpression pred = new PredicateExpression(fk.getDestinationTable(), operands);

				IExpression result;
				if (!quantifiedVars.isEmpty())
					result = new QuantifiedExpression(ExistentialQuantifier.getInstance(), new LinkedHashSet<Variable>(quantifiedVars), pred);
				else
					result = pred;

				return new BooleanQuery(result);

			}

		} else if (statement instanceof IDeleteStatement) {

			if (c instanceof ForeignKeyConstraint) {
				ForeignKeyConstraint fk = (ForeignKeyConstraint) c;
				ITable sourceTable = null;

				// note that fk.getDestinationTable() == statement.getRelationSchema()

				for (ITable t : systemState.getTables()) {
					if (t.getConstraints().contains(fk)) {
						sourceTable = t;
						break;
					}
				}

				List<Variable> vars = sourceTable.getFreeVariables();
				List<IExpression> disjuncts = new Vector<IExpression>();

				for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
					if (fk.getDestinationFields().contains(statement.getRelationSchema().getFields().get(i))) {
						// the i-th value in the tuple is part of the foreign key
						// get the corresponding field in sourceTable

						String correspondingField = fk.getFields().get(
								fk.getDestinationFields().indexOf(statement.getRelationSchema().getFields().get(i)));

						// get the corresponding variable in sourceTable
						Variable var = null;
						for (Variable v : vars)
							if (v.getIdentifier().equals(correspondingField))
								var = v;

						// extract the corresponding value in the tuple
						IOperand value = statement.getTuple().getContent().get(i);

						// create the corresponding inequality
						disjuncts.add(new UnaryExpression(NotOperator.getInstance(), new EqualityExpression(var, value)));
					} else {
						// the i-th value is not part of the foreign key
						// do nothing
					}

				// for (int i = 0; i < statement.getRelationSchema().getFields().size(); i++)
				// if (fk.getFields().contains(statement.getRelationSchema().getFields().get(i))) {
				// // the i-th value in the tuple is part of the foreign key
				// // get the corresponding field in destinationTable
				//
				// String correspondingField = fk.getDestinationFields().get(
				// fk.getFields().indexOf(statement.getRelationSchema().getFields().get(i)));
				//
				// // get the corresponding variable in destinationTable
				// Variable var = null;
				// for (Variable v : vars)
				// if (v.getIdentifier().equals(correspondingField))
				// var = v;
				//
				// // extract the corresponding value in the tuple
				// IOperand value = statement.getTuple().getContent().get(i);
				//
				// // create the corresponding inequality
				// disjuncts.add(new UnaryExpression(NotOperator.getInstance(), new EqualityExpression(var, value)));
				//
				// } else {
				// // the i-th value is not part of the foreign key
				// // do nothing
				// }

				PredicateExpression pred = new PredicateExpression(sourceTable, new Vector<IOperand>(vars));
				disjuncts.add(new UnaryExpression(NotOperator.getInstance(), pred));

				IExpression expr;
				if (disjuncts.size() > 1)
					expr = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
				else
					expr = disjuncts.get(0);

				QuantifiedExpression univ = new QuantifiedExpression(UniversalQuantifier.getInstance(), new LinkedHashSet<Variable>(vars), expr);

				ITable destinationTable = ((ForeignKeyConstraint) c).getDestinationTable();
				List<IOperand> operands1 = new Vector<IOperand>();
				LinkedHashSet<Variable> vars1 = new LinkedHashSet<Variable>();
				List<IExpression> disjuncts1 = new Vector<IExpression>();
				for (int i = 0; i < destinationTable.getFields().size(); i++)
					if (fk.getDestinationFields().contains(destinationTable.getFields().get(i))) {
						// the i-th value in the tuple is part of the foreign key
						// get the i-th value in the tuple

						IOperand value = statement.getTuple().getContent().get(i);
						operands1.add(value);

					} else {
						// the i-th value is not part of the foreign key

						Variable var = destinationTable.getFreeVariables().get(i);

						IOperand value = statement.getTuple().getContent().get(i);

						operands1.add(var);
						vars1.add(var);

						// create the disjunct
						EqualityExpression eq = new EqualityExpression(var, value);
						UnaryExpression not = new UnaryExpression(NotOperator.getInstance(), eq);
						disjuncts1.add(not);
					}

				PredicateExpression pred1 = new PredicateExpression(destinationTable, operands1);
				IExpression disj1 = null;
				if (disjuncts1.size() > 1)
					disj1 = new PolyadicExpression(OrOperator.getInstance(), disjuncts1);
				else if (disjuncts1.size() == 1)
					disj1 = disjuncts1.get(0);

				List<IExpression> conjuncts = new Vector<IExpression>();
				conjuncts.add(pred1);
				if (disj1 != null)
					conjuncts.add(disj1);
				IExpression conj;
				if (conjuncts.size() > 1)
					conj = new PolyadicExpression(AndOperator.getInstance(), conjuncts);
				else
					conj = pred1;

				IExpression exist;
				if (!vars1.isEmpty())
					exist = new QuantifiedExpression(ExistentialQuantifier.getInstance(), vars1, conj);
				else
					exist = conj;

				List<IExpression> disjuncts2 = new Vector<IExpression>();
				disjuncts2.add(univ);
				disjuncts2.add(exist);
				PolyadicExpression result = new PolyadicExpression(OrOperator.getInstance(), disjuncts2);
				return new BooleanQuery(result);
			} else
				assert (false);

		} else
			assert (false);

		return null;
	}

	/**
	 * Get all the constraints that might leak information when the statement is execute.
	 * 
	 * @param statement
	 * @param systemState
	 * @return
	 */
	private List<IConstraint> getConstraints(IUpdateStatement statement, IState systemState) {
		Vector<IConstraint> cstrs = new Vector<IConstraint>();

		if (statement instanceof IInsertStatement) {
			for (IConstraint c : statement.getRelationSchema().getConstraints()) {
				if (c instanceof PrimaryKeyConstraint || c instanceof UniqueConstraint || c instanceof ForeignKeyConstraint)
					cstrs.add(c);
			}
		} else if (statement instanceof IDeleteStatement) {
			for (ITable t : systemState.getTables()) {
				for (IConstraint c : t.getConstraints())
					if (c instanceof ForeignKeyConstraint && ((ForeignKeyConstraint) c).getDestinationTable().equals(statement.getRelationSchema()))
						cstrs.add(c);
			}
		} else
			assert (false);

		return cstrs;
	}

	/**
	 * Extract the formula representing what the adversary can from the action. (see paper)
	 * 
	 * @param statement
	 * @param c
	 * @return The formula.
	 * @throws MalformedExpressionException
	 * @throws MalformedQueryException
	 */
	private IQuery fromAction(IUpdateStatement statement) throws MalformedQueryException, MalformedExpressionException {
		IQuery query = null;
		if (statement instanceof IInsertStatement)
			query = new BooleanQuery(new UnaryExpression(NotOperator.getInstance(), new PredicateExpression(statement.getRelationSchema(), statement
					.getTuple().getContent())));
		else if (statement instanceof IDeleteStatement)
			query = new BooleanQuery(new PredicateExpression(statement.getRelationSchema(), statement.getTuple().getContent()));
		else
			assert (false);
		return query;
	}

	/**
	 * Implements the pec function in the paper.
	 * 
	 * @param statement
	 * @param systemState
	 * @param user
	 * @return True if the statement preserves the equivalence class. False otherwise.
	 * @throws ExecutionException
	 * @throws MalformedPrivilege
	 * @throws MalformedUserException
	 * @throws MalformedStatementException
	 * @throws MalformedStateException
	 */
	private boolean pec(IUpdateStatement statement, IState systemState, IUser user) throws ExecutionException, MalformedStatementException,
			MalformedUserException, MalformedPrivilege, MalformedStateException {

		if (logging)
			logger.startMeasure(TimeProperty.CONFIDENTIALITY_ADDITIONAL_CHECKS);

		boolean ok = true;
		for (IView view : systemState.getViews()) {
			if (authorizedTypes(systemState, user).contains(view)) {

				Set<ITable> det = determines(view, systemState);

				if (det.contains(statement.getRelationSchema()))
					for (ITable type : det) {
						if (!authorizedTypes(systemState, user).contains(type))
							ok = false;
					}
			}
		}
		if (ok) {
			if (logging)
				logger.endMeasure(TimeProperty.CONFIDENTIALITY_ADDITIONAL_CHECKS);
			return true;
		}
		ok = true;
		for (IView view : systemState.getViews()) {
			if (authorizedTypes(systemState, user).contains(view)) {

				Set<IPositiveGrantStatement> newPolicy = new HashSet<IPositiveGrantStatement>();
				for (ITable table : systemState.getTables()) {
					if (authorizedTypes(systemState, user).contains(table) && !table.equals(statement.getRelationSchema())) {
						newPolicy.add((IPositiveGrantStatement) stmtFactory.createChangePolicyStatement(userFactory.getAdmin(),
								CHANGE_POLICY_MODE.GRANT, new SelectPrivilege(table), user));
					}
				}

				// check whether the materialization of the view is secure or not
				// create the new state in which user can read only the tables in authTables

				IState newState = stateFactory.createNewState(systemState.getUsers(), systemState.getTables(), newPolicy, systemState.getViews(),
						systemState.getTriggers(), systemState.getInvoker(), systemState.getRollbackState(), systemState.getScheduledTriggers(),
						systemState.getTuple(), systemState.getCurrentTrigger());

				ok = ok && secure(user, view.getSelectQuery().getQuery(), newState);

			}
		}
		if (logging)
			logger.endMeasure(TimeProperty.CONFIDENTIALITY_ADDITIONAL_CHECKS);
		if (ok)
			return true;

		return false;
	}

	private Set<ITable> determines(IView view, IState systemState) {

		Set<ITable> tables = systemState.getTables();

		IExpression expression = view.getSelectQuery().getQuery().getExpression();

		// unique ids
		IncrementalIdentifiers vInc = new IncrementalIdentifiers();
		expression.accept(vInc);

		expression = vInc.getRenamedExpression();

		// normalize
		NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), vInc.getMaximumValue(), PRIVILEGE_MODE.INVOKER);
		expression.accept(vNorm);

		expression = vNorm.getNormalizedExpression();

		List<Set<ITable>> pw = Util.powerset(tables);
		for (Set<ITable> t : pw) {
			// check determines
			DeterminesVisitor vDet = new DeterminesVisitor(t, this.stmtFactory);
			expression.accept(vDet);

			if (vDet.getResult())
				return t;
		}
		return null;

	}

	/**
	 * The secure sub-procedure (see paper).
	 * 
	 * @param user
	 *            - The user.
	 * @param query
	 *            - The query
	 * @param state
	 *            - The system's state
	 * @return True if expression is secure for user in state. False otherwise.
	 * @throws ExecutionException
	 */
	private boolean secure(IUser user, IQuery query, IState state) throws ExecutionException {
		try {
			if (logging)
				if (query instanceof INonBooleanQuery)
					logger.startMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_REWRITING);
				else
					logger.startMeasure(TimeProperty.CONFIDENTIALITY_BOOLEAN_QUERY_REWRITING);

			IExpression expression = null;
			if (query instanceof INonBooleanQuery) {
				ISelectStatement select = stmtFactory.createSelectStatement(userFactory.getAdmin(), query);

				if(logging)
					logger.startMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_MATERIALIZATION);
				expression = engine.materializeQuery(select, state).getExpression();
				if(logging)
					logger.endMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_MATERIALIZATION);
			} else
				expression = query.getExpression();

			// rewrite the expression using incremental identifiers
			IncrementalIdentifiers vInc = new IncrementalIdentifiers();
			expression.accept(vInc);

			// normalize the expression
			int min = vInc.getMaximumValue();

			IExpression old = null;

			IExpression normalized = vInc.getRenamedExpression();
			do {
				NormalizeVisitor vNorm = new NormalizeVisitor(IncrementalIdentifiers.getPrefix(), min, PRIVILEGE_MODE.INVOKER);
				normalized.accept(vNorm);

				old = normalized;
				normalized = vNorm.getNormalizedExpression();
				min = vNorm.getMaximumValue();
			} while (!old.equals(normalized));

			// add anonymous views
			if (anonymousViews) {
				old = null;

				do {
					AddAnonymousViewsVisitor vAnonymViews = new AddAnonymousViewsVisitor();
					normalized.accept(vAnonymViews);

					old = normalized;
					normalized = vAnonymViews.getTransformed();
				} while (!old.equals(normalized));
			}

			// check security
			SecureRewritingVisitor v = new SecureRewritingVisitor(state, user, stmtFactory, IncrementalIdentifiers.getPrefix(), min);
			normalized.accept(v);

			IExpression rewritten = v.getTransformed();

			vInc = new IncrementalIdentifiers();
			rewritten.accept(vInc);

			rewritten = vInc.getRenamedExpression();

			// remove anonymous views
			if (anonymousViews) {
				old = null;
				min = vInc.getMaximumValue();
				normalized = rewritten;
				do {
					RemoveAnonymousViewsVisitor vAnonymViews = new RemoveAnonymousViewsVisitor(min);
					normalized.accept(vAnonymViews);

					old = normalized;
					normalized = vAnonymViews.getTransformed();
					min = vAnonymViews.getMinValue();
				} while (!old.equals(normalized));
				rewritten = normalized;
			}

			ISelectStatement select = stmtFactory.createSelectStatement(userFactory.getAdmin(), new BooleanQuery(rewritten));

			if (logging)
				if (query instanceof INonBooleanQuery)
					logger.endMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_REWRITING);
				else
					logger.endMeasure(TimeProperty.CONFIDENTIALITY_BOOLEAN_QUERY_REWRITING);

			if (logging)
				if (query instanceof INonBooleanQuery)
					logger.startMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_EXECUTION);
				else
					logger.startMeasure(TimeProperty.CONFIDENTIALITY_BOOLEAN_QUERY_EXECUTION);

			if (engine.evaluate(select, state) == false) {
				if (logging)
					if (query instanceof INonBooleanQuery)
						logger.endMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_EXECUTION);
					else
						logger.endMeasure(TimeProperty.CONFIDENTIALITY_BOOLEAN_QUERY_EXECUTION);
				return true;
			} else {
				if (logging)
					if (query instanceof INonBooleanQuery)
						logger.endMeasure(TimeProperty.CONFIDENTIALITY_NON_BOOLEAN_QUERY_EXECUTION);
					else
						logger.endMeasure(TimeProperty.CONFIDENTIALITY_BOOLEAN_QUERY_EXECUTION);
				return false;
			}
		} catch (MalformedQueryException | MalformedStatementException | MalformedUserException e) {
			throw new ExecutionException(e);
		}

	}

	/**
	 * Compute the set of authorized types. (See permission function in the paper).
	 * 
	 * @param state
	 * @param user
	 * @return The set of types for which the user has a direct grant.
	 */
	public Set<IType> authorizedTypes(IState state, IUser user) {
		Set<IType> auth = new HashSet<IType>();

		for (IPositiveGrantStatement g : state.getPolicy()) {
			if (g.getTarget().equals(user) && g.getPrivilege() instanceof SelectPrivilege) {
				auth.add(((SelectPrivilege) g.getPrivilege()).getObj());
			}
		}

		return auth;
	}

}
