package prototype.pdp.secure;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.AnonymousView;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;

/**
 * Visitor that takes a value V (true or false) and by visiting a formula Phi produces a formula Psi such that (1) Psi is secure, and (2) if the value
 * of Psi on the database is V, then the value of Phi on the database is V. Works only for SRNF expressions with incremental identifiers.
 * 
 * 
 *
 */
public class FixedSecureResultVisitor implements IVisitor {

	/**
	 * True if everything is ok.
	 */
	private boolean ok;
	/**
	 * The value.
	 */
	private boolean value;
	/**
	 * The state.
	 */
	private IState state;
	/**
	 * The user.
	 */
	private IUser user;
	/**
	 * The resulting expression.
	 */
	private IExpression expression;
	private Map<IType, Set<IType>> equivalent;

	/**
	 * Constructor.
	 * 
	 * @param user
	 *            - The user.
	 * @param state
	 *            - The state.
	 * @param flag
	 *            - The value.
	 */
	public FixedSecureResultVisitor(IUser user, IState state, boolean value) {
		if (user != null && state != null && state.getUsers().contains(user)) {
			this.ok = true;
			this.user = user;
			this.state = state;
			this.value = value;

			Map<IType, Set<IType>> oldEquivalent = null;
			Map<IType, Set<IType>> equivalent = new HashMap<IType, Set<IType>>();

			do {
				if (equivalent != null)
					oldEquivalent = new HashMap<IType, Set<IType>>(equivalent);
				for (IView v : state.getViews()) {
					if (v.getMode() == PRIVILEGE_MODE.OWNER) {
						if (v.getSelectQuery().getQuery().getExpression() instanceof PredicateExpression) {
							if (equivalent.containsKey(v)) {
								equivalent.get(v).add(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType());

								if (equivalent.containsKey(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType()))
									equivalent.get(v).addAll(
											equivalent.get(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType()));

							} else {
								Set<IType> types = new HashSet<IType>();
								types.add(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType());

								equivalent.put(v, types);
							}

							if (equivalent.containsKey(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType())) {
								equivalent.get(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType()).add(v);

								if (equivalent.containsKey(v))
									equivalent.get(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType()).addAll(
											equivalent.get(v));
							} else {
								Set<IType> types = new HashSet<IType>();
								types.add(v);

								equivalent.put(((PredicateExpression) v.getSelectQuery().getQuery().getExpression()).getType(), types);
							}
						}
					}
				}

			} while (!equivalent.equals(oldEquivalent));
			this.equivalent = equivalent;

		} else
			this.ok = false;

	}

	/**
	 * Private constructor.
	 * 
	 * @param user
	 *            - The user.
	 * @param state
	 *            - The state.
	 * @param flag
	 *            - The value.
	 */
	private FixedSecureResultVisitor(IUser user, IState state, boolean value, Map<IType, Set<IType>> equivalent) {
		this.user = user;
		this.state = state;
		this.value = value;
		this.equivalent = equivalent;
		this.ok = true;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok) {
			if (authorized(expr.getType())) {
				try {
					this.expression = expr.copy();
				} catch (MalformedExpressionException e) {
					assert (false);
				}
			} else if (value == true)
				this.expression = FalseExpression.getInstance();
			else if (value == false)
				this.expression = TrueExpression.getInstance();
		}
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok) {
			try {
				this.expression = expr.copy();
			} catch (MalformedExpressionException e) {
				assert (false);
			}
		}
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			if (expr.getQuantifier() instanceof ExistentialQuantifier) {
				if (this.value == false) {
					if (boundByAuthorized(expr.getVariables(), expr.getExpression())) {
						FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, equivalent);
						expr.getExpression().accept(v);

						try {
							this.expression = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.expression);
						} catch (MalformedExpressionException e) {
							assert (false);
						}
					} else
						this.expression = TrueExpression.getInstance();

				} else if (this.value == true) {

					if (allAuthorized(expr.getExpression()))
						try {
							this.expression = expr.copy();
						} catch (MalformedExpressionException e) {
							assert (false);
						}
					else
						this.expression = FalseExpression.getInstance();

				}
			} else if (expr.getQuantifier() instanceof UniversalQuantifier) {
				if (this.value == true) {
					if (boundByAuthorized(expr.getVariables(), expr.getExpression())) {
						FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, equivalent);
						expr.getExpression().accept(v);

						try {
							this.expression = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.expression);
						} catch (MalformedExpressionException e) {
							assert (false);
						}
					} else
						this.expression = FalseExpression.getInstance();

				} else if (this.value == false) {

					if (allAuthorized(expr.getExpression()))
						try {
							this.expression = expr.copy();
						} catch (MalformedExpressionException e) {
							assert (false);
						}
					else
						this.expression = TrueExpression.getInstance();

				}
			} else
				assert (false); // unreachable
		}
	}

	private boolean boundByAuthorized(Set<Variable> variables, IExpression expression) {
		boolean bound = true;

		Set<IType> types = new HashSet<IType>();
		for (ITable t : state.getTables())
			if (authorized(t)) {
				types.add(t);

				// add the equivalent types
				if (equivalent.containsKey(t))
					types.addAll(equivalent.get(t));
			}
		for (IView v : state.getViews())
			if (authorized(v) && v.getMode() == PRIVILEGE_MODE.OWNER) {
				types.add(v);

				// add the equivalent types
				if (equivalent.containsKey(v))
					types.addAll(equivalent.get(v));
			}

		for (Variable var : variables) {
			IsBoundByAuthorizedPredicateVisitor v = new IsBoundByAuthorizedPredicateVisitor(var, types);
			expression.accept(v);

			if (!v.getResult()) {
				bound = false;
				break;
			}
		}

		return bound;
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, !value, equivalent);
			expr.getExpression().accept(v);

			try {
				this.expression = new UnaryExpression(expr.getOperator(), v.expression);
			} catch (MalformedExpressionException e) {
				assert (false);
			}
		}
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			if (expr.getOperator() instanceof AndOperator) {

				if (this.value == false) {
					List<IExpression> exprs = new Vector<IExpression>();
					for (IExpression e : expr.getExpressions()) {
						FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, equivalent);
						e.accept(v);

						exprs.add(v.expression);
					}
					try {
						this.expression = new PolyadicExpression(expr.getOperator(), exprs);
					} catch (MalformedExpressionException e1) {
						assert (false);
					}
				} else if (this.value == true) {
					boolean authorized = true;
					for (IExpression e : expr.getExpressions()) {
						if (!allAuthorized(e)) {
							authorized = false;
							break;
						}
					}

					if (authorized)
						try {
							this.expression = expr.copy();
						} catch (MalformedExpressionException e1) {
							assert (false);
						}
					else
						this.expression = FalseExpression.getInstance();
				}

			} else if (expr.getOperator() instanceof OrOperator) {

				if (this.value == true) {
					List<IExpression> exprs = new Vector<IExpression>();
					for (IExpression e : expr.getExpressions()) {
						FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, equivalent);
						e.accept(v);

						exprs.add(v.expression);
					}
					try {
						this.expression = new PolyadicExpression(expr.getOperator(), exprs);
					} catch (MalformedExpressionException e1) {
						assert (false);
					}
				} else if (this.value == false) {
					boolean authorized = true;
					for (IExpression e : expr.getExpressions()) {
						if (!allAuthorized(e)) {
							authorized = false;
							break;
						}
					}

					if (authorized)
						try {
							this.expression = expr.copy();
						} catch (MalformedExpressionException e1) {
							assert (false);
						}
					else
						this.expression = TrueExpression.getInstance();
				}

			} else
				assert (false); // unreachable
		}
	}

	private boolean allAuthorized(IExpression e) {

		AllAuthorizedVisitor v = new AllAuthorizedVisitor(user, state);
		e.accept(v);

		return v.allAuthorized();
	}

	/**
	 * Check whether the give user is authorized to read the content of the given type in the given state.
	 * 
	 * @param type
	 *            - The type.
	 * @return True if user is authorized to read the content of type in state. False otherwise.
	 */
	private boolean authorized(IType type) {
		for (IPositiveGrantStatement g : state.getPolicy())
			if (g.getTarget().equals(user) && (g.getPrivilege() instanceof SelectPrivilege)) {
				if (((SelectPrivilege) g.getPrivilege()).getObj().equals(type))
					return true;
				else if (equivalent.containsKey(((SelectPrivilege) g.getPrivilege()).getObj())
						&& equivalent.get(((SelectPrivilege) g.getPrivilege()).getObj()).contains(type)) {
					return true;
				} else if (type instanceof AnonymousView) {
					// remove the anonymous views in the definition
					IncrementalIdentifiers vInc = new IncrementalIdentifiers();
					((AnonymousView) type).getSelectQuery().getExpression().accept(vInc);

					IExpression rewritten = vInc.getRenamedExpression();

					IExpression old = null;
					int min = vInc.getMaximumValue();
					IExpression normalized = rewritten;
					do {
						RemoveAnonymousViewsVisitor vAnonymViews = new RemoveAnonymousViewsVisitor(min);
						normalized.accept(vAnonymViews);

						old = normalized;
						normalized = vAnonymViews.getTransformed();
						min = vAnonymViews.getMinValue();
					} while (!old.equals(normalized));
					rewritten = normalized;

					if (rewritten instanceof QuantifiedExpression
							&& ((QuantifiedExpression) rewritten).getQuantifier() instanceof ExistentialQuantifier
							&& ((QuantifiedExpression) rewritten).getExpression() instanceof PredicateExpression) {
						return authorized(((PredicateExpression) ((QuantifiedExpression) rewritten).getExpression()).getType());
					}
				}

			}
		return false;
	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			this.expression = expr.copy();
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			this.expression = expr.copy();
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	/**
	 * Get the associated formula.
	 * 
	 * @return The resulting formula.
	 */
	public IExpression getFormula() {
		return this.expression;
	}

}
