package prototype.pdp.secure;

import java.util.HashSet;
import java.util.Set;

import prototype.objects.IType;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that checks whether a variable is bound by predicates included in a given set. It works only for formulae with unique identifiers and in
 * SRNF.
 * 
 * 
 *
 */
public class IsBoundByAuthorizedPredicateVisitor implements IVisitor {

	/**
	 * The result.
	 */
	private boolean result;
	private boolean ok;
	/**
	 * The variable.
	 */
	private Variable var;
	/**
	 * The set of types.
	 */
	private Set<IType> types;

	/**
	 * If strict is true, then we ignore equality expressions for checking whether var is bound or not.
	 */
	private boolean strict = true;

	public IsBoundByAuthorizedPredicateVisitor(Variable var, Set<IType> types) {
		if (var != null && types != null) {
			this.ok = true;
			this.var = var;
			this.types = types;
			this.result = false;
		} else
			this.ok = false;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok) 
			if (expr.getOperands().contains(var) && types.contains(expr.getType()))
				this.result = true;
		
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (!strict)
			if (expr.getOp1() instanceof Variable && expr.getOp2() instanceof Constant && expr.getOp1().equals(var))
				this.result = true;
			else if (expr.getOp2() instanceof Variable && expr.getOp1() instanceof Constant && expr.getOp2().equals(var))
				this.result = true;
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			if (expr.getVariables().contains(var))
				assert (false); // the assumption is that the formula has unique identifiers.

			IsBoundByAuthorizedPredicateVisitor v = new IsBoundByAuthorizedPredicateVisitor(var, types);
			expr.getExpression().accept(v);

			this.result = v.getResult();
		}

	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok)
			this.result = false;
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			if (expr.getOperator() instanceof OrOperator) {

				boolean isBound = true;

				for (IExpression e : expr.getExpressions()) {
					IsBoundByAuthorizedPredicateVisitor v = new IsBoundByAuthorizedPredicateVisitor(var, types);
					e.accept(v);

					isBound = isBound && v.getResult();

					if (!isBound)
						break;
				}

				this.result = isBound;

			} else if (expr.getOperator() instanceof AndOperator) {
				// we treat the expression as the semantically equivalent expression phi_{1} AND ... AND phi_{n} AND x_{1} = y_{1} AND ... x_{m} =
				// y_{m}

				Set<Variable> toCheck = new HashSet<Variable>();
				toCheck.add(var);
				if (!strict)
					for (IExpression e : expr.getExpressions()) {

						if (e instanceof EqualityExpression && ((EqualityExpression) e).getOp1() instanceof Variable
								&& ((EqualityExpression) e).getOp2() instanceof Variable) {

							if (toCheck.contains(((EqualityExpression) e).getOp1()) || toCheck.contains(((EqualityExpression) e).getOp2())) {
								toCheck.add((Variable) ((EqualityExpression) e).getOp1());
							}
						}
					}

				boolean isBound = false;

				for (IExpression e : expr.getExpressions()) {
					boolean isEqualityVariables = e instanceof EqualityExpression && ((EqualityExpression) e).getOp1() instanceof Variable
							&& ((EqualityExpression) e).getOp2() instanceof Variable;
					if (!isEqualityVariables) {

						for (Variable var : toCheck) {
							IsBoundByAuthorizedPredicateVisitor v = new IsBoundByAuthorizedPredicateVisitor(var, types);
							e.accept(v);

							isBound = isBound || v.getResult();

							if (isBound)
								break;
						}
					}
				}

				this.result = isBound;

			} else
				assert (false);// do nothing

		}
	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			this.result = false;
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			this.result = false;
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	public boolean getResult() {
		return this.result;
	}

}
