package prototype.pdp.secure;

import java.util.Vector;

import prototype.objects.IView;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.AnonymousView;
import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.INonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that visits and expression and does one step of normalization.
 * 
 * 
 *
 */
public class NormalizeVisitor implements IVisitor {

	/**
	 * The resulting expression.
	 */
	private IExpression normalized;

	/**
	 * The OK flag.
	 */
	private boolean ok;

	/**
	 * The prefix.
	 */
	private String prefix;

	/**
	 * The minValue.
	 */
	private int minValue;

	/**
	 * The privilege mode for the views that must be replaced by the visitor.
	 */
	private PRIVILEGE_MODE mode;

	/**
	 * Constructor.
	 * 
	 * @param prefix
	 *            - The prefix pattern for the "new variables".
	 * @param minValue
	 *            - The minimum identifier for the new variables.
	 * @param mode
	 *            - The privilege mode for the views that must be replaced by the visitor.
	 */
	public NormalizeVisitor(String prefix, int minValue, PRIVILEGE_MODE mode) {

		if (prefix != null && !prefix.isEmpty() && mode != null) {
			ok = true;
			this.prefix = prefix;
			this.minValue = minValue;
			this.mode = mode;
		} else
			ok = false;

	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok) {
			if (((expr.getType() instanceof IView) && ((IView) expr.getType()).getMode() == this.mode) || expr.getType() instanceof AnonymousView) {
				// normalize
				INonBooleanQuery definition = (INonBooleanQuery) ((IView) expr.getType()).getSelectQuery().getQuery();

				// replace free variables (and take care of possible clashes)
				ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(definition.getOutput(), expr.getOperands(), true, prefix, minValue);
				definition.getExpression().accept(v);

				this.normalized = v.getResult();
				this.minValue = v.getMinValue();

			} else
				try {
					this.normalized = expr.copy();
				} catch (MalformedExpressionException e) {
					assert (false);
				}
		}
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok)
			try {
				this.normalized = new EqualityExpression(expr.getOp1().copy(), expr.getOp2().copy());
			} catch (MalformedExpressionException e) {
				assert (false);
			}

	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			this.normalized = expr.copy();
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			this.normalized = expr.copy();
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {

			NormalizeVisitor v = new NormalizeVisitor(prefix, minValue, mode);
			expr.getExpression().accept(v);

			try {
				this.normalized = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.getNormalizedExpression());
				this.minValue = v.minValue;
			} catch (MalformedExpressionException e) {
				assert (false);
			}
		}
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			NormalizeVisitor v = new NormalizeVisitor(prefix, minValue, mode);
			expr.getExpression().accept(v);

			try {
				this.normalized = new UnaryExpression(NotOperator.getInstance(), v.getNormalizedExpression());
				this.minValue = v.minValue;
			} catch (MalformedExpressionException e) {
				assert (false);
			}
		}
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			Vector<IExpression> exprList = new Vector<IExpression>();
			for (IExpression e : expr.getExpressions()) {
				NormalizeVisitor v = new NormalizeVisitor(prefix, minValue, mode);
				e.accept(v);
				if (v.getNormalizedExpression() != null)
					exprList.add(v.getNormalizedExpression());
				else
					ok = false;
				this.minValue = v.minValue;
			}
			if (ok == true)
				try {
					this.normalized = new PolyadicExpression(expr.getOperator(), exprList);
				} catch (MalformedExpressionException e1) {
					assert (false);
				}

		}
	}

	/**
	 * Get the (one step) normalized expression.
	 * 
	 * @return The expression.
	 */
	public IExpression getNormalizedExpression() {
		return normalized;
	}

	public int getMaximumValue() {
		return minValue;
	}
}
