package prototype.pdp.secure;

import java.util.List;
import java.util.Vector;

import prototype.objects.implementation.AnonymousView;
import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that replaces SRNF-expressions of the form R_{xQ1,...xQl}(xQ1,...xQl), where R_{xQ1,...xQl} is an anonymous view, with their definitions.
 * 
 * The visitor is part of the preprocessing steps for the "secure" procedure.
 * 
 * 
 *
 */
public class RemoveAnonymousViewsVisitor implements IVisitor {

	/**
	 * The transformed expression.
	 */
	private IExpression transformed;
	private int minValue;

	public RemoveAnonymousViewsVisitor(int minValue) {
		this.minValue = minValue;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (!(expr.getType() instanceof AnonymousView))
			this.transformed = expr;
		else {
			// replace with definition
			NonBooleanQuery q = ((AnonymousView) expr.getType()).getSelectQuery();
			IncrementalIdentifiers vInc = new IncrementalIdentifiers(this.minValue);
			q.getExpression().accept(vInc);

			IExpression e1 = vInc.getRenamedExpression();

			// replace the variables in e1 with the operands of expr
			List<IOperand> replacements = new Vector<IOperand>();
			List<Variable> toReplace = new Vector<Variable>();
			
			for(int i = 0; i < expr.getType().getCardinality(); i++)
			{
				try {
					toReplace.add(new Variable(IncrementalIdentifiers.getPrefix() + vInc.getAssignment().get(q.getOutput().get(i))));
				} catch (MalformedExpressionException e) {
					assert(false);
				}
				replacements.add(expr.getOperands().get(i));
			}

			ReplaceFreeVariablesVisitor vRepl = new ReplaceFreeVariablesVisitor(toReplace, replacements, true, IncrementalIdentifiers.getPrefix(),
					vInc.getMaximumValue());
			e1.accept(vRepl);

			this.minValue = vRepl.getMinValue();
			this.transformed = vRepl.getResult();
		}

	}

	@Override
	public void visit(EqualityExpression expr) {
		this.transformed = expr;
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		// visit the sub-expression
		RemoveAnonymousViewsVisitor v = new RemoveAnonymousViewsVisitor(this.minValue);
		expr.getExpression().accept(v);

		try {
			this.transformed = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.transformed);
			this.minValue = v.minValue;
		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(UnaryExpression expr) {

		RemoveAnonymousViewsVisitor v = new RemoveAnonymousViewsVisitor(this.minValue);
		expr.getExpression().accept(v);

		try {
			this.transformed = new UnaryExpression(expr.getOperator(), v.transformed);
			this.minValue = v.minValue;
		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {
		List<IExpression> exprs = new Vector<IExpression>();

		for (IExpression ex : expr.getExpressions()) {
			RemoveAnonymousViewsVisitor v = new RemoveAnonymousViewsVisitor(this.minValue);
			ex.accept(v);

			exprs.add(v.transformed);
			this.minValue = v.minValue;
		}

		try {
			this.transformed = new PolyadicExpression(expr.getOperator(), exprs);
		} catch (MalformedExpressionException e) {
			assert (false);
		}
	}

	@Override
	public void visit(FalseExpression expr) {
		this.transformed = expr;
	}

	@Override
	public void visit(TrueExpression expr) {
		this.transformed = expr;
	}

	/**
	 * Get the transformed expression.
	 * 
	 * @return The transformed expression.
	 */
	public IExpression getTransformed() {
		return transformed;
	}
	
	public int getMinValue() {
		return minValue;
	}
}