package prototype.pdp.secure;

import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;
import prototype.util.Pair;

/**
 * Visitor implementing the rewriting procedure for the 'secure' approximation (see our paper section 6.2).
 * 
 * It works only for formulae in SRNF on which has been previously applied the IncrementalIdentifiers visitor.
 * 
 * 
 *
 */
public class SecureRewritingVisitor implements IVisitor {

	/**
	 * The resulting transformed formula.
	 */
	private IExpression transformed;

	/**
	 * The current system state.
	 */
	private IState state;

	/**
	 * The user.
	 */
	private IUser user;

	/**
	 * True if the visit is ok. False otherwise.
	 */
	private boolean ok;

	/**
	 * The authorization graph.
	 */
	// private AuthorizationGraph authGraph;
	private ImprovedAuthorizationGraph authGraph;

	/**
	 * The statement factory.
	 */
	private IStatementFactory factory;

	private String prefix;

	private int minValue;

	/**
	 * Constructor.
	 * 
	 * @param state
	 *            - The state in which the security of the formula is going to be checked.
	 * @param user
	 *            - The state for which the security of the formula is going to be checked.
	 * @param factory
	 *            - The statement factory.
	 */
	public SecureRewritingVisitor(IState state, IUser user, IStatementFactory factory, String prefix, int minValue) {
		if (state != null && user != null && state.getUsers().contains(user)) {
			this.state = state;
			this.user = user;
			this.ok = true;
			this.prefix = prefix;
			this.minValue = minValue;

			try {
				authGraph = new ImprovedAuthorizationGraph(state, factory); 
				System.out.println(authGraph.toString());
			} catch (MalformedExpressionException | MalformedQueryException | MalformedDbObjectException e) {
				assert (false);
				this.ok = false;
			}

		} else
			this.ok = false;
	}

	/**
	 * Constructor. It skips the creation of the authorization graph.
	 * 
	 * @param state
	 *            - The state in which the security of the formula is going to be checked.
	 * @param user
	 *            - The state for which the security of the formula is going to be checked.
	 * @param factory
	 *            - The statement factory.
	 * @param authGraph
	 *            - The authorization graph.
	 */
	private SecureRewritingVisitor(IState state, IUser user, IStatementFactory factory, ImprovedAuthorizationGraph authGraph, String prefix,
			int minValue) {
		if (state != null && user != null && state.getUsers().contains(user)) {
			this.state = state;
			this.user = user;
			this.ok = true;

			this.prefix = prefix;
			this.minValue = minValue;

			this.authGraph = authGraph;
			this.factory = factory;

		} else
			this.ok = false;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok)
			try {
				this.transformed = createResultingFormula(expr, new UnaryExpression(NotOperator.getInstance(), getConstraint(expr, user)));
			} catch (MalformedExpressionException e) {
				assert (false);
			}
	}

	/**
	 * Get the constraint associated with the given type in the given state for the given user.
	 * 
	 * @param type
	 *            - The type.
	 * @param user
	 *            - The user for which we want to check the security.
	 * @param list
	 *            - The list of operands that must be used to bind the constraint.
	 * @return The constraint.
	 */
	// private IExpression getConstraint(IType type, IUser user, List<IOperand> list) {
	// IExpression constraint = authGraph.getConstraint(type, user);
	//
	// // we must replace the variables in the constraint
	//
	// List<Variable> toReplace = null;
	// if (type instanceof IView)
	// toReplace = ((NonBooleanQuery) ((IView) type).getSelectQuery().getQuery()).getOutput();
	// else if (type instanceof ITable)
	// toReplace = ((ITable) type).getFreeVariables();
	// else
	// assert (false);
	// ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(toReplace, list, true, prefix, minValue);
	// constraint.accept(v);
	// this.minValue = v.getMinValue();
	// return v.getResult();
	//
	// }

	private IExpression getConstraint(IExpression expression, IUser user) {

		try {
			Pair<IExpression, Integer> pair = authGraph.getConstraint(expression, user, minValue);
			this.minValue = pair.getSecond();
			return pair.getFirst();
		} catch (MalformedQueryException | MalformedExpressionException e) {
			assert (false);
			this.ok = false;
			return null;
		}
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok)
			// we must return false.
			this.transformed = createResultingFormula(expr, FalseExpression.getInstance());
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			// we must return false.
			this.transformed = FalseExpression.getInstance();
	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			// we must return false.
			this.transformed = FalseExpression.getInstance();
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			SecureRewritingVisitor v = new SecureRewritingVisitor(state, user, factory, authGraph, prefix, minValue);
			expr.getExpression().accept(v);

			IExpression rw = v.getTransformed();
			
			

			try {
				Vector<IExpression> expList = new Vector<IExpression>();
//				for (Variable x : expr.getVariables())
//					expList.add(getAdomSec(state, user, x));
				expList.add(new UnaryExpression(NotOperator.getInstance(), rw));
				if (expr.getQuantifier() instanceof ExistentialQuantifier)
					expList.add(expr.getExpression().copy());
				else
					expList.add(new UnaryExpression(NotOperator.getInstance(), expr.getExpression().copy()));
			
				this.minValue = v.minValue;
				PolyadicExpression expression = new PolyadicExpression(AndOperator.getInstance(), expList);

				List<Variable> toReplace = new Vector<Variable>(expr.getVariables());
				List<IOperand> replacements = new Vector<IOperand>();
				LinkedHashSet<Variable> quantVars = new LinkedHashSet<Variable>();
				for (int i = 0; i < toReplace.size(); i++) {
					Variable newVar = new Variable(getNewIdentifier());
					replacements.add(newVar);
					quantVars.add(newVar);
				}

				ReplaceFreeVariablesVisitor vRep = new ReplaceFreeVariablesVisitor(toReplace, replacements, true, this.prefix, this.minValue);
				expression.accept(vRep);
				
				

				QuantifiedExpression expr2 = new QuantifiedExpression(ExistentialQuantifier.getInstance(), quantVars, vRep.getResult());

				this.transformed = createResultingFormula(expr, new UnaryExpression(NotOperator.getInstance(), expr2));
				this.minValue = vRep.getMinValue();
			} catch (MalformedExpressionException e) {
				assert (false);
			}

		}

	}

	/**
	 * Get the expression representing adomSec(x).
	 * 
	 * @param state2
	 * @param user2
	 * @param x
	 * @return The expression.
	 */
	private IExpression getAdomSec(IState state, IUser user, Variable x) {

		Set<IType> types = new HashSet<IType>();
		
		for (IView v : state.getViews())
			if (v.getMode() == PRIVILEGE_MODE.OWNER && authorized(user, v, state))
				types.add(v);
		for (ITable t : state.getTables())
			if (authorized(user, t, state))
				types.add(t);

		try {
			IExpression expr =  getAdomOnTypes(state, x, types);
			
			return expr;
		} catch (MalformedExpressionException e) {
			assert (false);
			return null;
		}
	}

	/**
	 * Compute the formula representing the active domain associated with the set of types given as input. Formally, adom(v) is satisfied iff v is the
	 * active domain of one of the types given as input.
	 * 
	 * @param state
	 * @param x
	 * @param types
	 * @return
	 * @throws MalformedExpressionException
	 */
	private IExpression getAdomOnTypes(IState state, Variable x, Set<IType> types) throws MalformedExpressionException {
		Vector<IExpression> exprList = new Vector<IExpression>();
		for (IType type : types) {
			LinkedHashSet<Variable> vars = new LinkedHashSet<Variable>();

			List<IOperand> operands = new Vector<IOperand>();
			for (int i = 0; i < type.getCardinality(); i++) {
				Variable v = new Variable(getNewIdentifier());
				operands.add(v);
				vars.add(v);
			}
			PredicateExpression predicate = new PredicateExpression(type, operands);

			Vector<IExpression> equalities = new Vector<IExpression>();
			for (int i = 0; i < type.getCardinality(); i++)
				equalities.add(new EqualityExpression(x, operands.get(i)));

			IExpression disjunction;
			if (equalities.size() > 1)
				disjunction = new PolyadicExpression(OrOperator.getInstance(), equalities);
			else
				disjunction = equalities.get(0); // there is always something in equalities.

			Vector<IExpression> conjuncts = new Vector<IExpression>();
			conjuncts.add(predicate);
			conjuncts.add(disjunction);
			PolyadicExpression conjunction = new PolyadicExpression(AndOperator.getInstance(), conjuncts);

			exprList.add(new QuantifiedExpression(ExistentialQuantifier.getInstance(), vars, conjunction));

		}
		if (exprList.size() == 0)
			return FalseExpression.getInstance();
		else if (exprList.size() == 1)
			return exprList.get(0);
		else
			return new PolyadicExpression(OrOperator.getInstance(), exprList);
	}

	private String getNewIdentifier() {
		minValue++;
		return prefix + minValue;
	}

	/**
	 * Check whether there is a grant that grants to the given user the SELECT permission over the given type.
	 * 
	 * @param user
	 *            - The user
	 * @param type
	 *            - The type
	 * @param state
	 *            - The state in which we check the permission.
	 * @return True if in state there is a grant statement giving to user the SELECT permission over type. False otherwise.
	 */
	private boolean authorized(IUser user, IType type, IState state) {
		for (IPositiveGrantStatement g : state.getPolicy())
			if (g.getTarget().equals(user) && (g.getPrivilege() instanceof SelectPrivilege)
					&& ((SelectPrivilege) g.getPrivilege()).getObj().equals(type))
				return true;
		return false;
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			SecureRewritingVisitor v = new SecureRewritingVisitor(this.state, this.user, factory, authGraph, prefix, minValue);
			expr.getExpression().accept(v);
			this.transformed = createResultingFormula(expr, v.getTransformed());
			this.minValue = v.minValue;
		}
	}

	private IExpression createResultingFormula(IExpression originalFormula, IExpression rewrittenFormula) {
		//return rewrittenFormula;
		try {
			FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, true);
			originalFormula.accept(v);

			IExpression trueFixedSec = v.getFormula();

			v = new FixedSecureResultVisitor(user, state, false);
			originalFormula.accept(v);

			IExpression falseFixedSec = v.getFormula();

			List<IExpression> conjuncts = new Vector<IExpression>();
			conjuncts.add(rewrittenFormula);
			conjuncts.add(new UnaryExpression(NotOperator.getInstance(), trueFixedSec));
			conjuncts.add(falseFixedSec);

			PolyadicExpression result = new PolyadicExpression(AndOperator.getInstance(), conjuncts);

			return result;

		} catch (MalformedExpressionException e) {
			assert (false); // unreachable
			return null;
		}
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			// we first rewrite the expression expr from \phi_1 AND \phi_2 ... AND \phi_N to expr1 := \phi_1 AND (\phi_2 AND \phi_3 ... AND \phi_N)

			Vector<IExpression> expList = new Vector<IExpression>();
			for (int i = 1; i < expr.getExpressions().size(); i++)
				expList.add(expr.getExpressions().get(i));

			IExpression subExpr = null;
			try {
				if (expList.size() > 1)
					subExpr = new PolyadicExpression(expr.getOperator(), expList);
				else
					subExpr = expList.get(0);
			} catch (MalformedExpressionException e) {
				assert (false);
			}

			// we transform expr1.getExpressions().get(0)
			SecureRewritingVisitor v = new SecureRewritingVisitor(state, user, factory, authGraph, prefix, minValue);
			expr.getExpressions().get(0).accept(v);
			IExpression rw1 = v.getTransformed();

			// we transform subExpr
			v = new SecureRewritingVisitor(state, user, factory, authGraph, prefix, v.minValue);
			subExpr.accept(v);
			IExpression rw2 = v.getTransformed();

			// we construct the transformed formula
			Vector<IExpression> expList1 = new Vector<IExpression>();
			try {
				expList1.add(rw1.copy());
				expList1.add(rw2.copy());
				PolyadicExpression expr1 = new PolyadicExpression(OrOperator.getInstance(), expList1);

//				PolyadicExpression expr1 = new PolyadicExpression(AndOperator.getInstance(), expList1);
//
//				Vector<IExpression> expList2 = new Vector<IExpression>();
//
//				if (expr.getOperator().equals(AndOperator.getInstance()))
//
//					expList2.add(expr.getExpressions().get(0).copy());
//
//				else
//					expList2.add(new UnaryExpression(NotOperator.getInstance(), expr.getExpressions().get(0).copy()));
//
//				expList2.add(new UnaryExpression(NotOperator.getInstance(), rw1.copy()));
//				expList2.add(rw2.copy());
//
//				PolyadicExpression expr2 = new PolyadicExpression(AndOperator.getInstance(), expList2);
//
//				Vector<IExpression> expList3 = new Vector<IExpression>();
//
//				if (expr.getOperator().equals(AndOperator.getInstance()))
//
//					expList3.add(subExpr.copy());
//
//				else
//					expList3.add(new UnaryExpression(NotOperator.getInstance(), subExpr.copy()));
//
//				expList3.add(new UnaryExpression(NotOperator.getInstance(), rw2.copy()));
//				expList3.add(rw1.copy());
//
//				PolyadicExpression expr3 = new PolyadicExpression(AndOperator.getInstance(), expList3);
//
//				Vector<IExpression> expList4 = new Vector<IExpression>();
//
//				expList4.add(expr1);
//				expList4.add(expr2);
//				expList4.add(expr3);
//
//				this.transformed = createResultingFormula(expr, new PolyadicExpression(OrOperator.getInstance(), expList4));
				this.transformed = createResultingFormula(expr, expr1);
				this.minValue = v.minValue;
			} catch (MalformedExpressionException e) {
				assert (false);
			}
		}
	}

	/**
	 * Get the transformed expression.
	 * 
	 * @return The transformed expression.
	 */
	public IExpression getTransformed() {
		if (ok)
			return transformed;
		else
			return null;
	}

}
