package prototype.pdp.secure;

import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Vector;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that visits an expression with unique identifiers and replace them with variables of the form x_i. Differently from IncrementalIdentifier,
 * given two expressions that are equivalent modulo the variable identifiers, the visitor produces the same results.
 * 
 * 
 *
 */
public class UniformVariableAssignmentVisitor implements IVisitor {

	/**
	 * The common prefix.
	 */
	private static String prefix = "x_";

	/**
	 * The map from variables to integers.
	 */
	private Map<Variable, Integer> assignment;

	/**
	 * The current value.
	 */
	private int current;

	/**
	 * The resulting expression.
	 */
	private IExpression transformed;

	

	public UniformVariableAssignmentVisitor() {
		assignment = new HashMap<Variable, Integer>();
		current = 0;
	}

	public UniformVariableAssignmentVisitor(int min) {
		assignment = new HashMap<Variable, Integer>();
		current = min;
	}

	private UniformVariableAssignmentVisitor(Map<Variable, Integer> assignment, int current) {
		this.assignment = new HashMap<Variable, Integer>(assignment);
		this.current = current;
	}

	@Override
	public void visit(PredicateExpression expr) {
		try {
			Vector<IOperand> newOperands = new Vector<IOperand>();
			for (IOperand o : expr.getOperands()) {
				if (o instanceof Constant)

					newOperands.add(o.copy());

				else if (o instanceof Variable) {
					newOperands.add(getVariable((Variable) o));
				} else
					assert (false); // unreachable

			}

			this.transformed = new PredicateExpression(expr.getType(), newOperands);
		} catch (MalformedExpressionException e1) {
			assert (false);
		}
	}

	/**
	 * Get the new variable.
	 * 
	 * @param v
	 * @return
	 * @throws MalformedExpressionException
	 */
	private Variable getVariable(Variable v) throws MalformedExpressionException {
		if (!assignment.containsKey(v)) {
			// add the new value
			current++;
			assignment.put(v, current);
		}
		return new Variable(prefix + assignment.get(v).toString());
	}

	@Override
	public void visit(EqualityExpression expr) {
		try {
			IOperand op1 = null;
			IOperand op2 = null;
			if (expr.getOp1() instanceof Constant)
				op1 = expr.getOp1().copy();
			else if (expr.getOp1() instanceof Variable)
				op1 = getVariable((Variable) expr.getOp1());
			else
				assert (false); // unreachable

			if (expr.getOp2() instanceof Constant)
				op2 = expr.getOp2().copy();
			else if (expr.getOp2() instanceof Variable)
				op2 = getVariable((Variable) expr.getOp2());
			else
				assert (false); // unreachable

			this.transformed = new EqualityExpression(op1, op2);
		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(FalseExpression expr) {
		this.transformed = expr.copy();
	}

	@Override
	public void visit(TrueExpression expr) {
		this.transformed = expr.copy();
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		try {
			LinkedHashSet<Variable> newVars = new LinkedHashSet<Variable>();
			// adjust the assignments
			for (Variable v : expr.getVariables()) {
				// remove the old assignment
				assignment.remove(v);
			}

			UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor(assignment, current);
			expr.getExpression().accept(v);

			this.assignment.putAll(v.assignment);
			this.current = v.current;
			for (Variable var : expr.getVariables()) {
				newVars.add(getVariable(var));
				this.assignment.remove(var);
			}

			this.transformed = new QuantifiedExpression(expr.getQuantifier(), newVars, v.transformed);

		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(UnaryExpression expr) {
		UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor(assignment, current);
		expr.getExpression().accept(v);

		try {
			this.transformed = new UnaryExpression(expr.getOperator(), v.transformed);
		} catch (MalformedExpressionException e) {
			assert (false);
		}
		this.assignment.putAll(v.assignment);
		this.current = v.current;
	}

	@Override
	public void visit(Variable var) {
		assert (false);
	}

	@Override
	public void visit(Constant expr) {
		assert (false);
	}

	@Override
	public void visit(AndOperator op) {
		assert (false);
	}

	@Override
	public void visit(Equality op) {
		assert (false);
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false);
	}

	@Override
	public void visit(NotOperator op) {
		assert (false);
	}

	@Override
	public void visit(OrOperator op) {
		assert (false);
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false);
	}

	@Override
	public void visit(PolyadicExpression expr) {
		Vector<IExpression> newExprs = new Vector<IExpression>();

		int current = this.current;
		for (IExpression e : expr.getExpressions()) {
			UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor(assignment, current);
			e.accept(v);

			newExprs.add(v.transformed);
			current = v.current;

			this.assignment.putAll(v.assignment);
		}

		this.current = current;
		try {
			this.transformed = new PolyadicExpression(expr.getOperator(), newExprs);
		} catch (MalformedExpressionException e1) {
			assert (false);
		}

	}

	/**
	 * Get the resulting expression.
	 * 
	 * @return The new expression.
	 */
	public IExpression getRenamedExpression() {

		return this.transformed;
	}

	/**
	 * Get the maximum value.
	 * 
	 * @return The maximul value.
	 */
	public int getMaximumValue() {
		return current;
	}

	public static String getPrefix() {
		return prefix;
	}

	/**
	 * Get the map assigning the variables to the integer identifiers.
	 * 
	 * @return The assignment.
	 */
	public Map<Variable, Integer> getAssignment() {
		return assignment;
	}

}
