package prototype.pdp.secure.newPdp;

import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.constraints.ForeignKeyConstraint;
import prototype.objects.constraints.IConstraint;
import prototype.pdp.secure.NormalizeVisitor;
import prototype.pdp.secure.RemoveAnonymousViewsVisitor;
import prototype.pdp.secure.UniformVariableAssignmentVisitor;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.OrOperator;
import prototype.util.Pair;

/**
 * This class can be used to check whether a formula is contained in another formula.
 * 
 * 
 *
 */
public class ContainmentChecker {

	/**
	 * The singleton instance.
	 */
	private static ContainmentChecker singleton_;

	/**
	 * Get the singleton.
	 */
	public static ContainmentChecker getInstance() {
		if (singleton_ == null)
			singleton_ = new ContainmentChecker();
		return singleton_;
	}

	/**
	 * Class constructor.
	 * 
	 * @param state
	 *            - The current state.
	 */
	private ContainmentChecker() {
	}

	/**
	 * Checks whether the query q1 is contained in the query q2.
	 * 
	 * @param contained
	 *            - The contained query.
	 * @param container
	 *            - The contains query.
	 * @return True if q1 \subseteq q2. False otherwise.
	 */
	public boolean contains(NonBooleanQuery contained, NonBooleanQuery container) {
		// check rule 1
		if (checkContainmentConjunction(contained, container))
			return true;

		// check rule 2
		if (checkContainmentDisjunction(contained, container))
			return true;

		// check rule 4
		if (equivalentModuloVariableIdentifiers(contained, container))
			return true;

		// check rule 5
		if (checkContainmentIntegrityConstraints(contained, container))
			return true;

		// check rule 3
		if (checkContainmentProjection(contained, container))
			return true;

		return false;
	}

	/**
	 * Check whether q1 is contained in q2 due to existential quantifiers. Implementation of rule 3.
	 * 
	 * @param q1
	 *            - The contained query.
	 * @param q2
	 *            - The container query.
	 * @return True if q1 \subseteq q2 due to existential quantifiers. False otherwise.
	 */
	private boolean checkContainmentProjection(NonBooleanQuery contained, NonBooleanQuery container) {

		IExpression containedExpr = contained.getExpression();
		IExpression containerExpr = container.getExpression();

		if (containedExpr instanceof QuantifiedExpression && ((QuantifiedExpression) containedExpr).getQuantifier() instanceof ExistentialQuantifier
				&& containerExpr instanceof QuantifiedExpression
				&& ((QuantifiedExpression) containerExpr).getQuantifier() instanceof ExistentialQuantifier)
			for (Variable var1 : ((QuantifiedExpression) containedExpr).getVariables())
				for (Variable var2 : ((QuantifiedExpression) containerExpr).getVariables()) {

					try {
						// build the new contained query
						List<Variable> output1 = new Vector<Variable>(contained.getOutput());
						output1.add(var1);
						IExpression expression1 = null;
						if (((QuantifiedExpression) containedExpr).getVariables().size() > 1) {
							LinkedHashSet<Variable> variables = new LinkedHashSet<Variable>(((QuantifiedExpression) containedExpr).getVariables());
							variables.remove(var1);

							expression1 = new QuantifiedExpression(ExistentialQuantifier.getInstance(), variables,
									((QuantifiedExpression) containedExpr).getExpression());

						} else
							expression1 = ((QuantifiedExpression) containedExpr).getExpression();

						NonBooleanQuery newContained = new NonBooleanQuery(output1, expression1);

						// build the new container query
						List<Variable> output2 = new Vector<Variable>(container.getOutput());
						output2.add(var2);
						IExpression expression2 = null;
						if (((QuantifiedExpression) containerExpr).getVariables().size() > 1) {
							LinkedHashSet<Variable> variables = new LinkedHashSet<Variable>(((QuantifiedExpression) containerExpr).getVariables());
							variables.remove(var2);
							expression2 = new QuantifiedExpression(ExistentialQuantifier.getInstance(), variables,
									((QuantifiedExpression) containerExpr).getExpression());

						} else
							expression2 = ((QuantifiedExpression) containerExpr).getExpression();

						NonBooleanQuery newContainer = new NonBooleanQuery(output2, expression2);

						// check containment recursively
						if (contains(newContained, newContainer))
							return true;

					} catch (MalformedQueryException | MalformedExpressionException e) {
						assert (false);
						return false;
					}

				}

		return false;
	}

	/**
	 * Check whether q1 is contained in q2 due to a disjunction. Implementation of rule 2.
	 * 
	 * @param q1
	 *            - The contained query.
	 * @param q2
	 *            - The container query.
	 * @return True if q1 \subseteq q2 due to a disjunction. False otherwise.
	 */
	private boolean checkContainmentDisjunction(NonBooleanQuery q1, NonBooleanQuery q2) {
		Pair<IExpression, Map<Variable, Integer>> pair1 = normalize(q1);
		IExpression containedExpr = pair1.getFirst();
		Map<Variable, Integer> containedAssignment = pair1.getSecond();

		Pair<IExpression, Map<Variable, Integer>> pair2 = normalize(q2);
		IExpression container = pair2.getFirst();
		Map<Variable, Integer> containerAssignment = pair2.getSecond();

		if (container instanceof PolyadicExpression && ((PolyadicExpression) container).getOperator() instanceof OrOperator) {
			for (IExpression expr : ((PolyadicExpression) container).getExpressions())
				if (expr.equals(containedExpr)) {
					Map<Variable, Variable> map = new HashMap<Variable, Variable>();

					for (Variable from : containedAssignment.keySet())
						INNER: for (Variable to : containerAssignment.keySet())
							if (containerAssignment.get(to) == containedAssignment.get(from)) {
								map.put(from, to);
								break INNER;
							}

					// check whether the variables used in the two queries also agree with the mapping

					List<Variable> keyVars = new Vector<Variable>();
					for (Variable var : q1.getOutput())
						keyVars.add(map.get(var));

					List<Variable> candidateVars = q2.getOutput();

					if (candidateVars.equals(keyVars))
						return true;
				}

		}
		return false;
	}

	/**
	 * Check whether q1 is contained in q2 due to integrity constraints. Implementation of rule 5.
	 * 
	 * @param q1
	 *            - The contained query.
	 * @param q2
	 *            - The container query.
	 * @return True if q1 \subseteq q2 due to an integrity constraint. False otherwise.
	 */
	private boolean checkContainmentIntegrityConstraints(NonBooleanQuery q1, NonBooleanQuery q2) {
		Pair<IExpression, Map<Variable, Integer>> pair1 = normalize(q1);
		IExpression contained = pair1.getFirst();
		Map<Variable, Integer> containedAssignment = pair1.getSecond();

		Pair<IExpression, Map<Variable, Integer>> pair2 = normalize(q2);
		IExpression container = pair2.getFirst();
		Map<Variable, Integer> containerAssignment = pair2.getSecond();

		if (contained instanceof QuantifiedExpression && ((QuantifiedExpression) contained).getQuantifier() instanceof ExistentialQuantifier
				&& ((QuantifiedExpression) contained).getExpression() instanceof PredicateExpression
				&& ((PredicateExpression) ((QuantifiedExpression) contained).getExpression()).getType() instanceof ITable)
			if (container instanceof QuantifiedExpression && ((QuantifiedExpression) container).getQuantifier() instanceof ExistentialQuantifier
					&& ((QuantifiedExpression) container).getExpression() instanceof PredicateExpression
					&& ((PredicateExpression) ((QuantifiedExpression) container).getExpression()).getType() instanceof ITable) {

				LinkedHashSet<Variable> variablesContained = ((QuantifiedExpression) contained).getVariables();
				PredicateExpression predicateContained = (PredicateExpression) ((QuantifiedExpression) contained).getExpression();
				ITable tableContained = (ITable) predicateContained.getType();

				LinkedHashSet<Variable> variablesContainer = ((QuantifiedExpression) container).getVariables();
				PredicateExpression predicateContainer = (PredicateExpression) ((QuantifiedExpression) container).getExpression();
				ITable tableContainer = (ITable) predicateContainer.getType();

				for (IConstraint cstr : tableContained.getConstraints())
					if (cstr instanceof ForeignKeyConstraint) {
						ForeignKeyConstraint fk = (ForeignKeyConstraint) cstr;
						if (fk.getDestinationTable().equals(tableContainer)) {
							boolean ok = true;
							INNER1: for (String field : tableContained.getFields())
								if (!fk.getFields().contains(field)) {
									int index = tableContained.getFields().indexOf(field);
									IOperand op = predicateContained.getOperands().get(index);
									if (op instanceof Constant) {
										ok = false;
										break INNER1; // a non-key vars must be one of the quantified variables
									} else if (op instanceof Variable && !variablesContained.contains(op)) {
										ok = false;
										break INNER1; // a non-key vars must be one of the quantified variables
									}
								}

							if (ok)
								INNER2: for (String field : tableContainer.getFields())
									if (!fk.getDestinationFields().contains(field)) {
										int index = tableContainer.getFields().indexOf(field);
										IOperand op = predicateContainer.getOperands().get(index);
										if (op instanceof Constant) {
											ok = false;
											break INNER2; // a non-key vars must be one of the quantified variables
										} else if (op instanceof Variable && !variablesContainer.contains(op)) {
											ok = false;
											break INNER2; // a non-key vars must be one of the quantified variables
										}
									}

							if (ok) {
								Map<Variable, Variable> map = new HashMap<Variable, Variable>();

								for (Variable from : containedAssignment.keySet())
									INNER: for (Variable to : containerAssignment.keySet())
										if (containerAssignment.get(to) == containedAssignment.get(from)) {
											map.put(from, to);
											break INNER;
										}

								// check whether the variables used in the two queries also agree with the mapping

								List<Variable> keyVars = new Vector<Variable>();
								for (Variable var : q1.getOutput())
									keyVars.add(map.get(var));

								List<Variable> candidateVars = q2.getOutput();

								if (candidateVars.equals(keyVars))
									return true;

							}
						}
					}
			}
		return false;
	}

	/**
	 * Check whether q1 is contained in q2 due to a conjunction. Implementation of rule 1.
	 * 
	 * @param q1
	 *            - The contained query.
	 * @param q2
	 *            - The container query.
	 * @return True if q1 \subseteq q2 due to a conjunction. False otherwise.
	 */
	private boolean checkContainmentConjunction(NonBooleanQuery q1, NonBooleanQuery q2) {

		Pair<IExpression, Map<Variable, Integer>> pair1 = normalize(q1);
		IExpression containedExpr = pair1.getFirst();
		Map<Variable, Integer> containedAssignment = pair1.getSecond();

		Pair<IExpression, Map<Variable, Integer>> pair2 = normalize(q2);
		IExpression container = pair2.getFirst();
		Map<Variable, Integer> containerAssignment = pair2.getSecond();

		if (containedExpr instanceof PolyadicExpression && ((PolyadicExpression) containedExpr).getOperator() instanceof AndOperator) {
			for (IExpression expr : ((PolyadicExpression) containedExpr).getExpressions())
				if (expr.equals(container)) {
					Map<Variable, Variable> map = new HashMap<Variable, Variable>();

					for (Variable from : containedAssignment.keySet())
						INNER: for (Variable to : containerAssignment.keySet())
							if (containerAssignment.get(to) == containedAssignment.get(from)) {
								map.put(from, to);
								break INNER;
							}

					// check whether the variables used in the two queries also agree with the mapping

					List<Variable> keyVars = new Vector<Variable>();
					for (Variable var : q1.getOutput())
						keyVars.add(map.get(var));

					List<Variable> candidateVars = q2.getOutput();

					if (candidateVars.equals(keyVars))
						return true;
				}

		}
		return false;
	}

	/**
	 * Checks whether the queries q1 and q2 are equivalent modulo the variable identifiers. Implements rule 4.
	 * 
	 * @param q1
	 *            - The first query.
	 * @param q2
	 *            - The second query.
	 * @return True if q1 can be obtained from q2 by renaming the variable identifiers. False otherwise.
	 */
	private boolean equivalentModuloVariableIdentifiers(NonBooleanQuery q1, NonBooleanQuery q2) {

		Pair<IExpression, Map<Variable, Integer>> pair1 = normalize(q1);
		IExpression key = pair1.getFirst();
		Map<Variable, Integer> keyMap = pair1.getSecond();

		Pair<IExpression, Map<Variable, Integer>> pair2 = normalize(q2);

		IExpression candidate = pair2.getFirst();
		Map<Variable, Integer> candidateMap = pair2.getSecond();

		if (key.equals(candidate)) {
			Map<Variable, Variable> map = new HashMap<Variable, Variable>();

			for (Variable from : keyMap.keySet())
				INNER: for (Variable to : candidateMap.keySet())
					if (candidateMap.get(to) == keyMap.get(from)) {
						map.put(from, to);
						break INNER;
					}

			// check whether the variables used in the two queries also agree with the mapping

			List<Variable> keyVars = new Vector<Variable>();
			for (Variable var : q1.getOutput())
				keyVars.add(map.get(var));

			List<Variable> candidateVars = q2.getOutput();

			if (candidateVars.equals(keyVars)) // the two queries are identical modulo variable identifiers :-)
				return true;

		}
		return false;
	}

	/**
	 * Normalize the expression q.
	 * 
	 * @param q
	 *            - The expression.
	 * @return A pair containing the normalized expression and the mapping from variables to integers.
	 */
	private Pair<IExpression, Map<Variable, Integer>> normalize(NonBooleanQuery q) {
		UniformVariableAssignmentVisitor v = new UniformVariableAssignmentVisitor();
		q.getExpression().accept(v);

		// we remove anonymous views
		IExpression old = null;
		int min = v.getMaximumValue();
		IExpression normalized = v.getRenamedExpression();
		do {
			RemoveAnonymousViewsVisitor vAnonymViews = new RemoveAnonymousViewsVisitor(min);
			normalized.accept(vAnonymViews);

			NormalizeVisitor vNorm1 = new NormalizeVisitor(UniformVariableAssignmentVisitor.getPrefix(), vAnonymViews.getMinValue(),
					PRIVILEGE_MODE.INVOKER);
			vAnonymViews.getTransformed().accept(vNorm1);

			NormalizeVisitor vNorm2 = new NormalizeVisitor(UniformVariableAssignmentVisitor.getPrefix(), vNorm1.getMaximumValue(),
					PRIVILEGE_MODE.OWNER);
			vNorm1.getNormalizedExpression().accept(vNorm2);

			old = normalized;
			normalized = vNorm2.getNormalizedExpression();
			min = vNorm2.getMaximumValue();
		} while (!old.equals(normalized));

		// rename again
		v = new UniformVariableAssignmentVisitor();
		normalized.accept(v);

		return new Pair<IExpression, Map<Variable, Integer>>(v.getRenamedExpression(), v.getAssignment());
	}
}
