package prototype.pdp.secure.newPdp;

import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import prototype.objects.ITable;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.MalformedDbObjectException;
import prototype.objects.PRIVILEGE_MODE;
import prototype.objects.implementation.AnonymousView;
import prototype.privileges.SelectPrivilege;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.NonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.interfaces.IState;
import prototype.users.IUser;
import prototype.util.Util;

/**
 * Visitor that takes a value V (true or false) and by visiting a formula Phi produces a formula Psi such that (1) Psi is secure, and (2) if the value
 * of Psi on the database is V, then the value of Phi on the database is V. Works only for SRNF expressions with incremental identifiers.
 * 
 * 
 *
 */
public class FixedSecureResultVisitor implements IVisitor {

	/**
	 * True if everything is ok.
	 */
	private boolean ok;
	/**
	 * The value.
	 */
	private boolean value;
	/**
	 * The state.
	 */
	private IState state;
	/**
	 * The user.
	 */
	private IUser user;
	/**
	 * The resulting expression.
	 */
	private IExpression expression;

	/**
	 * The set of variables that are range restricted in the resulting formula.
	 */
	private Set<Variable> rangeRestricted;

	/**
	 * If strict is true, then we ignore equality expressions for computing the range restricted variables.
	 */
	private boolean strict = true;
	/**
	 * The set of authorized types.
	 */
	private Set<IType> types;

	/**
	 * Constructor.
	 * 
	 * @param user
	 *            - The user.
	 * @param state
	 *            - The state.
	 * @param flag
	 *            - The value.
	 */
	public FixedSecureResultVisitor(IUser user, IState state, boolean value) {
		if (user != null && state != null && state.getUsers().contains(user)) {
			this.ok = true;
			this.user = user;
			this.state = state;
			this.value = value;

			// compute the authorized types
			types = new HashSet<IType>();
			for (ITable t : state.getTables())
				if (authorized(t))
					types.add(t);

			for (IView v : state.getViews())
				if (authorized(v) && v.getMode() == PRIVILEGE_MODE.OWNER)
					types.add(v);

			try {
				// extend it with the anonymous views
				Set<IType> additionalTypes = new HashSet<IType>();
				for (IType type : types) {

					List<IOperand> operands = null;
					LinkedHashSet<Variable> vars = null;
					if (type instanceof IView) {
						operands = new Vector<IOperand>(((NonBooleanQuery) ((IView) type).getSelectQuery().getQuery()).getOutput());
						vars = new LinkedHashSet<Variable>(((NonBooleanQuery) ((IView) type).getSelectQuery().getQuery()).getOutput());
					} else if (type instanceof ITable) {
						operands = new Vector<IOperand>(((ITable) type).getFreeVariables());
						vars = new LinkedHashSet<Variable>(((ITable) type).getFreeVariables());
					} else if (type instanceof AnonymousView) {
						operands = new Vector<IOperand>((((AnonymousView) type).getSelectQuery()).getOutput());
						vars = new LinkedHashSet<Variable>((((AnonymousView) type).getSelectQuery()).getOutput());
					}

					PredicateExpression pred = new PredicateExpression(type, operands);

					for (Set<Variable> quantVars : Util.powerset(vars))
						if (!quantVars.isEmpty() && quantVars.size() != type.getCardinality()) {
							Set<Variable> freeVars = new LinkedHashSet<Variable>(vars);

							// remove the quantified variables
							freeVars.removeAll(quantVars);

							NonBooleanQuery query;
							if (!quantVars.isEmpty())
								query = new NonBooleanQuery(new Vector<Variable>(freeVars), new QuantifiedExpression(
										ExistentialQuantifier.getInstance(), new LinkedHashSet<Variable>(quantVars), pred));
							else
								query = new NonBooleanQuery(new Vector<Variable>(freeVars), pred);

							AnonymousView v = createAnonymousView(type.getId(), new Vector<Variable>(freeVars), query);
							additionalTypes.add(v);
						}

				}

				types.addAll(additionalTypes);
			} catch (MalformedQueryException | MalformedExpressionException e) {
				e.printStackTrace();
				this.ok = false;
			}
		} else
			this.ok = false;

	}

	private FixedSecureResultVisitor(IUser user, IState state, boolean value, Set<IType> authTypes) {
		this.ok = true;
		this.user = user;
		this.state = state;
		this.value = value;
		this.types = authTypes;
	}

	/**
	 * Get the associated formula.
	 * 
	 * @return The resulting formula.
	 */
	public IExpression getRewrittenFormula() {
		return this.expression;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok) {

			// check containment
			Set<IType> selection = new HashSet<IType>();

			NonBooleanQuery query = constructNonBooleanQuery(expr.getType());

			for (IType type : types) {
				if (value) {
					// check whether type is contained in query
					NonBooleanQuery contained = constructNonBooleanQuery(type);
					if (ContainmentChecker.getInstance().contains(contained, query))
						selection.add(type);

				} else {
					// check whether type contains query
					NonBooleanQuery container = constructNonBooleanQuery(type);
					if (ContainmentChecker.getInstance().contains(query, container))
						selection.add(type);
				}
			}

			Vector<IExpression> exprs = new Vector<IExpression>();
			for (IType type : selection) {
				try {
					PredicateExpression pred = new PredicateExpression(type, expr.getOperands());
					exprs.add(pred);
				} catch (MalformedExpressionException e) {
					assert (false);
				}
			}

			// construct the resulting formula
			if (value)
				try {
					if (exprs.size() > 1)
						this.expression = new PolyadicExpression(OrOperator.getInstance(), exprs);
					else if (exprs.size() == 1)
						this.expression = exprs.get(0);
					else
						this.expression = FalseExpression.getInstance();
				} catch (MalformedExpressionException e) {
					assert (false);
				}
			else
				try {
					if (exprs.size() > 1)
						this.expression = new PolyadicExpression(AndOperator.getInstance(), exprs);
					else if (exprs.size() == 1)
						this.expression = exprs.get(0);
					else
						this.expression = TrueExpression.getInstance();
				} catch (MalformedExpressionException e) {
					assert (false);
				}

			// set the range restricted variables
			this.rangeRestricted = new HashSet<Variable>();
			if (!(expression instanceof TrueExpression || expression instanceof FalseExpression))
				for (IOperand op : expr.getOperands())
					if (op instanceof Variable)
						rangeRestricted.add((Variable) op);
		}
	}

	/**
	 * Create an anonymous view associated with the query.
	 * 
	 * @param id
	 *            - The original type's id.
	 * @param query
	 *            - The query associated with the view.
	 * @return The anonymous view.
	 */
	private AnonymousView createAnonymousView(String id, List<Variable> vars, NonBooleanQuery query) {

		List<String> fields = new Vector<String>();
		String viewId = id;
		for (int i = 0; i < query.getOutput().size(); i++) {
			if (vars != null) {
				if (i == 0)
					viewId = viewId + "_(";
				else
					viewId = viewId + ", ";

				viewId += vars.get(i).toString();
			}

			fields.add(query.getOutput().get(i).toString());

			if (vars != null && i == query.getOutput().size() - 1)
				viewId = viewId + ")";
		}

		try {
			return new AnonymousView(viewId, fields, query);
		} catch (MalformedDbObjectException e) {
			assert (false);
		}
		return null;
	}

	/**
	 * Construct the non-boolean query associated with the given type.
	 * 
	 * @param type
	 *            - The type.
	 * @return The non-boolean query.
	 */
	private NonBooleanQuery constructNonBooleanQuery(IType type) {
		NonBooleanQuery query = null;
		if (type instanceof ITable) {
			Vector<IOperand> operands = new Vector<IOperand>();
			for (Variable var : ((ITable) type).getFreeVariables()) {
				operands.add(var);
			}
			try {
				PredicateExpression pred = new PredicateExpression(type, operands);
				query = new NonBooleanQuery(((ITable) type).getFreeVariables(), pred);
			} catch (MalformedExpressionException | MalformedQueryException e) {
				e.printStackTrace();
				assert (false);
			}
		} else if (type instanceof IView)
			query = (NonBooleanQuery) ((IView) type).getSelectQuery().getQuery();
		else if (type instanceof AnonymousView)
			query = ((AnonymousView) type).getSelectQuery();

		return query;
	}

	/**
	 * Check whether the give user is authorized to read the content of the given type in the given state.
	 * 
	 * @param type
	 *            - The type.
	 * @return True if user is authorized to read the content of type in state. False otherwise.
	 */
	private boolean authorized(IType type) {
		for (IPositiveGrantStatement g : state.getPolicy())
			if (g.getTarget().equals(user) && (g.getPrivilege() instanceof SelectPrivilege)) {
				if (((SelectPrivilege) g.getPrivilege()).getObj().equals(type))
					return true;
			}
		return false;
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok) {
			this.expression = expr;

			rangeRestricted = new HashSet<Variable>();

			if (!strict)
				if (expr.getOp1() instanceof Variable && expr.getOp2() instanceof Constant)
					rangeRestricted.add((Variable) expr.getOp1());
				else if (expr.getOp2() instanceof Variable && expr.getOp1() instanceof Constant)
					rangeRestricted.add((Variable) expr.getOp2());
		}
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, types);
			expr.getExpression().accept(v);

			if (v.rangeRestricted.containsAll(expr.getVariables())) {
				// rewrite the formula
				rangeRestricted = new HashSet<Variable>(v.rangeRestricted);
				rangeRestricted.removeAll(expr.getVariables());

				try {
					this.expression = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.getRewrittenFormula());
				} catch (MalformedExpressionException e) {
					assert (false);
				}

			} else {
				// replace the sub-expression with a constant
				if (value)
					this.expression = FalseExpression.getInstance();
				else
					this.expression = TrueExpression.getInstance();

				rangeRestricted = new HashSet<Variable>();
			}

		}
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, !value, types);
			expr.getExpression().accept(v);
			try {
				this.expression = new UnaryExpression(NotOperator.getInstance(), v.getRewrittenFormula());
			} catch (MalformedExpressionException e) {
				assert (false);
			}
			this.rangeRestricted = new HashSet<Variable>();
		}
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			Vector<IExpression> exprs = new Vector<IExpression>();
			for (IExpression e : expr.getExpressions()) {
				FixedSecureResultVisitor v = new FixedSecureResultVisitor(user, state, value, types);
				e.accept(v);

				exprs.add(v.getRewrittenFormula());
				if (this.rangeRestricted == null)
					this.rangeRestricted = new HashSet<Variable>(v.rangeRestricted);
				else if (expr.getOperator() instanceof AndOperator)
					this.rangeRestricted.addAll(v.rangeRestricted);
				else if (expr.getOperator() instanceof OrOperator)
					this.rangeRestricted.retainAll(v.rangeRestricted);
			}

			try {
				this.expression = new PolyadicExpression(expr.getOperator(), exprs);
			} catch (MalformedExpressionException e) {
				assert (false);
			}

		}
	}

	@Override
	public void visit(FalseExpression expr) {
		if (ok)
			this.expression = expr.copy();
	}

	@Override
	public void visit(TrueExpression expr) {
		if (ok)
			this.expression = expr.copy();
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

}
