package prototype.queries.domain.relational.calculus;

import prototype.interfaces.CopyException;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.visitors.FreeVariableVisitor;
import prototype.queries.domain.relational.calculus.visitors.HasUniqueIdentifiersVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.domain.relational.calculus.visitors.OneStepSrnfTransformationVisitor;
import prototype.queries.domain.relational.calculus.visitors.RangeRestrictedVisitor;
import prototype.queries.domain.relational.calculus.visitors.ReplacePredicatesWithConstantsVisitor;

/**
 * Class representing boolean queries in the domain relational calculus.
 * 
 * 
 *
 */
public class BooleanQuery implements IQuery {

	/**
	 * The expression representing the query.
	 */
	private IExpression expression;

	/**
	 * Constructor.
	 * 
	 * @param expression
	 *            - The expression representing the query.
	 * @throws MalformedQueryException
	 *             - if expression is null or is not a boolean expression.
	 */
	public BooleanQuery(IExpression expression) throws MalformedQueryException {
		if (expression != null) {
			FreeVariableVisitor v = new FreeVariableVisitor();
			expression.accept(v);

			if (v.getFreeVariables().isEmpty()) {
				HasUniqueIdentifiersVisitor v1 = new HasUniqueIdentifiersVisitor();
				expression.accept(v1);
				if (v1.hasUniqueIdentifiers()) {

					// write the expression using unique incremental identifiers
					IncrementalIdentifiers vInc = new IncrementalIdentifiers();
					expression.accept(vInc);

					IExpression withIncrIds = vInc.getRenamedExpression();

					// replace all the occurrences of P(v1,...,vn,c_1, ...,c_m), where c_i are constants) with Exists x_1, ..., x_m. P(v1,...,vn, x_1,
					// ..., x_m) AND x_1 = c_1 AND ... AND x_m = c_m
					ReplacePredicatesWithConstantsVisitor vRepl = new ReplacePredicatesWithConstantsVisitor(IncrementalIdentifiers.getPrefix(), vInc.getMaximumValue());
					withIncrIds.accept(vRepl);
					
					IExpression replaced = vRepl.getTransformed();

					// check whether the expression is range restricted.
					try {
						IExpression srnf = isRangeRestricted(replaced);
						if (srnf != null)
							this.expression = srnf.copy();
						else
							throw new MalformedQueryException("The expression is not range restricted.");
					} catch (MalformedExpressionException e) {
						throw new MalformedQueryException(e);
					}
				} else
					throw new MalformedQueryException(
							"The expression associated with the query declares the same variable identifier more than once.");
			} else
				throw new MalformedQueryException("The expression associated with a boolean query cannot have free variables.");

		} else
			throw new MalformedQueryException("The expression associated with the query cannot be null");
	}

	/**
	 * Get the expression associated with the query.
	 * 
	 * @return The expression.
	 */
	public IExpression getExpression() {
		return expression;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return expression.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof BooleanQuery) {
			if (obj == this)
				return true;
			else
				return ((BooleanQuery) obj).getExpression().equals(this.getExpression());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getExpression().hashCode();
		return result;
	}

	/**
	 * Check whether an expression (with unique variable identifiers) is range restricted.
	 * 
	 * @param expr
	 *            - The expression (not null).
	 * @return The formula in Safe-Range Normal Form if expr is range restricted, null otherwise.
	 * @throws MalformedExpressionException
	 */
	private IExpression isRangeRestricted(IExpression expr) throws MalformedExpressionException {

		IExpression transformedOld = null;
		IExpression srnf = expr.copy();
		do {
			transformedOld = srnf;
			OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
			transformedOld.accept(v);
			srnf = v.getTransformed();
		} while (!transformedOld.equals(srnf));

		// here transformedNew contains a formula equivalent to expr which is in Safe-Range Normal Form.
		RangeRestrictedVisitor v = new RangeRestrictedVisitor();
		srnf.accept(v);

		if (v.getRangeRestricted() != null && v.getRangeRestricted().isEmpty())
			return srnf;
		else
			return null;

	}

	@Override
	public IQuery copy() throws CopyException {
		try {
			return new BooleanQuery(getExpression().copy());
		} catch (MalformedQueryException | MalformedExpressionException e) {
			throw new CopyException(e);
		}
	}

}
