package prototype.queries.domain.relational.calculus;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Vector;

import prototype.interfaces.CopyException;
import prototype.pdp.secure.UniformVariableAssignmentVisitor;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.FreeVariableVisitor;
import prototype.queries.domain.relational.calculus.visitors.HasUniqueIdentifiersVisitor;
import prototype.queries.domain.relational.calculus.visitors.IncrementalIdentifiers;
import prototype.queries.domain.relational.calculus.visitors.OneStepSrnfTransformationVisitor;
import prototype.queries.domain.relational.calculus.visitors.RangeRestrictedVisitor;
import prototype.queries.domain.relational.calculus.visitors.ReplacePredicatesWithConstantsVisitor;
import prototype.util.Util;

/**
 * Class representing non-boolean queries in the domain relational calculus.
 * 
 * 
 *
 */
public class NonBooleanQuery implements INonBooleanQuery {

	/**
	 * The expression representing the query.
	 */
	private IExpression expression;

	/**
	 * The output variables.
	 */
	private List<Variable> output;

	/**
	 * Constructor.
	 * 
	 * @param output
	 *            - the output variables
	 * @param expression
	 *            - The expression representing the query.
	 * @throws MalformedQueryException
	 * 
	 */
	public NonBooleanQuery(List<Variable> output, IExpression expression) throws MalformedQueryException {
		if (expression != null) {
			FreeVariableVisitor v = new FreeVariableVisitor();
			expression.accept(v);

			HasUniqueIdentifiersVisitor v1 = new HasUniqueIdentifiersVisitor(new LinkedHashSet<Variable>(v.getFreeVariables()));
			expression.accept(v1);
			if (v1.hasUniqueIdentifiers()) {
				if (output != null ) {
					if (v.getFreeVariables().containsAll(output) && output.containsAll(v.getFreeVariables())) {

						// write the expression using unique incremental identifiers
						//IncrementalIdentifiers vInc = new IncrementalIdentifiers();
						UniformVariableAssignmentVisitor vInc = new UniformVariableAssignmentVisitor();
						expression.accept(vInc);

						IExpression withIncrIds = vInc.getRenamedExpression();
						
						// replace all the occurrences of P(v1,...,vn,c_1, ...,c_m), where c_i are constants) with Exists x_1, ..., x_m. P(v1,...,vn, x_1,
						// ..., x_m) AND x_1 = c_1 AND ... AND x_m = c_m
						ReplacePredicatesWithConstantsVisitor vRepl = new ReplacePredicatesWithConstantsVisitor(IncrementalIdentifiers.getPrefix(), vInc.getMaximumValue());
						withIncrIds.accept(vRepl);
						
						IExpression replaced = vRepl.getTransformed();

						// check whether the expression is range restricted.
						try {
							IExpression srnf = isRangeRestricted(replaced);
							if (srnf != null) {
								this.expression = srnf.copy();

								this.output = new Vector<Variable>();
								for (Variable var : output)
									this.output.add(new Variable(IncrementalIdentifiers.getPrefix() + vInc.getAssignment().get(var)));

							} else
								throw new MalformedQueryException("The expression is not range restricted.");
						} catch (MalformedExpressionException e) {
							throw new MalformedQueryException(e);
						}
					} else
						throw new MalformedQueryException("The only free variable in expression must be the output variables.");
				} else
					throw new MalformedQueryException("The output variable cannot be null.");
			} else
				throw new MalformedQueryException("The expression associated with the query declares the same variable identifier more than once.");

		} else
			throw new MalformedQueryException("The expression associated with the query cannot be null");
	}

	/**
	 * Get the expression associated with the query.
	 * 
	 * @return The expression.
	 */
	public IExpression getExpression() {
		return expression;
	}

	/**
	 * Get the output variables.
	 * 
	 * @return The output variables.
	 */
	public List<Variable> getOutput() {
		return output;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {

		String vars = "";
		for (Variable var : output)
			vars += var + ", ";
		vars = Util.replaceLast(vars, ", ", "");

		return "{" + vars + " | " + expression.toString() + "}";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof NonBooleanQuery) {
			if (obj == this)
				return true;
			else
				return ((NonBooleanQuery) obj).getExpression().equals(this.getExpression())
						&& ((NonBooleanQuery) obj).getOutput().equals(this.getOutput());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getExpression().hashCode();
		result = 31 * result + getOutput().hashCode();
		return result;
	}

	/**
	 * Check whether an expression (with unique variable identifiers) is range restricted.
	 * 
	 * @param expr
	 *            - The expression (not null).
	 * @return The Safe-Range Normal Form equivalent expression if expr is range restricted, null otherwise.
	 * @throws MalformedExpressionException
	 */
	private IExpression isRangeRestricted(IExpression expr) throws MalformedExpressionException {

		IExpression transformedOld = null;
		IExpression srnf = expr.copy();
		do {
			transformedOld = srnf;
			OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
			transformedOld.accept(v);
			srnf = v.getTransformed();
		} while (!transformedOld.equals(srnf));

		// here transformedNew contains a formula equivalent to expr which is in Safe-Range Normal Form.
		RangeRestrictedVisitor v = new RangeRestrictedVisitor();
		srnf.accept(v);

		FreeVariableVisitor v1 = new FreeVariableVisitor();
		expr.accept(v1);

		if (v.getRangeRestricted() != null && v.getRangeRestricted().equals(v1.getFreeVariables()))
			return srnf;
		else
			return null;

	}

	@Override
	public INonBooleanQuery copy() throws CopyException {
		try {
			List<Variable> copyOutput = new Vector<Variable>();
			for (Variable v : getOutput())
				copyOutput.add((Variable) v.copy());
			return new NonBooleanQuery(output, getExpression().copy());
		} catch (MalformedExpressionException | MalformedQueryException e) {
			throw new CopyException(e);
		}
	}

}
