package prototype.queries.domain.relational.calculus.syntax.expressions;

import java.util.List;
import java.util.Vector;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.Operator;
import prototype.util.Util;

/**
 * Class representing a binary expression of the form expr1 op expr2 op ... op exprN
 * 
 * 
 *
 */
public class PolyadicExpression implements IExpression {

	/**
	 * List of expressions.
	 */
	private List<IExpression> expr;

	/**
	 * The binary operator.
	 */
	private Operator op;

	/**
	 * Constructor.
	 * 
	 * @param op
	 *            - Polyadic operator
	 * @param expressions
	 *            - List of expressions
	 * @throws MalformedExpressionException
	 */
	public PolyadicExpression(Operator op, List<IExpression> expressions) throws MalformedExpressionException {
		if (expressions == null || expressions.isEmpty())
			throw new MalformedExpressionException("Expressions cannot be null or empty.");
		else if (op == null)
			throw new MalformedExpressionException("Op cannot be null.");
		else if (expressions.size() < op.getMinArity())
			throw new MalformedExpressionException("The number of expressions is less than the minimum arity of the operator.");
		else if (op.getMaxArity() < expressions.size())
			throw new MalformedExpressionException("The number of expressions is more than the maximum arity of the operator.");
		else {
			for (IExpression e : expressions)
				if (e == null)
					throw new MalformedExpressionException("Expressions cannot contain null values");

			this.op = op;
			this.expr = expressions;
		}
	}

	/**
	 * Get the expression's list.
	 * 
	 * @return The expression's list.
	 */
	public List<IExpression> getExpressions() {
		return expr;
	}

	/**
	 * Get the operator.
	 * 
	 * @return The operator.
	 */
	public Operator getOperator() {
		return op;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String res = "(";
		String op = " " + this.op.toString() + " ";
		for (IExpression e : this.expr)
			res += e + op;
		res = Util.replaceLast(res, op, "");
		return res + ")";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IDrcVisitable#accept(guarnieri.marco.ethz.queries.visitors.IDrcVisitor)
	 */
	@Override
	public void accept(IVisitor visitor) {
		visitor.visit(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof PolyadicExpression) {
			if (obj == this)
				return true;
			else
				return ((PolyadicExpression) obj).getExpressions().equals(this.getExpressions())
						&& ((PolyadicExpression) obj).getOperator().equals(this.getOperator());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getOperator().hashCode();
		result = 31 * result + getExpressions().hashCode();
		return result;
	}

	@Override
	public PolyadicExpression copy() throws MalformedExpressionException {
		List<IExpression> copiedExprs = new Vector<IExpression>();
		for (IExpression e : expr)
			copiedExprs.add(e.copy());
		return new PolyadicExpression(op, copiedExprs);
	}
}
