package prototype.queries.domain.relational.calculus.syntax.expressions;

import java.util.LinkedHashSet;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.Operator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.util.Util;

/**
 * Class representing expressions of the form QUANTIFIER var \in T. Expr
 * 
 * 
 *
 */
public class QuantifiedExpression implements IExpression {

	/**
	 * The quantifier.
	 */
	private Operator quantifier;
	/**
	 * The list of variable.
	 */
	private LinkedHashSet<Variable> variables;

	/**
	 * The expression.
	 */
	private IExpression expression;

	/**
	 * Constructor.
	 * 
	 * @param quantifier
	 *            - the quantifier
	 * @param variable
	 *            - the quantified variable.
	 * 
	 * @param expression
	 *            - the expression.
	 * @throws MalformedExpressionException
	 */
	public QuantifiedExpression(Operator quantifier, LinkedHashSet<Variable> variables, IExpression expression) throws MalformedExpressionException {
		if (quantifier == null)
			throw new MalformedExpressionException("Quantifier cannot be null.");
		else if (!(quantifier instanceof ExistentialQuantifier) && !(quantifier instanceof UniversalQuantifier))
			throw new MalformedExpressionException("Quantifier must be either an existential quantifier or a universal quantifier.");
		else if (variables == null || variables.isEmpty())
			throw new MalformedExpressionException("Variables cannot be null or empty.");
		else if (expression == null)
			throw new MalformedExpressionException("Expression cannot be null.");
		else {
			this.quantifier = quantifier;
			this.variables = variables;
			this.expression = expression;
		}
	}

	/**
	 * Get the quantifier.
	 * 
	 * @return The quantifier.
	 */
	public Operator getQuantifier() {
		return quantifier;
	}

	/**
	 * Get the variables.
	 * 
	 * @return The variables.
	 */
	public LinkedHashSet<Variable> getVariables() {
		return variables;
	}

	/**
	 * Get the expression.
	 * 
	 * @return the expression.
	 */
	public IExpression getExpression() {
		return expression;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String vars = "";

		for (Variable v : this.getVariables())
			vars += v + ", ";

		vars= Util.replaceLast(vars, ", ", "");

		return quantifier + " " + vars + ". (" + expression + ")";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IDrcVisitable#accept(guarnieri.marco.ethz.queries.visitors.IDrcVisitor)
	 */
	@Override
	public void accept(IVisitor visitor) {
		visitor.visit(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof QuantifiedExpression) {
			if (obj == this)
				return true;
			else
				return ((QuantifiedExpression) obj).getVariables().equals(this.getVariables())
						&& ((QuantifiedExpression) obj).getQuantifier().equals(this.getQuantifier())
						&& ((QuantifiedExpression) obj).getExpression().equals(this.getExpression());
		} else
			return false;
	}
	
	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getVariables().hashCode();
		result = 31 * result + getExpression().hashCode();
		result = 31 * result + getQuantifier().hashCode();
		return result;
	}
	
	@Override
	public QuantifiedExpression copy() throws MalformedExpressionException {
		LinkedHashSet<Variable> copiedVars = new LinkedHashSet<Variable>();
		for(Variable var : variables)
			copiedVars.add((Variable) var.copy());
		return new QuantifiedExpression(quantifier, copiedVars, expression.copy());
	}
}
