package prototype.queries.domain.relational.calculus.visitors;

import java.util.HashSet;
import java.util.Set;

import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that retrieves the set of all free variables.
 * 
 * 
 *
 */
public class FreeVariableVisitor implements IVisitor {

	/**
	 * The set of free variables.
	 */
	private Set<Variable> freeVariables;

	/**
	 * Constructor.
	 */
	public FreeVariableVisitor() {
		freeVariables = new HashSet<Variable>();
	}

	/**
	 * Get the set of free variables.
	 * 
	 * @return The set of free variables.
	 */
	public Set<Variable> getFreeVariables() {
		return this.freeVariables;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * guarnieri.marco.ethz.queries.domain.relational.calculus.visitors.IDrcVisitor#visit(guarnieri.marco.ethz.queries.domain.relational.calculus.
	 * syntax.expressions.PredicateExpression)
	 */
	@Override
	public void visit(PredicateExpression expr) {
		for (IOperand o : expr.getOperands()) {
			FreeVariableVisitor v = new FreeVariableVisitor();

			o.accept(v);

			this.freeVariables.addAll(v.getFreeVariables());
		}

	}

	// /*
	// * (non-Javadoc)
	// *
	// * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.BinaryExpression)
	// */
	// @Override
	// public void visit(BinaryExpression expr) {
	// DrcFreeVariableVisitor v1 = new DrcFreeVariableVisitor();
	// DrcFreeVariableVisitor v2 = new DrcFreeVariableVisitor();
	//
	// expr.getExpression1().accept(v1);
	// expr.getExpression2().accept(v2);
	//
	// this.freeVariables.addAll(v1.getFreeVariables());
	// this.freeVariables.addAll(v2.getFreeVariables());
	// }

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * guarnieri.marco.ethz.queries.domain.relational.calculus.visitors.IDrcVisitor#visit(guarnieri.marco.ethz.queries.domain.relational.calculus.
	 * syntax.expressions.PolyadicExpression)
	 */
	@Override
	public void visit(PolyadicExpression expr) {
		for (IExpression e : expr.getExpressions()) {
			FreeVariableVisitor v = new FreeVariableVisitor();

			e.accept(v);

			this.freeVariables.addAll(v.getFreeVariables());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.EqualityExpression)
	 */
	@Override
	public void visit(EqualityExpression expr) {
		FreeVariableVisitor v1 = new FreeVariableVisitor();
		FreeVariableVisitor v2 = new FreeVariableVisitor();

		expr.getOp1().accept(v1);
		expr.getOp2().accept(v2);

		this.freeVariables.addAll(v1.getFreeVariables());
		this.freeVariables.addAll(v2.getFreeVariables());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.QuantifiedExpression)
	 */
	@Override
	public void visit(QuantifiedExpression expr) {
		FreeVariableVisitor v = new FreeVariableVisitor();

		expr.getExpression().accept(v);

		// add all the free variables in expr
		this.freeVariables.addAll(v.getFreeVariables());

		// remove expr.getVariable() from the set of free variables
		this.freeVariables.removeAll(expr.getVariables());

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.UnaryExpression)
	 */
	@Override
	public void visit(UnaryExpression expr) {
		FreeVariableVisitor v = new FreeVariableVisitor();

		expr.getExpression().accept(v);

		this.freeVariables.addAll(v.getFreeVariables());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.Variable)
	 */
	@Override
	public void visit(Variable var) {
		this.freeVariables.add(var);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.expressions.Constant)
	 */
	@Override
	public void visit(Constant expr) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.AndOperator)
	 */
	@Override
	public void visit(AndOperator op) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.Equality)
	 */
	@Override
	public void visit(Equality op) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.ExistentialQuantifier)
	 */
	@Override
	public void visit(ExistentialQuantifier op) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.NotOperator)
	 */
	@Override
	public void visit(NotOperator op) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.OrOperator)
	 */
	@Override
	public void visit(OrOperator op) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.queries.visitors.IVisitor#visit(guarnieri.marco.ethz.queries.syntax.operators.UniversalQuantifier)
	 */
	@Override
	public void visit(UniversalQuantifier op) {
		// do nothing
	}

	@Override
	public void visit(FalseExpression expr) {
		// do nothing
	}

	@Override
	public void visit(TrueExpression expr) {
		// do nothing
	}

}
