package prototype.queries.domain.relational.calculus.visitors;

import java.util.List;
import java.util.Map;
import java.util.Vector;

import prototype.objects.IType;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.util.Util;

/**
 * Visitor that translates a Safe-Range Normal Form domain-relational calculus query to a SQL query.
 * 
 * 
 *
 */
public class MySqlTranslator implements IVisitor {

	/**
	 * The sql expression.
	 */
	private String sqlExpression;

	/**
	 * List of all types
	 */
	private List<IType> types;

	/**
	 * Map assigning variables to names.
	 */
	private Map<Variable, String> variablesToSqlNames;

	// /**
	// * Equalities' list.
	// */
	// private List<String> equalities;

	/**
	 * The nesting level.
	 */
	private int nestingLevel;

	/**
	 * The isPositive flag.
	 */
	private boolean isPositive;

	// /**
	// * True if the translation should be in the boolean semantics or in the set semantics. The flag is needed for the translation.
	// */
	// private boolean booleanSemantics;

	public MySqlTranslator(Map<Variable, String> variablesToSqlNames) {
		this.sqlExpression = "";
		this.types = new Vector<IType>();
		this.variablesToSqlNames = variablesToSqlNames;
		this.nestingLevel = 0;
		this.isPositive = true;

	}

	// public MySqlTranslator(Map<Variable, String> variablesToSqlNames, boolean booleanSemantics) {
	// this.sqlExpression = "";
	// this.types = new Vector<IType>();
	// this.variablesToSqlNames = variablesToSqlNames;
	// this.nestingLevel = 0;
	// this.isPositive = true;
	//
	// }

	private MySqlTranslator(List<IType> types, Map<Variable, String> variablesToSqlNames, int nestingLevel, boolean isPositive) {
		this.sqlExpression = "";
		this.types = types;
		this.variablesToSqlNames = variablesToSqlNames;
		this.nestingLevel = nestingLevel;
		this.isPositive = isPositive;
	}

	@Override
	public void visit(PredicateExpression expr) {

		// add to types
		this.types.add(expr.getType());

		// compute how many occurrences of the type expr.getType() are in typesToSqlNames
		int i = 0;
		for (IType type : types)
			if (type.equals(expr.getType()))
				i++;
		List<String> equalities = new Vector<String>();
		// check whether the variables have been already assigned to names
		for (int j = 0; j < expr.getOperands().size(); j++) {
			String var = getTableIdentifier(expr.getType(), nestingLevel, i) + "." + expr.getType().getFields().get(j);
			if (expr.getOperands().get(j) instanceof Variable) {
				// the j-th operand is a variable
				if (variablesToSqlNames.containsKey(expr.getOperands().get(j))) {
					// add the equality. The variable represents a join.
					String op = "=";

					if (!var.equals(variablesToSqlNames.get(expr.getOperands().get(j)))) {
						String equality = var + " " + op + " " + variablesToSqlNames.get(expr.getOperands().get(j));
						equalities.add(equality);
					}

				}
			} else {
				// the j-th operand is a constant. This introduces an equality

				String op = "=";

				equalities.add(var + " " + op + " " + expr.getOperands().get(j).toString());
			}
		}

		if (!equalities.isEmpty()) {
			this.sqlExpression = "(";
			for (String eq : equalities) {
				sqlExpression += eq + " AND ";
			}
			sqlExpression = Util.replaceLast(sqlExpression, " AND ", ")");
		} else
			this.sqlExpression = "";

	}

	private String getTableIdentifier(IType type, int nestingLevel, int i) {
		return MySqlVariableVisitor.getTableIdentifier(type, nestingLevel, i);
	}

	@Override
	public void visit(PolyadicExpression expr) {

		this.sqlExpression = "(";

		String op = "";

		if (expr.getOperator() instanceof AndOperator)
			op = "AND";
		else if (expr.getOperator() instanceof OrOperator)
			op = "OR";
		else
			assert (false); // unreachable

		for (IExpression e : expr.getExpressions()) {
			MySqlTranslator v = new MySqlTranslator(new Vector<IType>(this.types), this.variablesToSqlNames, this.nestingLevel, this.isPositive);
			e.accept(v);

			for (int i = this.types.size(); i < v.getTypes().size(); i++)
				// temp.add(v.getTypes().get(i));
				types.add(v.getTypes().get(i));
			if (!v.getSqlExpression().isEmpty())
				this.sqlExpression += v.getSqlExpression() + " " + op + " ";
		}
		sqlExpression = Util.replaceLast(sqlExpression, " " + op + " ", "");
		sqlExpression += ")";
		// types = temp;
	}

	@Override
	public void visit(EqualityExpression expr) {

		String op = "=";
		// if (isPositive)
		// // op is =
		// op = "=";
		// else
		// // op is !=
		// op = "!=";

		if (expr.getOp1() instanceof Variable && expr.getOp2() instanceof Variable) {

			if (variablesToSqlNames.containsKey(expr.getOp1())) {
				if (variablesToSqlNames.containsKey(expr.getOp2())) {
					this.sqlExpression = variablesToSqlNames.get(expr.getOp1()) + " " + op + " " + variablesToSqlNames.get(expr.getOp2());
				} else
					assert (false); // the variable should always be in the map
			} else
				assert (false); // the variable should always be in the map

		} else if (expr.getOp1() instanceof Variable && expr.getOp2() instanceof Constant) {

			if (variablesToSqlNames.containsKey(expr.getOp1())) {
				this.sqlExpression = variablesToSqlNames.get(expr.getOp1()) + " " + op + " " + expr.getOp2().toString();
			} else
				assert (false); // the variable should always be in the map

		} else if (expr.getOp1() instanceof Constant && expr.getOp2() instanceof Variable) {

			if (variablesToSqlNames.containsKey(expr.getOp2())) {
				this.sqlExpression = expr.getOp1().toString() + " " + op + " " + variablesToSqlNames.get(expr.getOp2());
			} else
				assert (false); // the variable should always be in the map

		} else if (expr.getOp1() instanceof Constant && expr.getOp2() instanceof Constant) {
			this.sqlExpression = expr.getOp1().toString() + " " + op + " " + expr.getOp2().toString();
		} else
			assert (false); // unreachable
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (expr.getQuantifier() instanceof ExistentialQuantifier) {

			MySqlGetTypes v1 = new MySqlGetTypes();
			expr.getExpression().accept(v1);

			List<IType> newTypes = new Vector<IType>();
			// if (this.booleanSemantics) //// let's try to always get the empty vector.
			// newTypes = new Vector<IType>();
			// else
			// newTypes = this.types;

			MySqlTranslator v = new MySqlTranslator(newTypes, this.variablesToSqlNames, this.nestingLevel + 1, true);
			expr.getExpression().accept(v);

			List<IType> usedTypes = v.getTypes();
			// List<String> equals = v.getEqualities();

			String selectedVars = "";
			for (Variable var : expr.getVariables()) {
				selectedVars += variablesToSqlNames.get(var) + " , ";
			}
			selectedVars = Util.replaceLast(selectedVars, " , ", "");

			String tables = "";
			for (int i = 0; i < usedTypes.size(); i++) {
				if (i != 0)
					tables += " , ";

				int number = 0;
				for (int j = 0; j <= i; j++)
					if (usedTypes.get(j).equals(usedTypes.get(i)))
						number++;

				tables += usedTypes.get(i).getId() + " AS " + getTableIdentifier(usedTypes.get(i), nestingLevel + 1, number);
			}

			String whereExpr = "";
			// for (String eq : equals)
			// whereExpr += eq + " AND ";
			if (v.getSqlExpression() != null && !v.getSqlExpression().isEmpty())
				whereExpr += v.getSqlExpression();
			else
				whereExpr = Util.replaceLast(whereExpr, " AND ", "");

			this.sqlExpression = "EXISTS ( SELECT DISTINCT " + selectedVars + " FROM " + tables + " WHERE " + whereExpr + ")";

			this.types.addAll(usedTypes);
		} else if (expr.getQuantifier() instanceof UniversalQuantifier) {
			assert (false);// unreachable for Safe-Range Normal Form expressions
		} else {
			assert (false);// unreachable
		}

	}

	@Override
	public void visit(UnaryExpression expr) {

		MySqlTranslator v = new MySqlTranslator(this.types, this.variablesToSqlNames, this.nestingLevel, !isPositive);
		expr.getExpression().accept(v);

		if (!v.getSqlExpression().isEmpty())
			this.sqlExpression = "NOT (" + v.getSqlExpression() + ")";
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	/**
	 * Get all the types.
	 * 
	 * @return The types.
	 */
	public List<IType> getTypes() {
		return types;
	}

	/**
	 * Get the mapping from variables to names.
	 * 
	 * @return The mapping.
	 */
	public Map<Variable, String> getVariablesToSqlNames() {
		return variablesToSqlNames;
	}

	/**
	 * Get the SQL expression corresponding to the visited expression.
	 * 
	 * @return The translation.
	 */
	public String getSqlExpression() {
		return sqlExpression;
	}

	@Override
	public void visit(FalseExpression expr) {
		this.sqlExpression = expr.toString();
	}

	@Override
	public void visit(TrueExpression expr) {
		this.sqlExpression = expr.toString();
	}
}
