package prototype.queries.domain.relational.calculus.visitors;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import prototype.objects.IType;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that assigns to each variable in a positive Predicate a unique MySql identifier. Works only for SRNF expressions (with unique variable
 * identifiers).
 * 
 * 
 *
 */
public class MySqlVariableVisitor implements IVisitor {

	/**
	 * Map assigning variables to names.
	 */
	private Map<Variable, String> variablesToSqlNames;
	/**
	 * The nesting level.
	 */
	private int nestingLevel;

	/**
	 * List of all types
	 */
	private List<IType> types;

	/**
	 * The isPositive flag.
	 */
	private boolean isPositive;

	private Map<Variable, Integer> varToNestingLevel;

	// public MySqlVariableVisitor() {
	// this.variablesToSqlNames = new HashMap<Variable, String>();
	// this.nestingLevel = 0;
	// this.isPositive = true;
	// this.types = new Vector<IType>();
	// }

	private MySqlVariableVisitor(Map<Variable, String> variablesToSqlNames, int nestingLevel, boolean isPositive, List<IType> types,
			Map<Variable, Integer> varToNestingLevel) {
		this.variablesToSqlNames = variablesToSqlNames;
		this.nestingLevel = nestingLevel;
		this.isPositive = isPositive;
		this.types = types;
		this.varToNestingLevel = varToNestingLevel;

	}

	public MySqlVariableVisitor(Map<Variable, Integer> varToNestingLevel) {
		this.variablesToSqlNames = new HashMap<Variable, String>();
		this.nestingLevel = 0;
		this.isPositive = true;
		this.types = new Vector<IType>();
		this.varToNestingLevel = varToNestingLevel;
	}

	@Override
	public void visit(PredicateExpression expr) {
		// add to types
		this.types.add(expr.getType());

		if (isPositive) {
			// if the predicate is positive, then we can assign the name

			// compute how many occurrences of the type expr.getType() are in typesToSqlNames
			int i = 0;
			for (IType type : types)
				if (type.equals(expr.getType()))
					i++;

			// check whether the variables have been already assigned to names
			for (int j = 0; j < expr.getOperands().size(); j++)
				if (expr.getOperands().get(j) instanceof Variable) {
					// the j-th operand is a variable
					if (!variablesToSqlNames.containsKey(expr.getOperands().get(j))) {

						// there is no name yet. Add the new name

						if (varToNestingLevel.containsKey(expr.getOperands().get(j)))
							variablesToSqlNames.put((Variable) expr.getOperands().get(j),
									getTableIdentifier(expr.getType(), varToNestingLevel.get(expr.getOperands().get(j)).intValue(), i) + "."
											+ expr.getType().getFields().get(j));
						else
							variablesToSqlNames.put((Variable) expr.getOperands().get(j), getTableIdentifier(expr.getType(), nestingLevel, i) + "."
									+ expr.getType().getFields().get(j));

					}
				}
		}

	}

	@Override
	public void visit(EqualityExpression expr) {
		// do nothing
	}
	
	@Override
	public void visit(FalseExpression expr) {
		// do nothing
	}
	
	@Override
	public void visit(TrueExpression expr) {
		// do nothing
	}

	@Override
	public void visit(QuantifiedExpression expr) {

		MySqlGetTypes v1 = new MySqlGetTypes();
		expr.getExpression().accept(v1);

		boolean booleanSemantics;

		if (v1.getTypes().isEmpty())
			// we are in the set semantics
			booleanSemantics = false;
		else
			// we are in the boolean semantics
			booleanSemantics = true;

		Map<Variable, Integer> freeVars;
		if (booleanSemantics)
			freeVars = new HashMap<Variable, Integer>(this.varToNestingLevel);
		else {
			freeVars = new HashMap<Variable, Integer>(this.varToNestingLevel);
			for (Variable v : expr.getVariables())
				freeVars.put(v, this.nestingLevel + 1);
		}
		MySqlVariableVisitor v = new MySqlVariableVisitor(variablesToSqlNames, nestingLevel + 1, true, new Vector<IType>(), freeVars);
		expr.getExpression().accept(v);
	}

	@Override
	public void visit(UnaryExpression expr) {
		MySqlVariableVisitor v = new MySqlVariableVisitor(variablesToSqlNames, nestingLevel, !isPositive, types, this.varToNestingLevel);
		expr.getExpression().accept(v);
	}

	@Override
	public void visit(PolyadicExpression expr) {
		for (IExpression expression : expr.getExpressions()) {
			MySqlVariableVisitor v = new MySqlVariableVisitor(variablesToSqlNames, nestingLevel, isPositive, types, new HashMap<Variable, Integer>(
					this.varToNestingLevel));
			expression.accept(v);
		}
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	/**
	 * Get the table identifier.
	 * 
	 * @param type
	 *            - The type.
	 * @param nestingLevel
	 *            - the nesting level
	 * @param occurrence
	 *            - the occurrence.
	 * @return The id.
	 */
	public static String getTableIdentifier(IType type, int nestingLevel, int occurrence) {
		return type.getId() + "_" + nestingLevel + "_" + occurrence;
	}

	/**
	 * Get the mapping variables to names.
	 * 
	 * @return The mapping variables to names.
	 */
	public Map<Variable, String> getVariablesToSqlNames() {
		return variablesToSqlNames;
	}

}