package prototype.queries.domain.relational.calculus.visitors;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Vector;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that implements the rewriting rules to transform a formula in Safe-Range Normal Form. Note that the visitor assumes that the variable
 * identifiers are unique.
 * 
 * The visitor visits a formula and builds the formula obtained by applying once the rewriting rules for the SRNF
 * 
 * 
 *
 */
public class OneStepSrnfTransformationVisitor implements IVisitor {

	/**
	 * The resulting transformed formula.
	 */
	private IExpression transformed;

	@Override
	public void visit(PredicateExpression expr) {
		try {
			this.transformed = expr.copy();
		} catch (MalformedExpressionException e) {
			assert (false); // UNREACHABLE
		}
	}

	@Override
	public void visit(EqualityExpression expr) {
		try {
			this.transformed = expr.copy();
		} catch (MalformedExpressionException e) {
			assert (false); // UNREACHABLE
		}
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (expr.getQuantifier() instanceof ExistentialQuantifier) {
			if (expr.getExpression() instanceof QuantifiedExpression
					&& ((QuantifiedExpression) expr.getExpression()).getQuantifier() instanceof ExistentialQuantifier) {
				// flatten existential quantifiers

				try {
					QuantifiedExpression exprCopy = expr.copy();
					QuantifiedExpression nestedCopy = (QuantifiedExpression) exprCopy.getExpression();
					LinkedHashSet<Variable> newVars = expr.copy().getVariables();
					// merge the variables
					newVars.addAll(nestedCopy.getVariables());

					// modify the sub-sub expression
					OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
					nestedCopy.getExpression().accept(v);

					this.transformed = new QuantifiedExpression(ExistentialQuantifier.getInstance(), newVars, v.getTransformed());
				} catch (MalformedExpressionException e) {
					assert (false); // UNREACHABLE
				}

			} else
				// no rule can be applied to expr. Just modify the subexpression.
				try {
					QuantifiedExpression exprCopy = expr.copy();

					OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
					exprCopy.getExpression().accept(v);

					this.transformed = new QuantifiedExpression(ExistentialQuantifier.getInstance(), exprCopy.getVariables(), v.getTransformed());
				} catch (MalformedExpressionException e) {
					assert (false); // UNREACHABLE
				}
		} else if (expr.getQuantifier() instanceof UniversalQuantifier) {
			// replace \forall X. \phi with \not \exists X. \not \phi

			try {
				QuantifiedExpression exprCopy = expr.copy();

				// modify the sub expression \phi
				OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
				exprCopy.getExpression().accept(v);

				// create \not SRNF(\phi)
				UnaryExpression notPhi = new UnaryExpression(NotOperator.getInstance(), v.getTransformed());

				// create \exists X. \not SRNF(\phi)
				QuantifiedExpression existsNotPhi = new QuantifiedExpression(ExistentialQuantifier.getInstance(), exprCopy.getVariables(), notPhi);

				// create \not \exists X. \not \phi
				UnaryExpression notExistsNotPhi = new UnaryExpression(NotOperator.getInstance(), existsNotPhi);

				this.transformed = notExistsNotPhi;
			} catch (MalformedExpressionException e) {
				assert (false); // UNREACHABLE
			}

		} else
			assert (false);// unreachable
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (expr.getExpression() instanceof UnaryExpression && ((UnaryExpression) expr.getExpression()).getOperator() instanceof NotOperator) {

			try {
				// replace \not\not \phi with \phi

				UnaryExpression exprCopy = expr.copy();
				IExpression phi = ((UnaryExpression) exprCopy.getExpression()).getExpression();

				OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
				phi.accept(v);

				this.transformed = v.getTransformed();

			} catch (MalformedExpressionException e) {
				assert (false); // UNREACHABLE
			}

		} else if (expr.getExpression() instanceof PolyadicExpression
				&& ((PolyadicExpression) expr.getExpression()).getOperator() instanceof AndOperator) {

			try {
				// replace \not (\phi1 AND ... AND \phiN) with (\not \phi1 OR ... OR \not \phiN)

				UnaryExpression exprCopy = expr.copy();
				List<IExpression> exprs = ((PolyadicExpression) exprCopy.getExpression()).getExpressions();

				// create \not SRNF(\phi_i)
				List<IExpression> transformedExprs = new Vector<IExpression>();
				for (IExpression phi : exprs) {
					OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
					phi.accept(v);

					UnaryExpression notPhi = new UnaryExpression(NotOperator.getInstance(), v.getTransformed());
					transformedExprs.add(notPhi);
				}

				this.transformed = new PolyadicExpression(OrOperator.getInstance(), transformedExprs);

			} catch (MalformedExpressionException e) {
				assert (false); // UNREACHABLE
			}

		} else if (expr.getExpression() instanceof PolyadicExpression
				&& ((PolyadicExpression) expr.getExpression()).getOperator() instanceof OrOperator) {

			try {
				// replace \not (\phi1 OR ... OR \phiN) with (\not \phi1 AND ... AND \not \phiN)

				UnaryExpression exprCopy = expr.copy();
				List<IExpression> exprs = ((PolyadicExpression) exprCopy.getExpression()).getExpressions();

				// create \not SRNF(\phi_i)
				List<IExpression> transformedExprs = new Vector<IExpression>();
				for (IExpression phi : exprs) {
					OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
					phi.accept(v);

					UnaryExpression notPhi = new UnaryExpression(NotOperator.getInstance(), v.getTransformed());
					transformedExprs.add(notPhi);
				}

				this.transformed = new PolyadicExpression(AndOperator.getInstance(), transformedExprs);

			} catch (MalformedExpressionException e) {
				assert (false); // UNREACHABLE
			}
		} else {
			// no rule can be applied to expr. Just modify the subexpression.
			try {
				UnaryExpression exprCopy = expr.copy();

				OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
				exprCopy.getExpression().accept(v);

				this.transformed = new UnaryExpression(NotOperator.getInstance(), v.getTransformed());
			} catch (MalformedExpressionException e) {
				assert (false); // UNREACHABLE
			}
		}
	}

	@Override
	public void visit(FalseExpression expr) {
		this.transformed = expr.copy();
	}
	
	@Override
	public void visit(TrueExpression expr) {
		this.transformed = expr.copy();
	}
	
	@Override
	public void visit(Variable var) {
		assert (false); // do nothing
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // do nothing
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(Equality op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // do nothing
	}

	@Override
	public void visit(PolyadicExpression polyadicExpression) {

		try {
			List<IExpression> transformedExprs = new Vector<IExpression>();

			for (IExpression e : polyadicExpression.copy().getExpressions()) {
				if (e instanceof PolyadicExpression && ((PolyadicExpression) e).getOperator().equals(polyadicExpression.getOperator())) {
					// flatten polyadic expression
					for (IExpression nested : ((PolyadicExpression) e).getExpressions()) {
						OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
						nested.accept(v);

						transformedExprs.add(v.getTransformed());
					}
				} else {
					// transform e
					OneStepSrnfTransformationVisitor v = new OneStepSrnfTransformationVisitor();
					e.accept(v);

					transformedExprs.add(v.getTransformed());
				}
			}

			this.transformed = new PolyadicExpression(polyadicExpression.getOperator(), transformedExprs);

		} catch (MalformedExpressionException e) {
			assert (false);// unreachable
		}

	}

	/**
	 * Get the rewritten expression.
	 * 
	 * @return The expression.
	 */
	public IExpression getTransformed() {
		return transformed;
	}

}
