package prototype.queries.domain.relational.calculus.visitors;

import java.util.HashSet;
import java.util.Set;

import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that computes the set of range-restricted variables. The visitor works just on expressions in the Safe-Range Normal Form.
 * 
 * 
 *
 */
public class RangeRestrictedVisitor implements IVisitor {

	/**
	 * The set of range restricted variables.
	 */
	private Set<Variable> rangeRestricted;

	/**
	 * If strict is true, then we ignore equality expressions for computing the range restricted variables.
	 */
	private boolean strict = true;

	/**
	 * Constructor.
	 */
	public RangeRestrictedVisitor() {
		rangeRestricted = null;
	}

	@Override
	public void visit(PredicateExpression expr) {
		rangeRestricted = new HashSet<Variable>();
		for (IOperand op : expr.getOperands())
			if (op instanceof Variable)
				rangeRestricted.add((Variable) op);
	}

	@Override
	public void visit(EqualityExpression expr) {
		rangeRestricted = new HashSet<Variable>();

		if (!strict)
			if (expr.getOp1() instanceof Variable && expr.getOp2() instanceof Constant)
				rangeRestricted.add((Variable) expr.getOp1());
			else if (expr.getOp2() instanceof Variable && expr.getOp1() instanceof Constant)
				rangeRestricted.add((Variable) expr.getOp2());
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (expr.getQuantifier() instanceof ExistentialQuantifier) {
			RangeRestrictedVisitor v = new RangeRestrictedVisitor();
			expr.getExpression().accept(v);

			if (v.getRangeRestricted() != null && v.getRangeRestricted().containsAll(expr.getVariables())) {
				rangeRestricted = new HashSet<Variable>(v.getRangeRestricted());
				rangeRestricted.removeAll(expr.getVariables());
			} else
				rangeRestricted = null;

		} else if (expr.getQuantifier() instanceof UniversalQuantifier)
			assert (false); // we consider just expressions in the Safe-Range Normal Form
		else
			assert (false); // unreachable code
	}

	@Override
	public void visit(FalseExpression expr) {
		rangeRestricted = new HashSet<Variable>();
	}

	@Override
	public void visit(TrueExpression expr) {
		rangeRestricted = new HashSet<Variable>();
	}

	@Override
	public void visit(UnaryExpression expr) {

		RangeRestrictedVisitor v = new RangeRestrictedVisitor();
		expr.getExpression().accept(v);

		if (v.getRangeRestricted() == null)
			this.rangeRestricted = null;// since there is an expression that returns null (\perp), then rr(expr) = null (\perp)
		else
			this.rangeRestricted = new HashSet<Variable>(); // range restricted is emptyset
	}

	@Override
	public void visit(Variable var) {
		assert (false);// do nothing
	}

	@Override
	public void visit(Constant expr) {
		assert (false);// do nothing
	}

	@Override
	public void visit(AndOperator op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(Equality op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(NotOperator op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(OrOperator op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false);// do nothing
	}

	@Override
	public void visit(PolyadicExpression polyadicExpression) {
		if (polyadicExpression.getOperator() instanceof OrOperator) {

			rangeRestricted = new HashSet<Variable>();

			// compute rr for the first element in polyadicExpression.getExpressions()
			IExpression e1 = polyadicExpression.getExpressions().get(0);

			RangeRestrictedVisitor v = new RangeRestrictedVisitor();
			e1.accept(v);

			if (v.getRangeRestricted() != null) {

				rangeRestricted.addAll(v.getRangeRestricted());

				for (IExpression e : polyadicExpression.getExpressions()) {
					v = new RangeRestrictedVisitor();
					e.accept(v);

					if (v.getRangeRestricted() != null) {
						// compute the intersection
						rangeRestricted.retainAll(v.getRangeRestricted());
					} else {
						// since there is an expression that returns null (\perp), then rr(polyadicExpression) = null (\perp)
						rangeRestricted = null;
						break;
					}
				}

			} else {
				// since there is an expression that returns null (\perp), then rr(polyadicExpression) = null (\perp)
				rangeRestricted = null;
			}

		} else if (polyadicExpression.getOperator() instanceof AndOperator) {
			// we treat the expression as the semantically equivalent expression phi_{1} AND ... AND phi_{n} AND x_{1} = y_{1} AND ... x_{m} = y_{m}
			rangeRestricted = new HashSet<Variable>();

			for (IExpression e : polyadicExpression.getExpressions()) {
				boolean isEqualityVariables = e instanceof EqualityExpression && ((EqualityExpression) e).getOp1() instanceof Variable
						&& ((EqualityExpression) e).getOp2() instanceof Variable;
				if (!isEqualityVariables) {
					RangeRestrictedVisitor v = new RangeRestrictedVisitor();
					e.accept(v);

					if (v.getRangeRestricted() != null)
						rangeRestricted.addAll(v.getRangeRestricted());
					else {
						rangeRestricted = null;
						break;
					}
				}
			}

			if (!strict)
				if (rangeRestricted != null) {
					// modify the result according to the sub-formulae of the form x_{i} = y_{i}

					for (IExpression e : polyadicExpression.getExpressions()) {

						if (e instanceof EqualityExpression && ((EqualityExpression) e).getOp1() instanceof Variable
								&& ((EqualityExpression) e).getOp2() instanceof Variable) {
							if (rangeRestricted.contains(((EqualityExpression) e).getOp1())
									|| rangeRestricted.contains(((EqualityExpression) e).getOp2())) {
								// if one of the two operands is range restricted, then both are range restricted
								rangeRestricted.add((Variable) ((EqualityExpression) e).getOp1());
								rangeRestricted.add((Variable) ((EqualityExpression) e).getOp1());
							}
						}
					}
				}

		} else
			assert (false);// do nothing
	}

	/**
	 * Get the set of range restricted variables. If the set is null, this denoted \perp (see algorithm definition at page 84
	 * "Foundations of Databases"). The method returns null iff there is a quantified variable that is not range restricted.
	 * 
	 * @return The set of range restricted variables.
	 */
	public Set<Variable> getRangeRestricted() {
		return rangeRestricted;
	}
}
