package prototype.queries.domain.relational.calculus.visitors;

import java.util.List;
import java.util.Vector;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that replaces some free variables with other IOperands.
 * 
 * 
 *
 */
public class ReplaceFreeVariablesVisitor implements IVisitor {

	/**
	 * List of variables to replace.
	 */
	private List<Variable> toReplace;

	/**
	 * List of replacements.
	 */
	private List<IOperand> replacements;

	/**
	 * Resulting transformed expression.
	 */
	private IExpression result;

	/**
	 * True iff toReplace contains unique identifiers and |toReplace| = |replacements|.
	 */
	private boolean ok;

	/**
	 * True if the visitor replaces clashes between replacements and quantified variables. If replacements contains just constants, replaceClashes is
	 * useless.
	 */
	private boolean replaceClashes;

	/**
	 * The prefix pattern for the variables created during clashes.
	 */
	private String prefix;

	/**
	 * The minimum id for the variables created during clashes.
	 */
	private int minimumId;

	/**
	 * Constructor. The visitor works iff toReplace contains unique identifiers and |toReplace| = |replacements|. If that's not the case, the
	 * resulting expression is null.
	 * 
	 * @param toReplace
	 *            - List of variables to replace.
	 * @param replacements
	 *            - List of replacements.
	 */
	public ReplaceFreeVariablesVisitor(List<Variable> toReplace, List<IOperand> replacements) {
		if (toReplace != null && replacements != null && toReplace.size() == replacements.size()) {

			// check whether toReplace contains just unique quantifiers
			for (int i = 0; i < toReplace.size(); i++)
				for (int j = 0; j < toReplace.size(); j++)
					if (i != j && toReplace.get(i).equals(toReplace.get(j))) {
						ok = false;
						break;
					}

			ok = true;
			this.toReplace = toReplace;
			this.replacements = replacements;

		} else {
			ok = false;
		}
	}

	/**
	 * Constructor. The visitor works iff toReplace contains unique identifiers and |toReplace| = |replacements|. If that's not the case, the
	 * resulting expression is null.
	 * 
	 * @param toReplace
	 *            - List of variables to replace.
	 * @param replacements
	 *            - List of replacements.
	 * @param replaceClashes
	 *            - if True, the visitor replaces clashes between replacements and quantified variables. If replacements contains just constants,
	 *            replaceClashes is useless.
	 * @param prefix
	 *            - The prefix pattern for the variables created during clashes.
	 * @param minimumId
	 *            - The minimum id for the variables created during clashes.
	 */
	public ReplaceFreeVariablesVisitor(List<Variable> toReplace, List<IOperand> replacements, boolean replaceClashes, String prefix, int minimumId) {
		this(toReplace, replacements);
		if (ok && replaceClashes && prefix != null && !prefix.isEmpty()) {
			this.replaceClashes = true;
			this.prefix = prefix;
			this.minimumId = minimumId;
		} else {
			this.ok = false;
			this.replaceClashes = false;
		}
	}

	/**
	 * Private constructor. It skips all the checks on toReplace and replacements.
	 * 
	 * @param toReplace
	 *            - List of variables to replace.
	 * @param replacements
	 *            - List of replacements.
	 * @param ok
	 *            - Ok flag.
	 */
	private ReplaceFreeVariablesVisitor(List<Variable> toReplace, List<IOperand> replacements, boolean ok, boolean replaceClashes, String prefix,
			int minimumId) {
		this.toReplace = toReplace;
		this.replacements = replacements;
		this.ok = ok;
		this.prefix = prefix;
		this.minimumId = minimumId;
	}

	@Override
	public void visit(PredicateExpression expr) {
		if (ok) {
			List<IOperand> ops = new Vector<IOperand>();

			for (IOperand op : expr.getOperands()) {
				if (toReplace.contains(op))
					ops.add(replacements.get(toReplace.indexOf(op)));
				else
					ops.add(op);
			}

			try {
				this.result = new PredicateExpression(expr.getType(), ops);
			} catch (MalformedExpressionException e) {
				assert (false); // unreachable code
			}
		}
	}

	@Override
	public void visit(EqualityExpression expr) {
		if (ok) {

			if (toReplace.contains(expr.getOp1())) {
				if (toReplace.contains(expr.getOp2())) {
					try {
						this.result = new EqualityExpression(replacements.get(toReplace.indexOf(expr.getOp1())), replacements.get(toReplace
								.indexOf(expr.getOp2())));
					} catch (MalformedExpressionException e) {
						assert (false); // unreachable code
					}
				} else {
					try {
						this.result = new EqualityExpression(replacements.get(toReplace.indexOf(expr.getOp1())), expr.getOp2());
					} catch (MalformedExpressionException e) {
						assert (false); // unreachable code
					}
				}
			} else {
				if (toReplace.contains(expr.getOp2())) {
					try {
						this.result = new EqualityExpression(expr.getOp1(), replacements.get(toReplace.indexOf(expr.getOp2())));
					} catch (MalformedExpressionException e) {
						assert (false); // unreachable code
					}
				} else {
					try {
						this.result = new EqualityExpression(expr.getOp1(), expr.getOp2());
					} catch (MalformedExpressionException e) {
						assert (false); // unreachable code
					}
				}
			}

		}
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (ok) {
			ReplaceFreeVariablesVisitor v = null;
			Vector<Variable> toReplace = new Vector<Variable>(this.toReplace);
			Vector<IOperand> replacements = new Vector<IOperand>(this.replacements);
			
			if (replaceClashes) {

				for (IOperand var : new Vector<IOperand>(replacements)) {
					if (var instanceof Variable) {
						if (expr.getVariables().contains(var)) {
							// replace the quantified variable with a new one.

							Variable newVar = null;
							try {
								newVar = new Variable(getNewIdentifier());
							} catch (MalformedExpressionException e) {
								assert (false);
							}
							toReplace.add((Variable) var);
							replacements.add(newVar);
						}
					}
				}
			}
			v = new ReplaceFreeVariablesVisitor(toReplace, replacements, ok, replaceClashes, prefix, this.minimumId);
			expr.getExpression().accept(v);

			try {
				this.result = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), v.getResult());

				this.minimumId = v.minimumId;
			} catch (MalformedExpressionException e) {
				assert (false); // unreachable code
			}
		}
	}

	private String getNewIdentifier() {
		minimumId++;
		return prefix+minimumId;
	}

	@Override
	public void visit(UnaryExpression expr) {
		if (ok) {
			ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(toReplace, replacements, ok, replaceClashes, prefix,minimumId);
			expr.getExpression().accept(v);
			this.minimumId = v.minimumId;
			try {
				this.result = new UnaryExpression(expr.getOperator(), v.getResult());
			} catch (MalformedExpressionException e) {
				assert (false); // unreachable code
			}
		}
	}
	
	@Override
	public void visit(FalseExpression expr) {
		this.result = expr.copy();
	}
	
	@Override
	public void visit(TrueExpression expr) {
		this.result = expr.copy();
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable code
	}

	@Override
	public void visit(PolyadicExpression expr) {
		if (ok) {
			List<IExpression> exprs = new Vector<IExpression>();
			for (IExpression e : expr.getExpressions()) {
				ReplaceFreeVariablesVisitor v = new ReplaceFreeVariablesVisitor(toReplace, replacements, ok, replaceClashes, prefix, minimumId);
				e.accept(v);

				exprs.add(v.getResult());

				this.minimumId = v.minimumId;
			}
			try {
				this.result = new PolyadicExpression(expr.getOperator(), exprs);
			} catch (MalformedExpressionException e) {
				assert (false); // unreachable code
			}
		}
	}

	/**
	 * Get the resulting expression in which all occurrences of variables in toReplace are replaced with operands in replacements. It returns null if
	 * toReplace and replacements are not compatible.
	 * 
	 * @return The resulting expression.
	 */
	public IExpression getResult() {
		return result;
	}

	public int getMinValue() {
		return minimumId;
	}
}
