package prototype.queries.domain.relational.calculus.visitors;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Vector;

import prototype.queries.MalformedExpressionException;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;

/**
 * Visitor that replaces predicates of the form P(var_1,...,var_n,const_1, ...,const_m) with Exists x_1, ..., x_m. P(var_1,...,var_n, x_1, ..., x_m)
 * AND x_1 = const_1 AND ... AND x_m = const_m. It works properly only for expressions with incremental identifiers up to value (passed by parameter
 * in the constructor).
 * 
 * 
 *
 */
public class ReplacePredicatesWithConstantsVisitor implements IVisitor {

	/**
	 * The current index.
	 */
	private int current;
	/**
	 * The prefix.
	 */
	private String prefix;

	/**
	 * The transformed expression.
	 */
	private IExpression transformed;

	public ReplacePredicatesWithConstantsVisitor(String prefix, int value) {
		this.prefix = prefix;
		this.current = value;
	}

	@Override
	public void visit(PredicateExpression expr) {
		List<IOperand> operands = new Vector<IOperand>();
		List<IExpression> equalities = new Vector<IExpression>();
		LinkedHashSet<Variable> variables = new LinkedHashSet<Variable>();

		for (IOperand o : expr.getOperands()) {
			if (o instanceof Variable)
				operands.add(o);
			else if (o instanceof Constant) {
				// create a new variable
				Variable newVar = getNewVariable();

				// add it to the operands
				operands.add(newVar);

				// add it to the set of quantified variables
				variables.add(newVar);

				// create the equality
				try {
					equalities.add(new EqualityExpression(newVar, o));
				} catch (MalformedExpressionException e) {
					assert (false); // unreachable
				}
			} else
				assert (false); // unreachable
		}

		try {
			PredicateExpression pred = new PredicateExpression(expr.getType(), operands);

			if (equalities.size() == 0)
				this.transformed = pred;
			else {
				equalities.add(pred);

				PolyadicExpression and = new PolyadicExpression(AndOperator.getInstance(), equalities);
				this.transformed = new QuantifiedExpression(ExistentialQuantifier.getInstance(), variables, and);
			}
		} catch (MalformedExpressionException e) {
			assert (false);
		}

	}

	@Override
	public void visit(EqualityExpression expr) {
		try {
			this.transformed = expr.copy();
		} catch (MalformedExpressionException e) {
			assert (false);
		}
	}

	@Override
	public void visit(QuantifiedExpression expr) {
		int v = this.current;

		ReplacePredicatesWithConstantsVisitor vR = new ReplacePredicatesWithConstantsVisitor(prefix, v);
		expr.getExpression().accept(vR);

		v = vR.getCurrent();

		try {
			this.transformed = new QuantifiedExpression(expr.getQuantifier(), expr.getVariables(), vR.getTransformed());
			this.current = v;
		} catch (MalformedExpressionException e1) {
			assert (false);
		}
	}

	@Override
	public void visit(UnaryExpression expr) {
		int v = this.current;

		ReplacePredicatesWithConstantsVisitor vR = new ReplacePredicatesWithConstantsVisitor(prefix, v);
		expr.getExpression().accept(vR);

		v = vR.getCurrent();

		try {
			this.transformed = new UnaryExpression(expr.getOperator(), vR.getTransformed());
			this.current = v;
		} catch (MalformedExpressionException e1) {
			assert (false);
		}
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(PolyadicExpression expr) {

		int v = this.current;
		List<IExpression> exprs = new Vector<IExpression>();
		for (IExpression e : expr.getExpressions()) {
			ReplacePredicatesWithConstantsVisitor vR = new ReplacePredicatesWithConstantsVisitor(prefix, v);
			e.accept(vR);

			exprs.add(vR.getTransformed());
			v = vR.getCurrent();
		}

		try {
			this.transformed = new PolyadicExpression(expr.getOperator(), exprs);
			this.current = v;
		} catch (MalformedExpressionException e1) {
			assert (false);
		}

	}

	@Override
	public void visit(FalseExpression expr) {
		this.transformed = expr.copy();
	}

	@Override
	public void visit(TrueExpression expr) {
		this.transformed = expr.copy();
	}

	/**
	 * Create a new variable.
	 * 
	 */
	private Variable getNewVariable() {
		try {
			current++;
			return new Variable(prefix + current);
		} catch (MalformedExpressionException e) {
			assert (false);// unreachable
			return null;
		}
	}

	public int getCurrent() {
		return current;
	}

	public IExpression getTransformed() {
		return transformed;
	}

}
