package prototype.queries.translation;

import java.util.Set;

import prototype.objects.IType;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.FalseExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PredicateExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.TrueExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.IVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.Equality;
import prototype.queries.operators.ExistentialQuantifier;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.util.Util;

/**
 * Visitor that translates a Safe-Range Normal Form domain-relational calculus query to a SQL query.
 * 
 * 
 *
 */
public class NaiveMySqlTranslator implements IVisitor {

	/**
	 * The sql expression.
	 */
	private String sqlExpression;

	/**
	 * The nesting level.
	 */
	private int nestingLevel;

	private int adomCount;

	private Set<IType> types;

	private Set<Constant> values;
	
	
	private final String SELECT_KEYWORD = "SELECT";

	public NaiveMySqlTranslator(Set<IType> types, Set<Constant> values) {
		this.sqlExpression = "";
		this.nestingLevel = 0;
		this.adomCount = 0;
		this.types = types;
		this.values = values;
	}

	public NaiveMySqlTranslator(Set<IType> types, Set<Constant> values, int adomCount) {
		this.sqlExpression = "";
		this.nestingLevel = 0;
		this.adomCount = adomCount;
		this.types = types;
		this.values = values;
	}

	private NaiveMySqlTranslator(Set<IType> types, Set<Constant> values, int nestingLevel, int adomCount) {
		this.sqlExpression = "";
		this.nestingLevel = nestingLevel;
		this.adomCount = adomCount;
		this.types = types;
		this.values = values;
	}

	@Override
	public void visit(PredicateExpression expr) {

		this.sqlExpression = "(";
		for (IOperand op : expr.getOperands())
			if (op instanceof Variable)
				sqlExpression += op+".value" + " ,";
			else
				sqlExpression += op + " ,";

		this.sqlExpression = Util.replaceLast(sqlExpression, " ,", ")");

		this.sqlExpression += " IN ("+SELECT_KEYWORD+" * FROM " + expr.getType().getId() + ")";

	}

	@Override
	public void visit(PolyadicExpression expr) {

		this.sqlExpression = "(";

		String op = "";

		if (expr.getOperator() instanceof AndOperator)
			op = "AND";
		else if (expr.getOperator() instanceof OrOperator)
			op = "OR";
		else
			assert (false); // unreachable

		for (IExpression e : expr.getExpressions()) {
			NaiveMySqlTranslator v = new NaiveMySqlTranslator(this.types, this.values, this.nestingLevel, this.adomCount);
			e.accept(v);

			this.sqlExpression += v.sqlExpression+ " " + op + " ";
			this.adomCount = v.adomCount;
		}
		sqlExpression = Util.replaceLast(sqlExpression, " " + op + " ", "");
		sqlExpression += ")";
	}

	@Override
	public void visit(EqualityExpression expr) {

		String op = "=";

		this.sqlExpression = expr.getOp1().toString() + ((expr.getOp1() instanceof Variable)? ".value" : "");
		this.sqlExpression += " " + op + " " + expr.getOp2().toString() + ((expr.getOp2() instanceof Variable)? ".value" : "");

	}

	@Override
	public void visit(QuantifiedExpression expr) {
		if (expr.getQuantifier() instanceof ExistentialQuantifier) {

			NaiveMySqlTranslator v = new NaiveMySqlTranslator(this.types, this.values, this.nestingLevel + 1, adomCount);
			expr.getExpression().accept(v);

			this.adomCount = v.adomCount;

			String tables = "";
			for (Variable var : expr.getVariables())
				tables += getAdomType(var.getIdentifier())+" ,";
			tables = Util.replaceLast(tables, " ,", "");

			String whereExpr = "";
			if (v.getSqlExpression() != null && !v.getSqlExpression().isEmpty())
				whereExpr += v.getSqlExpression();
			else
				whereExpr = Util.replaceLast(whereExpr, " AND ", "");

			String selectedVars = "";
			for (Variable var : expr.getVariables())
				selectedVars += var+ ".value"  + " , ";

			selectedVars = Util.replaceLast(selectedVars, " , ", "");

			// translate as a boolean query
			this.sqlExpression = "EXISTS ("+SELECT_KEYWORD+" " + selectedVars + " FROM " + tables + " WHERE " + whereExpr + ")";

		} else if (expr.getQuantifier() instanceof UniversalQuantifier) {
			assert (false);// unreachable for Safe-Range Normal Form expressions
		} else {
			assert (false);// unreachable
		}

	}

	@Override
	public void visit(UnaryExpression expr) {

		NaiveMySqlTranslator v = new NaiveMySqlTranslator(this.types, this.values, this.nestingLevel, this.adomCount);
		expr.getExpression().accept(v);

		if (!v.getSqlExpression().isEmpty())
			this.sqlExpression = "NOT (" + v.getSqlExpression() + ")";
		this.adomCount = v.adomCount;
	}

	@Override
	public void visit(Variable var) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Constant expr) {
		assert (false); // unreachable
	}

	@Override
	public void visit(AndOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(Equality op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(ExistentialQuantifier op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(NotOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(OrOperator op) {
		assert (false); // unreachable
	}

	@Override
	public void visit(UniversalQuantifier op) {
		assert (false); // unreachable
	}

	/**
	 * Get the SQL expression corresponding to the visited expression.
	 * 
	 * @return The translation.
	 */
	public String getSqlExpression() {
		return sqlExpression;
	}

	@Override
	public void visit(FalseExpression expr) {
		this.sqlExpression = expr.toString();
	}

	@Override
	public void visit(TrueExpression expr) {
		this.sqlExpression = expr.toString();
	}

	public static String getAdomType(String id) {
		String expr = "__ADOM__ AS " + id;
		return expr;
	}

}
