package prototype.statements.abstractImplementation;

import prototype.objects.IDbObject;
import prototype.objects.ITable;
import prototype.privileges.IPrivilege;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.statements.MalformedStatementException;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IChangePolicyStatement;
import prototype.statements.interfaces.create.ICreateStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.statements.modes.UPDATE_MODE;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Interface representing the Factory objects for generating statements.
 * 
 * 
 *
 */
public interface IStatementFactory {

	
	
	
	/**
	 * Factory method for ADD USER statements.
	 * 
	 * @param invoker
	 *            - The user executing the statement.
	 * @param toCreate
	 *            - The new user.
	 * @return The statement.
	 * @throws MalformedStatementException
	 */
	public IAddUser createAddUserStatement(IUser invoker, IUser toCreate) throws MalformedStatementException;
	
	
	/**
	 * Factory method for GRANT, GRANT WITH OPTION, and REVOKE statements.
	 * 
	 * @param invoker
	 *            - The user executing the statement.
	 * @param mode
	 *            - The mode.
	 * @param privilege
	 *            - The privilege.
	 * @param target
	 *            - The target of the statement.
	 * @return The statement.
	 * @throws MalformedStatementException
	 */
	public IChangePolicyStatement createChangePolicyStatement(IUser invoker, CHANGE_POLICY_MODE mode, IPrivilege privilege, IUser target)
			throws MalformedStatementException;

	/**
	 * Factory method for CREATE statements.
	 * 
	 * @param invoker
	 *            - The user executing the command.
	 * @param object
	 *            - The object to be created.
	 * @return The CREATE statement.
	 * @throws MalformedStatementException
	 */
	public ICreateStatement createCreateStatement(IUser invoker, IDbObject object) throws MalformedStatementException;

	/**
	 * Factory method for SELECT statements.
	 * 
	 * @param invoker
	 *            - the user executing the query.
	 * @param query
	 *            - the query.
	 * @return The SELECT statement.
	 * @throws MalformedStatementException
	 */
	public ISelectStatement createSelectStatement(IUser invoker, IQuery query) throws MalformedStatementException;

	/**
	 * Factory method for update statements.
	 * 
	 * @param invoker
	 *            - The user executing the statement.
	 * @param mode
	 *            - The update mode (INSERT or DELETE)
	 * @param schema
	 *            - The relation schema the statement is going to modify.
	 * @param tuple
	 *            - The tuple the statement is going to add or remove.
	 * @return The update statement.
	 * @throws MalformedStatementException
	 */
	public IUpdateStatement createUpdateStatement(IUser invoker, UPDATE_MODE mode, ITable schema, ITuple tuple) throws MalformedStatementException;

}
