package prototype.statements.abstractImplementation.update;

import prototype.interfaces.CopyException;
import prototype.objects.ITable;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.Statement;
import prototype.statements.interfaces.TranslationException;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.UPDATE_MODE;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Abstract class for update statements.
 * 
 * 
 *
 */
public abstract class UpdateStatement extends Statement implements IUpdateStatement {

	/**
	 * The update mode (either INSERT or DELETE)
	 */
	private UPDATE_MODE mode;

	/**
	 * The relation schema the statement is going to modify.
	 */
	private ITable schema;

	/**
	 * The tuple the statement is going to add/remove.
	 */
	private ITuple tuple;

	/**
	 * Constructor.
	 * 
	 * @param invoker
	 *            - The user executing the statement.
	 * @param mode
	 *            - The update mode (INSERT or DELETE)
	 * @param schema
	 *            - The relation schema the statement is going to modify.
	 * @param tuple
	 *            - The tuple the statement is going to add or remove.
	 * @throws MalformedStatementException
	 */
	public UpdateStatement(IUser invoker, UPDATE_MODE mode, ITable schema, ITuple tuple) throws MalformedStatementException {
		super(invoker);

		if (mode == null) {
			throw new MalformedStatementException("Mode cannot be null.");
		} else if (schema == null) {
			throw new MalformedStatementException("Schema cannot be null.");
		} else if (tuple == null) {
			throw new MalformedStatementException("Tuple cannot be null.");
		} else if (schema.getFields().size() != tuple.getSize()) {
			throw new MalformedStatementException("Schema size and tuple size are different.");
		} else {
			this.mode = mode;
			try {
				this.schema = schema.copy();
				this.tuple = tuple.copy();
			} catch (CopyException e) {
				throw new MalformedStatementException(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.interfaces.IUpdateStatement#getRelationSchema()
	 */
	@Override
	public ITable getRelationSchema() {
		return this.schema;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.interfaces.IUpdateStatement#getTuple()
	 */
	@Override
	public ITuple getTuple() {
		return this.tuple;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.implementation.Statement#toSQL()
	 */
	@Override
	public abstract String toSQL() throws TranslationException;

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.interfaces.IUpdateStatement#getMode()
	 */
	@Override
	public UPDATE_MODE getMode() {
		return this.mode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		else if (obj instanceof UpdateStatement) {
			if (obj == this)
				return true;
			else
				return ((UpdateStatement) obj).getInvoker().equals(this.getInvoker())
						&& ((UpdateStatement) obj).getRelationSchema().equals(this.getRelationSchema())
						&& ((UpdateStatement) obj).getTuple().equals(this.getTuple()) && ((UpdateStatement) obj).getMode().equals(this.getMode());
		} else
			return false;
	}

	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + getInvoker().hashCode();
		result = 31 * result + getRelationSchema().hashCode();
		result = 31 * result + getTuple().hashCode();
		result = 31 * result + getMode().hashCode();
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.interfaces.IStatement#getDefinition()
	 */
	@Override
	public String getDefinition() {
		String columns = "[ ";
		for (int i = 0; i < this.getRelationSchema().getCardinality(); i++) {
			if (i != 0)
				columns += ", ";
			columns += this.getRelationSchema().getFields().get(i);
		}
		columns += "]";

		String values = "( ";
		for (int i = 0; i < this.getTuple().getSize(); i++) {
			if (i != 0)
				values += ", ";
			values += this.getTuple().getContent().get(i);
		}
		values += ")";

		return this.getInvoker() + ", " + this.getMode().toString() + ", " + this.getRelationSchema().getId() + columns + ", " + values;

	}
}
