package prototype.statements.implementation.mysql;

import prototype.objects.IDbObject;
import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IView;
import prototype.privileges.IPrivilege;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.implementation.mysql.add.MySqlAddUser;
import prototype.statements.implementation.mysql.changePolicy.MySqlGrantStatement;
import prototype.statements.implementation.mysql.changePolicy.MySqlGrantWithOptionStatement;
import prototype.statements.implementation.mysql.changePolicy.MySqlRevokeStatement;
import prototype.statements.implementation.mysql.create.MySqlCreateTableStatement;
import prototype.statements.implementation.mysql.create.MySqlCreateTriggerStatement;
import prototype.statements.implementation.mysql.create.MySqlCreateViewStatement;
import prototype.statements.implementation.mysql.select.MySqlSelectStatement;
import prototype.statements.implementation.mysql.update.MySqlDeleteStatement;
import prototype.statements.implementation.mysql.update.MySqlInsertStatement;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IChangePolicyStatement;
import prototype.statements.interfaces.create.ICreateStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.statements.modes.CHANGE_POLICY_MODE;
import prototype.statements.modes.UPDATE_MODE;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Class representing the MySql statement factory.
 * 
 * 
 *
 */
public class MySqlStatementFactory implements IStatementFactory {

	/**
	 * The singleton instance.
	 */
	private static MySqlStatementFactory singleton_;

	/**
	 * Get the unique instance of MySqlStatementFactory
	 * 
	 * @return
	 */
	public static MySqlStatementFactory getInstance() {
		if (singleton_ == null)
			singleton_ = new MySqlStatementFactory();
		return singleton_;
	}

	/**
	 * Private constructor.
	 */
	private MySqlStatementFactory() {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.abstractImplementation.IStatementFactory#createAddUserStatement(guarnieri.marco.ethz.users.IUser,
	 * guarnieri.marco.ethz.users.IUser)
	 */
	@Override
	public IAddUser createAddUserStatement(IUser invoker, IUser toCreate) throws MalformedStatementException {
		return new MySqlAddUser(invoker, toCreate);
	}

	@Override
	public IChangePolicyStatement createChangePolicyStatement(IUser invoker, CHANGE_POLICY_MODE mode, IPrivilege privilege, IUser target)
			throws MalformedStatementException {
		switch (mode) {
		case GRANT:
			return new MySqlGrantStatement(invoker, mode, privilege, target);
		case GRANT_WITH_OPTION:
			return new MySqlGrantWithOptionStatement(invoker, mode, privilege, target);
		case REVOKE:
			return new MySqlRevokeStatement(invoker, mode, privilege, target);
		default:
			assert (false); // unreachable
			return null;
		}
	}

	@Override
	public ICreateStatement createCreateStatement(IUser invoker, IDbObject object) throws MalformedStatementException {
		if (object instanceof ITrigger)
			return new MySqlCreateTriggerStatement(invoker, (ITrigger) object);
		else if (object instanceof IView)
			return new MySqlCreateViewStatement(invoker, (IView) object);
		else if (object instanceof ITable)
			return new MySqlCreateTableStatement(invoker, (ITable) object);
		else {
			assert (false); // unreachable
			return null;
		}

	}

	@Override
	public ISelectStatement createSelectStatement(IUser invoker, IQuery query) throws MalformedStatementException {
		return new MySqlSelectStatement(invoker, query);
	}

	@Override
	public IUpdateStatement createUpdateStatement(IUser invoker, UPDATE_MODE mode, ITable schema, ITuple tuple) throws MalformedStatementException {
		switch (mode) {
		case INSERT:
			return new MySqlInsertStatement(invoker, mode, schema, tuple);
		case DELETE:
			return new MySqlDeleteStatement(invoker, mode, schema, tuple);
		default:
			assert (false); // unreachable
			return null;
		}
	}

}
