package prototype.statements.implementation.mysql.select;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import prototype.interfaces.CopyException;
import prototype.objects.IType;
import prototype.queries.domain.relational.calculus.INonBooleanQuery;
import prototype.queries.domain.relational.calculus.IQuery;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.MySqlGetTypes;
import prototype.queries.domain.relational.calculus.visitors.MySqlTranslator;
import prototype.queries.domain.relational.calculus.visitors.MySqlVariableVisitor;
import prototype.queries.translation.GetConstants;
import prototype.queries.translation.GetTypes;
import prototype.queries.translation.NaiveMySqlTranslator;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.select.SelectStatement;
import prototype.statements.interfaces.TranslationException;
import prototype.users.IUser;
import prototype.util.Util;

/**
 * The concrete class representing SELECT statements for the MySql database.
 * 
 * 
 *
 */
public class MySqlSelectStatement extends SelectStatement {

	boolean naiveTranslation = true;

	/**
	 * Constructor.
	 * 
	 * @param invoker
	 *            - The user executing the query.
	 * @param query
	 *            - The actual query.
	 * @throws MalformedStatementException
	 */
	public MySqlSelectStatement(IUser invoker, IQuery query) throws MalformedStatementException {
		super(invoker, query);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.statements.implementation.abstraction.select.SelectStatement#toSQL()
	 */
	@Override
	public String toSQL() throws TranslationException {

		if (this.getQuery() instanceof INonBooleanQuery) {

			if (naiveTranslation) {
				GetConstants vConsts = new GetConstants();
				this.getQuery().getExpression().accept(vConsts);

				GetTypes vTypes = new GetTypes();
				this.getQuery().getExpression().accept(vTypes);

				NaiveMySqlTranslator v = new NaiveMySqlTranslator(vTypes.getTypes(), vConsts.getConstants());
				this.getQuery().getExpression().accept(v);
				
				String tables = "";
				for (Variable var : ((INonBooleanQuery) this.getQuery()).getOutput()){
					tables += NaiveMySqlTranslator.getAdomType(var.getIdentifier())+" ,";
				}
				tables = Util.replaceLast(tables, " ,", "");

				String whereExpr = "";
				if (v.getSqlExpression() != null && !v.getSqlExpression().isEmpty())
					whereExpr += v.getSqlExpression();
				else
					whereExpr = Util.replaceLast(whereExpr, " AND ", "");

				String selectedVars = "";
				for (Variable var : ((INonBooleanQuery) this.getQuery()).getOutput())
					selectedVars += var+".value" + " , ";

				selectedVars = Util.replaceLast(selectedVars, " , ", "");

				// translate as a boolean query
				String expr = "SELECT DISTINCT " + selectedVars + " FROM " + tables + " WHERE " + whereExpr;

				return expr;
			} else {

				// non-boolean query

				boolean booleanSemantics;

				MySqlGetTypes v1 = new MySqlGetTypes();
				this.getQuery().getExpression().accept(v1);

				if (v1.getTypes().isEmpty())
					// we are in the set semantics
					booleanSemantics = false;
				else
					// we are in the boolean semantics
					booleanSemantics = true;

				Map<Variable, Integer> varToNestLevel = new HashMap<Variable, Integer>();
				if (!booleanSemantics) {
					// add the free variables to varToNestLevel
					for (Variable var : ((INonBooleanQuery) this.getQuery()).getOutput()) {
						varToNestLevel.put(var, 0);
					}
				}

				MySqlVariableVisitor vv = new MySqlVariableVisitor(varToNestLevel);
				this.getQuery().getExpression().accept(vv);
				Map<Variable, String> variablesToSqlNames = vv.getVariablesToSqlNames();

				MySqlTranslator v = new MySqlTranslator(variablesToSqlNames);
				this.getQuery().getExpression().accept(v);

				List<IType> usedTypes = v.getTypes();
				// List<String> equals = v.getEqualities();

				String selectedVars = "";
				for (Variable var : ((INonBooleanQuery) this.getQuery()).getOutput()) {
					selectedVars += variablesToSqlNames.get(var) + " , ";
				}
				selectedVars = Util.replaceLast(selectedVars, " , ", "");

				String tables = "";
				for (int i = 0; i < usedTypes.size(); i++) {
					if (i != 0)
						tables += " , ";

					int number = 0;
					for (int j = 0; j <= i; j++)
						if (usedTypes.get(j).equals(usedTypes.get(i)))
							number++;

					tables += usedTypes.get(i).getId() + " AS " + MySqlVariableVisitor.getTableIdentifier(usedTypes.get(i), 0, number);
				}

				String whereExpr = "";
				// for (String eq : equals)
				// whereExpr += eq + " AND ";
				if (v.getSqlExpression() != null && !v.getSqlExpression().isEmpty())
					whereExpr += v.getSqlExpression();
				else
					whereExpr = Util.replaceLast(whereExpr, " AND ", "");

				// if (booleanSemantics)
				return "SELECT DISTINCT " + selectedVars + " FROM " + tables + " WHERE " + whereExpr + ";";
				// else
				// return "SELECT DISTINCT " + selectedVars + " FROM (" + v.getSqlExpression() + ") AS temp";
			}
		} else {

			if (naiveTranslation) {
				GetConstants vConsts = new GetConstants();
				this.getQuery().getExpression().accept(vConsts);

				GetTypes vTypes = new GetTypes();
				this.getQuery().getExpression().accept(vTypes);

				NaiveMySqlTranslator v = new NaiveMySqlTranslator(vTypes.getTypes(), vConsts.getConstants());
				this.getQuery().getExpression().accept(v);
				
				String expr = "SELECT CASE WHEN " + v.getSqlExpression() + " THEN 'true' ELSE 'false' END";

				return expr;
			} else {
				// boolean query

				MySqlVariableVisitor vv = new MySqlVariableVisitor(new HashMap<Variable, Integer>());

				this.getQuery().getExpression().accept(vv);
				Map<Variable, String> variablesToSqlNames = vv.getVariablesToSqlNames();

				MySqlTranslator v = new MySqlTranslator(variablesToSqlNames);
				this.getQuery().getExpression().accept(v);

				return "SELECT CASE WHEN " + v.getSqlExpression() + " THEN 'true' ELSE 'false' END";
			}

		}
	}

	@Override
	public MySqlSelectStatement copy() throws CopyException {

		try {
			return new MySqlSelectStatement(getInvoker().copy(), getQuery().copy());
		} catch (MalformedStatementException e) {
			throw new CopyException(e);
		}

	}

}
