package prototype.system.implementation;

import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import prototype.objects.IDbObject;
import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IView;
import prototype.results.actionResults.IActionResult;
import prototype.results.implementation.BooleanActionResult;
import prototype.results.implementation.MalformedResultException;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IGrantStatement;
import prototype.statements.interfaces.changePolicy.IGrantWithOptionStatement;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.changePolicy.IRevokeStatement;
import prototype.statements.interfaces.create.ICreateStatement;
import prototype.statements.interfaces.create.ICreateTableStatement;
import prototype.statements.interfaces.create.ICreateTriggerStatement;
import prototype.statements.interfaces.create.ICreateViewStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.system.MalformedStateException;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.interfaces.IState;
import prototype.system.interfaces.IStateFactory;
import prototype.users.IUser;
import prototype.util.Pair;

/**
 * Abstract class implementing the Execution Engine. It implements the execution of all statements but SELECT, INSERT, and DELETE. They are left to
 * the concrete implementation, which depends on the database.
 * 
 * 
 *
 */
public abstract class AbstractExecutionEngine implements IDatabaseEngine {

	/**
	 * The state factory.
	 */
	private IStateFactory stateFactory;

	/**
	 * Constructor.
	 * 
	 * @param factory
	 *            - The state factory.
	 */
	public AbstractExecutionEngine(IStateFactory factory) {
		this.stateFactory = factory;
	}

	@Override
	public Pair<IActionResult, IState> execute(IAddUser stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {
					try {
						// modify the system's state
						Set<IUser> newUsers = state.getUsers();
						newUsers.add(stmt.getUser());

						// add user cannot be executed in transactions.. invoker = null, rollback state = null, and scheduledTriggers = \emptyset

						IState newState = stateFactory.createNewState(newUsers, state.getTables(), state.getPolicy(), state.getViews(),
								state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);

						IActionResult newResult = new BooleanActionResult(true, false, null);

						return new Pair<IActionResult, IState>(newResult, newState);

					} catch (MalformedStateException | MalformedResultException e) {
						throw new ExecutionException(e);
					}

				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	@Override
	public Pair<IActionResult, IState> execute(IGrantStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {
					if (state.getUsers().contains(stmt.getTarget())) {
						try {
							// modify the system's state
							Set<IPositiveGrantStatement> newPolicy = state.getPolicy();
							newPolicy.add(stmt);

							// grant statements can be executed both in transactions and outside transactions.
							IState newState = null;
							if (state.isTransaction()) {
								// in transaction
								// invoker = state.getInvoker()
								// rollbackstate = state.getRollbackstate()

								// we are executing the trigger's action
								// drop the first trigger in state.getScheduledTriggers() (the one we are currently executing)
								// set it as the current trigger.
								SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>(state.getScheduledTriggers());
								ITrigger old = state.getScheduledTriggers().first();
								scheduled.remove(old);

								ITrigger toExecute;
								if (!scheduled.isEmpty())
									toExecute = scheduled.first();
								else
									toExecute = null;

								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), state.getInvoker(), state.getRollbackState(), scheduled,
										state.getTuple(), toExecute);
							} else {
								// outside transactions
								// invoker = null
								// rollbacks = null
								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);
							}

							IActionResult newResult = new BooleanActionResult(true, false, null);

							return new Pair<IActionResult, IState>(newResult, newState);

						} catch (MalformedStateException | MalformedResultException e) {
							throw new ExecutionException(e);
						}
					} else
						throw new ExecutionException("Target is not in the state.");
				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	@Override
	public Pair<IActionResult, IState> execute(IGrantWithOptionStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {
					if (state.getUsers().contains(stmt.getTarget())) {
						try {
							// modify the system's state
							Set<IPositiveGrantStatement> newPolicy = state.getPolicy();
							newPolicy.add(stmt);

							IState newState = null;
							if (state.isTransaction()) {
								// in transaction
								// invoker = state.getInvoker()
								// rollbackstate = state.getRollbackstate()
								// we are executing the trigger's action
								// drop the first trigger in state.getScheduledTriggers() (the one we are currently executing)
								// set it as the current trigger.
								SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>(state.getScheduledTriggers());
								ITrigger old = state.getScheduledTriggers().first();
								scheduled.remove(old);

								ITrigger toExecute;
								if (!scheduled.isEmpty())
									toExecute = scheduled.first();
								else
									toExecute = null;

								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), state.getInvoker(), state.getRollbackState(), scheduled,
										state.getTuple(), toExecute);
							} else {
								// outside transactions
								// invoker = null
								// rollbacks = null
								// scheduledTriggers = \emptyset
								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);
							}

							IActionResult newResult = new BooleanActionResult(true, false, null);

							return new Pair<IActionResult, IState>(newResult, newState);

						} catch (MalformedStateException | MalformedResultException e) {
							throw new ExecutionException(e);
						}
					} else
						throw new ExecutionException("Target is not in the state.");
				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	@Override
	public Pair<IActionResult, IState> execute(IRevokeStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {
					if (state.getUsers().contains(stmt.getTarget())) {
						try {
							// modify the system's state
							Set<IPositiveGrantStatement> newPolicy = revoke(state.getPolicy(), stmt);

							IState newState = null;
							if (state.isTransaction()) {
								// in transaction
								// invoker = state.getInvoker()
								// rollbackstate = state.getRollbackstate()
								// we are executing the trigger's action
								// drop the first trigger in state.getScheduledTriggers() (the one we are currently executing)
								// set it as the current trigger.
								SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>(state.getScheduledTriggers());
								ITrigger old = state.getScheduledTriggers().first();
								scheduled.remove(old);

								ITrigger toExecute;
								if (!scheduled.isEmpty())
									toExecute = scheduled.first();
								else
									toExecute = null;

								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), state.getInvoker(), state.getRollbackState(), scheduled,
										state.getTuple(), toExecute);
							} else {
								// outside transactions
								// invoker = null
								// rollbacks = null
								// scheduledTriggers = \emptyset
								newState = stateFactory.createNewState(state.getUsers(), state.getTables(), newPolicy, state.getViews(),
										state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);
							}

							IActionResult newResult = new BooleanActionResult(true, false, null);

							return new Pair<IActionResult, IState>(newResult, newState);

						} catch (MalformedStateException | MalformedResultException e) {
							throw new ExecutionException(e);
						}
					} else
						throw new ExecutionException("Target is not in the state.");
				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	/**
	 * Compute the policy after the revoke statement. Implements dependOn(S,u,p,u') (see page Appendix A of our paper).
	 * 
	 * @param policy
	 *            - The initial policy.
	 * @param stmt
	 *            - The revoke statement.
	 * @return The new policy.
	 * 
	 */
	@Override
	public Set<IPositiveGrantStatement> revoke(Set<IPositiveGrantStatement> policy, IRevokeStatement stmt) {

		Set<IPositiveGrantStatement> lDep = oneStepDepend(policy, new HashSet<IPositiveGrantStatement>(), stmt);

		Set<IPositiveGrantStatement> newPolicy = new HashSet<IPositiveGrantStatement>(policy);

		for (IPositiveGrantStatement g1 : lDep) {
			if (g1.getPrivilege().equals(stmt.getPrivilege())) {

				boolean remove = true;

				INNER: for (IPositiveGrantStatement g2 : policy) {
					if (g2.getPrivilege().equals(stmt.getPrivilege()) && g2.getTarget().equals(g1.getInvoker()) && !lDep.contains(g2)) {
						remove = false;
						break INNER;
					}
				}

				if (remove)
					newPolicy.remove(g1);
			}
		}

		return newPolicy;
	}

	/**
	 * Iterative procedure that computes all the grants that might be affected by a revoke statement. Implements lDep(S,u,p,u') (see page Appendix A
	 * of our paper).
	 * 
	 * @param initialPolicy
	 *            - The initial policy.
	 * @param previousStep
	 *            - The set of grants at the previous step.
	 * @param stmt
	 *            - The revoke statement.
	 * @return The set of grants that might be affected by a revoke statement (depends on initialPolicy, previousStep, and stmt).
	 */
	private Set<IPositiveGrantStatement> oneStepDepend(Set<IPositiveGrantStatement> initialPolicy, Set<IPositiveGrantStatement> previousStep,
			IRevokeStatement stmt) {
		Set<IPositiveGrantStatement> oneStepDependOn = new HashSet<IPositiveGrantStatement>(previousStep);

		for (IPositiveGrantStatement g : initialPolicy) {
			if (g.getInvoker().equals(stmt.getInvoker()) && g.getTarget().equals(stmt.getTarget()) && g.getPrivilege().equals(stmt.getPrivilege()))
				oneStepDependOn.add(g);
		}

		for (IPositiveGrantStatement g1 : previousStep) {
			if (g1 instanceof IGrantWithOptionStatement) {
				for (IPositiveGrantStatement g2 : initialPolicy) {
					if (g2.getInvoker().equals(g1.getTarget()) && g1.getPrivilege().equals(g2.getPrivilege()))
						oneStepDependOn.add(g2);
				}
			}
		}

		if (oneStepDependOn.equals(previousStep))
			return oneStepDependOn;
		else
			return oneStepDepend(initialPolicy, oneStepDependOn, stmt);
	}

	@Override
	public Pair<IActionResult, IState> execute(ICreateTableStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {

					for (IDbObject obj : state.getTables())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getViews())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getTriggers())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					try {
						// modify the system's state
						Set<ITable> newTables = state.getTables();
						newTables.add(stmt.getObject());

						executeCreateOnDatabase(stmt, state);

						// create table cannot be executed in transactions.. invoker = null and rollback state = null
						// scheduledTriggers = \emptyset
						IState newState = stateFactory.createNewState(state.getUsers(), newTables, state.getPolicy(), state.getViews(),
								state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);

						IActionResult newResult = new BooleanActionResult(true, false, null);

						return new Pair<IActionResult, IState>(newResult, newState);

					} catch (MalformedStateException | MalformedResultException e) {
						throw new ExecutionException(e);
					}

				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	@Override
	public Pair<IActionResult, IState> execute(ICreateTriggerStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {

					for (IDbObject obj : state.getTables())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getViews())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getTriggers())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					try {
						// modify the system's state
						Set<ITrigger> newTriggers = state.getTriggers();
						newTriggers.add(stmt.getObject());

						// we do not execute the CREATE TRIGGER statement on the database because we handle all the triggers manually

						// create trigger cannot be executed in transactions.. invoker = null and rollback state = null
						// scheduledTriggers = \emptyset
						IState newState = stateFactory.createNewState(state.getUsers(), state.getTables(), state.getPolicy(), state.getViews(),
								newTriggers, null, null, new TreeSet<ITrigger>(), null, null);

						IActionResult newResult = new BooleanActionResult(true, false, null);

						return new Pair<IActionResult, IState>(newResult, newState);

					} catch (MalformedStateException | MalformedResultException e) {
						throw new ExecutionException(e);
					}

				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	@Override
	public Pair<IActionResult, IState> execute(ICreateViewStatement stmt, IState state) throws ExecutionException {
		if (stmt != null)
			if (state != null) {
				if (state.getUsers().contains(stmt.getInvoker())) {

					for (IDbObject obj : state.getTables())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getViews())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					for (IDbObject obj : state.getTriggers())
						if (obj.getId().equalsIgnoreCase(stmt.getObject().getId()))
							throw new ExecutionException("The id already exists in the current state");

					try {
						// modify the system's state
						Set<IView> newViews = state.getViews();
						newViews.add(stmt.getObject());

						executeCreateOnDatabase(stmt, state);

						// create view cannot be executed in transactions.. invoker = null and rollback state = null
						// scheduledTriggers = \emptyset
						IState newState = stateFactory.createNewState(state.getUsers(), state.getTables(), state.getPolicy(), newViews,
								state.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);

						IActionResult newResult = new BooleanActionResult(true, false, null);

						return new Pair<IActionResult, IState>(newResult, newState);

					} catch (MalformedStateException | MalformedResultException e) {
						throw new ExecutionException(e);
					}

				} else
					throw new ExecutionException("Invoker is not in the state.");
			} else
				throw new ExecutionException("State cannot be null.");
		else
			throw new ExecutionException("Stmt cannot be null.");
	}

	/**
	 * Execute the actual create state on the database.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @throws ExecutionException
	 */
	protected abstract void executeCreateOnDatabase(ICreateStatement stmt, IState state) throws ExecutionException;

	@Override
	public abstract Pair<IActionResult, IState> execute(ISelectStatement stmt, IState state) throws ExecutionException;

	@Override
	public abstract Pair<IActionResult, IState> execute(IInsertStatement stmt, IState state) throws ExecutionException;

	@Override
	public abstract Pair<IActionResult, IState> execute(IDeleteStatement stmt, IState state) throws ExecutionException;

}
