package prototype.system.implementation;

import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IView;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.system.MalformedStateException;
import prototype.system.interfaces.IState;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Class representing abstract system states. The abstract class takes care of everything but the representation (and roll-back) of the database
 * content. The database content should be dealt with in a system-dependent way, e.g., for MySql the database content can be represented just by the
 * id of a save-point in a transaction.
 * 
 * 
 * 
 *
 */
public abstract class AbstractState implements IState {

	/**
	 * The users in the state.
	 */
	private Set<IUser> users;

	/**
	 * The tables.
	 */
	private Set<ITable> tables;

	/**
	 * The policy.
	 */
	private Set<IPositiveGrantStatement> policy;

	/**
	 * The views.
	 */
	private Set<IView> views;

	/**
	 * The triggers.
	 */
	private Set<ITrigger> triggers;

	/**
	 * The roll-back state.
	 */
	private IState rollbackState;

	/**
	 * The invoker.
	 */
	private IUser invoker;

	/**
	 * The scheduled triggers.
	 */
	private SortedSet<ITrigger> scheduledTriggers;

	/**
	 * The current tuple.
	 */
	private ITuple tuple;

	/**
	 * The trigger we are currently executing.
	 */
	private ITrigger currentTrigger;

	/**
	 * Constructor.
	 * 
	 * @param users
	 *            - The users.
	 * @param tables
	 *            - The tables.
	 * @param policy
	 *            - The policy.
	 * @param views
	 *            - The set of views.
	 * @param triggers
	 *            - The set of triggers.
	 * @throws MalformedStateException
	 */
	public AbstractState(Set<IUser> users, Set<ITable> tables, Set<IPositiveGrantStatement> policy, Set<IView> views, Set<ITrigger> triggers)
			throws MalformedStateException {
		if (tables != null) {
			if (users != null && !users.isEmpty()) {
				if (policy != null) {
					if (views != null) {
						if (triggers != null) {

							// check all users used in policy, views, triggers are in users
							for (IView v : views)
								if (!users.contains(v.getOwner()))
									throw new MalformedStateException("All view's owners must be in users.");

							for (ITrigger v : triggers)
								if (!users.contains(v.getOwner()))
									throw new MalformedStateException("All trigger's owners must be in users.");

							for (ITable v : tables)
								if (!users.contains(v.getOwner()))
									throw new MalformedStateException("All table's owners must be in users.");

							for (IPositiveGrantStatement v : policy)
								if (!users.contains(v.getInvoker()) || !users.contains(v.getTarget()))
									throw new MalformedStateException("All users in the policy must be in users.");

							this.users = users;
							this.tables = tables;
							this.policy = policy;
							this.triggers = triggers;
							this.views = views;
							this.invoker = null;
							this.rollbackState = null;
							this.scheduledTriggers = new TreeSet<ITrigger>();
							this.tuple = null;
							this.currentTrigger = null;
						} else
							throw new MalformedStateException("Triggers cannot be null.");
					} else
						throw new MalformedStateException("Views cannot be null.");

				} else
					throw new MalformedStateException("Policy cannot be null.");
			} else
				throw new MalformedStateException("Users cannot be null or empty.");
		} else
			throw new MalformedStateException("Tables cannot be null");
	}

	/**
	 * Constructor.
	 * 
	 * @param users
	 *            - The users.
	 * @param tables
	 *            - The tables.
	 * @param policy
	 *            - The policy.
	 * @param views
	 *            - The set of views.
	 * @param triggers
	 *            - The set of triggers.
	 * @param invoker
	 *            - The invoker.
	 * @param rollback
	 *            - The rollback state
	 * @param scheduledTriggers
	 *            - the set of scheduled triggers
	 * @param tuple
	 *            - the current tuple.
	 * @throws MalformedStateException
	 */
	public AbstractState(Set<IUser> users, Set<ITable> tables, Set<IPositiveGrantStatement> policy, Set<IView> views, Set<ITrigger> triggers,
			IUser invoker, IState rollback, SortedSet<ITrigger> scheduledTriggers, ITuple tuple, ITrigger trigger) throws MalformedStateException {
		this(users, tables, policy, views, triggers);
		if ((invoker != null && rollback != null && scheduledTriggers != null && tuple != null )
				|| (invoker == null && rollback == null && scheduledTriggers != null && tuple == null )) {

			for (ITrigger v : scheduledTriggers)
				if (!triggers.contains(v))
					throw new MalformedStateException("All scheduled trigger must be in triggers.");

			this.invoker = invoker;
			this.rollbackState = rollback;
			this.scheduledTriggers = scheduledTriggers;
			this.tuple = tuple;
			this.currentTrigger = trigger;
		} else
			throw new MalformedStateException(
					"Rollback, invoker, and tuple can either be all  null or not null. Scheduled Trigger cannot be null.");
	}

	@Override
	public Set<IUser> getUsers() {
		return new HashSet<IUser>(this.users);
	}

	@Override
	public Set<IPositiveGrantStatement> getPolicy() {
		return new HashSet<IPositiveGrantStatement>(this.policy);
	}

	@Override
	public Set<IView> getViews() {
		return new HashSet<IView>(views);
	}

	@Override
	public Set<ITrigger> getTriggers() {
		return new HashSet<ITrigger>(triggers);
	}

	@Override
	public boolean isTransaction() {
		return !scheduledTriggers.isEmpty();
	}

	@Override
	public IState getRollbackState() {
		return rollbackState;
	}

	@Override
	public IUser getInvoker() {
		return invoker;
	}

	@Override
	public Set<ITable> getTables() {
		return new HashSet<ITable>(tables);
	}

	@Override
	public SortedSet<ITrigger> getScheduledTriggers() {
		return new TreeSet<ITrigger>(scheduledTriggers);
	}

	@Override
	public ITuple getTuple() {
		return this.tuple;
	}

	@Override
	public ITrigger getCurrentTrigger() {
		return this.currentTrigger;
	}

}
