package prototype.system.implementation;

import prototype.objects.ITrigger;
import prototype.pdp.IAccessControlMechanism;
import prototype.results.IResult;
import prototype.results.accessControlResult.IAccessControlResult;
import prototype.results.accessControlResult.IActionAccessControlResult;
import prototype.results.actionResults.IActionResult;
import prototype.results.actionResults.IBooleanActionResult;
import prototype.results.implementation.ActionAccessControlResult;
import prototype.results.implementation.Explanation;
import prototype.results.implementation.MalformedResultException;
import prototype.results.implementation.Result;
import prototype.results.implementation.TriggerAccessControlResult;
import prototype.results.implementation.TriggerResult;
import prototype.statements.MalformedStatementException;
import prototype.statements.abstractImplementation.IStatementFactory;
import prototype.statements.interfaces.IStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.system.MalformedStateException;
import prototype.system.interfaces.IDatabaseEngine;
import prototype.system.interfaces.IState;
import prototype.system.interfaces.ISystem;
import prototype.util.Logger;
import prototype.util.Pair;
import prototype.util.Logger.TimeProperty;

/**
 * Class representing the system. It deals with everything.
 * 
 * 
 *
 */
public class System implements ISystem {

	/**
	 * The system's current state.
	 */
	private IState currentState;
	/**
	 * The access control mechanism.
	 */
	private IAccessControlMechanism mechanism;

	/**
	 * The statement factory.
	 */
	private IStatementFactory statementFactory;

	/**
	 * The execution engine.
	 */
	private IDatabaseEngine executionEngine;
	/**
	 * The logger.
	 */
	private Logger logger;
	/**
	 * Logging Flag.
	 */
	private boolean logging;

	/**
	 * Constructor.
	 * 
	 * @param mechanism
	 *            - The access control mechanism used in the system.
	 * @param engine
	 *            - The execution engine used in the system.
	 * @param statementFactory
	 *            -the statement factory.
	 * @throws MalformedStateException
	 */
	public System(IAccessControlMechanism mechanism, IDatabaseEngine engine, IStatementFactory statementFactory, Logger logger)
			throws ExecutionException {
		if (mechanism != null) {
			if (engine != null) {
				if (statementFactory != null) {
					this.mechanism = mechanism;
					this.statementFactory = statementFactory;
					this.executionEngine = engine;
					this.currentState = engine.getInitialState();
					if (logger != null)
						this.logging = true;
					else
						this.logging = false;
					this.logger = logger;
				} else
					throw new ExecutionException("The statement factory cannot be null.");
			} else
				throw new ExecutionException("The Execution engine cannot be null.");
		} else
			throw new ExecutionException("The mechanism cannot be null.");

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.ISystem#getSystemState()
	 */
	@Override
	public IState getSystemState() {
		return this.currentState;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.ISystem#getAccessControlMechanism()
	 */
	@Override
	public IAccessControlMechanism getAccessControlMechanism() {
		return this.mechanism;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.ISystem#execute(guarnieri.marco.ethz.statements.interfaces.IStatement)
	 */
	@Override
	public IResult execute(IStatement statement) throws ExecutionException {
		if (statement != null) {
			// we can execute a statement iff all the "scheduled triggers" have been executed
			if (currentState.getScheduledTriggers().isEmpty()) {
				
				
				if(logging)
					logger.startNewStep(statement.getDefinition());
				
				
				// check the PDP result and execute the statement.
				if (mechanism.authorizes(statement, currentState)) {
					// compute the access control result
					try {
						IAccessControlResult accResult = new ActionAccessControlResult(true, null);

						if(logging)
							logger.startMeasure(TimeProperty.COMMAND_EXECUTION);
						// execute the statement
						Pair<IActionResult, IState> pair = statement.execute(executionEngine, currentState);
						
						if(logging)
							logger.endMeasure(TimeProperty.COMMAND_EXECUTION);

						currentState = pair.getSecond(); // update the current state

						IResult result = new Result(pair.getFirst(), accResult);
						if(logging)
							logger.write();
						
						return result;
					} catch (MalformedResultException e) {
						throw new ExecutionException(e);
					}

				} else {
					// the statement is not authorized
					try {
						IAccessControlResult accResult = new ActionAccessControlResult(false, new Explanation(statement.toString()
								+ " is not authorized"));

						IResult result = new Result(null, accResult);

						// in this case there is no need to roll-back.. nothing has been done

						if(logging)
							logger.write();
						return result;

					} catch (MalformedResultException e) {
						throw new ExecutionException(e);
					}
				}
				
				
			} else
				throw new ExecutionException("Execute first the scheduled triggers, then submit a new statement.");

		} else
			throw new ExecutionException("statement cannot be null.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.ISystem#executeScheduledTrigger()
	 */
	@Override
	public Pair<ITrigger, IResult> executeScheduledTrigger() throws ExecutionException {
		if (currentState != null) {
			if (!currentState.getScheduledTriggers().isEmpty()) {
				try {
					
					
					ITrigger toExecute = currentState.getScheduledTriggers().first();

					IActionAccessControlResult conditionAccResult = null;
					IBooleanActionResult conditionResult = null;
					IActionAccessControlResult actionAccResult = null;
					IBooleanActionResult actionResult = null;

					
					
					// execute the condition
					ISelectStatement condition = toExecute.getConcreteCondition(currentState.getInvoker(), currentState.getTuple(), statementFactory);
					
					if(logging)
						logger.startNewStep(toExecute.getId()+" CONDITION "+condition);
					
					// check the PDP result and execute the statement.
					if (mechanism.authorizes(condition, currentState)) {
						// compute the access control result
						try {
							conditionAccResult = new ActionAccessControlResult(true, null);

							if(logging)
								logger.startMeasure(TimeProperty.COMMAND_EXECUTION);
							
							// execute the statement
							Pair<IActionResult, IState> pair = condition.execute(executionEngine, currentState);
							
							if(logging)
								logger.endMeasure(TimeProperty.COMMAND_EXECUTION);

							currentState = pair.getSecond(); // update the current state

							conditionResult = (IBooleanActionResult) pair.getFirst();
							
							if(logging)
								logger.write();

						} catch (MalformedResultException e) {
							throw new ExecutionException(e);
						}

					} else {
						// the statement is not authorized
						try {
							conditionAccResult = new ActionAccessControlResult(false, new Explanation(condition.toString() + " is not authorized"));

							conditionResult = null;
							
							if(logging)
								logger.write();

							// we must roll-back to currentState.getRollBackState()
							currentState = executionEngine.rollbackTo(currentState.getRollbackState());

						} catch (MalformedResultException e) {
							throw new ExecutionException(e);
						}
					}

					if (conditionAccResult.getResult() && conditionResult != null && conditionResult.getResult()) {
						// the condition is satisfied and the trigger is enabled.
						// execute the condition.

						IStatement action = toExecute.getConcreteAction(currentState.getInvoker(), currentState.getTuple(), statementFactory);
						
						if(logging)
							logger.startNewStep(toExecute.getId()+" ACTION "+action);
						
						// check the PDP result and execute the statement.
						if (mechanism.authorizes(action, currentState)) {
							// compute the access control result
							try {
								actionAccResult = new ActionAccessControlResult(true, null);

								if(logging)
									logger.startMeasure(TimeProperty.COMMAND_EXECUTION);
								
								// execute the statement
								Pair<IActionResult, IState> pair = action.execute(executionEngine, currentState);
								
								
								if(logging)
									logger.endMeasure(TimeProperty.COMMAND_EXECUTION);


								currentState = pair.getSecond(); // update the current state

								actionResult = (IBooleanActionResult) pair.getFirst();
								

								if(logging)
									logger.write();

							} catch (MalformedResultException e) {
								throw new ExecutionException(e);
							}

						} else {
							// the statement is not authorized
							try {
								actionAccResult = new ActionAccessControlResult(false, new Explanation(action.toString() + " is not authorized"));
								actionResult = null;


								if(logging)
									logger.write();
								
								// we must roll-back to currentState.getRollBackState()
								currentState = executionEngine.rollbackTo(currentState.getRollbackState());

							} catch (MalformedResultException e) {
								throw new ExecutionException(e);
							}
						}
					}

					// generate the IResult
					IResult result = null;

					if (conditionAccResult != null && conditionResult != null && actionAccResult != null && actionResult != null)
						result = new Result(new TriggerResult(conditionResult.getResult(), actionResult.getResult(), actionResult.exception(),
								actionResult.getExceptionExplanation()), new TriggerAccessControlResult(conditionAccResult.getResult(),
								conditionAccResult.getExplanation(), actionAccResult.getResult(), actionAccResult.getExplanation()));
					else if (conditionAccResult != null && conditionResult != null && actionAccResult != null && actionResult == null)
						result = new Result(new TriggerResult(conditionResult.getResult(), false, false, null), new TriggerAccessControlResult(
								conditionAccResult.getResult(), conditionAccResult.getExplanation(), actionAccResult.getResult(),
								actionAccResult.getExplanation()));
					else if (conditionAccResult != null && conditionResult != null && actionAccResult == null && actionResult == null)
						result = new Result(new TriggerResult(conditionResult.getResult(), false, false, null), new TriggerAccessControlResult(
								conditionAccResult.getResult(), conditionAccResult.getExplanation()));
					else if (conditionAccResult != null && conditionResult == null && actionAccResult == null && actionResult == null)
						result = new Result(new TriggerResult(false, false, false, null), new TriggerAccessControlResult(
								conditionAccResult.getResult(), conditionAccResult.getExplanation()));

					return new Pair<ITrigger, IResult>(toExecute, result);
				} catch (MalformedStatementException | MalformedResultException e) {
					throw new ExecutionException(e);
				}
			} else
				throw new ExecutionException("There are no scheduled triggers in the current state.");
		} else
			throw new ExecutionException("CurrentStatement is null.");
	}

	@Override
	public boolean noScheduledTriggers() throws ExecutionException {
		if (currentState != null) {
			return currentState.getScheduledTriggers().isEmpty();
		} else
			throw new ExecutionException("CurrentStatement is null.");
	}

	@Override
	public void close() throws ExecutionException {
		this.executionEngine.close();
	}
}
