package prototype.system.interfaces;

import java.util.Set;

import prototype.queries.domain.relational.calculus.BooleanQuery;
import prototype.results.actionResults.IActionResult;
import prototype.statements.interfaces.add.IAddUser;
import prototype.statements.interfaces.changePolicy.IGrantStatement;
import prototype.statements.interfaces.changePolicy.IGrantWithOptionStatement;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.changePolicy.IRevokeStatement;
import prototype.statements.interfaces.create.ICreateTableStatement;
import prototype.statements.interfaces.create.ICreateTriggerStatement;
import prototype.statements.interfaces.create.ICreateViewStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.system.implementation.ExecutionException;
import prototype.util.Pair;

/**
 * Interface for the database engine, which executes the statements.
 * 
 * 
 *
 */
public interface IDatabaseEngine {

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 * @throws ExecutionException
	 */
	public Pair<IActionResult, IState> execute(IAddUser stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(IGrantStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(IGrantWithOptionStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(IRevokeStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(ICreateTableStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(ICreateTriggerStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(ICreateViewStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(ISelectStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(IInsertStatement stmt, IState state) throws ExecutionException;

	/**
	 * Execute the statement and modify the state.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The statement's result.
	 */
	public Pair<IActionResult, IState> execute(IDeleteStatement stmt, IState state) throws ExecutionException;

	/**
	 * Get the initial state.
	 * 
	 * @return The initial state.
	 * @throws ExecutionException
	 */
	public IState getInitialState() throws ExecutionException;

	/**
	 * Roll-back to the given state. Return the new state.
	 * 
	 * @param state
	 *            - The state to roll-back
	 * @return The new state.
	 * @throws ExecutionException
	 */
	public IState rollbackTo(IState state) throws ExecutionException;

	/**
	 * Close the database connection.
	 * 
	 * @throws ExecutionException
	 */
	public void close() throws ExecutionException;

	/**
	 * Evaluate a boolean query. Note that it bypasses all the checks and it does not update the state.
	 * 
	 * @param query
	 *            - The query to evaluate.
	 * @param state
	 *            - The state.
	 * @return The query's result.
	 * @throws ExecutionException
	 */
	public boolean evaluate(ISelectStatement query, IState state) throws ExecutionException;

	/**
	 * Materialize non-boolean query.
	 * 
	 * @param query
	 *            - the non-boolean query
	 * @param state
	 *            - the state.
	 * @return The non-boolean query representing the materialization of query's result in state.
	 * @throws ExecutionException 
	 */
	public BooleanQuery materializeQuery(ISelectStatement query, IState state) throws ExecutionException;

	
	/**
	 * Compute the policy after the revoke statement. 
	 * 
	 * @param policy
	 *            - The initial policy.
	 * @param stmt
	 *            - The revoke statement.
	 * @return The new policy.
	 * 
	 */
	public Set<IPositiveGrantStatement> revoke(Set<IPositiveGrantStatement> policy, IRevokeStatement stmt);
}
