package prototype.system.interfaces;

import java.util.Set;
import java.util.SortedSet;

import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IView;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.tuple.ITuple;
import prototype.users.IUser;

/**
 * Interface representing the system's state.
 * 
 * 
 *
 */
public interface IState {

	/**
	 * Get the users in the state.
	 * 
	 * @return The set of users.
	 */
	public Set<IUser> getUsers();

	/**
	 * Get the policy in the state.
	 * 
	 * @return The policy.
	 */
	public Set<IPositiveGrantStatement> getPolicy();

	/**
	 * Get the views in the state.
	 * 
	 * @return The set of views.
	 */
	public Set<IView> getViews();

	/**
	 * Get the triggers in the state.
	 * 
	 * @return The set of triggers.
	 */
	public Set<ITrigger> getTriggers();

	/**
	 * Check whether the state is part of a transaction, i.e., inside the execution of triggers.
	 * 
	 * @return True if the state is in a transaction, False otherwise.
	 */
	public boolean isTransaction();

	/**
	 * Get the roll-back state.
	 * 
	 * @return The roll-back state if we are in a transaction, i.e., isTransaction() == true. Null otherwise.
	 */
	public IState getRollbackState();

	/**
	 * Get the user who started the transaction.
	 * 
	 * @return The user who started the transaction if we are in a transaction, i.e., isTransaction() == true. Null otherwise.
	 */
	public IUser getInvoker();

	/**
	 * Get the tables.
	 * 
	 * @return The tables.
	 */
	public Set<ITable> getTables();

	/**
	 * Get the list of scheduled triggers (the ones that we still have to execute).
	 * 
	 * @return The list of scheduled triggers.
	 */
	public SortedSet<ITrigger> getScheduledTriggers();

	/**
	 * Get the current tuple.
	 * 
	 * @return The current tuple if we are in a transaction, i.e., isTransaction() == true. Null otherwise.
	 */
	public ITuple getTuple();

	/**
	 * Get the trigger we are executing right now.
	 * 
	 * @return The trigger we are executing right now or null.
	 */
	public ITrigger getCurrentTrigger();

}
