package prototype.system.interfaces;

import prototype.objects.ITrigger;
import prototype.pdp.IAccessControlMechanism;
import prototype.results.IResult;
import prototype.statements.interfaces.IStatement;
import prototype.system.implementation.ExecutionException;
import prototype.util.Pair;

/**
 * Interface representing the system. The system is in charge of updating the system's state and executing the statements.
 * 
 * 
 *
 */
public interface ISystem {

	/**
	 * Get the current system's state.
	 * 
	 * @return The current system's state.
	 */
	public IState getSystemState();

	/**
	 * Get the access control mechanism used in the system.
	 * 
	 * @return The access control mechanism.
	 */
	public IAccessControlMechanism getAccessControlMechanism();

	/**
	 * Execute the statement.
	 * 
	 * @param statement
	 *            - The statement to execute.
	 * @return The statement's result.
	 * @throws ExecutionException
	 */
	public IResult execute(IStatement statement) throws ExecutionException;

	/**
	 * Execute the first scheduled trigger in the system.
	 * 
	 * @return A pair containing the trigger and its result.
	 * @throws ExecutionException
	 */
	public Pair<ITrigger, IResult> executeScheduledTrigger() throws ExecutionException;

	/**
	 * Check whether there are scheduled triggers.
	 * 
	 * @return True if there are scheduled triggers, false otherwise.
	 * @throws ExecutionException
	 */
	public boolean noScheduledTriggers() throws ExecutionException;

	/**
	 * Close the database connection.
	 * 
	 * @throws ExecutionException
	 */
	public void close() throws ExecutionException;

}
