package prototype.system.mysql;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import prototype.database.mysql.MySqlConfiguration;
import prototype.database.mysql.MySqlConnectionManager;
import prototype.objects.ITable;
import prototype.objects.ITrigger;
import prototype.objects.IType;
import prototype.objects.IView;
import prototype.objects.implementation.triggers.events.EVENT;
import prototype.pdp.secure.UniformVariableAssignmentVisitor;
import prototype.queries.MalformedExpressionException;
import prototype.queries.MalformedQueryException;
import prototype.queries.domain.relational.calculus.BooleanQuery;
import prototype.queries.domain.relational.calculus.INonBooleanQuery;
import prototype.queries.domain.relational.calculus.syntax.IExpression;
import prototype.queries.domain.relational.calculus.syntax.IOperand;
import prototype.queries.domain.relational.calculus.syntax.expressions.Constant;
import prototype.queries.domain.relational.calculus.syntax.expressions.EqualityExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.PolyadicExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.QuantifiedExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.UnaryExpression;
import prototype.queries.domain.relational.calculus.syntax.expressions.Variable;
import prototype.queries.domain.relational.calculus.visitors.ReplaceFreeVariablesVisitor;
import prototype.queries.operators.AndOperator;
import prototype.queries.operators.NotOperator;
import prototype.queries.operators.OrOperator;
import prototype.queries.operators.UniversalQuantifier;
import prototype.queries.translation.GetConstants;
import prototype.results.actionResults.IActionResult;
import prototype.results.explanation.IIntegrityExplanation;
import prototype.results.implementation.BooleanActionResult;
import prototype.results.implementation.Explanation;
import prototype.results.implementation.MalformedResultException;
import prototype.results.implementation.NonBooleanActionResult;
import prototype.statements.interfaces.TranslationException;
import prototype.statements.interfaces.changePolicy.IPositiveGrantStatement;
import prototype.statements.interfaces.create.ICreateStatement;
import prototype.statements.interfaces.create.ICreateTableStatement;
import prototype.statements.interfaces.create.ICreateViewStatement;
import prototype.statements.interfaces.select.ISelectStatement;
import prototype.statements.interfaces.update.IDeleteStatement;
import prototype.statements.interfaces.update.IInsertStatement;
import prototype.statements.interfaces.update.IUpdateStatement;
import prototype.system.MalformedStateException;
import prototype.system.implementation.AbstractExecutionEngine;
import prototype.system.implementation.ExecutionException;
import prototype.system.interfaces.IState;
import prototype.tuple.ITuple;
import prototype.tuple.Tuple;
import prototype.users.IUser;
import prototype.users.MalformedUserException;
import prototype.users.UserFactory;
import prototype.util.Pair;
import prototype.util.Util;

/**
 * Database engine for the MySql database.
 * 
 * 
 *
 */
public class MySqlDatabaseEngine extends AbstractExecutionEngine {

	/**
	 * The connection manager.
	 */
	private MySqlConnectionManager manager;
	/**
	 * State factory.
	 */
	private MySqlStateFactory stateFactory;
	/**
	 * Set of types (tables and views) created in the database.
	 */
	private Vector<IType> createdTypes;

	/**
	 * If true, creates (and keeps updated) a view with name "__ADOM__" representing the active domain.
	 */
	private boolean keepActiveDomain = true;

	/**
	 * Constructor.
	 * 
	 * @param configuration
	 *            - The database configuration.
	 * @throws ExecutionException
	 */
	public MySqlDatabaseEngine(MySqlConfiguration configuration, MySqlStateFactory factory) throws ExecutionException {
		super(factory);
		if (configuration != null) {
			this.manager = new MySqlConnectionManager(configuration);
			this.manager.setAutoCommit(false); // we always proceed inside a transaction.
			this.stateFactory = factory;
			factory.setConnectionManager(this.manager);
			createdTypes = new Vector<IType>();

			// create active domain
			createActiveDomain();
		} else
			throw new ExecutionException("Configuration cannot be null.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.abstractImplementation.AbstractExecutionEngine#execute(guarnieri.marco.ethz.statements.interfaces.select.
	 * ISelectStatement, guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public Pair<IActionResult, IState> execute(ISelectStatement stmt, IState state) throws ExecutionException {
		if (stmt != null && state != null) {

			if (state.isTransaction()) {

				if (state instanceof MySqlState && state.getRollbackState() instanceof MySqlState) {
					// we are executing the trigger's condition

					return executeSelectStatement(stmt, (MySqlState) state, (MySqlState) state.getRollbackState(), state.getInvoker(),
							state.getRollbackState(), state.getScheduledTriggers(), state.getTuple(), state.getCurrentTrigger(), true);
				} else
					throw new ExecutionException("State must be a MySqlState.");
			} else {
				if (state instanceof MySqlState) {
					MySqlState currentState = (MySqlState) state;

					// we are executing an action directly issued by a user
					// set next invoker and next rollback state to null (SELECT statements cannot schedule triggers)
					return executeSelectStatement(stmt, currentState, currentState, null, null, new TreeSet<ITrigger>(), null, null, false);

				} else
					throw new ExecutionException("State must be a MySqlState.");
			}

		} else
			throw new ExecutionException("Stmt and state cannot be null");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.abstractImplementation.AbstractExecutionEngine#execute(guarnieri.marco.ethz.statements.interfaces.update.
	 * IInsertStatement, guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public Pair<IActionResult, IState> execute(IInsertStatement stmt, IState state) throws ExecutionException {
		if (stmt != null && state != null) {

			if (state.isTransaction()) {

				if (state instanceof MySqlState && state.getRollbackState() instanceof MySqlState) {

					// we are executing the trigger's action
					// drop the first trigger in state.getScheduledTriggers() (the one we are currently executing)
					// set it as the current trigger.
					SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>(state.getScheduledTriggers());
					ITrigger old = state.getScheduledTriggers().first();
					scheduled.remove(old);

					ITrigger toExecute;
					if (!scheduled.isEmpty())
						toExecute = scheduled.first();
					else
						toExecute = null;

					return executeUpdateStatement(stmt, (MySqlState) state, (MySqlState) state.getRollbackState(), state.getInvoker(),
							state.getRollbackState(), scheduled, state.getTuple(), toExecute);
				} else
					throw new ExecutionException("State must be a MySqlState.");
			} else {
				if (state instanceof MySqlState) {
					MySqlState currentState = (MySqlState) state;

					SortedSet<ITrigger> scheduled = getScheduledTriggers(stmt, currentState);

					// we are executing an action directly issued by a user
					if (scheduled.isEmpty() || !isDatabaseModified(stmt))
						// set next invoker and next rollback state to null
						return executeUpdateStatement(stmt, currentState, currentState, null, null, new TreeSet<ITrigger>(), null, null);
					else
						// set next invoker and next rollback state to the proper values
						return executeUpdateStatement(stmt, currentState, currentState, stmt.getInvoker(), currentState, scheduled, stmt.getTuple(),
								scheduled.first());

				} else
					throw new ExecutionException("State must be a MySqlState.");
			}

		} else
			throw new ExecutionException("Stmt and state cannot be null");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.abstractImplementation.AbstractExecutionEngine#execute(guarnieri.marco.ethz.statements.interfaces.update.
	 * IDeleteStatement, guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public Pair<IActionResult, IState> execute(IDeleteStatement stmt, IState state) throws ExecutionException {
		if (stmt != null && state != null) {

			if (state.isTransaction()) {

				if (state instanceof MySqlState && state.getRollbackState() instanceof MySqlState) {
					// we are executing the trigger's action
					// drop the first trigger in state.getScheduledTriggers() (the one we are currently executing)
					// set it as the current trigger.
					SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>(state.getScheduledTriggers());
					ITrigger old = state.getScheduledTriggers().first();
					scheduled.remove(old);

					ITrigger toExecute;
					if (!scheduled.isEmpty())
						toExecute = scheduled.first();
					else
						toExecute = null;

					return executeUpdateStatement(stmt, (MySqlState) state, (MySqlState) state.getRollbackState(), state.getInvoker(),
							state.getRollbackState(), scheduled, state.getTuple(), toExecute);
				} else
					throw new ExecutionException("State must be a MySqlState.");
			} else {
				if (state instanceof MySqlState) {
					MySqlState currentState = (MySqlState) state;

					SortedSet<ITrigger> scheduled = getScheduledTriggers(stmt, currentState);

					// we are executing an action directly issued by a user
					if (scheduled.isEmpty() || !isDatabaseModified(stmt))
						// set next invoker and next rollback state to null
						return executeUpdateStatement(stmt, currentState, currentState, null, null, new TreeSet<ITrigger>(), null, null);
					else
						// set next invoker and next rollback state to the proper values
						return executeUpdateStatement(stmt, currentState, currentState, stmt.getInvoker(), currentState, scheduled, stmt.getTuple(),
								scheduled.first());

				} else
					throw new ExecutionException("State must be a MySqlState.");
			}

		} else
			throw new ExecutionException("Stmt and state cannot be null");
	}

	/**
	 * Compute the set of triggers scheduled in response to the execution of stmt.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The state.
	 * @return The set of scheduled triggers.
	 */
	private SortedSet<ITrigger> getScheduledTriggers(IUpdateStatement stmt, IState state) {
		SortedSet<ITrigger> scheduled = new TreeSet<ITrigger>();
		for (ITrigger trigger : state.getTriggers()) {
			if (trigger.getEvent().getEvent() == EVENT.INSERT && stmt instanceof IInsertStatement
					&& trigger.getEvent().getRelationSchema().equals(stmt.getRelationSchema()))
				scheduled.add(trigger);
			if (trigger.getEvent().getEvent() == EVENT.DELETE && stmt instanceof IDeleteStatement
					&& trigger.getEvent().getRelationSchema().equals(stmt.getRelationSchema()))
				scheduled.add(trigger);
		}
		return scheduled;
	}

	/**
	 * Execute the update statement.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The initial state.
	 * @param rollbackState
	 *            - The rollback state.
	 * @param invoker
	 *            - The invoker that must be set in the resulting state.
	 * @param nextRollbackState
	 *            - The rollback state that must be set in the resulting state.
	 * @param scheduledTriggers
	 *            - The set of scheduled triggers.
	 * @param tuple
	 *            - The current tuple.
	 * @return The result.
	 * @throws ExecutionException
	 */
	private Pair<IActionResult, IState> executeUpdateStatement(IUpdateStatement stmt, MySqlState state, MySqlState rollbackState, IUser invoker,
			IState nextRollbackState, SortedSet<ITrigger> scheduledTriggers, ITuple tuple, ITrigger toExecute) throws ExecutionException {
		Statement sqlStmt = manager.createStatement();

		boolean exception = false;
		IIntegrityExplanation expl = null;
		// execute the statement
		try {

			// if stmt actually modifies the database (in the relational calculus where relations are sets), then we actually execute it.
			// otherwise we just create the new state (the database will be the same and there will be no exceptions).
			if (isDatabaseModified(stmt)) {
				sqlStmt.executeUpdate(stmt.toSQL());
			}

		} catch (SQLException e) {
			// check whether the execution has caused an integrity constraint violation.
			if (e.getSQLState().equals(MySqlConnectionManager.getIntegrityExceptionCode())) {
				// integrity exception

				// create the explanation
				expl = new Explanation(e.getMessage());
				exception = true;

				// rollback to the savepoint just created
				manager.rollback(rollbackState.getSavePoint());

			} else {
				// other issues
				throw new ExecutionException(e);
			}

		} catch (TranslationException e) {
			throw new ExecutionException(e);
		} finally {

			try {
				sqlStmt.close();
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		}

		try {
			IActionResult result = new BooleanActionResult(!exception, exception, expl);

			// create the new state
			MySqlState newState = null;

			if (!exception)
				// propagate rollback state and invoker
				newState = stateFactory.createNewState(state.getUsers(), state.getTables(), state.getPolicy(), state.getViews(), state.getTriggers(),
						invoker, nextRollbackState, scheduledTriggers, tuple, toExecute);
			else
				// exception -> rollback. Set rollBackstate to null
				newState = stateFactory.createNewState(rollbackState.getUsers(), rollbackState.getTables(), rollbackState.getPolicy(),
						rollbackState.getViews(), rollbackState.getTriggers(), null, null, new TreeSet<ITrigger>(), null, null);

			Pair<IActionResult, IState> pair = new Pair<IActionResult, IState>(result, newState);

			return pair;
		} catch (MalformedResultException | MalformedStateException e) {
			throw new ExecutionException(e);
		}
	}

	/**
	 * Check whether the statement's execution actually modifies the database or not. Note that it does not modify the database.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @return True if the database will be modified by stmt's execution. False otherwise.
	 * @throws ExecutionException
	 */
	private boolean isDatabaseModified(IUpdateStatement stmt) throws ExecutionException {
		Statement sqlStmt = manager.createStatement();
		ResultSet rs = null;

		try {

			// first check whether the tuple is already in the database. If so, avoid the actual insert.

			String whereExpr = "";
			for (int i = 0; i < stmt.getTuple().getSize(); i++) {
				if (stmt.getTuple().getContent().get(i) instanceof Constant) {
					whereExpr += stmt.getRelationSchema().getFields().get(i) + " = " + stmt.getTuple().getContent().get(i) + " AND ";

				} else
					throw new ExecutionException("The tuple contains free variables.");

			}
			whereExpr = Util.replaceLast(whereExpr, " AND ", "");
			String check = "SELECT CASE WHEN EXISTS( SELECT * FROM " + stmt.getRelationSchema().getId() + " WHERE " + whereExpr
					+ " ) THEN 'true' ELSE 'false' END";

			rs = sqlStmt.executeQuery(check);
			// boolean queries return a string (either true or false)

			while (rs.next()) {
				if (rs.getString(1).equalsIgnoreCase("true"))
					if (stmt instanceof IInsertStatement)
						return false; // the tuple is already in the database => the INSERT has no effect.
					else if (stmt instanceof IDeleteStatement)
						return true; // the tuple is in the database => the DELETE has effect
					else
						assert (false); // unreachable

				else if (rs.getString(1).equalsIgnoreCase("false"))
					if (stmt instanceof IInsertStatement)
						return true; // the tuple is not in the database => the INSERT has effect.
					else if (stmt instanceof IDeleteStatement)
						return false; // the tuple is not in the database => the DELETE has no effect
					else
						assert (false); // unreachable
				else
					assert (false); // unreachable

			}

		} catch (SQLException e) {
			throw new ExecutionException(e);
		} finally {

			if (rs != null)
				try {
					rs.close();
				} catch (SQLException e) {
					e.printStackTrace();
				}

			try {
				sqlStmt.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		assert (false); // unreachable
		return false;
	}

	/**
	 * Execute the select statement.
	 * 
	 * @param stmt
	 *            - The statement.
	 * @param state
	 *            - The initial state.
	 * @param rollbackState
	 *            - The rollback state.
	 * @param invoker
	 *            - The invoker that must be set in the resulting state.
	 * @param nextRollbackState
	 *            - The rollback state that must be set in the resulting state.
	 * @param scheduledTriggers
	 *            - The set of scheduled triggers.
	 * @param tuple
	 *            - The current tuple.
	 * @param trigger
	 *            - The trigger we are executing.
	 * @return The result.
	 * @throws ExecutionException
	 */
	private Pair<IActionResult, IState> executeSelectStatement(ISelectStatement stmt, MySqlState state, MySqlState rollbackState, IUser invoker,
			IState nextRollbackState, SortedSet<ITrigger> scheduledTriggers, ITuple tuple, ITrigger trigger, boolean inTransaction)
			throws ExecutionException {
		Statement sqlStmt = manager.createStatement();
		ResultSet rs = null;
		// execute the statement
		try {

			// updateActiveDomain(stmt.getQuery().getExpression(), state);

			rs = sqlStmt.executeQuery(stmt.toSQL());

			IActionResult result = null;
			if (stmt.getQuery() instanceof INonBooleanQuery) {
				// non boolean query
				Set<ITuple> tuples = new HashSet<ITuple>();

				// iterate through the resultset
				while (rs.next()) {
					List<IOperand> tupleContent = new Vector<IOperand>();
					for (int i = 0; i < rs.getMetaData().getColumnCount(); i++) {
						try {
							tupleContent.add(new Constant(rs.getString(i + 1)));
						} catch (MalformedExpressionException e) {
							assert (false); // unreachable code
						}
					}
					tuples.add(new Tuple(tupleContent));
				}

				result = new NonBooleanActionResult(tuples, false, null);

			} else {
				// boolean queries return a string (either true or false)
				boolean res = false;
				while (rs.next()) {
					if (rs.getString(1).equalsIgnoreCase("true"))
						res = true;
					else if (rs.getString(1).equalsIgnoreCase("false"))
						res = false;
					else
						assert (false); // unreachable

				}
				result = new BooleanActionResult(res, false, null);
			}
			rs.close();
			sqlStmt.close();

			// create the new state
			if ((result instanceof BooleanActionResult) && !((BooleanActionResult) result).getResult() && inTransaction) {
				// we have just executed a trigger's condition and the trigger is not enabled

				ITrigger old = scheduledTriggers.first();
				scheduledTriggers.remove(old);

				if (!scheduledTriggers.isEmpty())
					trigger = scheduledTriggers.first();
				else
					trigger = null;

			}
			MySqlState newState = stateFactory.createNewState(state.getUsers(), state.getTables(), state.getPolicy(), state.getViews(),
					state.getTriggers(), invoker, nextRollbackState, scheduledTriggers, tuple, trigger);

			Pair<IActionResult, IState> pair = new Pair<IActionResult, IState>(result, newState);
			return pair;
		} catch (MalformedResultException | MalformedStateException | SQLException | TranslationException e) {
			throw new ExecutionException(e);
		} finally {
			try {
				// rs.close();
				sqlStmt.close();
			} catch (SQLException e) {
				e.printStackTrace();// throw new ExecutionException(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.IExecutionEngine#getInitialState()
	 */
	@Override
	public IState getInitialState() throws ExecutionException {
		try {
			Set<IUser> initialUsers = new HashSet<IUser>();
			try {
				initialUsers.add(UserFactory.getInstance().getAdmin());
			} catch (MalformedUserException e) {
				throw new ExecutionException(e);
			}

			return stateFactory.createNewState(initialUsers, new HashSet<ITable>(), new HashSet<IPositiveGrantStatement>(), new HashSet<IView>(),
					new HashSet<ITrigger>(), null, null, new TreeSet<ITrigger>(), null, null);
		} catch (MalformedStateException e) {
			throw new ExecutionException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see guarnieri.marco.ethz.system.interfaces.IExecutionEngine#rollbackTo(guarnieri.marco.ethz.system.interfaces.IState)
	 */
	@Override
	public IState rollbackTo(IState state) throws ExecutionException {
		if (state instanceof MySqlState) {

			// rollback to the savepoint
			manager.rollback(((MySqlState) state).getSavePoint());

			try {
				return stateFactory.createNewState(state.getUsers(), state.getTables(), state.getPolicy(), state.getViews(), state.getTriggers(),
						null, null, new TreeSet<ITrigger>(), null, null);
			} catch (MalformedStateException e) {
				throw new ExecutionException(e);
			}

		} else
			throw new ExecutionException("State must be a MySqlState.");
	}

	@Override
	public void close() throws ExecutionException {
		manager.rollback(null); // delete everything (note that this does not delete the tables and views)

		List<IType> reversedList = new Vector<IType>(createdTypes);
		Collections.reverse(reversedList);
		// drop all created views and tables
		for (IType type : reversedList) {
			Statement sqlStmt = manager.createStatement();

			String stmt = null;
			if (type instanceof IView)
				stmt = "DROP VIEW IF EXISTS " + type.getId() + ";";
			else if (type instanceof ITable)
				stmt = "DROP TABLE IF EXISTS " + type.getId() + ";";
			else {
				assert (false); // unreachable
			}

			// execute the statement
			try {
				sqlStmt.executeUpdate(stmt);
			} catch (SQLException e) {

				// other issues
				throw new ExecutionException(e);

			} finally {
				try {
					sqlStmt.close();
				} catch (SQLException e) {
					throw new ExecutionException(e);
				}
			}

		}

		Statement sqlStmt = manager.createStatement();

		// execute the statement
		try {
			sqlStmt.executeUpdate("DROP VIEW IF EXISTS __ADOM__;");
		} catch (SQLException e) {

			// other issues
			throw new ExecutionException(e);

		} finally {
			try {
				sqlStmt.close();
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		}
		sqlStmt = manager.createStatement();
		// execute the statement
		try {
			sqlStmt.executeUpdate("DROP TABLE IF EXISTS __ADOM__EXPR__VALUES__;");
		} catch (SQLException e) {

			// other issues
			throw new ExecutionException(e);

		} finally {
			try {
				sqlStmt.close();
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		}

		manager.close();
	}

	@Override
	protected void executeCreateOnDatabase(ICreateStatement stmt, IState state) throws ExecutionException {
		Statement sqlStmt = manager.createStatement();

		if (stmt instanceof ICreateViewStatement || stmt instanceof ICreateTableStatement)
			createdTypes.add((IType) stmt.getObject());

		// execute the statement
		try {
			sqlStmt.executeUpdate(stmt.toSQL());

			if (keepActiveDomain) {
				String expr = "CREATE OR REPLACE VIEW __ADOM__ (value) AS ";
				if (((IType) stmt.getObject()) instanceof ITable)
					for (String field : ((IType) stmt.getObject()).getFields())
						expr += "(SELECT " + field + " AS value FROM " + ((IType) stmt.getObject()).getId() + ") UNION ";
				for (IType table : state.getTables())
					for (String field : table.getFields())
						expr += "(SELECT " + field + " AS value FROM " + table.getId() + ") UNION ";

				Set<Constant> constants = new LinkedHashSet<Constant>();
				for (IView view : state.getViews()) {
					GetConstants v = new GetConstants();
					view.getSelectQuery().getQuery().getExpression().accept(v);

					constants.addAll(v.getConstants());
				}
				if (((IType) stmt.getObject()) instanceof IView) {
					GetConstants v = new GetConstants();
					((IView) stmt.getObject()).getSelectQuery().getQuery().getExpression().accept(v);
					constants.addAll(v.getConstants());
				}

				for (Constant c : constants)
					expr += "(SELECT " + c + " AS value) UNION ";

				expr += "(SELECT value FROM __ADOM__EXPR__VALUES__)";

				sqlStmt.executeUpdate(expr);
			}

		} catch (SQLException | TranslationException e) {

			// other issues
			throw new ExecutionException(e);

		} finally {
			try {
				sqlStmt.close();
			} catch (SQLException e) {
				throw new ExecutionException(e);
			}
		}

	}

	@Override
	public boolean evaluate(ISelectStatement query, IState state) throws ExecutionException {
		if (query.getQuery() instanceof INonBooleanQuery)
			throw new ExecutionException("Query must be a boolean query");

		Statement sqlStmt = manager.createStatement();
		ResultSet rs = null;

		try {

			// updateActiveDomain(query.getQuery().getExpression(), state);

			rs = sqlStmt.executeQuery(query.toSQL());
			// boolean queries return a string (either true or false)

			while (rs.next()) {
				if (rs.getString(1).equalsIgnoreCase("true"))
					return true;
				else if (rs.getString(1).equalsIgnoreCase("false"))
					return false;
				else
					assert (false); // unreachable

			}

		} catch (SQLException | TranslationException e) {
			throw new ExecutionException(e);
		} finally {

			if (rs != null)
				try {
					rs.close();
				} catch (SQLException e) {
					e.printStackTrace();
				}

			try {
				sqlStmt.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		assert (false); // unreachable
		return false;
	}

	@Override
	public BooleanQuery materializeQuery(ISelectStatement query, IState state) throws ExecutionException {

		if (!(query.getQuery() instanceof INonBooleanQuery))
			throw new ExecutionException("Query must be a non-boolean query");

		Statement sqlStmt = manager.createStatement();
		ResultSet rs = null;

		try {

			updateActiveDomain(query.getQuery().getExpression(), state);

			rs = sqlStmt.executeQuery(query.toSQL());

			List<Tuple> tuples = new Vector<Tuple>();
			while (rs.next()) {
				List<IOperand> values = new Vector<IOperand>();
				for (int i = 0; i < ((INonBooleanQuery) query.getQuery()).getOutput().size(); i++)
					values.add(new Constant(rs.getString(i + 1)));

				tuples.add(new Tuple(values));
			}

			// create the expression
			// IncrementalIdentifiers vInc = new IncrementalIdentifiers();
			UniformVariableAssignmentVisitor vInc = new UniformVariableAssignmentVisitor();
			query.getQuery().getExpression().accept(vInc);

			IExpression expr = vInc.getRenamedExpression();

			// get the assignments for the output variables
			int min = vInc.getMaximumValue();
			List<Variable> output = new Vector<Variable>();
			// List<Variable> quantified = new Vector<Variable>();
			for (Variable v : ((INonBooleanQuery) query.getQuery()).getOutput()) {
				output.add(new Variable(UniformVariableAssignmentVisitor.getPrefix() + vInc.getAssignment().get(v)));
				// min++;
				// quantified.add(new Variable(IncrementalIdentifiers.getPrefix() + min));
			}

			List<IExpression> conjuncts = new Vector<IExpression>();
			List<IExpression> disjuncts = new Vector<IExpression>();
			disjuncts.add(new UnaryExpression(NotOperator.getInstance(), expr));
			for (Tuple t : tuples) {
				ReplaceFreeVariablesVisitor vRepl = new ReplaceFreeVariablesVisitor(output, t.getContent());
				expr.accept(vRepl);

				vInc = new UniformVariableAssignmentVisitor(min);
				vRepl.getResult().accept(vInc);
				min = vInc.getMaximumValue();

				conjuncts.add(vInc.getRenamedExpression());

				List<IExpression> conjuncts1 = new Vector<IExpression>();
				for (int i = 0; i < output.size(); i++) {
					conjuncts1.add(new EqualityExpression(output.get(i), t.getContent().get(i)));
				}
				if (conjuncts1.size() > 1)
					disjuncts.add(new PolyadicExpression(AndOperator.getInstance(), conjuncts1));
				else
					disjuncts.add(conjuncts1.get(0)); // this does not cause an outOfBoundException because we materialize only non-boolean queries,
														// which always have at least one free variable.

			}
			IExpression or;
			if (disjuncts.size() > 1)
				or = new PolyadicExpression(OrOperator.getInstance(), disjuncts);
			else
				or = disjuncts.get(0);
			conjuncts.add(new QuantifiedExpression(UniversalQuantifier.getInstance(), new LinkedHashSet<Variable>(output), or));

			if (conjuncts.size() > 1) {
				return new BooleanQuery(new PolyadicExpression(AndOperator.getInstance(), conjuncts));
			} else
				return new BooleanQuery(conjuncts.get(0));

		} catch (SQLException | MalformedExpressionException | TranslationException | MalformedQueryException e) {
			throw new ExecutionException(e);
		} finally {

			if (rs != null)
				try {
					rs.close();
				} catch (SQLException e) {
					e.printStackTrace();
				}

			try {
				sqlStmt.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Updates the __ADOM__ view if keepActiveDomain is true. It does nothing otherwise.
	 * 
	 * @param state
	 *            - The state for which we want to create the active domain.
	 * 
	 * @param expr
	 *            - The expression from which we extract the constants in the active domain.
	 * @throws ExecutionException
	 */
	private void updateActiveDomain(IExpression expression, IState state) throws ExecutionException {
		if (keepActiveDomain) {
			Statement sqlStmt = manager.createStatement();
			// update the active domain
			try {
				// String expr = "CREATE OR REPLACE VIEW __ADOM__ (value) AS ";
				//
				// for (IType table : state.getTables())
				// for (String field : table.getFields())
				// expr += "(SELECT " + field + " AS value FROM " + table.getId() + ") UNION ";
				//
				Set<Constant> constants = new LinkedHashSet<Constant>();
				for (IView view : state.getViews()) {
					GetConstants v = new GetConstants();
					view.getSelectQuery().getQuery().getExpression().accept(v);

					constants.addAll(v.getConstants());
				}

				GetConstants v = new GetConstants();
				expression.accept(v);
				constants.addAll(v.getConstants());
				String expr = "";
				for (Constant c : constants) {
					expr += "INSERT INTO __ADOM__EXPR__VALUES__  VALUES (" + c + ");";
					
				}
				if (!constants.isEmpty())
					sqlStmt.executeUpdate(expr);
			} catch (SQLException e) {
				// other issues
				throw new ExecutionException(e);

			} finally {
				try {
					sqlStmt.close();
				} catch (SQLException e) {
					throw new ExecutionException(e);
				}
			}
		}

	}

	/**
	 * Create the __ADOM__EXPR__VALUES__.
	 * 
	 * @throws ExecutionException
	 */
	private void createActiveDomain() throws ExecutionException {
		if (keepActiveDomain) {
			Statement sqlStmt = manager.createStatement();
			// update the active domain
			try {
				String expr = "CREATE TABLE __ADOM__EXPR__VALUES__ ( value VARCHAR(200) NOT NULL) ENGINE InnoDB; ";
				sqlStmt.executeUpdate(expr);
			} catch (SQLException e) {
				// other issues
				throw new ExecutionException(e);

			} finally {
				try {
					sqlStmt.close();
				} catch (SQLException e) {
					throw new ExecutionException(e);
				}
			}
		}
	}

}
