package prototype.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import prototype.system.implementation.ExecutionException;

/**
 * Logger class. Saves timestamps using CSV format.
 * 
 * 
 *
 */
public class Logger {

	/**
	 * The name of the current step.
	 */
	private String stepName;
	/**
	 * The printwriter for the logs.
	 */
	private PrintWriter logs;

	interface Property {

	}

	/**
	 * The properties that we can measure.
	 * 
	 * 
	 */
	public enum TimeProperty implements Property {
		INTEGRITY_TOTAL_TIME, CONFIDENTIALITY_TOTAL_TIME, CONFIDENTIALITY_ADDITIONAL_CHECKS, CONFIDENTIALITY_BOOLEAN_QUERY_REWRITING, CONFIDENTIALITY_BOOLEAN_QUERY_EXECUTION, COMMAND_EXECUTION, CONFIDENTIALITY_NON_BOOLEAN_QUERY_REWRITING, CONFIDENTIALITY_NON_BOOLEAN_QUERY_EXECUTION, CONFIDENTIALITY_NON_BOOLEAN_QUERY_MATERIALIZATION;
	}

	/**
	 * The boolean properties.
	 * 
	 * 
	 */
	public enum BooleanProperty implements Property {
		INTEGRITY_RESULT, CONFIDENTIALITY_RESULT
	}

	/**
	 * Array of measurements (FROM)
	 */
	private long[] from;
	/**
	 * Array of measurements (TO)
	 */
	private long[] to;

	/**
	 * The time for each phase.
	 */
	private long[] time;
	/**
	 * Vector of all time properties.
	 */
	private Vector<TimeProperty> timePropertyVector;

	/**
	 * The value for each boolean property.
	 */
	private boolean[] booleanValues;
	/**
	 * Vector of all boolean properties.
	 */
	private Vector<BooleanProperty> booleanPropertyVector;

	/**
	 * Class constructor.
	 * 
	 * @param destination
	 *            - The file in which we store the logs.
	 * @param append
	 *            - If true, the logger appends the logs to destination. Overwrite destination otherwise.
	 * @throws ExecutionException
	 */
	public Logger(File destination, boolean append) throws ExecutionException {
		this.timePropertyVector = new Vector<TimeProperty>();
		for (TimeProperty p : TimeProperty.values())
			timePropertyVector.add(p);

		this.booleanPropertyVector = new Vector<BooleanProperty>();
		for (BooleanProperty p : BooleanProperty.values())
			booleanPropertyVector.add(p);

		try {
			this.logs = new PrintWriter(new FileWriter(destination, append));
		} catch (IOException e) {
			throw new ExecutionException(e);
		}

		reset();

		writeHeader();
	}

	private void writeHeader() {
		String log = "HEADER";
		log += ", ";
		for (TimeProperty p : timePropertyVector) {
			log += p + ", ";
		}
		for (BooleanProperty p : booleanPropertyVector) {
			log += p + ", ";
		}
		log = Util.replaceLast(log, ", ", "");

		// write to the log
		this.logs.println(log);
	}

	/**
	 * Start a new step in the measurement, i.e., a new action.
	 */
	public void startNewStep(String stepName) {
		this.stepName = "\"" + stepName + "\"";
	}

	/**
	 * Start the clock for the phase.
	 * 
	 * @param phase
	 *            - The phase.
	 */
	public void startMeasure(TimeProperty phase) {
		this.from[timePropertyVector.indexOf(phase)] = System.nanoTime();// System.currentTimeMillis();
	}

	/**
	 * Stop the clock for the phase.
	 * 
	 * @param phase
	 *            - The phase.
	 */
	public void endMeasure(TimeProperty phase) {
		this.to[timePropertyVector.indexOf(phase)] = System.nanoTime();// System.currentTimeMillis();

		this.time[timePropertyVector.indexOf(phase)] += this.to[timePropertyVector.indexOf(phase)] - this.from[timePropertyVector.indexOf(phase)];
	}

	/**
	 * Write the current measurements on the log.
	 */
	public void write() throws ExecutionException {

		// build the string
		String log = stepName;
		log += ", ";
		for (TimeProperty p : timePropertyVector) {
			log += (this.time[timePropertyVector.indexOf(p)]) + ", ";
		}
		for (BooleanProperty p : booleanPropertyVector) {
			log += (this.booleanValues[booleanPropertyVector.indexOf(p)]) + ", ";
		}
		log = Util.replaceLast(log, ", ", "");

		// write to the log
		this.logs.println(log);

		reset();
	}

	private void reset() {
		// reset the parameters
		this.stepName = null;
		this.from = new long[TimeProperty.values().length];
		this.to = new long[TimeProperty.values().length];
		this.time = new long[TimeProperty.values().length];
		this.booleanValues = new boolean[BooleanProperty.values().length];
	}

	/**
	 * Close the logs.
	 */
	public void close() {
		logs.flush();
		logs.close();
	}

	public void changeFile(File destination, boolean append) throws ExecutionException {
		this.close();
		this.timePropertyVector = new Vector<TimeProperty>();
		for (TimeProperty p : TimeProperty.values())
			timePropertyVector.add(p);
		this.booleanPropertyVector = new Vector<BooleanProperty>();
		for (BooleanProperty p : BooleanProperty.values())
			booleanPropertyVector.add(p);
		try {
			this.logs = new PrintWriter(new FileWriter(destination, append));
		} catch (IOException e) {
			throw new ExecutionException(e);
		}

		writeHeader();
		reset();
	}

	public void setBooleanValue(BooleanProperty property, boolean value) {
		this.booleanValues[booleanPropertyVector.indexOf(property)] = value;
	}
}
